/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * AgentComponent.java
 ***************************************************************************************/

package agent.base;

import java.awt.*;
import java.util.*;
import java.io.Serializable;

import agent.simplest.State;

/**
 * Generic component used to base other components.  This is a useful
 * file to read, as you need to follow the conventions outlined
 * in this class to be JAF-compliant.
 * <P>
 * Specifically, the Control component will usually provide a runtime
 * organization consisting of several different phases, during which
 * different activities are expected to be performed.  These phases
 * will correspond to particular functions called in your component,
 * which are provided in this base AgentComponent class.  For example,
 * during the "initialization" phase of control, the init() function
 * of your component will be called.  During the "ending" phase, end()
 * will be called, and so on.  So if your component needed to do
 * something during the init phase, you would override the init()
 * method to perform this action.  Note that if you do not need to
 * take any actions in a phase, you do not need to extend the
 * corrpesponding method.
 * <P>
 * Read through the method comments below for a more detailed view
 * of the different methods.  See agent.simplest.Control for more
 * information.
 *
 * @see agent.simplest.Control 
 */
public abstract class AgentComponent extends Panel implements Serializable {
  static final long serialVersionUID = 1234567890L; // kludge static ser_id.  Bad bad Bryan!
  private StringArray descriptor;
  private Vector dependencies = new Vector(10);
  private boolean hasgui = true;

  protected static int MIN_WIDTH = 300;
  protected static int MIN_HEIGHT = 23;

  /**
   * Default constructor.  Registers the component with State.  Individual
   * component constructors should only modify things that are "constant"
   * as far as the thread of execution is concerned and are independent
   * of other installed components.  This will typically include setting
   * up GUIs and calculating environmental constants that aren't user-
   * configurable.  You'll also want to note your dependencies and 
   * configuration parameters here.
   * @see #addDependency
   * @see agent.simplest.State#addParameterInfo
   */
  public AgentComponent() {
    setDescriptor(new StringArray(new String[] {getClass().getName()}));

    if (System.getProperty("APPLETFRAME") != null) {
        if (System.getProperty("APPLETFRAME").equalsIgnoreCase("false")) {
            setHasGUI(false);
        } else {
            setHasGUI(true);
        }
    }

    if (hasGUI())
        setFont(new Font("Dialog", Font.PLAIN, 12));
  }

  /**
   * Returns a simple String describing the component.  This
   * is generated from the normal descriptor StringArray.
   * @return The descriptor string
   * @see #setDescriptor
   */
  public String getDescriptorString() {
    String[] descriptor;
    String label = ""; 
    descriptor = getDescriptor().value();
    for (int i = 0; i < descriptor.length; i++) {
      label = descriptor[i] + " " + label;
    }
    return label;
  }

  /**
   * Returns an array of string tokens describing the component
   * (e.g. {"Executor"} or {"Learning", "Cache"}).  This descriptor
   * is used by other components to find components offering
   * a given service.
   * <P>
   * Defaults to the class name
   * @return An array of descriptor strings.
   */
  public StringArray getDescriptor() {
    return descriptor;
  }

  /**
   * Sets the descriptor array of the component.  When resetting the
   * descriptor array, keep in mind that any agent components you
   * extend may also have custom descriptors, and you may want to
   * integrate these into your new setting.  This function is 
   * responsible for notifying the State component about the 
   * descriptor change.
   * @param d An array of descriptor strings.
   * @see agent.simplest.State#removeComponent
   * @see agent.simplest.State#addComponent
   */
  public void setDescriptor(StringArray d) {

    if (!d.equals(descriptor)) {
      State.removeComponent(this);
      descriptor = d;
      State.addComponent(this);
    }
  }

  /**
   * Returns an array of string tokens describing the components which
   * this component depends on.  For instance, if the Problem Solving
   * class relies on a Execute and State classes to work, those should
   * both be listed here. 
   * <P>
   * The control class is responsible for checking component dependencies
   * after the component construction phase is complete.
   */
  public String[] getDependencies() {
    Enumeration e = dependencies.elements();
    String[] deps = new String[dependencies.size()];
    int i = 0;

    while(e.hasMoreElements()) {
      deps[i++] = (String)e.nextElement();
    }

    return deps;
  }

  /**
   * Adds a dependency to the dependency list.  For instance, if the 
   * Problem Solving class relies on Execute and State classes to
   * work, those should  both be added here.  Duplicate additions
   * are ignored.
   * <P>
   * Dependancies should be specified in the component's constructor.
   * <P>
   * Note that you only have to declare a dependancy on a class if
   * the class <U>must</U> be available for the component to work.
   * If you wrap your uses of a component with checks for its
   * presence (with State.hasComponent), and handle its absence
   * gracefully, you do not need to specify a dependancy on that
   * component.
   * @see #getDependencies
   * @see agent.simplest.State#hasComponent
   */
  public void addDependency(String dep) {
    if (!dependencies.contains(dep)) {
      dependencies.insertElementAt(dep, 0);
      if (!State.hasComponent(dep))
	utilities.Log.getDefault().log("  Unsatisfied dependancy added: " + dep, 3);
    }
  }

  /**
   * Called when the system state is stable (constructors have completed).
   * Typically, this should be used to setup component links and initialize
   * any transient structures the component makes use of (open files, sockets,
   * etc.)  Relatively permanent inter-component relationships can also be set
   * up here (i.e., it's kosher to have a static addListener statement hooking
   * the State component to Communications, rather than hand specifying it
   * in the builder tool.) This will be called just once, upon system start up.
   * @see agent.simplest.Control#initComponents
   */
  public void init() { }

  /**
   * Called when the component is ready to execute.  Threads, if any, should 
   * be started here, along with any other of the "real" services a given
   * component is responsible for.
   * <P>
   * I didn't call this method "start" because it could conflict in
   * components extending Thread, even though it sounds better.
   * @see agent.simplest.Control#startComponents
   */
  public void begin() { }

  /**
   * Called when the component is ended, which basically only happens
   * when the agent is terminated.  This should do any cleanup or post-execution
   * processing you might need.  Note that this is separate from Java's 
   * finalize() method, and was added because finalize() is not always
   * called upon program termination - you can still have regular finalize()
   * methods.
   * <P>
   * You have a begin - you have an end, hence the name.
   * @see agent.simplest.Control#stopComponents
   */
  public void end() { }

  /**
   * Called when the system has been reset (not on initialization).  This
   * should basically clear all the state from each component and put
   * it into a state where the begin method can be called.  You shouldn't
   * reset your listener list or set of componenent references unless you're
   * generating these dynamically.  The begin method for your component should
   * be called by the Control component following the reset method.
   * @see agent.simplest.Control#resetComponents
   */
  public void reset() { }

  /**
   * Called periodically by the Control component, either in reponse
   * to actual passage of time or in response to some other stimuli
   * it percieves as a time indicator (such as a pulse from a time
   * controller).  Use this method as you would the contents of an
   * unbounded while loop - they will be called periodically forever,
   * and thus can be used to effect the normal moment-to-moment
   * behavior of a given component.
   */
  public void pulse() { }

  /**
   * Returns true if the component has a popupmenu.  Override this
   * in components which do (this returns false by default).
   * @return True if the component has a popup menu
   */
  public boolean hasPopupMenu() {
      return false;
  }

  /**
   * Returns true if the component does something else if you
   * click on it (like show a dialog box).  Override this
   * in components which do (this returns false by default).
   * <p>
   * Note that you may also simply add(new Button(..)) to your
   * component, and use the regular AWT button widget.  In
   * this case, you do not need to override this function.
   * @return True if the component does visible with clicks.
   */
  public boolean hasClickAction() {
      return false;
  }

  /**
   * This is an event firing helper.  Basically, it is a firing wrapper,
   * which queues events in such a manner that all components should
   * receive them in order.  Call this function when you wish to fire
   * an event, and implement the fireEvent method to control how that
   * firing actually proceeds.
   * @param event The event to fire
   * @see #fireEvent(AgentEvent)
   */
  public void startFireEvent(AgentEvent event) {
    
    // Start by adding it to the queue
    ev_queue.addElement(event);

    // Stop if someone else is already firing
    if (ev_firing)
      return;

    // Critical section
    synchronized (ev_lock) {
      // We be firing
      ev_firing = true;
      
      // Sent each of the known event out
      while (!ev_queue.isEmpty()) {
	AgentEvent ev = (AgentEvent)ev_queue.firstElement();
	ev_queue.removeElement(ev);
	fireEvent(ev);
      }
      
      // Done
      ev_firing = false;
    }
  }
  private boolean ev_firing = false;
  private Vector ev_queue = new Vector();
  private Integer ev_lock = new Integer(0);

  /**
   * This is where your actual event firing should take place.  Generally,
   * this involves traversing through a list of listeners and sending the
   * event to each object listed therin.  You should not have to call this
   * yourself, instead call startFireEvent, which manages the firing 
   * process in a cleaner manner than would normally take place.
   * @see #startFireEvent
   */
  protected void fireEvent(AgentEvent event) {
    utilities.Log.getDefault().log("Warning - No-op fireEvent was called.", 1);
  }

    /**
     * Sets a flag indicating that the component should make its GUI
     * or not, theoretically allowing people to run an agent sans
     * display.
     * @param s Set to true to show the GUI
     */
    public void setHasGUI(boolean s) { hasgui = s; }

    /**
     * Returns true if the component should perform GUI things.
     */
    public boolean hasGUI() { return hasgui; }

  /**
   * Stuff below is the default appearence methods. Override paint or add
   * an AWT component to your component to have a different appearence.
   * Some of this below is cheesed from the button demo that comes with
   * the Sun BDK.
   */

  /**
   */
  public Dimension getPreferredSize() { 
    
    // Use default if something's been added
    if (getComponentCount() != 0) {
      return super.getPreferredSize();
    }

    // Get the string
    String label = getDescriptorString(); 

    // Measure it
    FontMetrics fm = getFontMetrics(getFont());
    Dimension d = new Dimension(Math.max(MIN_WIDTH, fm.stringWidth(label) + 12),
                                Math.max(MIN_HEIGHT, fm.getMaxAscent() + fm.getMaxDescent() + 8));

    return d;
  }

  /**
   * The label is centered in both dimensions.
   */
  public synchronized void paint(Graphics g) {
    int width = getSize().width;
    int height = getSize().height;

    // Use default if something's been added
    if (getComponentCount() != 0) {
      super.paint(g);
      return;
    }

    // Get the string
    String label = getDescriptorString(); 

    // Draw it
    g.setColor(getBackground());
    g.setColor(getForeground());
    g.setFont(getFont());
    g.drawRect(2, 2, width - 4, height - 4);
    FontMetrics fm = g.getFontMetrics();
    g.drawString(label, (width - fm.stringWidth(label)) / 2,
		 (height + fm.getMaxAscent() - fm.getMaxDescent()) / 2);

    if (hasPopupMenu()) {
        Polygon p = new Polygon();
        p.addPoint(width - 16, height / 2 - 3);
        p.addPoint(width - 10, height / 2 - 3);
        p.addPoint(width - 13, height / 2 + 3);
        g.fillPolygon(p);
    }

    if (hasClickAction()) {
        g.fillOval(10, height / 2 - 3, 6, 6);
    }
  }

  /**
   */
  public void setFont(Font f) {
    super.setFont(f);
    sizeToFit();
  }

  /**
   */
  private void sizeToFit() {
    Dimension d = getSize();
    Dimension pd = getPreferredSize();
    
    if (pd.width > d.width || pd.height > d.height) {
      int width = d.width;
      if (pd.width > width) {
	width = pd.width;
      }
      int height = d.height;
      if (pd.height > height) {
	height = pd.height;
      }
      setSize(width, height);
      
      Component p = getParent();
      if (p != null) {
	p.invalidate();
	p.validate(); 
      }
    }
  }
}

