/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * ListenerVector.java
 ***************************************************************************************/

package agent.base;

import java.util.*;
import java.io.Serializable;

import agent.base.AgentListener;

/**
 * Prioritized listener table
 */
public class ListenerVector implements Serializable {
  private static final int DEF_LENTH = 10;
  private static final int DEF_WIDTH = 5;

  private Vector list;

  /**
   * Default constructor
   */
  public ListenerVector() {
    this(DEF_LENTH, DEF_WIDTH);
  }

  /**
   * Parameterized constructor
   *
   * @param length The "length" of the list, i.e., the nuber of priorities
   * @param width The length of each sub-priority list
   */
  public ListenerVector(int length, int width) {

    list = new Vector(length);

    // Generate the list of lists
    for (int i = 0; i < length; i++) {
      list.insertElementAt(new Vector(width), i);
    }
  }

  /**
   * Adds a listener to the list, no priority.  This function should
   * not be used in conjunction with the prioritized version.
   *
   * @param a The listener to add to the list
   * @param priority The priority slot to add it to
   */
  public synchronized void add(AgentListener a) {
    add(a,0);
  }

  /**
   * Adds a listener to the list
   *
   * @param a The listener to add to the list
   * @param priority The priority slot to add it to
   */
  public synchronized void add(AgentListener a, int priority) {
    Vector l = (Vector)list.elementAt(priority);
    l.addElement(a);
  }

  /**
   * Removes a listener from the list
   *
   * @param a The listener to remove from the list
   */
  public synchronized boolean remove(AgentListener a) {
    Vector l; 

    // Search in each list for the listener
    for (int i = 0; i < list.size(); i++) {
      l = (Vector)list.elementAt(i);
      if (l.removeElement(a)) return true;
    }

    return false;
  }

  /**
   * Removes all the listeners from the list
   */
  public synchronized void removeAllElements() {
    Vector l; 

    // Search in each list for the listener
    for (int i = 0; i < list.size(); i++) {
      l = (Vector)list.elementAt(i);
      l.removeAllElements();
    }
  }

  /**
   * Returns a sorted enumeration of the listeners (note that the
   * relative order of similar priorities is undefined, but different
   * priorities will be sorted, unlike a regular enumeration).
   */
  public Enumeration elements() {
    return new ListenerEnumeration(list);
  }

  /**
   * Returns the length of the list
   */
  public int length() { return list.size(); }

  /**
   * Returns the size of the list
   * @see #length()
   */
   public int size() { return list.size(); }

  /**
   * Determines if the listener already exists in the list
   */
  public boolean contains(Object e) {
    Vector l;

    for (int i = 0; i < list.size(); i++) {
      l = (Vector)list.elementAt(i);
      if (l.contains(e))
          return true;
    }

    return false;
  }
}

final class ListenerEnumeration implements Enumeration {
  Vector vector;
  Object next = null;
  int curp = 0;
  int curl = 0;

  ListenerEnumeration(Vector v) {
    vector = v;
    findNext();
  }

  public boolean hasMoreElements() {
    return (next != null);
  }

  public Object nextElement() {
    Object cur = next;

    findNext();
    return cur;
  }

  private synchronized void findNext() {
    Vector l;

    // Look at each priority
    while(curp < vector.size()) {
      l = (Vector)vector.elementAt(curp); 

      // And at each listener in the slot
      try {
          if (curl < l.size())
              next = l.elementAt(curl);
          else
              next = null;
      } catch (ArrayIndexOutOfBoundsException e) {
          next = null;
      }

      // Ok?
      if (next == null) {
          curl = 0;
          curp++;
	
      } else {
          curl++;
          return;
      }
    }

    next = null;
  }
}
