/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * Communicate.java
 *
 * manifest: Name: agent/mass/Communicate.class
 * manifest: Depends-On: agent/simplest/Communicate.class
 * manifest: Java-Bean: True
 ***************************************************************************************/

package agent.mass;

import java.io.*;
import java.awt.*;
import java.util.*;
import java.net.*;

import agent.base.*;
import agent.simplest.*;
import utilities.Message;
import utilities.KQMLMessage;
import utilities.Connection;

/**
 * An extension of the simplest communicate class which work
 * with the simulator and speaks KQML
 * @see agent.simplest.Communicate
 */
public class Communicate extends agent.simplest.Communicate implements PropertyEventListener {
  static final long serialVersionUID = 1234567890L; // kludge static ser_id.  Bad bad Bryan!

  /**
   * Constructor
   */
  public Communicate() { 
    super();

    // Redefine these parameters and defaults
    State.addParameterInfo(HOST, null, "The ip address of the machine hosting the Mass simultor.", "localhost");
    State.addParameterInfo(PORT, null, null, new Integer(25000));
  }

  /**
   * Inits me
   */
  public void init() {
    super.init();

    // Observation stuff
    if (State.hasComponent("Observe")) {
        Observe observe = (Observe)State.findComponent("Observe");

        observe.addObserver(new agent.simplest.Observer("Real Messages Sent") {
                public boolean checkEvent(AgentEvent e) {
                    if ((e instanceof MessageEvent) && (e.getID() == MessageEvent.SENT)) {
                        if (((MessageEvent)e).getMessage().getDestAddr().equalsIgnoreCase("simulator")) return false;
                        return true;
                    }
                    return false;
                }
            });
        observe.addObserver(new agent.simplest.Observer("Real Messages Received") {
                public boolean checkEvent(AgentEvent e) {
                    if ((e instanceof MessageEvent) && (e.getID() == MessageEvent.RECEIVED)) {
                        if (((MessageEvent)e).getMessage().getSourceAddr().equalsIgnoreCase("simulator")) return false;
                        return true;
                    }
                    return false;
                }
            });
    }
  }

    /**
     * Begin
     */
    public void begin() {

        super.begin();

        if (State.hasComponent("LogViewer")) {
            LogViewer viewer = (LogViewer)State.findComponent("LogViewer");
            
            LogViewerFilter filter = viewer.findFilter("Msg Sent");
            if (filter != null)
                viewer.removeFilter(filter);

            viewer.addFilter(new LogViewerFilter("Msg Sent", false) {
                    public boolean isMyLogEntry(String s) {
                        //           if (s.indexOf(":type control") != -1) return false;
                        if (s.indexOf("(ack pulse)") != -1) return false;
                        return s.indexOf("Message Sent: ") != -1;
                    }
                    public String formatMyLogEntry(String s) {
                        return "From: " + getAgent(s) + "\n" +
                            "To: " + getTo(s) + "\n" +
                            "Mesg: " + s.substring(s.indexOf('"'));
                    }                
                    public void drawToken(Graphics g, int x, int y, LogViewerEntrySlot v) {
                        Color c = g.getColor();
                        if (v.size() == 1) {
                            LogViewerEntry e = (LogViewerEntry)v.firstElement();
                            String to = getTo(e.getData());
                            g.setColor(LogViewer.getColor(to));
                        }
                        drawTokenType(g, x, y, 0);
                        g.setColor(c);
                    }
                    public void drawTokenType(Graphics g, int x, int y, int type) {
                        super.drawTokenType(g, x, y, 2);
                    }
                    public String getTokenType(int type) {
                        switch (type) {
                        case 0:
                            return "Message Sent (color is destination)";
                        }
                        return null;
                    }
                    public String getTo(String s) {
                        int i = s.indexOf("(to ") + 4;
                        return s.substring(i, s.indexOf(')', i));
                    }
                });
        }
    }

  /**
   * Resets me
   */
  public void reset() {
      log.log("mass.Communicate does not reset", Log.LOG_WARNING, FAC_COM);
      //super.reset();
  }

  /**
   * This derivation checks to see if the contacted host is the Mass simulator.
   * by comparing host to the Host property.  If it is the simulator, the
   * handshake is performed.
   * @param host The address of the remote host to connect to
   * @param port The port number to connect to
   * @param type The connection type
   */
  public void open(String host, int port, int type) {

    // Fix the connection type
    if (state.hasProperty(HOST) && host.equals(state.getProperty(HOST))) {
      log.log("Changing simulator connection type to KQML...", Log.LOG_INFO, FAC_COM);
      type = Connection.KQML;
    }

    // Make the connection
    super.open(host, port, type);

    // Update stuff if its the simulator
    if (state.hasProperty(HOST) && host.equals(state.getProperty(HOST))) {
      // Make it the default
      Connection c = findConnection(host);
      if (c == null) {
	log.log("Whoops, looks like there was an error connecting to the simulator (" + host + ":" + port + ").  Quit the agent, and make sure you have started Mass and that it is listening on the correct port before trying again.", Log.LOG_ERR, FAC_COM);
	return;
      }
      setDefaultConnection(c);
      ready = true;

      // Simulator handshake
      KQMLMessage m = new KQMLMessage("register", "", "simulator");
      m.addField("type", "control");
      sendMessage(m);

      // Name isn't set yet
      String name = (String)state.getProperty(State.NAME);
      state.setProperty(State.NAME, name + "?");

      // Send queued messages
      flushOQueue();
      log.log("Connection to simulator established.", Log.LOG_INFO, FAC_COM);
    }
  }

  /**
   * Determines if it is ok to send a message yet, checks the
   * name as well as connection state.
   * @return true if messages can be sent
   */
  public boolean ready() {
    if (state == null) return false;

    String name = (String)state.getProperty(State.NAME);

    return (super.ready() && (!name.endsWith("?")));
  }

  /**
   * Finds a connection in the connection list (actually justs
   * switches to default connection).
   * @param host The host's connection to find
   * @return The connection, or null if not found
   */
  public Connection findConnection(String host) {
    Connection c = findDefaultConnection();

    if (c == null)
	return super.findConnection(host);
    else
        return c;
  }

  /**
   * Sends a KQML message
   * @param m The message you want to send
   * @see Message
   */
  public boolean sendMessage(Message m) {

    if ((m.getSourceAddr() == null) || (m.getSourceAddr().equals("")))
      m.setSourceAddr((String)state.getProperty(State.NAME));

    return super.sendMessage(m);
  }

  /**
   * Monitor messages that arrive before they are dispatched.  Handles
   * clock inc and echo reply.  Queues non-control messages.
   * @param m The message event to look at
   * @param c The connection it came in on
   * @return true if the Message was fired
   */
    public boolean messagePreprocess(Message m, Connection c) {

        log.log("messagePreprocess: " + m, Log.LOG_SPAM, FAC_COM);

        // Check for other message types
        if (c.getType() != Connection.KQML) {
            log.log("Message received on unknown connection type, handing off to parent.", Log.LOG_DEBUG2, FAC_COM);
            return super.messagePreprocess(m, c);
        }

        String perf = ((KQMLMessage)m).getPerformative();
        String content = (String)m.getContent();
        String type = (String)((KQMLMessage)m).getField("type");
        KQMLMessage reply;

        // Fill in the time
        m.setReceiveTime(((Integer)state.getProperty(State.TIME)).intValue());

        m.setConnection(c);

        if (type == null) { type = ""; }

        if (type.equalsIgnoreCase("control")) {
            if (perf.equalsIgnoreCase("ask")) {
                if (m.contentWord().equalsIgnoreCase("pulse")) {
                    // Clock pulse
                    log.log("Pulse", Log.LOG_SPAM, FAC_COM);

                    // Pulse
                    control.doPulse();

                    // We're done here
                    return true;
                }

            } else if (perf.equalsIgnoreCase("echo")) {
                // Ping
                reply = new KQMLMessage("echo", m.getContent(), m.getSourceAddr());
                reply.addField("in-reply-to", "echo");
                sendMessage(reply);
            }

            // Everything else
            log.log("Firing control message", Log.LOG_SPAM, FAC_COM);
            startFireEvent(new MessageEvent(this, m, c, MessageEvent.RECEIVED));
            return true;

        } else {
            // Queue it
            synchronized (iqueue) {
                log.log("Queuing message for delivery", Log.LOG_SPAM, FAC_COM);
                iqueue.addElement(new MessageEvent(this, m, c, MessageEvent.RECEIVED));
            }
        }

        return false;
    }

  /**
   * Monitor messages that arrive before they are dispatched.  Handles
   * pulse acknowledgement.
   * @param m The message event to look at
   * @param c The connection it came in on
   */
  public void messagePostprocess(Message m, Connection c) {

    // Check for other message types
    if (c.getType() != Connection.KQML) {
      super.messagePostprocess(m, c);
      return;
    }

    // Check if we can send
    flushOQueue();
  }
}





