/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * CondRule.java
 ***************************************************************************************/

package agent.mass;

import java.io.*;
import java.util.*;

import agent.simplest.*;

/**
 * Used to store conditioning rules.  Extend as necessary.
 * This class is really not much more than a convenience structure,
 * containing parsing and matching functions.
 * <P>
 * @depricated We don't really use these anymore, as AsggOrgDesign is
 * obsolete.
 */
public class CondRule {
  protected static Log log; // hack
  protected static int FAC_CONDRULE; // hack

  /**
   * Multi-purpose codes
   */
  public static final int UNKNOWN = -9999;
  public static final int PERMANENT = 30000;

  /**
   * Action codes
   */
  public static final int REMOVE = 100;
  public static final int ADD = 101;
  public static final int REPLACE = 102;
  public static final int MODIFY = 103;

  /**
   * Target codes
   */
  public static final int GRAMMAR = 200;
  public static final int TAEMS = 201;

  /**
   * View codes
   */
  public static final int OBJECTIVE = 300;
  public static final int SUBJECTIVE = 301;
  public static final int CONDITIONED = 302;
  
  // General
  public int start = UNKNOWN;      // Rule start time
  public int duration = UNKNOWN;   // Rule duration
  public int action = UNKNOWN;     // Rule action code
  public int target = UNKNOWN;     // Rule target code

  // Grammar specific
  public String rule = null;       // Grammar rule string

  // Taems-specific
  public int view = UNKNOWN;       // View code (e.g. objective, subjective..)
  public String type = null;       // Node type (e.g. interrelationship, task...)
  public String subtype = null;    // Interrelationship subtype (e.h. enables, limits...)
  public String src = null;        // Interrelationship source
  public String tgt = null;        // Interrelationship target
  public String node = null;       // Task/Method node name

  /**
   * Rule translation constructor.  A string containing a valid rule is 
   * parsed and stored.  The string format is as follows:
   * <P>
   * All rules start with the following prefix:
   * <P>
   *    <DD><TT>START DURATION TARGET ACTION VIEW</TT>
   * <P>
   * Meaning, the start time, duration of rule's effect, the desired target
   * (e.g. the grammar or Taems object) and the action (e.g. remove, add, etc.).
   * If the duration field is filled by an asterik "*" the rule's effects
   * are enforced indefinately once started.
   * <P>
   * Following this prefix will be the action description, which can take 
   * multiple forms.  Grammar rules can take one of the following forms:
   * <P>
   *    <DD><TT>START DURATION GRAMMAR REPLACE VIEW GRAMMAR_FILE</TT>
   *    <DD><TT>START DURATION GRAMMAR REPLACE VIEW GRAMMAR</TT>
   *    <DD><TT>START DURATION GRAMMAR MODIFY VIEW LHS RHS</TT>
   * <P>
   * Where the grammar file is stored in src, the grammar uses the same syntax
   * as an asgg grammar, and the modify rule allows you to change the RHS of a
   * single given rule with lhs LHS.
   * <P>
   * TTaems can take two forms, one for interrelationships and another 
   * for tasks and methods:
   * <P>
   *    <DD><TT>START DURATION TAEMS ACTION VIEW TYPE SUBTYPE SOURCE TARGET</TT><BR>
   *    <DD><TT>START DURATION TAEMS ACTION VIEW TYPE NODE</TT>
   * <P>
   * Where the view is the structure type (e.g. objective, subjective, etc.),
   * type is the affected node type (e.g interrelationship, method), subtype
   * is the interrelationship type (e.g. enables, limits), source is the 
   * IR source node name, target the IR target node's name and node the node
   * name of the task or method to modify.  Note that the target, action and view
   * fields should be filled with the appropriate constants defined  in this class.
   *
   * @param str The string to parse, presumably containing a 
   *   conditioning rule.
   */
  public CondRule(String str) {
    String tmp;

    // Tokenit
    StringTokenizer tok = new StringTokenizer(str);
    if (tok.countTokens() < 5) {
      log.log("Error: Conditioning rule syntax error (" + str + ")", Log.LOG_ERR, FAC_CONDRULE);
      return;
    }

    // Start translating
    start = Integer.parseInt(tok.nextToken());
    tmp = tok.nextToken();
    if (tmp.indexOf("*") >= 0)
      duration = PERMANENT;
    else
      duration = Integer.parseInt(tmp);
    target = transTarget(tok.nextToken());
    action = transAction(tok.nextToken());
    view = transView(tok.nextToken());

    if (target == GRAMMAR) {
      // We got us a grammar rule
      if (action == REPLACE) {
	// New grammar
	if (tok.countTokens() == 1) {
	  // File
	  src = tok.nextToken();

	} else {
	  // Grammar
	  String spot = tok.nextToken();
	  rule = str.substring(str.indexOf(spot, str.length()));
	}

      } else if (tok.countTokens() >= 2) {
	// Modify rule
	src = tok.nextToken();
	tgt = "";
	while (tok.hasMoreElements())
	  tgt += tok.nextToken() + " ";
      }

    } else if (target == TAEMS) {
      // That's a taems rule that is
      if (tok.countTokens() >= 4) {
	// IR rule
	type = tok.nextToken();
	subtype = tok.nextToken();
	src = tok.nextToken();
	tgt = tok.nextToken();

      } else if (tok.countTokens() >= 2) {
	// Task/Meth rule
	type = tok.nextToken();
	node = tok.nextToken();

      } else {
	log.log("Error: Error parsing taems conditioning rule", Log.LOG_ERR, FAC_CONDRULE);
      }

    } else {
      // Eh?
      log.log("Error: Uknown rule target", Log.LOG_ERR, FAC_CONDRULE);
      return;
    }

    // Extra junk?
    if (tok.countTokens() > 0) {
      String junk = "";
      while (tok.hasMoreElements()) junk = junk + tok.nextToken();
      log.log("Unrecognized junk at end of rule (" + junk + ")", Log.LOG_ERR, FAC_CONDRULE);
    }
  }

  /**
   * Grammar rule constructor.
   * @param s Start time
   * @param d Duration time
   * @param t Target code
   * @param a Action code
   * @param v View code
   * @param r Grammar rule
   */
  public CondRule(int s, int d, int t, int a, int v, String r) {

    start = s;
    duration = d;
    target = t;
    action = a;
    view = v;
    rule = r;
  }

  /**
   * TTaems interrelationship rule constructor.
   * @param s Start time
   * @param d Duration time
   * @param t Target code
   * @param a Action code
   * @param v View code
   * @param ty Type, should be set to "Interrelationship"
   * @param st IR subtype
   * @param sr IR source node
   * @param tg IR target node
   */
  public CondRule(int s, int d, int t, int a, int v, String ty, String st, String sr, String tg) {

    start = s;
    duration = d;
    target = t;
    action = a;
    view = v;
    type = ty;
    subtype = st;
    src = sr;
    tgt = tg;
  }

  /**
   * TTaems task/method and grammar modify rule constructor.
   * @param s Start time
   * @param d Duration time
   * @param t Target code
   * @param a Action code
   * @param v View code
   * @param p1 For Taems this is node type, should be set to "Task" or "Method".
   *    For Grammar this is rhs (stored in src)
   * @param p2 For Taems this is the node name
   *    For Grammar this is the space delimited rhs (stored in tgt)
   */
  public CondRule(int s, int d, int t, int a, int v, String p1, String p2) {

    start = s;
    duration = d;
    target = t;
    action = a;
    view = v;

    if (target == GRAMMAR) {
      src = p1;
      tgt = p2;
    } else {
      type = p1;
      node = p2;
    }
  }

  /**
   * Translates the action code
   * @param the action code in string form
   * @return The action code
   */
  public int transAction(String a) {

    if (a.equalsIgnoreCase("REMOVE"))
      return REMOVE;
    else if (a.equalsIgnoreCase("ADD"))
      return ADD;
    else if (a.equalsIgnoreCase("REPLACE"))
      return REPLACE;
    else if (a.equalsIgnoreCase("MODIFY"))
      return MODIFY;
    else
      return UNKNOWN;
  }

  /**
   * Translates the target code
   * @param the target code in string form
   * @return The target code
   */
  public int transTarget(String t) {

    if (t.equalsIgnoreCase("GRAMMAR"))
      return GRAMMAR;
    else if (t.equalsIgnoreCase("TAEMS"))
      return TAEMS;
    else
      return UNKNOWN;
  }

  /**
   * Translates the view code
   * @param the view code in string form
   * @return The view code
   */
  public int transView(String v) {

    if (v.equalsIgnoreCase("OBJECTIVE"))
      return OBJECTIVE;
    else if (v.equalsIgnoreCase("SUBJECTIVE"))
      return SUBJECTIVE;
    else if (v.equalsIgnoreCase("CONDITIONED"))
      return CONDITIONED;
    else
      return UNKNOWN;
  }

  /**
   * Checks if the supplied rule "matches" this rule.  Matching is
   * essentially a watered down version of equals - an object
   * is considered to be matching if it equals the comparison
   * object in a specified subset of attributes.
   * <P>
   * Use UNKNOWN for int attributes you don't care
   * about, or null for String ones. 
   * @param The rule to compare against
   * @return True if the two match, false otherwise
   */
  public boolean matches(CondRule compare) {

    if ((compare.start != CondRule.UNKNOWN) && (compare.start != start))
      return false;
    if ((compare.duration != CondRule.UNKNOWN) && (compare.duration != duration))
      return false;
    if ((compare.action != CondRule.UNKNOWN) && (compare.action != action))
      return false;
    if ((compare.target != CondRule.UNKNOWN) && (compare.target != target))
      return false;
    if ((compare.view != CondRule.UNKNOWN) && (compare.view != view))
      return false;
    
    if ((compare.rule != null) && (!compare.rule.equalsIgnoreCase(rule)))
      return false;
    if ((compare.type != null) && (!compare.type.equalsIgnoreCase(type)))
      return false;
    if ((compare.subtype != null) && (!compare.subtype.equalsIgnoreCase(subtype)))
      return false;
    if ((compare.src != null) && (!compare.src.equalsIgnoreCase(src)))
      return false;
    if ((compare.tgt != null) && (!compare.tgt.equalsIgnoreCase(tgt)))
      return false;
    if ((compare.node != null) && (!compare.node.equalsIgnoreCase(node)))
      return false;
    
    return true;
  }

  /**
   * String-ify
   */
  public String toString() {
    String str = "[CondRule] (" + start + "-" + duration + ") [" + target + ":" + action + "] " + view + " ";

    if (target == GRAMMAR) {
      if (rule != null)
	str = str + rule;
      else
	str = str + " " + src + " => " + tgt;

    } else if (target == TAEMS) {
      if (node != null)
	 str = str + type + " " + node;
      else
	str = str + type + " " + subtype + " " + src + " " + tgt;
    } else {
      str = str + "unkown";
    }

    return str;
  }
}
