/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * ConflictRecord.java
 ***************************************************************************************/

package agent.mass;

import java.io.*;
import java.awt.*;
import java.util.*;

import taems.*;

/**
 */
public class ConflictRecord {
    public static final int REQUEST = 0;
    public static final int RESPONSE = 1;
    public static final int DELAY = 2;

    protected ConflictListener source;
    protected int type;
    private static int mid = 0;
    protected int id = mid++;
    protected ConflictRecord record;
    protected Taems taems;
    protected Hashtable attributes = new Hashtable();

    /**
     * Default constructor.
     * @param source The source of the record
     * @param taems The taems object in conflict
     */
    public ConflictRecord(ConflictListener source, Taems taems) {
        setSource(source);
        setType(REQUEST);
        setTaems(taems);
    }

    /**
     * Response constructor.  This assigns the records id to
     * match that of the instigating record - the ConflictRecord
     * which the new record is addressing.
     * @param record The instigating record
     * @param source The source of the record
     * @param taems The resolved taems object 
     */
    public ConflictRecord(ConflictListener source, ConflictRecord record, Taems taems) {
        setRecord(record);
        setSource(source);
        setType(RESPONSE);
        setTaems(taems);
    }

    /**
     * Delay constructor.  This assigns the records id to
     * match that of the instigating record - the ConflictRecord
     * which the new record is addressing.  This indicates to
     * the originator that there will be a delay before the resolution
     * results are available - essentially that the process will be
     * asynchronous, and that they may wish to wait before proceeding.
     * Typically, the original conflict source should be a ConflictListener,
     * so the resolver can eventually fire the results back to it.
     * @param record The instigating record
     * @param source The source of the record
     */
    public ConflictRecord(ConflictListener source, ConflictRecord record) {
        setRecord(record);
        setSource(source);
        setType(DELAY);
    }

    /**
     * Accessors
     */
    protected void setRecord(ConflictRecord r) { 
        record = r;
        if (getRecord() != null)
            setID(getRecord().getID());
    }
    public ConflictRecord getRecord() { return record; }
    protected void setSource(ConflictListener s) { source = s; }
    public ConflictListener getSource() { return source; }
    protected void setType(int t) { type = t; }
    public int getType() { return type; }
    protected void setTaems(Taems t) { taems = t; }
    public Taems getTaems() { return taems; }
    protected void setID(int i) { id = i; }
    public int getID() { return id; }

    /**
     * Note that this uses the toString() method for the key
     * to actually store the object.
     * @param k The key identifying the desired attribute
     * @return The object, or null if not found
     */
    public Object getAttribute(Object k) {
        return attributes.get(k.toString());
    }

    /**
     * Sets an attribute's data.
     * <P>
     * If a key is added with a null data object, the key will
     * be removed from the attribute set.
     * @param k The key identifying the desired attribute
     * @param d The attribute data
     */
    public void setAttribute(Object k, Object d) { 
        if (d == null)
            removeAttribute(k);
        else {
            Object o = getAttribute(k);
            if ((o == null) || (!o.toString().equals(d.toString()))) {
                attributes.put(k.toString(), d);
            }
        }
    }

    /**
     * Removes an attribute field from the object.
     * @param k The attribute to remove.
     */
    public void removeAttribute(Object k) {
        attributes.remove(k.toString());
    }

    /**
     * Returns the attribute names
     * @return An enumeration of all the attribute keys
     */
    public Enumeration getAttributes() { return attributes.keys(); }

    /**
     * Determines if the node has a particular attribute
     * @return True if the attribute is present
     */
    public boolean hasAttribute(Object k) {
        return attributes.containsKey(k.toString());
    }

    /**
     * Returns a reference to the raw attribute table.  Not
     * for public consumption.
     * @return The attribute table, or null if not found
     */
    protected Hashtable getAttributesTable() {
        return attributes;
    }

    /**
     * Returns the number of attributes.
     * @return The number of attributes posessed by the object
     */
    public int numAttributes() {
        return attributes.size();
    }
}
