/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * ConflictResolution.java
 *
 * manifest: Name: agent/mass/ConflictResolution.class
 * manifest: Java-Bean: True
 ***************************************************************************************/

package agent.mass;

import java.io.*;
import java.awt.*;
import java.util.*;

import agent.base.*;
import agent.simplest.*;
import utilities.*;
import taems.*;
import agent.simplest.Log;

/**
 */
public class ConflictResolution extends AgentComponent implements ConflictListener {
    static final long serialVersionUID = 1234567890L; // kludge static ser_id.  Bad bad Bryan!

    protected Log log;
    protected State state;
    protected Vector resolvers = new Vector();

    protected int FAC_CR;

    /**
     * Default constructor.
     */
    public ConflictResolution() {
        super();

        addDependency("Log");
        addDependency("State");
    }

    /**
     * Init
     */
    public void init() {

        log = (Log)State.findComponent("Log");
        state = (State)State.findComponent("State");

        FAC_CR = log.getFacilityID(this);

        super.init();
    }

    /**
     * Begin
     */
    public void begin() {
        addResolver(new NodeReplacementResolver());
        addResolver(new PreRescheduleResolver());
        addResolver(new RescheduleResolver());
        addResolver(new AltScheduleResolver());
    }

    /**
     * Registers a new conflict resolving object.
     */
    public void addResolver(ConflictListener l) {
        if (!resolvers.contains(l))
            resolvers.addElement(l);
    }

    /**
     * Un-registers a conflict resolving object.
     */
    public void removeResolver(ConflictListener l) {
        resolvers.removeElement(l);
    }

    /**
     * Gets the resolvers
     */
    public Enumeration getResolvers() {
        return resolvers.elements();
    }

    /**
     * Called when a conflict needs to be resolved.
     * This should return a vector of ConflictRecords, which
     * describe different ways the conflict could be resolved.
     * <p>
     * Do not modify the contents of the original conflict
     * record.
     */
    public Vector conflictRequest(ConflictRecord cr) {
        log.log("Performing conflict resoltion analysis", Log.LOG_INFO, FAC_CR);

        Vector results = new Vector();
        for (int i = 10; i >= 0; i--) {
            log.log("Checking resolvers for level " + i, Log.LOG_INFO, FAC_CR);
            Enumeration e = getResolvers();
            while (e.hasMoreElements()) {
                ConflictListener cl = (ConflictListener)e.nextElement();
                if (cl.getResolverPriority() != i) continue;
                log.log("Checking resolver " + cl.toString(), Log.LOG_DEBUG, FAC_CR);
                Vector v = cl.conflictRequest(cr);
                if (v == null) continue;
                Enumeration ve = v.elements();
                while (ve.hasMoreElements())
                    results.addElement(ve.nextElement());
            }
            if (!results.isEmpty()) break;
        }
        return results;
    }

    /**
     * Called when a conflict response is received
     */
    public void conflictResponse(ConflictRecord cr) { }
    public int getResolverPriority() { return 0; }

    /**
     * Looks for a prespecified node replacement comment within
     * the failed method, and replaces it.  The modified structure
     * is then returned.  This allows the designer to prespecify
     * a failure reaction directly.
     */
    public class NodeReplacementResolver implements ConflictListener {
        SimpleTaemsReader reader;
        String mycode = "ReplaceNode";

        public NodeReplacementResolver() {
            reader = (SimpleTaemsReader)State.findComponent("SimpleTaemsReader");
        }

        public int getResolverPriority() { return 6; }

        /**
         * Called when a conflict needs to be resolved.
         * This should return a vector of ConflictRecords, which
         * describe different ways the conflict could be resolved.
         * <p>
         * Do not modify the contents of the original conflict
         * record.
         */
        public Vector conflictRequest(ConflictRecord cr) {
            Vector results = new Vector();

            if ((cr.getType() == ConflictRecord.REQUEST) &&
                (cr.getTaems() != null)) {
                Enumeration e = cr.getTaems().findNodes(new Method());
                while (e.hasMoreElements()) {
                    TaskBase t = (TaskBase)e.nextElement();
                    if (t.hasFinished() && t.getCurrentQuality() == 0) {
                        log.log("Looking for conflict stuff on " + t.getLabel(), Log.LOG_INFO, FAC_CR);
                        while (t != null) {
                            if (t.hasAttribute("NoQualityException")) {
                                String r = (String)t.getAttribute("NoQualityException");
                                if (r.startsWith(mycode)) {
                                    Taems n = generateReplaceNodeStructure(cr.getTaems(), r);
                                    if (n != null)
                                        results.add(new ConflictRecord(this, cr, n));
                                }
                                break;
                            }
                            t = t.firstSupertask();
                        }
                    }
                }
            }

            return results;
        }

        /**
         * Called when a conflict response is received
         */
        public void conflictResponse(ConflictRecord cr) { }

        /**
         * Replace node reaction
         * ReplaceNode Node1 NewTaems Node2
         */
        public Taems generateReplaceNodeStructure(Taems t, String r) {
            StringTokenizer tok = new StringTokenizer(r);
            log.log("Generating replace node structure " + r, Log.LOG_INFO, FAC_CR);

            // Get the parameters
            if (!tok.nextToken().equals(mycode)) {
                log.log("Whoops, this isn't a " + mycode + " directive", Log.LOG_WARNING, FAC_CR);
                return null;
            }
            String onn = tok.nextToken();
            String ntn = tok.nextToken();
            String nnn = tok.nextToken();

            Taems ot = (Taems)t.clone();
            Taems nt = reader.readTTaems(ntn);

            // Find the participating nodes
            Node on = ot.findNode(new Node(onn));
            Node nn = nt.findNode(new Node(nnn));

            // Then perform the replacement, mangling the read
            // in task structure in the process
            if ((ot != null) && (on != null) && (nn != null)) {
                log.log("Performing replacement", Log.LOG_INFO, FAC_CR);
                Node vn = ot.exciseNode(on);
                String name = nn.getLabel();
                nt.removeNode(nn);
                nn.setLabel(vn.getLabel());
                ot.addNode(nn);
                ot.retargetVirtuals();
                nn.setLabel(name);
                ot.setScheduleVector(null);
            } else {
                log.log("Could not generate replace node structure", Log.LOG_ERR, FAC_CR);
                return null;
            }

            return ot;
        }
    }

    /**
     * Allows the modeler to directly specify an alternate schedule
     * location to use in case of failure.
     */
    public class PreRescheduleResolver implements ConflictListener {
        SimpleTaemsReader reader;
        String mycode = "PreReschedule";

        public PreRescheduleResolver() {
            reader = (SimpleTaemsReader)State.findComponent("SimpleTaemsReader");
        }

        public int getResolverPriority() { return 6; }

        /**
         * Called when a conflict needs to be resolved.
         * This should return a vector of ConflictRecords, which
         * describe different ways the conflict could be resolved.
         * <p>
         * Do not modify the contents of the original conflict
         * record.
         */
        public Vector conflictRequest(ConflictRecord cr) {
            Vector results = new Vector();

            if ((cr.getType() == ConflictRecord.REQUEST) &&
                (cr.getTaems() != null)) {
                Enumeration e = cr.getTaems().findNodes(new Method());
                while (e.hasMoreElements()) {
                    TaskBase t = (TaskBase)e.nextElement();
                    if (t.hasFinished() && t.getCurrentQuality() == 0) {
                        log.log("Looking for conflict stuff on " + t.getLabel(), Log.LOG_INFO, FAC_CR);
                        while (t != null) {
                            if (t.hasAttribute("NoQualityException")) {
                                String r = (String)t.getAttribute("NoQualityException");
                                if (r.startsWith(mycode)) {
                                    Taems n = generatePreRescheduleStructure(cr.getTaems(), r);
                                    if (n != null)
                                        results.add(new ConflictRecord(this, cr, n));
                                }
                                break;
                            }
                            t = t.firstSupertask();
                        }
                    }
                }
            }

            return results;
        }

        /**
         * Called when a conflict response is received
         */
        public void conflictResponse(ConflictRecord cr) { }

        /**
         * Replace node reaction
         * Reschedule StartMethod NewTaems
         */
        public Taems generatePreRescheduleStructure(Taems t, String r) {
            StringTokenizer tok = new StringTokenizer(r);
            log.log("Generating prereschedule node structure " + r, Log.LOG_INFO, FAC_CR);

            // Get the parameters
            if (!tok.nextToken().equals(mycode)) {
                log.log("Whoops, this isn't a " + mycode + " directive", Log.LOG_WARNING, FAC_CR);
                return null;
            }
            if (t.numSchedules() <= 0) {
                log.log("Taems structure " + t.getLabel() + " has no schedule to update", Log.LOG_WARNING, FAC_CR);
                return null;
            }
            t = (Taems)t.clone();
            Schedule s = t.getFirstSchedule();

            // Delete from start onwards
            String start = tok.nextToken();
            ScheduleElement el = s.getScheduleElement(start);
            if (el == null) {
                log.log("Could not find start node " + start + " in schedule", Log.LOG_WARNING, FAC_CR);
                return null;
            }
            for (int i = s.getScheduleElementIndex(el); s.size() > i; ) {
                s.removeScheduleElement(s.getScheduleElement(i));
            } 

            // Read in the new schedule
            Taems nt = reader.readTTaems(tok.nextToken());
            Schedule ns = nt.getFirstSchedule();

            // Now add the rest
            int time = ((Integer)state.getProperty(State.TIME)).intValue();
            for (Enumeration e = ns.getElements(); e.hasMoreElements();) {
                el = (ScheduleElement)e.nextElement();
                el.setStartTime(el.getStartTime().applyOffset(time));
                el.setFinishTime(el.getFinishTime().applyOffset(time));
                s.addScheduleElement(el);
            }
            t.setFirstSchedule(s);
            t.retargetSchedules();

            return t;
        }
    }

    /**
     * Submits the task structure back to the scheduler, for a complete
     * replanning.
     */
    public class RescheduleResolver implements ConflictListener {

        public RescheduleResolver() {
        }

        public int getResolverPriority() { return 4; }

        /**
         * Called when a conflict needs to be resolved.
         * This should return a vector of ConflictRecords, which
         * describe different ways the conflict could be resolved.
         * <p>
         * Do not modify the contents of the original conflict
         * record.
         */
        public Vector conflictRequest(ConflictRecord cr) {
            Vector results = new Vector();

            if ((cr.getType() == ConflictRecord.REQUEST) &&
                (cr.getTaems() != null)) {
                Enumeration e = cr.getTaems().findNodes(new Method());
                while (e.hasMoreElements()) {
                    TaskBase t = (TaskBase)e.nextElement();
                    if (t.hasFinished() && t.getCurrentQuality() == 0) {
                        Taems n = generateRescheduleStructure(cr.getTaems());
                        if (n != null)
                            results.add(new ConflictRecord(this, cr, n));
                    }
                }
            }

            return results;
        }

        /**
         * Called when a conflict response is received
         */
        public void conflictResponse(ConflictRecord cr) { }

        /**
         * Replace node reaction
         * Reschedule StartMethod NewTaems
         */
        public Taems generateRescheduleStructure(Taems t) {
            log.log("Generating reschedule node structure", Log.LOG_INFO, FAC_CR);

            Scheduler scheduler = (Scheduler)state.findComponent("Scheduler");
            t = (Taems)t.clone();
            t.setScheduleVector(null);
            Vector v = scheduler.evaluateTaems(t, true);
            if ((v != null) && (!v.isEmpty())) {
                t.setScheduleVector(v);
            } else {
                return null;
            }

            return t;
        }
    }

    /**
     * Tries to use existing alternate plans (returned by the prior planning act)
     * to resolve the conflict.
     */
    public class AltScheduleResolver implements ConflictListener {

        public AltScheduleResolver() {
        }

        public int getResolverPriority() { return 5; }

        /**
         * Called when a conflict needs to be resolved.
         * This should return a vector of ConflictRecords, which
         * describe different ways the conflict could be resolved.
         * <p>
         * Do not modify the contents of the original conflict
         * record.
         */
        public Vector conflictRequest(ConflictRecord cr) {
            Vector results = new Vector();

            if ((cr.getType() == ConflictRecord.REQUEST) &&
                (cr.getTaems() != null)) {
                Enumeration e = cr.getTaems().findNodes(new Method());
                while (e.hasMoreElements()) {
                    TaskBase t = (TaskBase)e.nextElement();
                    if (t.hasFinished() && t.getCurrentQuality() == 0) {
                        Taems n = generateAltScheduleStructure(cr.getTaems());
                        if (n != null)
                            results.add(new ConflictRecord(this, cr, n));
                    }
                }
            }

            return results;
        }

        /**
         * Called when a conflict response is received
         */
        public void conflictResponse(ConflictRecord cr) { }

        /**
         * Replace node reaction
         * Reschedule StartMethod NewTaems
         */
        public Taems generateAltScheduleStructure(Taems t) {
            log.log("Generating altschedule node structure", Log.LOG_INFO, FAC_CR);

            Schedule best = null;
            t = (Taems)t.clone();
            Enumeration e = t.getSchedules();
            nextSchedule: while (e.hasMoreElements()) {
                Schedule s = (Schedule)e.nextElement();
                if (s.size() <= 0) continue;
                Enumeration es = s.getElements();
                while (es.hasMoreElements()) {
                    // Set the start and finish so the new schedule
                    // won't try to re-execute things
                    ScheduleElement se = (ScheduleElement)es.nextElement();
                    if (se.getMethod().getStartTime() != Integer.MIN_VALUE) {
                        se.setActualStart(se.getMethod().getStartTime());
                    }
                    if (se.getMethod().getFinishTime() != Integer.MIN_VALUE) {
                        se.setActualFinish(se.getMethod().getFinishTime());
                        if (se.getMethod().getCurrentQuality() <= 0)
                            continue nextSchedule;
                    }
                }
                // Best is currently the first one that isn't broken
                best = s;
            }

            if (best != null) {
                t.setFirstSchedule(best);
                return t;
            }

            return null;
        }
    }
}
