/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * Constraint.java
 ***************************************************************************************/

package agent.mass;

import java.io.Serializable;
import java.util.*;

/**
 * Constraint class.  Use this to define environmental constraints
 * that won't necessarily be explicitly specified in the task 
 * structure which have bearing on the agent.
 */
public class Constraint implements Serializable {
  private static final int TABLE_SIZE = 5;
  /**
   * Orientation constants
   */
  public static final int AT     = 0;
  public static final int ABOVE  = 1;
  public static final int BELOW  = 2;
  public static final int AFTER  = ABOVE;
  public static final int BEFORE = BELOW;

  private String task, resource;
  private int start, duration, orientation, priority;
  private float quantity;
  private long id;
  private Hashtable related;
  
  /**
   * Default constructor, sets the (hopefully) unique id number.
   */
  public Constraint() {
    setID(State.getRandom().nextLong());

    related = new Hashtable(TABLE_SIZE);
  }
  
  /**
   * Resource constraint constructor
   * @param r The resource's name
   * @param q The quantity (level) of the resource
   * @param s The start time
   * @param d The duration time
   * @param o The orientation (at, above, below)
   * @param p The priority
   */
  public Constraint(String r, float q, int s, int d, int o, int p) {
    this();
    setResource(r);
    setQuantity(q);
    setStart(s);
    setDuration(d);
    setOrientation(o);
    setPriority(p);
  }
  
  /**
   * Task constraint constructor
   * @param t The task to perform
   * @param s The start time
   * @param o The orientation (at, after, before)
   * @param p The priority
   */
  public Constraint(String t, int s, int o, int p) {
    this();
    setTask(t);
    setStart(s);
    setOrientation(o);
    setPriority(p);
  }

  /**
   * Static parsing method.  Note that this does not parse the
   * output of the toString method, but a simple space deliminated
   * string of the parameters you would send to a constructor.
   * If we want to parse the toString output that can be added.
   * @param str The string version of the constraint
   * @return A new constraint
   */
  public static Constraint parseConstraint(String str) {
    StringTokenizer t = new StringTokenizer(str);

    if (t.countTokens() == 6) {        // resource
      return new Constraint(t.nextToken(),
			    Float.valueOf(t.nextToken()).floatValue(),
			    Integer.parseInt(t.nextToken()),
			    Integer.parseInt(t.nextToken()),
			    Integer.parseInt(t.nextToken()),
			    Integer.parseInt(t.nextToken())
			    );
    } else if (t.countTokens() == 4) { // task
      return new Constraint(t.nextToken(),
			    Integer.parseInt(t.nextToken()),
			    Integer.parseInt(t.nextToken()),
			    Integer.parseInt(t.nextToken())
			    );
    } else {
      System.err.println("Error: Constraint parse error (" + str + ")");
      return null;
    }
  }

  /**
   * Gets the constraint task name
   * @return The task to perform
   */
  public String getTask() { return task; }

  /**
   * Sets the constraint task name
   * @param t The task to perform
   */
  public void setTask(String t) { task = t; }
  
  /**
   * Gets the constraint resource type
   * @return The resource's name
   */
  public String getResource() { return resource; }
  
  /**
   * Sets the constraint resource type
   * @param r The resource's name
   */
  public void setResource(String r) { resource = r; }
  
  /**
   * Gets the constraint quantity. 
   * @return The resource's constraint quantity
   */
  public float getQuantity() { return quantity; }
  
  /**
   * Sets the constraint resource quantity
   * @param q The resource's usage quantity
   */
  public void setQuantity(float q) { quantity = q; }

  /**
   * Gets the constraint start time
   * @return The start time
   */
  public int getStart() { return start; }
  
  /**
   * Sets the constraint start time
   * @param s The start time
   */
  public void setStart(int s) { start = s; }
  
  /**
   * Gets the constraint duration time
   * @return The duration time
   */
  public int getDuration() { return duration; }
  
  /**
   * Sets the constraint duration time
   * @param d The duration time
   */
  public void setDuration(int d) { duration = d; }

  /**
   * Gets the constraint orientation
   * @return The orientation
   */
  public int getOrientation() { return orientation; }

  /**
   * Sets the constraint orientation
   * @return The orientation
   */
  public void setOrientation(int o) { orientation = o; }
  
  /**
   * Gets the constraint priority
   * @return The priority
   */
  public int getPriority() { return priority; }
  
  /**
   * Sets the constraint priority
   * @param p The priority
   */
  public void setPriority(int p) { priority = p; }

  /**
   * Returns the unique constraint id number.  This may be useful
   * during inter-agent constraint discussing or for identifying
   * local constraints.
   * @return The id number
   */
  public long getID() { return id; }

  /**
   * Sets the constraint id number.
   * @param The new id number
   */
  public void setID(long i) { id = i; }

  /**
   * Gets a reference to the related (constraint) table.  You can
   * use this to store related constraints you have to satisfy,
   * or whatever else you want to store.
   */
  public Hashtable getRelated() { return related; }

  /**
   * Returns a string representation of the constraint
   * @return A string
   */
  public String toString() {
    if (resource != null) {
      return "[Constraint-" + id + "] " +
        quantity + " " + resource + " (" + orientation +
	") from " + start + " for " + duration;
    } else {
      return "[Constraint-" + id + "] " +
	" perform " + task + " (" + orientation +
	") from " + start;
    }
  }
}
