/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * Control.java
 *
 * manifest: Name: agent/mass/Control.class
 * manifest: Depends-On: agent/simplest/Control.class
 * manifest: Java-Bean: True
 ***************************************************************************************/

package agent.mass;

import java.io.*;
import java.util.*;
import java.awt.*;
import java.applet.*;

import agent.base.*;
import agent.simplest.*;
import utilities.*;
import agent.simplest.Log;

/**
 * This class adds to the simplest Control by listening for messages
 * from the simulator
 */
public class Control extends agent.simplest.Control implements MessageEventListener {
  static final long serialVersionUID = 1234567890L; // kludge static ser_id.  Bad bad Bryan!
  protected Communicate communicate;

  /**
   * Constructor.
   */
  public Control() {
    super();

    addDependency("Communicate");

    // Override the defaults on simplest.Control parameters
    State.addParameterInfo(AUTOPULSE, null, null, Boolean.FALSE);
    State.addParameterInfo(AUTOPULSEDELAY, null, null, new Integer(0));
  }

  /**
   * Sets up component helpers.
   */
  public void init() {
    super.init();
    communicate = (Communicate)State.findComponent("Communicate");

    communicate.addMessageEventListener(this);
  }

  /**
   * Quits the agent, telling the simulator about it, too
   */
  public void quit() {
    KQMLMessage reply;

    if ((communicate != null) && communicate.ready()) {
      reply = new KQMLMessage("reply", "(ack disconnect)", "simulator");
      reply.addField("in-reply-to", "disconnect");
      communicate.sendMessage(reply);
    } else {
      log.log("Cannot send disconnect to simulator - communicate not ready", Log.LOG_WARNING, FAC_CTRL);
    }

    super.quit();
  }


  /**
   * This is called after each pulse
   */
  public void pulsePostprocess() {
      super.pulsePostprocess();
      sendPulseAcknowledgement();
  }

  /**
   * Spin waits in the function until a certain time has arrived.
   * Should halt all other events from taking place in the meantime.
   * @param npt The time to wait for
   */
    public void waitForTime(int npt) {
        log.log("Starting to wait for time " + npt, Log.LOG_INFO, FAC_CTRL);

        state.setProperty("NextPulseTime", new Integer(npt));
        sendPulseAcknowledgement();
        
        while (true) {
            int time = ((Integer)state.getProperty(State.TIME)).intValue();
            if (time >= npt)
                break;
            try {
                Thread.yield();
                Thread.currentThread().sleep(pulsedelay);
            } catch (InterruptedException e) { }
        }
    }

  /**
   * Should trigger the Control thread to perform a pulse.  This
   * should only be used if AutoExecute is false.
   */
  public void doPulse() {
      
      // Set and get the time
      Integer time = (Integer)state.getProperty(State.TIME);
      if (time == null) { time = new Integer(0); }
      time = new Integer(time.intValue() + 1);
      log.log("Setting time to " + time, Log.LOG_INFO, FAC_CTRL);
      state.setProperty(State.TIME, time);

      // Any NextPulseTime?
      Integer npt = (Integer)state.getProperty("NextPulseTime");

      if (isPaused()) {
          log.log("Agent is paused, pulse avoided", Log.LOG_INFO, FAC_CTRL);
          sendExtendExecutions(1);
          sendPulseAcknowledgement();

      } if (npt == null || npt.intValue() < time.intValue()) {
          super.doPulse();

      } else if (npt.intValue() > time.intValue()) {
          log.log("NextPulseTime is " + npt + ", pulse avoided", Log.LOG_INFO, FAC_CTRL);
          sendExtendExecutions(1);
          sendPulseAcknowledgement();

      } else if (npt.intValue() == time.intValue()) {
          log.log("NextPulseTime is " + npt + ", pulse should be continuing", Log.LOG_INFO, FAC_CTRL);
      }
  }

    /** 
     * Send an execution extend message for all currently
     * running actions.  This is typically done while the agent
     * is paused or otherwise occupied, to reflect the idea
     * that other actions should be delayed.
     * @param i The amount to extend by
     */
    public void sendExtendExecutions(int i) {
        KQMLMessage reply;

        reply = new KQMLMessage("error", "(ExecuteExtend " + i + ")", "simulator");
        communicate.sendMessage(reply);
    }

    /** 
     * Send the pulse acknowledgement back to the simulator.  It will
     * not send more than one ack pulse for a given time.
     */
    int lpa = -1;
    public void sendPulseAcknowledgement() {
        KQMLMessage reply;

        int time = ((Integer)state.getProperty(State.TIME)).intValue();
        if (time <= lpa) return;
        lpa = time;

        reply = new KQMLMessage("reply", "(ack pulse)", "simulator");
        reply.addField("in-reply-to", "pulse");
        communicate.sendMessage(reply);
    }

  /**
   * Examines messages.  Handles pulse, reset and disconnect messages.
   * @param m The message event to look at
   */
  public void messageSent(MessageEvent m) {}
    public void messageReceived(MessageEvent m) {

        // Check for other message types
        if (m.getConnection().getType() != Connection.KQML) {
            return;
        }

        String perf = ((KQMLMessage)m.getMessage()).getPerformative();
        String content = (String)m.getMessage().contentWord();
        KQMLMessage reply;

        if (perf.equalsIgnoreCase("ask")) {
            if (content.equalsIgnoreCase("pulse")) {
                // Pulse
                //if (!isPaused())
                //    pulseComponents();

            } else if (content.equalsIgnoreCase("reset")) {
                // Reset
                log.log("Reset message received", Log.LOG_WARNING, FAC_CTRL);
	
                reply = new KQMLMessage("reply", "(" + state.getProperty(State.NAME) + ")", m.getMessage().getSourceAddr());
                reply.addField("in-reply-to", "reset");
                communicate.sendMessage(reply);
                resetComponents();
       
            } else if (content.equalsIgnoreCase("disconnect")) {
                // Disconnect
                log.log("Disconnect message received", Log.LOG_WARNING, FAC_CTRL);
	
                quit();
            }
        }
    }
}
