/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * Execute.java
 *
 * manifest: Name: agent/mass/Execute.class
 * manifest: Depends-On: agent/simplest/Execute.class
 * manifest: Depends-On: agent/simplest/TaemsAction.class
 * manifest: Java-Bean: True
 ***************************************************************************************/

package agent.mass;

import java.io.*;
import java.util.*;
import java.awt.*;

import agent.base.*;
import agent.simplest.*;
import taems.*;
import utilities.*;
import agent.simplest.Log;
import agent.simplest.Action;
import agent.simplest.State;
import agent.simplest.Communicate;
import agent.simplest.Observer;

/**
 * Responsible for executing things.  It does this by sending execution
 * requests to the simulation controller and waiting for a response
 * when something happens.  This process should be transparent to you,
 * so in most other respects this is identical to the simplest.Execute
 * class.
 * <P>
 * Other features notable to this class:
 * <UL>
 * <LI> If AutoExecute is set, this class will attempt to automatically
 * execute tasks from taems.Schedule objects.  See the pulse function
 * for more details on how this works.
 * <LI> If a TaemsAction is used to given to executeAction,
 * the Taems task structure will be updated with the execution
 * characteristics when the action completes if UpdateDist is set.
 * <LI> If Display and RunDisplay are set, the picture locations
 * will be sent to the simulation controller at the correct times.
 * </UL>
 * <P>
 * This class also has minimal support for "local" actions - Actions
 * which are performed and completed by the agent, rather than simulated
 * via Mass or some other simulator.  To perform a "local" action, 
 * the Action object must first have a non-null "Local" data field (e.g.
 * using Action.setData("Local"), or having a "Local" attribute
 * in the method when using TaemsAction).  When your component recieved
 * the actionStarted event for the local action, it should do whatever
 * is necessary to perform the action.  When the task is completed, the
 * Action should be updated with its final quality and cost (Action.setQuality
 * and setCost), and its complete() should be called.  On the next pulse,
 * Execute should recognize the action has completed and process it.
 * Note that if the ExecutionMessages parameter is set to false, all
 * Actions are implicitly local, because Mass is not being used to simulate
 * their execution.
 * <P>
 * See simplest.Execute for an example of how to use this class.
 * @see agent.simplest.Execute
 * @see #pulse
 */
public class Execute extends agent.simplest.Execute implements MessageEventListener, PropertyEventListener, ConflictListener {
  static final long serialVersionUID = 1234567890L; // kludge static ser_id.  Bad bad Bryan!
  protected State state;
  protected Communicate communicate;
  protected Scheduler scheduler;
  protected ConflictResolution conflicts;
  protected CachedResolutionResolver conflictcache;
  protected SimpleTaemsReader reader;
  private String laststatus;
  private boolean autoexec;
  private boolean seqexec;
  private boolean execmessages;

  public static final String DISPLAY = "Display";
  public static final String RUNDISPLAY = "RunDisplay";
  public static final String UPDATEDIST = "UpdateDist";
  public static final String AUTOEXECUTE = "AutoExecute";
  public static final String EXECUTIONMESSAGES = "ExecutionMessages";
  public static final String SEQUENTIALEXECUTION = "SequentialExecution";

  private static Class ta_cargs[];
  static {
      try {
          ta_cargs = new Class[2];
          ta_cargs[0] = Class.forName("taems.ScheduleElement");
          ta_cargs[1] = Class.forName("taems.Schedule");
      } catch (ClassNotFoundException e) {
      }
  }

  /**
   * Default constructor
   */
  public Execute() {
    super();
    addDependency("State");
    addDependency("Communicate");

    State.addParameterInfo(DISPLAY, "String", "URL to a picture of the agent.  This is used by Mass to show the agent when idle.");
    State.addParameterInfo(RUNDISPLAY, "String", "URL to a picture of the agent running.  This is used by Mass to show the agent when it is performing a method.");
    State.addParameterInfo(UPDATEDIST, "Boolean", "If true, Execute will update CQD distributions on a method when it completed", Boolean.TRUE);
    State.addParameterInfo(AUTOEXECUTE, "Boolean", "If true, Execute will automatically execute scheduled items.  The getSchedules function is used to find the schedules", Boolean.TRUE);
    State.addParameterInfo(EXECUTIONMESSAGES, "Boolean", "If true, send execution messages to the Mass simulator", Boolean.TRUE);
    State.addParameterInfo(SEQUENTIALEXECUTION, "Boolean", "When AutoExec is true, Execute will perform only one action at a time if this parameter is true", Boolean.FALSE);
  }

  /**
   * Sets up component helpers
   */
    public void init() {
        state = (agent.simplest.State)State.findComponent("State");
        communicate = (agent.simplest.Communicate)State.findComponent("Communicate");
        scheduler = (Scheduler)State.findComponent("Scheduler");
        conflicts = (ConflictResolution)State.findComponent("ConflictResolution");
        reader = (SimpleTaemsReader)State.findComponent("SimpleTaemsReader");

        communicate.addMessageEventListener(this);
        state.addPropertyEventListener(this);

        super.init();

        // Conflict stuff
        if (conflicts != null) {
            conflicts.addResolver(conflictcache = new CachedResolutionResolver());
        }

        // Observation stuff
        if (State.hasComponent("Observe")) {
            Observe observe = (Observe)State.findComponent("Observe");
            observe.addObserver(new Observer("Deadlined Methods") {
                    public boolean checkEvent(AgentEvent e) {
                        if ((e instanceof ActionEvent) && (e.getID() == ActionEvent.COMPLETE) && (((ActionEvent)e).getAction() instanceof TaemsAction))
                            return true;
                        return false;
                    }
                    public void addEvent(AgentEvent e) {	  
                        if (checkEvent(e)) {
                            TaemsAction a = (TaemsAction)((ActionEvent)e).getAction();
                            if (a.getMethod().getDeadline() >= 0)
                                updateCount(1);
                        }
                    }
                });
            observe.addObserver(new Observer("Failed Deadlined Methods") {
                    public boolean checkEvent(AgentEvent e) {
                        if ((e instanceof ActionEvent) && (e.getID() == ActionEvent.COMPLETE) && (((ActionEvent)e).getAction() instanceof TaemsAction))
                            return true;
                        return false;
                    }
                    public void addEvent(AgentEvent e) {	  
                        if (checkEvent(e)) {
                            TaemsAction a = (TaemsAction)((ActionEvent)e).getAction();
                            int deadline = a.getMethod().getDeadline();
                            if ((deadline >= 0) && (a.getMethod().getActualLatestFinishTime() < a.getFinish())) {
                                updateCount(1);
                            }
                        }
                    }
                });
        }
    }

    /**
     * If AutoExecute is true, Execute will search for schedules that it
     * can run, using the getSchedules method. For each schedule
     * found, the getNextTasks() method is called to obtain an Enumeration
     * of theoretically runnable ScheduleElements.  Each ScheduleElement
     * is used to find the method it represent, and each method is used
     * to create a TaemsAction.  This TaemsAction is then passed to
     * executeAction.
     * <P>
     * This function also searches for "local" execution objects and
     * manages their completion.  It is debateable if this actually works
     * right, however.
     * @see #getSchedules
     */
    public void pulse() {
        KQMLMessage message;
        boolean gotone = false;
    
        // Check for local completions
        Enumeration e = actions.elements();
        while (e.hasMoreElements()) {
            Action a = (Action)e.nextElement();
            if ((a.getData("Local") != null) && a.isDone() && (a.getData("NoMessage") == null) && execmessages) {
                Object []rsrc = (Object [])a.getData("Resources");
                String rsrcstr = "";
                if (rsrc != null) { // make used string
                    rsrcstr += "Used ";
                    for (int i = 0; i < rsrc.length; i++) {
                        rsrcstr += rsrc[i].toString() + " " + rsrc[++i].toString() + " ";
                    }
                }
                message  = new KQMLMessage("ask",
                                           "(ExecuteEnd " + a.getName() + " " +
                                           a.getTime() + " " +
                                           a.getCost() + " " +
                                           a.getQuality() + " " +
                                           rsrcstr +
                                           ")",
                                           "simulator");
                message.addField("reply-with", Long.toString(a.getID()));
                message.addField("type", "control");
                communicate.sendMessage(message);

            } else if (((a.getData("Local") != null) || !execmessages) && a.isDone()) {
                log.log("Completing action " + a.getName(), Log.LOG_INFO, FAC_EXECUTE);
                completeAction(a);
            }
        }

        // Grab an action and perform it
        if (autoexec && (!seqexec || !isExecuting())) {
            int time = ((Integer)state.getProperty(State.TIME)).intValue();
            Schedule schedule;

            // Check for methods that are taking too long
            handleLateActions();

            // Find and run schedules
            e = getSchedules();
            while (e.hasMoreElements()) {
                Schedule sched = (Schedule)e.nextElement();
                gotone |= (runSchedule(sched) > 0);
                if (gotone && seqexec) break;
            }            
            if (!gotone)
                log.log("No new tasks available for Execute to perform", Log.LOG_SPAM, FAC_EXECUTE);

        } else {
            if (!autoexec)
                log.log("Autoexecution is not enabled", Log.LOG_SPAM, FAC_EXECUTE);
        }

        super.pulse();
    }

    /**
     * Looks for local TaemsActions which have taken too long (according
     * to their ScheduleElement's finish time, and attempts to delay their
     * finish time to the current time.
     */
    public void handleLateActions() {
        int time = ((Integer)state.getProperty(State.TIME)).intValue();
        
        log.log("Execute searching for actions that are taking too long", Log.LOG_DEBUG, FAC_EXECUTE);

        // First update the resources - these serve as temporary placeholders
        Enumeration e = actions();
        while (e.hasMoreElements()) {
            Action a = (Action)e.nextElement();
            if (a instanceof TaemsAction) {
                TaemsAction ta = (TaemsAction)a;
                ScheduleElement se = ta.getScheduleElement();
                if (se != null) {
                    if (se.getFinishTime().calculateMax() <= time) {
                        float offset = time - se.getFinish() + 1;

                        if (scheduler != null) {
                            log.log("Delaying schedule element resources " + se.getLabel() + " for " + offset + " clicks", Log.LOG_DEBUG, FAC_EXECUTE);
                            scheduler.updateResourceModeler(se, (int)offset);
                        }
                    }
                }
            }
        }

        // Now perform the real delays, using those placeholders in case of interdependencies
        e = actions();
        while (e.hasMoreElements()) {
            Action a = (Action)e.nextElement();
            if (a instanceof TaemsAction) {
                TaemsAction ta = (TaemsAction)a;
                ScheduleElement se = ta.getScheduleElement();
                if (se != null) {
                    if (se.getFinishTime().calculateMax() <= time) {
                        float offset = time - se.getFinish() + 1;
                        log.log("Delaying schedule element finish " + se.getLabel() + " for " + offset + " clicks", Log.LOG_DEBUG, FAC_EXECUTE);

                        if (scheduler != null) {
                            log.log("Using scheduler to delay schedule element finish from " + se.getFinishTime().calculateMax() + " to " + (time + 1), Log.LOG_DEBUG2, FAC_EXECUTE);
                            try {
                                scheduler.delayFinishTime(ta.getSchedule(), se, (int)offset);
                            } catch(ScheduleErrorException see) {
                                log.log("Warning, could not delay schedule element finish " + se.getLabel() + " to time " + (int)(offset + time) + ": " + see, Log.LOG_WARNING, FAC_EXECUTE);
                                continue;
                            }
                        } else {
                            if (ta.getSchedule() != null) {
                                log.log("Using schedule method to delay schedule element finish from " + se.getFinishTime().calculateMax() + " to " + (time + 1), Log.LOG_DEBUG2, FAC_EXECUTE);
                                ta.getSchedule().delayFinishTime(se, (int)offset);
                            } else {
                                log.log("Sorry, I can't find a way to delay schedule element finish time", Log.LOG_WARNING, FAC_EXECUTE);
                            }
                        }

                    } else {
                        if (log.getLogLevel(FAC_EXECUTE) >= Log.LOG_DEBUG2) {
                            log.log("Schedule element " + se.getLabel() + " is ok with maximum finish time of " + se.getFinishTime().calculateMax(), Log.LOG_DEBUG2, FAC_EXECUTE);
                        }
                    }
                } else {
                    log.log("Cannot check lateness on action " + a.getName() + ", no schedule element found", Log.LOG_DEBUG, FAC_EXECUTE);
                }
            }
        }
    }

    /**
     * Attempts to run elements in a schedule
     * @param schedule The schedule to inspect
     * @return The number of items that were run
     */
    public int runSchedule(Schedule schedule) {
        int time = ((Integer)state.getProperty(State.TIME)).intValue();
        Vector runnable = new Vector();

        if (schedule == null)
            return 0;

        log.log("Looking for things to run in schedule containing " + schedule.size() + " elements", Log.LOG_DEBUG, FAC_EXECUTE);

        // Find things to run
        Enumeration e = new SafeEnumeration(schedule.getNextTasks(time));
        while (e.hasMoreElements()) {
            ScheduleElement t = (ScheduleElement)e.nextElement();

            log.log("Evaluating the runability of schedule element " + t.getLabel(), Log.LOG_DEBUG, FAC_EXECUTE);

            // Fix up start time if necessary
            if (t.getStart() < time) {
                float offset = time - t.getStart();
                log.log("Delaying schedule element " + t.getLabel() + " for " + offset + " clicks", Log.LOG_DEBUG, FAC_EXECUTE);

                if (scheduler != null) {
                    log.log("Using scheduler to delay schedule element start from " + t.getStart() + " to " + time, Log.LOG_DEBUG2, FAC_EXECUTE);
                    try {
                        scheduler.delayStartTime(schedule, t, (int)offset);
                    } catch(ScheduleErrorException see) {
                        log.log("Warning, could not delay schedule element start " + t.getLabel() + " to time " + (int)(offset + time) + ": " + see, Log.LOG_WARNING, FAC_EXECUTE);
                        return 0;
                    }
                } else {
                    log.log("Using schedule method to delay schedule element start from " + t.getStart() + " to " + time, Log.LOG_DEBUG2, FAC_EXECUTE);
                    schedule.delayStartTime(t, (int)offset);
                }

                log.log("New start time for " + t.getLabel() + ": " + t.getStart(), Log.LOG_DEBUG, FAC_EXECUTE);
            }

            // Check start time
            if (t.getStart() <= time) {
                runnable.addElement(t);
                if (seqexec) {
                    log.log("We're using sequential execution, so I'm just going to stop here", Log.LOG_DEBUG, FAC_EXECUTE);
                    break;
                }
            } else {
                log.log("Whoops, it's too early to execute " + t.getLabel(), Log.LOG_DEBUG, FAC_EXECUTE);
            }
        }

        // Now check them again
        Iterator i = runnable.iterator();
        while (i.hasNext()) {
            ScheduleElement t = (ScheduleElement)i.next();
            log.log("Performing final precondition check for " + t.getLabel(), Log.LOG_DEBUG, FAC_EXECUTE);
            if (t.checkPreconditions(time)) {
                // Dandy
            } else {
                log.log("Schedule element " + t.getLabel() + " no longer passes all preconditions", Log.LOG_DEBUG, FAC_EXECUTE);
                i.remove();
            }
        }

        // And run those that are left
        i = runnable.iterator();
        while (i.hasNext()) {
            ScheduleElement t = (ScheduleElement)i.next();
            Action a;
            if (t.getMethod().hasAttribute("Action")) {
                String name = (String)t.getMethod().getAttribute("Action");
                log.log("Trying to dynamically create action " + name, Log.LOG_DEBUG, FAC_EXECUTE);
                try {
                    Class c = Class.forName(name);
                    try {
                        // Is it a TaemsAction?
                        java.lang.reflect.Constructor con = c.getDeclaredConstructor(ta_cargs);
                        log.log("Using TaemsAction two param style constructor", Log.LOG_DEBUG2, FAC_EXECUTE);
                        Object[] args = new Object[2];
                        args[0] = (ScheduleElement)t;
                        args[1] = (Schedule)schedule;
                        a = (Action)con.newInstance(args);

                    } catch (Exception ex) {
                        // If not, just try the 0 arg
                        log.log("Using zero param style constructor", Log.LOG_DEBUG2, FAC_EXECUTE);
                        a = (Action)c.newInstance();
                    }

                } catch (ClassNotFoundException ex) {
                    log.log("Error: Class " + name + " was not found\n" + ex.toString(), Log.LOG_ERR, FAC_EXECUTE);
                    continue;
                } catch (InstantiationException ex) {
                    log.log("Error: Class " + name + " could not be instantiated\n" + ex.toString(), Log.LOG_ERR, FAC_EXECUTE);
                    continue;
                } catch (IllegalAccessException ex) {
                    log.log("Error: Class " + name + " could not be instantiated\n" + ex.toString(), Log.LOG_ERR, FAC_EXECUTE);
                    continue;
                } catch (ClassCastException ex) {
                    log.log("Error: Class " + name + " does not appear to be an Action\n" + ex.toString(), Log.LOG_ERR, FAC_EXECUTE);
                    continue;
                }
            } else {
                a = new TaemsAction(t, schedule);
            }
            if (a != null)
                executeAction(a);
            else
                log.log("Error: Action for schedule element " + t.getLabel() + " was null", Log.LOG_ERR, FAC_EXECUTE);
        }

        log.log("Done looking at schedule, " + runnable.size() + " elements executed", Log.LOG_DEBUG, FAC_EXECUTE);

        return runnable.size();
    }

  /**
   * Starts to execute the given action by sending a request off to
   * the simulator (we assume that some sort of objective Taems view has
   * previously been sent to the simulator so it know what we're talking
   * about.  Note this does not wait for action completion, listen for a
   * actionCompleted or actionAborted message to determine when the 
   * action terminates.
   * @param a The action to perform
   * @see SimpleTaemsReader
   */
  public void executeAction(Action a) {
    boolean update;
    KQMLMessage reply = null;

    // TaemsAction?
    if (a instanceof TaemsAction) {
      TaemsAction ta = (TaemsAction)a;

      // Check for NL-methods
      Method method = ta.getMethod();
      method.setStatus(Node.EXECUTING);
      if (method.isNonLocal())
          a.setData("NonLocal", "true");

      if (ta.getMethod().isVirtual()) {
          log.log("Warning: Executing action " + a.getName() + " is virtual", Log.LOG_WARNING, FAC_EXECUTE);
      }
    }

    if ((a.getData("NonLocal") != null) || (a.getData("NoMessage") != null)) {
      // Do nothing, should already be taken care of

    } else if (a.getData("Local") != null) {
      // Local real methods
      reply  = new KQMLMessage("ask",
			       "(ExecuteReal " + a.getName() + ")",
			       "simulator");

    } else {
      // Regular simulated methods
      reply  = new KQMLMessage("ask",
			       "(ExecuteStart " + a.getName() + ")",
			       "simulator");
    }

    // Send message if we should
    if ((reply != null) && execmessages) {
      reply.addField("reply-with", Long.toString(a.getID()));
      communicate.sendMessage(reply);
    }

    super.executeAction(a);
    update = ! isExecuting();

    // Update display?
    if (update) {
      Object newdisp = state.getProperty(RUNDISPLAY);
      if (newdisp != null) {
          state.setProperty("TempDisplay", state.getProperty(DISPLAY));
          state.setProperty(DISPLAY, newdisp);
      }
    }
  }

  /**
   * This is called when the action has been completed
   * @param a The action
   */
    public void completeAction(Action a) {
      Method method = null;

    // TaemsAction
    if (a instanceof TaemsAction) {
      TaemsAction ta = (TaemsAction)a;

      // Check for the method
      method = ta.getMethod();
    }

    // Search if we didn't find it
    if (method == null) {
        log.log("Searching for completed method: " + a.getName(), Log.LOG_DEBUG, FAC_EXECUTE);
        Enumeration e = getTaems();
        while (e.hasMoreElements()) {
            Taems task = (Taems)e.nextElement();
            method = (Method)task.findNode(new Method(a.getName(), null));
            if (method != null) break;
        }
    }

    // Check for virtual methods
    if ((method != null) && method.isVirtual()) {
        log.log("Warning: Completed action " + method.getLabel() + " references a virtual method", Log.LOG_WARNING, FAC_EXECUTE);
    }

    // This needs to be done first
    if (a.getFinish() < 0)
        a.setFinish(((Integer)state.getProperty(State.TIME)).intValue());

    // Ok, now update it
    if (method != null) {
        
        log.log("Completing method:\n" + method.toString(), Log.LOG_SPAM, FAC_EXECUTE);

        // Check duration
        if (log.getLogLevel(FAC_EXECUTE) >= Log.LOG_DEBUG) {
            Outcome o = method.getGlobalOutcome();
            if (o != null) {
                Distribution outs = o.getDuration();
                if (outs.containsValue((float)a.getDuration(), 0.05f)) {
                    log.log("Action " + a.getName() + " duration of " + a.getDuration() + " fell in " + outs, Log.LOG_DEBUG, FAC_EXECUTE);
                } else {
                    log.log("Action " + a.getName() + " duration of " + a.getDuration() + " fell outside " + outs, Log.LOG_DEBUG, FAC_EXECUTE);
                }
            } else {
                log.log("Warning: method " + method.getLabel() + " has no global outcome", Log.LOG_WARNING, FAC_EXECUTE);
            }
        }
            
        // Update distributions if we should
        if (state.hasProperty(UPDATEDIST) && ((Boolean)state.getProperty(UPDATEDIST)).booleanValue()) {
            method.removeAllOutcomes();
            method.addOutcome(new Outcome("Final_Results",
                                          new Distribution(new Float(a.getQuality()), new Float(1.0)),
                                          new Distribution(new Float(a.getDuration()), new Float(1.0)),
                                          new Distribution(new Float(a.getCost()), new Float(1.0)),
                                          (float)1.0
                                          ));
        }
        method.setStatus(Node.COMPLETED);

    } else {
        if (a instanceof TaemsAction)
            log.log("Error: The completed method " + a.getName() + " could not be found", Log.LOG_ERR, FAC_EXECUTE);
    }

    super.completeAction(a);

    // Check for conflicts
    checkConflicts(method);

    // Update display?
    if (!isExecuting()) {
      Object newdisp = state.getProperty("TempDisplay");
      if (newdisp != null) {
          state.setProperty(DISPLAY, newdisp);
      }
    }
  }

    /**
     * This performs the conflict checking of methods, using the 
     * ConflictResolution module.  Note that currently, it only supports
     * swapping out of task structures, which may not be compatible with
     * all problem solver designs (especially if you use a common schedule
     * object, which this does not recognize).
     */
    public void checkConflicts(Method m) {
        if ((conflicts == null) || (reader == null) || (m == null)) return;
        
        if (m.hasFinished()) {
            if (m.getCurrentQuality() <= 0) {
                Taems taems = reader.findTaems(m);
                Vector v = conflicts.conflictRequest(new ConflictRecord(this, taems));
                if (!v.isEmpty()) {
                    ConflictRecord cr = (ConflictRecord)v.firstElement();
                    if (cr.getTaems() != null) {
                        log.log("Replacing current task structure with the one provided by " + cr.getSource().toString(), Log.LOG_INFO, FAC_EXECUTE);
                        conflictcache.addResolution(taems, cr);
                        reader.removeTaems(taems);
                        reader.storeTaems(SimpleTaemsReader.CONTAEMS, cr.getTaems());
                    }
                }
            }
        }
    }

    /**
     * Determines the status of an action by polling the simulation controller.
     * Currently, this fills in the following fields of the action:
     * @param a The action to check on
     * @return the action, with the STATUS data, and other elements filled in
     */
    public Action statusAction(Action a) {
        int timeout = 5; // 5 seconds

        if (!execmessages) {
            log.log("Warning: statusAction does not function without ExecutionMessages", Log.LOG_WARNING, FAC_EXECUTE);
            return a;
        }

        laststatus = null;
        KQMLMessage message = new KQMLMessage("ask", "(status " + a.getName() + ")", "simulator");
        message.addField("reply-with", Long.toString(a.getID()));
        communicate.sendMessage(message);
    
        // Wait for a reply
        timeout *= 10;
        while (laststatus == null) {
            timeout--;
            if (timeout <= 0) {
                log.log("Timeout waiting for action status (" + a.getName() + ") reply from simulator", Log.LOG_ERR, FAC_EXECUTE);
                break;
            }
            communicate.receiveMessage();
            try { Thread.currentThread().sleep(100); } catch (InterruptedException e) { }
        }
    
        // Fill in the data
        a.setData("Update", state.getProperty(State.TIME));
        if (laststatus == null) {
            a.setData("Status", "Unknown");
        } else {
            // Name status cost qual time Used rsrc1 qty1 rsrc2 qty2...
            StringTokenizer tok = new StringTokenizer(laststatus,":");
            tok.nextToken();
            a.setData("Status", tok.nextToken());
            a.setCost(Float.valueOf(tok.nextToken()).floatValue()); 
            a.setQuality(Float.valueOf(tok.nextToken()).floatValue()); 
            a.setTime(Integer.parseInt(tok.nextToken()) * -1); 
            Object rsrc[];
            if (tok.countTokens() > 0) {
                rsrc = new Object[tok.countTokens() - 1];
                tok.nextToken();  // Used
                for (int i = 0; tok.hasMoreTokens(); i++) {
                    rsrc[i] = tok.nextToken(); // rsrc name
                    rsrc[++i] = Float.valueOf(tok.nextToken()); // rsrc qty
                }
            } else {
                rsrc = new Object[0];
            }
            a.setData("Resources", rsrc);
        }

        return a;
    }

  /**
   * Extends the currently executing action
   * message to the simulator.
   * @param a The action
   * @param i The number of clicks to extend it by
   * @return true if the action was found and extended
   */
  public boolean extendAction(Action a, int i) {

      if (!execmessages) {
          log.log("Warning: extendAction does not function without ExecutionMessages", Log.LOG_WARNING, FAC_EXECUTE);
          return false;
      }

    log.log("Trying to extend action " + a + " by" + i + " clicks", Log.LOG_WARNING, FAC_EXECUTE);

    KQMLMessage message = new KQMLMessage("error", "(ExecuteExtend " + i + " " + a.getName() + ")", "simulator");
    message.addField("reply-with", Long.toString(a.getID()));
    communicate.sendMessage(message);

    return true;
  }

  /**
   * Cancels the currently executing action by sending an abort
   * message to the simulator.
   * <BR>
   * Note that cancelled actions will have zero quality, as
   * if they had failed, unless the ExecutionType attribute
   * on the Taems method is set to anytime.
   * @param a The action
   * @param reason Why the action was cancelled
   * @return true if the action was found and cancelled
   */
  public boolean cancelAction(Action a, String reason) {

      if (!execmessages) {
          log.log("Warning: cancelAction does not function without ExecutionMessages", Log.LOG_WARNING, FAC_EXECUTE);
          return super.cancelAction(a, reason);
      }

    log.log("Trying to cancel action " + a + " (" + reason + ")", Log.LOG_WARNING, FAC_EXECUTE);

    KQMLMessage message = new KQMLMessage("error", "(Abort " + a.getName() + ")", "simulator");
    message.addField("reply-with", Long.toString(a.getID()));
    communicate.sendMessage(message);

    return true;
  }

  /**
   * Handles the end of a completed cancellation.
   * @param a The action
   * @param reason Why the action was cancelled
   * @return true if the action was found and cancelled
   */
  protected boolean finishCancelAction(Action a, String reason) {
    boolean status = false;
    Method method = null;
    boolean anytime = false;

    if (a == null) return false;

    // TaemsAction
    if (a instanceof TaemsAction) {
      TaemsAction ta = (TaemsAction)a;

      // Check for the method
      method = ta.getMethod();

      if (method.hasAttribute("ExecutionType") &&
          ((String)method.getAttribute("ExecutionType")).equalsIgnoreCase("anytime")) {
          anytime = true;
      }
    }

    // Kill the quality if it isn't anytime
    if (!anytime) {
        log.log("Setting quality of aborted action " + a.getName() + " to zero.", Log.LOG_DEBUG, FAC_EXECUTE);
        a.setQuality(0);
    }

    // Search if we didn't find it
    if (method == null) {
        Enumeration e = getTaems();
        while (e.hasMoreElements()) {
            Taems task = (Taems)e.nextElement();
            method = (Method)task.findNode(new Method(a.getName(), null));
            if (method != null) break;
        }
    }
    
    // This needs to be done first
    if (a.getFinish() < 0)
        a.setFinish(((Integer)state.getProperty(State.TIME)).intValue());

    if (method != null) {
        // Update distributions if we should
        if (state.hasProperty(UPDATEDIST) && ((Boolean)state.getProperty(UPDATEDIST)).booleanValue()) {
            method.removeAllOutcomes();
            method.addOutcome(new Outcome("Final_Results",
                                          new Distribution(new Float(a.getQuality()), new Float(1.0)),
                                          new Distribution(new Float(a.getDuration()), new Float(1.0)),
                                          new Distribution(new Float(a.getCost()), new Float(1.0)),
                                          (float)1.0
                                          ));
        }

        method.setStatus(Node.COMPLETED);
    }

    status = actions.removeElement(a);
    if (status) {
        status = super.finishCancelAction(a, reason);
        if (status)
            log.log("Action " + a + " cancelled", Log.LOG_INFO, FAC_EXECUTE);
        else
            log.log("Action " + a + " not cancelled", Log.LOG_WARNING, FAC_EXECUTE);

    } else
      log.log("Error: Tried to cancel an action I don't know about", Log.LOG_ERR, FAC_EXECUTE);

    // Update display?
    if (!isExecuting()) {
      Object newdisp = state.getProperty("TempDisplay");
      if (newdisp != null) {
          state.setProperty(DISPLAY, newdisp);
      }
    }

    return status;
  }

    /**
     * Determines if the Execute component is "done", that is, all
     * known actions or schedules have completed.
     * @return true if Execute is done (as far as it can tell)
     */
    public boolean isDone() {
        if (! super.isDone()) return false;

        Enumeration e = getSchedules();
        while (e.hasMoreElements()) {
            Schedule s = (Schedule)e.nextElement();
            if (!s.isCompleted()) return false;
        }

        return true;
    }

    /**
     * This function overrides a given action's characteristics
     * on the simulator side.  It will set the start, finish,
     * cost and quality of the method if it is found in the global
     * task structure, and set the method as completed.
     * @param a  The action you want to override
     */
    public void overrideActionStatus(Action a) {

      if (!execmessages) {
          return;
      }

      KQMLMessage message;
      message  = new KQMLMessage("tell",
                                 "(ExecuteSet " + a.getName() + " " +
                                 a.getStart() + " " +
                                 a.getFinish() + " " +
                                 a.getCost() + " " +
                                 a.getQuality() + " " +
                                 ")",
                                 "simulator");
      message.addField("type", "control");
      communicate.sendMessage(message);
    }

    /**
     * These do nothing.
     */
    public Vector conflictRequest(ConflictRecord cr) { return null; }
    public void conflictResponse(ConflictRecord cr) { }
    public int getResolverPriority() { return 0; }
    
    /**
     * Watch for a messages.  Handles EndExecute messages.
     * @param me The message event to look at
     */
    public void messageReceived(MessageEvent me) {

        // Check for other message types
        if (me.getConnection().getType() != Connection.KQML) {
            return;
        }

        KQMLMessage m = (KQMLMessage)me.getMessage(); 
        String perf = m.getPerformative();
        String type = (String)((KQMLMessage)m).getField("type");

        if (type == null) { type = ""; }

        // Look for status replies
        if (type.equalsIgnoreCase("control")) {
            if (perf.equalsIgnoreCase("reply")) {
                if (findAction(m.contentWord()) != null) {
                    laststatus = m.contentData();
                }
            }
        }

        if (perf.equalsIgnoreCase("tell")) {
            if (m.contentWord().startsWith("ActivateInterrelationship")) {
                // Interrelationship activation
                // ActivateInterrelationship Name delay
                int time = ((Integer)state.getProperty(State.TIME)).intValue();
                StringTokenizer s = new StringTokenizer(m.contentData());
                if (s.countTokens() != 2) {
                    log.log("Error: Interrelationship message \"" + m.getContent() + "\" does not appear to be valid", Log.LOG_ERR, FAC_EXECUTE);
                    return;
                }
                String irName = s.nextToken();
                int delay = Integer.parseInt(s.nextToken());

                // Look for IR
                Interrelationship match = new Interrelationship(irName, null, null, null, null, null);
                Interrelationship ir = null;
                Enumeration e = getTaems();
                while (e.hasMoreElements()) {
                    Taems taems = (Taems)e.nextElement();
                    try {
                        ir = (Interrelationship)taems.findNode(match);
                    } catch (ClassCastException ex) {
                        ir = null;
                    }
                    if (ir != null) break;
                }

                // Check
                if (ir == null) {
                    log.log("Interrelationship (" + irName + ") could not be found.", Log.LOG_ERR, FAC_EXECUTE);
                } else {
                    // Finish up
                    log.log("Updating interrelationship " + ir.getLabel() + ".", Log.LOG_DEBUG, FAC_EXECUTE);
                    ir.setActive(true);
                    if (state.hasProperty(UPDATEDIST) && ((Boolean)state.getProperty(UPDATEDIST)).booleanValue()) {
                        Distribution d = new Distribution(new Float(delay), new Float(1.0));
                        ir.setDelay(d);
                    }
                }
            }
        }

        if (perf.equalsIgnoreCase("reply") || perf.equalsIgnoreCase("tell")) {
            if ((m.contentWord().startsWith("ExecuteEnd")) ||
                (m.contentWord().startsWith("ExecuteAbort"))) {
                // Method execution completed
                // ExecuteEnd Name duration cost quality Used rsrc1 qty1 rsrc2 qty2...
                // Find the action
                Action action;
                String id = m.getField("in-reply-to");
                if (perf.equalsIgnoreCase("tell")) {
                    action = findAction(id);
                } else {
                    action = findAction(Long.parseLong(id));
                }
                if (action != null)
                    log.log("Execte termination message received for " + action, Log.LOG_DEBUG2, FAC_EXECUTE);
                else {
                    log.log("Error: Could not find action matching id " + id, Log.LOG_ERR, FAC_EXECUTE);
                    return;
                }

                // Set the source
                action.setData("Source", m.getSourceAddr());

                // Get the method distributions
                int time = ((Integer)state.getProperty(State.TIME)).intValue();
                StringTokenizer s = new StringTokenizer(m.contentData());
                if (s.countTokens() < 4) {
                    log.log("Error: Execute message \"" + m.getContent() + "\" does not appear to be valid", Log.LOG_ERR, FAC_EXECUTE);
                    return;
                }
                String methodName = s.nextToken();
                int dumdur = (int)Float.parseFloat(s.nextToken());	// ignore me
                int duration = time - action.getStart();
                if (dumdur != duration)
                    log.log("Given duration (" + dumdur + ") does not match observed (" + duration + ")", Log.LOG_WARNING, FAC_EXECUTE);
                float cost = Float.valueOf(s.nextToken()).floatValue();
                float quality = Float.valueOf(s.nextToken()).floatValue();
                Object rsrc[];
                if (s.countTokens() > 0) {
                    rsrc = new Object[s.countTokens() - 1];
                    s.nextToken();  // Used
                    for (int i = 0; s.hasMoreTokens(); i++) {
                        rsrc[i] = s.nextToken(); // rsrc name
                        rsrc[++i] = Float.valueOf(s.nextToken()); // rsrc qty
                    }
                } else {
                    rsrc = new Object[0];
                }

                // Finish up with the id specified action
                if (action != null) {
                    action.setCost(cost);
                    action.setQuality(quality);
                    action.setTime(0);
                    action.setData("Resources", rsrc);
                } else {
                    log.log("Error: action id " + id + " not known to us.", Log.LOG_ERR, FAC_EXECUTE);
                }

                // That's it
                if (m.contentWord().startsWith("ExecuteAbort")) {
                    finishCancelAction(action, "Abort");
                } else {
                    completeAction(action);
                }
            }

        } else if (perf.equalsIgnoreCase("error")) {
            // Airoar
            if (m.contentWord().startsWith("ExecuteError")) {
                StringTokenizer s = new StringTokenizer(m.contentData());
                String name = s.nextToken();
                log.log("Error: " + m.contentData(), Log.LOG_ERR, FAC_EXECUTE);
                long id = Long.parseLong(m.getField("in-reply-to"));
                Action a = findAction(id);
                a.setQuality(0);
                a.setCost(0);
                finishCancelAction(a, "Error");
            }
        }
    }
    public void messageSent(MessageEvent m) { }

    /**
     * Gets the local Taems task structure.  Returns an enumeration
     * of them.  If conditional views are found, those are returned,
     * otherwise subjective ones are returned.
     * @return An enumeration of Taems structures.
     */
    protected Enumeration getTaems() {

        if (state.hasProperty(SimpleTaemsReader.CONTAEMS)) {
            Object o  = state.getProperty(SimpleTaemsReader.CONTAEMS);
            if (o instanceof Taems) {
                Vector v = new Vector(1);
                v.addElement(o);
                return v.elements();
            } else if (o instanceof Hashtable) {
                return new SafeEnumeration(((Hashtable)o).elements());
            }

        } else if (state.hasProperty(SimpleTaemsReader.SUBTAEMS)) {
            Object o  = state.getProperty(SimpleTaemsReader.SUBTAEMS);
            if (o instanceof Taems) {
                Vector v = new Vector(1);
                v.addElement(o);
                return v.elements();
            } else if (o instanceof Hashtable) {
                return new SafeEnumeration(((Hashtable)o).elements());
            }
        }

        return new Vector(0).elements();
    }

    /**
     * First, one or more Schedule objects are searched for, in this order:
     * <OL>
     * <LI>The Scheduler.SCHEDULE property from state
     * <LI>getFirstSchedule from each conditioned task structure(s)
     * <LI>getFirstSchedule from each subjective task structure(s)
     * </OL>
     * Only schedules from one step are used, so if a Scheduler.SCHEDULE entry is
     * found in State, only that object will be returned.
     * @return An enumeration of schedules
     */
    protected Enumeration getSchedules() {
        Vector v = new Vector();

        // Look in state first for a schedule
        if (state.hasProperty(Scheduler.SCHEDULE) && (state.getProperty(Scheduler.SCHEDULE) instanceof Schedule)) {
            log.log("Looking in master schedule", Log.LOG_DEBUG2, FAC_EXECUTE);
            v.addElement(state.getProperty(Scheduler.SCHEDULE));
            
        // ...if that isn't found, then look in the Taems structures
        } else {
            Enumeration en = getTaems();
            while (en.hasMoreElements()) {
                Taems task = (Taems)en.nextElement();
                log.log("Looking for schedule in structure " + task.getLabel(), Log.LOG_DEBUG2, FAC_EXECUTE);
                Schedule sched = task.getFirstSchedule();
                if (sched != null) {
                    v.addElement(sched);
                }
            }
        }

        return v.elements();
    }

  /**
   * Watch for interesting property changes.  Monitors Task changes.
   * @param e The event to look at
   */
  public void propertyChanged(PropertyEvent e) { 
    String key = e.getKey().toString();

    if (key.equalsIgnoreCase("SubTaemsTask")) {
      if (isExecuting()) {
          log.log("A new task structure has arrived, but I'm not cancelling your actions", Log.LOG_DEBUG, FAC_EXECUTE);
          //cancelAllActions("NewTask");
      }

    } else if (key.equalsIgnoreCase(AUTOEXECUTE)) {
        autoexec = ((Boolean)e.getProperty()).booleanValue();

    } else if (key.equalsIgnoreCase(SEQUENTIALEXECUTION)) {
        seqexec = ((Boolean)e.getProperty()).booleanValue();

    } else if (key.equalsIgnoreCase(EXECUTIONMESSAGES)) {
        execmessages = ((Boolean)e.getProperty()).booleanValue();
    }
  }
  public void propertyRemoved(PropertyEvent e) { propertyChanged(e); }
  public void propertyAdded(PropertyEvent e) { propertyChanged(e); }

  /**
   */
    public class CachedResolutionResolver implements ConflictListener {
        public Hashtable cache = new Hashtable();

        public CachedResolutionResolver() {
        }

        public int getResolverPriority() { return 9; }

        /**
         * Called when a conflict needs to be resolved.
         * This should return a vector of ConflictRecords, which
         * describe different ways the conflict could be resolved.
         * <p>
         * Do not modify the contents of the original conflict
         * record.
         */
        public Vector conflictRequest(ConflictRecord cr) {
            Vector results = new Vector();

            if ((cr.getType() == ConflictRecord.REQUEST) &&
                (cr.getTaems() != null)) {
                String key = generateKey(cr.getTaems());
                if (cache.containsKey(key)) {
                    results.add(((Taems)cache.get(key)).clone());
                }
            }

            return results;
        }

        /**
         * Called when a conflict response is received
         */
        public void conflictResponse(ConflictRecord cr) { }
        
        public void addResolution(Taems t, ConflictRecord cr) {
            String key = generateKey(t);
            cache.put(key, (Taems)cr.getTaems().clone());
            log.log("Added conflict resolution cache for " + key, Log.LOG_DEBUG, FAC_EXECUTE);
        }

        protected String generateKey(Taems taems) {
            StringBuffer buf = new StringBuffer();

            buf.append(scheduler.generateCacheKey(taems).toString());
            buf.append(generateScheduleKey(taems));

            return buf.toString();
        }

        protected String generateScheduleKey(Taems taems) {
            StringBuffer buf = new StringBuffer();

            Enumeration e = taems.getFirstSchedule().getElements();
            while (e.hasMoreElements()) {
                ScheduleElement elem = (ScheduleElement)e.nextElement();
                buf.append(elem.getLabel());
                if (elem.getMethod().hasFinished()) {
                    buf.append(String.valueOf(elem.getMethod().getCurrentQuality()));
                }
            }

            return buf.toString();
        }
    }
}
