/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * PreprocessTaemsReader.java
 *
 * manifest: Name: agent/mass/PreprocessTaemsReader.class
 * manifest: Java-Bean: True
 ***************************************************************************************/

package agent.mass;

import java.io.*;
import java.util.*;
import java.awt.*;

import agent.base.*;
import agent.simplest.*;
import taems.parser.*;
import taems.preprocessor.*;
import taems.*;

import utilities.*;
import agent.simplest.Log;

/**
 * This component takes in a marked up Taems file, as specified by the "PreTaems"
 * property.  In your agent configuration you have to add this line:<BR>
 * <CODE>PreTaems: [filename].ptaems </CODE> <P>
 * This .ptaems file is essentially textual Taems (as is used by
 * SimpleTaemsReader), with additional tags that allow you to perform some
 * on-the-fly generation in response to current conditions.  So, from this
 * one file, you can:
 * <UL>
 * <LI>Generate the objective, subjective and conditioned views.
 * <LI>Use the flexible assertion style to probabilistically construct task
 * elements.
 * <LI>Generate structures with recursive techniques
 * <LI>Fetch and parse arbitrary Taems elements (including schedules,
 * commitments, etc.)
 * <LI>Use given parameters to derive contextual structures.
 * </UL>
 * <P>
 * To have a better idea how it works, please check the PreProcessor
 * online documentation at <A HREF="http://mas.cs.umass.edu/research/taems/white/"> http://mas.cs.umass.edu/research/taems/white (section 7)</A>
 * <P>
 */
public class PreprocessTaemsReader extends agent.mass.SimpleTaemsReader {
    static final long serialVersionUID = 1234567890L; // kludge static ser_id.  Bad bad Bryan!
    protected String defaultPreprocessor = null;
    protected Hashtable defaultHT = new Hashtable();
    protected Hashtable preprocessorList = new Hashtable();

    public static final String PRETAEMS = "PreTaems";

    /**
     * Default constructor
     */
    public PreprocessTaemsReader() {
        super();

        State.addParameterInfo(PRETAEMS, "String", "Preprocessor-style Taems definition file");
    }

    /**
     * Sets up component helpers
     */
    public void init() {
        super.init();
	String name = (String)state.getProperty(PRETAEMS);
	if ((name != null) && (getPreprocessor(name) != null))
	    defaultPreprocessor = name;
    }
    

    /**
     * Read the file and generate the PreProcessor object
     */
    public PreProcessor getPreprocessor(String name) {
        if (preprocessorList.contains(name)){
            return((PreProcessor)preprocessorList.get(name));
        }
        BufferedReader file = state.getExtendedEntryReader(name);
        PreProcessor preprocessor = null;
        if (file != null) {
            try {
                PreProcessorParser parser = new PreProcessorParser(file);
                preprocessor = parser.Input();
                preprocessorList.put(name, preprocessor);
            } catch (Exception e) {      
                log.log("Error finding preprocessor input file "+ name,Log.LOG_ERR, FAC_ORGDES);
                e.printStackTrace();
            } 
        } else {
            log.log("Error accessing PreTaems file: " +
                    name, Log.LOG_ERR, FAC_ORGDES);
            return(null);
        }
        return(preprocessor);
    }

    /**
     * Read the preprocessing file
     */
    public void begin() { 
        super.begin();
    }
	    
    /**
     * Reset
     */
    public void reset() {

        defaultHT = new Hashtable();
        if (state.hasProperty(State.RANDOMSEED))
            defaultHT.put(State.RANDOMSEED, (Long)state.getProperty(State.RANDOMSEED));
	for (Enumeration e = preprocessorList.keys() ; e.hasMoreElements();){
	    PreProcessor pre = (PreProcessor)preprocessorList.get(e.nextElement());
	    pre.reset();
	}
        super.reset();
    }



    /**
     * You can see the default Hashtable, but you really really should know what
     * you are doing. If if you want to combine the default hashtable plus your own,
     * just call makeTaemsStructure function.
     * @param ht - Hashtable
     * @see #makeTaemsStructure
     */
    public void setDefaultHashtable(Hashtable ht) {
        defaultHT = ht;
    }

    /**
     * Returns a list of all the current hashtable keys
     * @return An enumeration
     */
    public Enumeration getHashtableKeys() {
        return defaultHT.keys();
    }

    /**
     * Returns a the current hashtable value corresponding to the 
     * given key.
     * @param k The key to check
     * @return The value, or null if not found
     */
    public Object getHashtableValue(Object k) {
        return defaultHT.get(k);
    }

    /**
     * Adds a key/value pair to the default hashtable.  Note that the
     * same hashtable is used for all three views (but during generation
     * the View key is update with the view's name).
     * @param k The key to add
     * @param v The key's value
     */
    public void setHashtableKey(Object k, Object v) {
        if (k == null) {
            log.log("Warning: null key hashtable add attempted", Log.LOG_WARNING, FAC_ORGDES);
            return;
        }

        if (v != null) {
            log.log("Adding hashtable key [" + k + ":" + v + "]", Log.LOG_DEBUG, FAC_ORGDES);
            defaultHT.put(k, v);
        } else {
            removeHashtableKey(k);
        }
    }

    /**
     * Removes a key/value pair from the default hashtable.  Note that the
     * same hashtable is used for all three views (but during generation
     * the View key is update with the view's name).
     * @param k The key to remove
     */
    public void removeHashtableKey(Object k) {
        if (k == null) {
            log.log("Warning: null key hashtable remove attempted", Log.LOG_WARNING, FAC_ORGDES);
            return;
        }

        log.log("Removing hashtable key [" + k + "]", Log.LOG_DEBUG, FAC_ORGDES);
        defaultHT.remove(k);
    }

    /**
     * Generate TTAEMS based on the hashtable definition. This hashtable merge with 
     * the default one.
     * @param ht - Hashtable
     * @return task - Taems structure, or null if an error occurs.
     */
    public Taems makeTaemsStructure(Hashtable ht) {
	return(makeTaemsStructure(ht,defaultPreprocessor));
    }

    /**
     * Generate TTAEMS based on the hashtable definition using the 
     * preprocessor file (given as an argument)
     * @param ht - Hashtable
     * @param file - String preprocessor file
     */
    public Taems makeTaemsStructure(Hashtable ht, String file) {

        PreProcessor preprocessor = getPreprocessor(file);
        if (preprocessor == null) {
            log.log("Warning: No preprocessor found", Log.LOG_ERR, FAC_ORGDES);
            return null;
        }

        setHashtableKey(State.TIME, (Integer)state.getProperty(State.TIME));
        for (Enumeration e = getHashtableKeys(); e.hasMoreElements(); ) {
            String key = (String)e.nextElement();
            ht.put(key, getHashtableValue(key));
        }
	
        log.log("Generating task structure", Log.LOG_DEBUG, FAC_ORGDES);

        if (preprocessor != null) {
            preprocessor.reset();
            try {
                Taems task = readTTaems(new java.io.StringReader(preprocessor.toTTaems(ht)));
                return task;
            } catch (Throwable e) {
                log.log("Error reading taems structure: " + e, Log.LOG_ERR, FAC_ORGDES);
                e.printStackTrace();
                System.err.println(preprocessor.toTTaems(ht));
            }
            return null;

        } else {
            log.log("Error, no PreProcessor file read.", 
                    Log.LOG_ERR, FAC_ORGDES);
            return null;
        }
    }

    /**
     * Generates the objective, subjective and conditional
     * structures.  These are just the default ones, and 
     * each is also sent or set to its designtated location.
     * This will typically only be called when the agent
     * is created or reset.
     * <P>
     * @param Hashtable ht - Hashtable function passed 
     * Also increments the un-labeled structure id.
     * @see #makeConditionalStructure
     * @see #makeSubjectiveStructure
     * @see #makeObjectiveStructure
     */
    public void makeAllStructures(Hashtable ht) {
	NUM++;
	
	makeObjectiveStructure(ht,true);
	makeSubjectiveStructure(ht,true);
	makeConditionalStructure(ht,true);
    }


    /**
     * Creates the conditioned version of the task structure.
     * @param set Set this to true if you want the conditional
     *    CONTAEMS object to be updated with this structure.
     * @param Hashtable pass by the agent controller.
     * @return The structure, or null if none generated.
     * @see #makeSubjectiveStructure
     */
    public Taems makeConditionalStructure(Hashtable ht, boolean set) {
	return(makeConditionalStructure(defaultPreprocessor,ht,set));
    }

    /**
     * Creates the conditioned version of the task structure.
     * @param set Set this to true if you want the conditional
     *    CONTAEMS object to be updated with this structure.
     * @param String - file name 
     * @param Hashtable pass by the agent controller.
     * @return The structure, or null if none generated.
     * @see #makeSubjectiveStructure
     */
    public  Taems makeConditionalStructure(String file, Hashtable ht, boolean set) {
	if (defaultPreprocessor != null) {
	    setHashtableKey("View", "Conditional");
	    log.log("Generating conditional task structure", Log.LOG_INFO, FAC_ORGDES);
	    Taems task = makeTaemsStructure(ht,file);
	    removeHashtableKey("View");
	    if (task != null) {
		if ((task.getLabel() == null) || task.getLabel().equals(""))
		    task.setLabel("Structure_" + NUM);
		
		// Set it?
		if (set) {
		    task = storeTaems(CONTAEMS, task);
		}
		
		log.log("Generated conditioned task structure", Log.LOG_SPAM, FAC_ORGDES);
		return task;
	    
	    } else {
		log.log("Error, taems generation failed.", Log.LOG_ERR, FAC_ORGDES);
	    }
	
	    log.log("Conditioned generation failed", Log.LOG_WARNING, FAC_ORGDES);
	    
	} else {
	    return(super.makeConditionalStructure(true));
	}
	
        return null;
    }

    public Taems makeConditionalStructure(boolean set) {
	return(makeConditionalStructure(defaultPreprocessor, new Hashtable(), set));
    }

    /**
     * Generates the subjective Taems task structure
     * @param set Set this to true if you want the subjective
     *    SubTaemsTask object to be updated with this structure.
     * @param Hashtable pass by the agent controller.
     * @return The structure, or null if none generated.
     * @see #makeObjectiveStructure
     */
    public Taems makeSubjectiveStructure(Hashtable ht, boolean set) {
	return(makeSubjectiveStructure(defaultPreprocessor,ht,set));
    }
    /**
     * Generates the subjective Taems task structure
     * @param set Set this to true if you want the subjective
     *    SubTaemsTask object to be updated with this structure.
     * @param String - file name 
     * @param Hashtable pass by the agent controller.
     * @return The structure, or null if none generated.
     * @see #makeObjectiveStructure
     */

    public Taems makeSubjectiveStructure(String file,Hashtable ht, boolean set) {
        if (defaultPreprocessor != null) {
            setHashtableKey("View", "Subjective");
            log.log("Generating subjective task structure", Log.LOG_INFO, FAC_ORGDES);
            Taems task = makeTaemsStructure(ht,file);
            removeHashtableKey("View");
            if (task != null) {
                if ((task.getLabel() == null) || task.getLabel().equals(""))
                    task.setLabel("Structure_" + NUM);

                // Set it?
                if (set) {
                    task = storeTaems(SUBTAEMS, task);
                }
	    
                log.log("Generated subjective task structure", Log.LOG_SPAM, FAC_ORGDES);
                return task;
		
            } else {
                log.log("Error, taems generation failed.", Log.LOG_ERR, FAC_ORGDES);
            }
	    
            log.log("Subjective generation failed", Log.LOG_WARNING, FAC_ORGDES);

        } else {
            return(super.makeSubjectiveStructure(true));
        }

        return null;
    }

    public Taems makeSubjectiveStructure(boolean set) {
	return(makeSubjectiveStructure(defaultPreprocessor,new Hashtable(), set));
    }

    /**
     * Generates the objective Taems task structure.
     * @param send Set this to true if you want the structure
     *   sent to the simulator
     * @param String - file name 
     * @param Hashtable pass by the agent controller.
     * @return The structure.  Note that you shouldn't really be
     *   looking at the objective view on the agent side.
     */
    public Taems makeObjectiveStructure(String file,Hashtable ht, boolean send) {
        if (defaultPreprocessor != null) {
            setHashtableKey("View", "Objective");
            log.log("Generating objective task structure", Log.LOG_INFO, FAC_ORGDES);
            Taems task = makeTaemsStructure(ht,file);
            removeHashtableKey("View");

            if (task != null) {
                if ((task.getLabel() == null) || task.getLabel().equals(""))
                    task.setLabel("Structure_" + NUM);
 
                // Handle it
                task = handleNewObjectiveStructure(task, send);

                log.log("Generated objective task structure", Log.LOG_SPAM, FAC_ORGDES);
                return task;
	    
            } else {
                log.log("Error, taems generation failed.", Log.LOG_ERR, FAC_ORGDES);
            }

            log.log("Objective generation failed", Log.LOG_WARNING, FAC_ORGDES);

        } else {
            return(super.makeObjectiveStructure(true));
        }

        return null;
    }

    public Taems makeObjectiveStructure(Hashtable ht, boolean send) {
	return(makeObjectiveStructure(defaultPreprocessor, ht, send));
    }

    public Taems makeObjectiveStructure(boolean send) {
	return(makeObjectiveStructure(defaultPreprocessor,new Hashtable(), send));
    }

  /**
   * Watch for messages
   * @param m The message event to look at
   */
  public void messageReceived(MessageEvent e) {
    super.messageReceived(e);

    // Check for other message types
    if (e.getConnection().getType() != Connection.KQML) {
      return;
    }

    KQMLMessage m = (KQMLMessage)e.getMessage(); 
    String perf = m.getPerformative();

    if (perf.equalsIgnoreCase("tell")) {
        String word = m.contentWord();
        if (word.equalsIgnoreCase("PreprocessKey")) {
            String key = m.contentData();
            String value = key.substring(key.indexOf(" ") + 1);
            key = key.substring(0, key.indexOf(" "));

            if ((value == null) || value.equalsIgnoreCase("null"))
                removeHashtableKey(key);
            else
                setHashtableKey(key, value);
        }
    }
  }

  /**
   * Watch for interesting property changes.  Monitors Task changes.
   * @param e The event to look at
   */
  public void propertyChanged(PropertyEvent e) { 
    String key = e.getKey().toString();

    if (key.equalsIgnoreCase(State.RANDOMSEED)) {
      setHashtableKey(State.RANDOMSEED, (Long)e.getProperty());

    } else if (key.equalsIgnoreCase(State.NAME)) {
      setHashtableKey(State.NAME, e.getProperty());
    }

    super.propertyChanged(e);
  }
}


