/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/****************************************************************
 * Scheduler.java
 *
 * manifest: Name: agent/mass/Scheduler.class
 * manifest: Java-Bean: True
 ****************************************************************/

package agent.mass;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;

import agent.base.*;
import agent.simplest.*;
import agent.simplest.Observer;
import taems.*;
import taems.parser.*;

/**
 * The format of the criteria vector is as follows:<P>
 * (quality_threshold)<BR> (value)
 * (cost_limit)<BR>
 * (duration_limit)<BR>
 *
 * (quality_certainty_threshold)<BR> (value between 0 - 1)
 * (cost_certainty_threshold)<BR>
 * (duration_certainty_threshold)<BR>
 *
 * (goodness_quality_slider)<BR> Sum = [0-1]
 * (goodness_cost_slider)<BR>
 * (goodness_duration_slider)<BR>
 *
 * (threshold_quality_slider)<BR> Sum = [0-1]
 * (threshold_cost_slider)<BR>
 * (threshold_duration_slider)<BR>
 *
 * (uncertainty_quality_slider)<BR>
 * (uncertainty_cost_slider)<BR>
 * (uncertainty_duration_slider)<BR>
 *
 * (threshold_certainty_quality_slider)<BR>
 * (threshold_certainty_cost_slider)<BR>
 * (threshold_certainty_duration_slider)<BR>
 *
 * (meta_goodness_slider)<BR>
 * (meta_threshold_slider)<BR>
 * (meta_uncertainty_slider)<BR>
 * (meta_uncertainty_threshold_slider)<BR>
 * <B> Version 1.0 </B>
 */

public class Scheduler extends AgentComponent implements PropertyEventListener, MouseListener {
  static final long serialVersionUID = 1234567890L; // kludge static ser_id.  Bad bad Bryan!
  protected ListenerVector listeners = new ListenerVector();
  public static final int CRITERIA_SIZE = 22;
  public static final int LABEL_SIZE=6;
  protected Criteria criteria = new Criteria(CRITERIA_SIZE);
  protected String[] defaultLabels = {"Q", "C", "D", "Q", "C", "D", "Q", "C", "D","Q", "C", "D","R", "T", "C", "C" };
  protected int[] defaultSliders = {60, 40, 0, 45, 45, 5, 0, 0, 100,0,0,0,100,0,0,0 };
  protected int[] defaultValues = {50, 100, 200, 0, 0, 0};
  protected Frame controls = null;
  protected ScrollbarPanel[] sliders;
  protected TextField[] values;
  protected ReadTTaems parser=null;
  protected boolean override = false;
  protected int FAC_SCHED;
  protected String SCHEDULER,SCHED_DIR;
  protected agent.simplest.State state;
  protected agent.simplest.Log log;
  protected boolean autoschedule, cacheschedules, autosetstarttime;
  protected int cachelimit = -1, numschedules = 15, maxschedules = -1;
  protected Hashtable cache = new Hashtable();
  protected Observer obsched, obcache, obschedtime;
  protected Reader lastreader;
  protected DTCNative dtc_native;

  public static final String AUTOSCHEDULE = "AutoSchedule";
  public static final String SCHEDULERCRITERIA = "SchedulerCriteria";
  public static final String DTC = "DTC";
  public static final String DTCOVERRIDE = "DTCOverride";
  public static final String NUMSCHEDULES = "NumSchedules";
  public static final String CACHESCHEDULES = "CacheSchedules";
  public static final String CACHELIMIT = "CacheLimit";
  public static final String AUTOSETSTARTTIME = "AutoSetStartTime";
  public static final String MAXREADSCHEDULES = "MaxReadSchedules";
  public static final String USEDTCNATIVE = "UseDTCNative";
  public static final String DTCLIBNAME = "DTCLibName";

  public static final String SCHEDULE = "Schedule";

  /**
   * Constructor. Nothing fancy here.
   */
  public Scheduler() {
    super();

    addDependency("State");
    addDependency("Log");

    State.addParameterInfo(AUTOSCHEDULE, "Boolean", "If true, automacticly reschedule if the TAEMS structure changes", Boolean.TRUE);
    State.addParameterInfo(SCHEDULERCRITERIA, "Vector", "The default scheduler evaluation criteria");
    State.addParameterInfo(DTC, "String", "Pathname of the DTC scheduler.", new String("modules/scheduler/dtc"));
    State.addParameterInfo(DTCOVERRIDE, "String", "DTC execution string, used in place of existing default.  The taems file and output directory will be appended to this to create the full execution command.  This overrides DTC and NumSchedules.  It does not currently work with UseDTCNative.");
    State.addParameterInfo(NUMSCHEDULES, "Integer", "Maximum number of schedules for DTC to return", new Integer(2));
    State.addParameterInfo(CACHESCHEDULES, "Boolean", "If true, tries to used cached results in place of running DTC", Boolean.TRUE);
    State.addParameterInfo(CACHELIMIT, "Integer", "If set and positive, sets a maximum limit on the number of cached schedules", new Integer(15));
    State.addParameterInfo(AUTOSETSTARTTIME, "Boolean", "If true, set the starting time of the schedule to now instead of 0 (DTC's default)", Boolean.TRUE);
    State.addParameterInfo(MAXREADSCHEDULES, "Integer", "Maximum number of schedules to read at a time (0 for all)", new Integer(0));
    State.addParameterInfo(USEDTCNATIVE, "Boolean", "Use native DTC code if available", Boolean.TRUE);
    State.addParameterInfo(DTCLIBNAME, "String", "Path to the DTC native interface library", "bin/libdtc-java.so");

    addMouseListener(this);

    sliders = new ScrollbarPanel[CRITERIA_SIZE - LABEL_SIZE];
    values = new TextField[LABEL_SIZE];
    
    for(int i = 0 ; i < LABEL_SIZE  ; i++) 
      criteria.insertElementAt(new Float(defaultValues[i]), i);
    for(int i = LABEL_SIZE ; i < CRITERIA_SIZE  ; i++) 
      criteria.insertElementAt(new Float(defaultSliders[i - LABEL_SIZE]), i);
  }

  /**
   * Sets up component helpers. Set the criteria using the default.
   */
  public void init() {
    super.init();
    override = false;

    log = (agent.simplest.Log)State.findComponent("Log");
    state = (agent.simplest.State)State.findComponent("State");

    FAC_SCHED = log.getFacilityID(this);
    if (state.hasProperty(State.MASSHOME))
        SCHEDULER = (String)state.getProperty(State.MASSHOME) + "/" + (String)state.getProperty(DTC);
    else 
        SCHEDULER = (String)state.getProperty(DTC);
    log.log("Scheduler is " + SCHEDULER, Log.LOG_INFO, FAC_SCHED);

    if (SCHEDULER != null) {
        SCHED_DIR = SCHEDULER.substring(0, SCHEDULER.lastIndexOf("/")) + "/";
        log.log("Scheduler output directory is " + SCHED_DIR, Log.LOG_INFO, FAC_SCHED);
    } else 
        log.log("Warning: Could not find path to scheduler", Log.LOG_WARNING, FAC_SCHED);

    state.addPropertyEventListener(this);

    if (hasGUI()) {
        controls = new Frame("Scheduler Evaluation Criteria");
        controls.setLayout(new BorderLayout());
        openCriteriaFrame(criteria);
        
        controls.pack();
    }

    dtc_native = new DTCNative();
    if (DTCNative.isNativeAvailable()) {
        log.log("DTC Native interface is in use.", Log.LOG_INFO, FAC_SCHED);
    } else {
        log.log("DTC Native interface is disabled.", Log.LOG_INFO, FAC_SCHED);
    }

    if (State.hasComponent("Observe")) {
        Observe observe = (Observe)State.findComponent("Observe");
        observe.addObserver(obsched = new Observer("Scheduler Calls") {
                public boolean usesEvents() { return false; }
            });
        observe.addObserver(obschedtime = new Observer("Scheduler Run Time") {
                public boolean usesEvents() { return false; }
            });
        observe.addObserver(obcache = new Observer("Schedule Cache Hits") {
                public boolean usesEvents() { return false; }
            });
    }    
  }

  /**
   * Reset. Guess what is resets the scheduler.
   */
  public void reset() {
    override = false;
    criteria.removeAllElements();
    for(int i = 0 ; i < LABEL_SIZE  ; i++) 
      criteria.insertElementAt(new Float(defaultValues[i]), i);
    for(int i = LABEL_SIZE ; i < CRITERIA_SIZE  ; i++) 
      criteria.insertElementAt(new Float(defaultSliders[i - LABEL_SIZE]), i);
  }
  
  /**
   * Finalizer.  Flushes and closes the file stream.
   */
  public void end() {
    if (controls != null)
      controls.dispose();
  }

  /**
   * Adds a listener to the list
   */
  public void addScheduleEventListener(ScheduleEventListener l) {
    //listeners.add(l, l.getScheduleEventPriority());
    listeners.add(l);
  }

  /**
   * Removes a listener
   */
  public void removeScheduleEventListener(ScheduleEventListener l) {
    listeners.remove(l);
  }

  /**
   * Sends an action event to each of the registered listeners
   * @param e The event to send
   */
  protected void fireEvent(AgentEvent event) {
    ScheduleEvent e = (ScheduleEvent)event;
    ScheduleEventListener l;
    Enumeration enumr = listeners.elements();

    log.log("Firing schedule event", Log.LOG_DEBUG2, FAC_SCHED);

    while (enumr.hasMoreElements()) {
      l = (ScheduleEventListener)enumr.nextElement();

      if (!e.isActive()) break;

      switch (e.getID()) {
      case ScheduleEvent.ADDED:
          l.scheduleAdded(e);
          break;
      case ScheduleEvent.EXCEPTION:
          l.scheduleException(e);
          break;
      default:
          break;
      }
    }
    
    if ((log != null) && (log.getLogLevel(FAC_SCHED) >= Log.LOG_INFO))
        log.log(e, Log.LOG_INFO, FAC_SCHED);

    log.log("Fired a schedule event", Log.LOG_SPAM, FAC_SCHED);
  }
  
  /**
   * Constructs the criteria control frame, using the supplied criteria.
   * Be careful it doesn't display it !!!!
   * @param The vector with the criteria
   */
    public void openCriteriaFrame(Vector vect) {
        /* Set data here */
        //    controls.setSize(400,400);
        controls.removeAll();
        GridBagLayout gridbag = new GridBagLayout();
        GridBagConstraints c = new GridBagConstraints();
        controls.setLayout( gridbag );
        c.fill = GridBagConstraints.BOTH;
        c.weightx = 1.0;  
        c.weighty = 1.0;
        c.gridwidth = 1;
        c.insets = new Insets(3,5,3,5);

        Panel p = new Panel();
        GridBagLayout gridbag4 = new GridBagLayout(); 
        GridBagConstraints c4 = new GridBagConstraints();
        Label l = new Label();
        boolean begin = false, end = false;

        for(int i = LABEL_SIZE ; i < CRITERIA_SIZE  ; i++) {
            if (i == LABEL_SIZE) {
                l = new Label("Raw-Goodness");
                begin = true;
            } else if (i == LABEL_SIZE+3) {
                l = new Label("Threshold/Limits");
                begin = true;
            } else if (i == LABEL_SIZE+6) {
                l = new Label("Raw-Certainty");
                begin = true;
            } else if (i == LABEL_SIZE+9) {
                l = new Label("Certainty Threshold");
                begin = true;
            } else if (i == LABEL_SIZE+12) {
                l = new Label("Meta");
                begin = true;
            }
      
            if (begin) {
                p = new Panel();
                gridbag4 = new GridBagLayout();
                c4 = new GridBagConstraints();
                p.setLayout( gridbag4);
                c4.weightx = 1.0;  
                c4.fill = GridBagConstraints.NORTH;
                c4.gridwidth = GridBagConstraints.REMAINDER;
                gridbag4.setConstraints(l, c4);
                p.add(l);
                begin = false;
            }

            c4.gridwidth = 1; // Reset to default
            c4.weightx = 1.0;  
            c4.weighty = 1.0;
            c4.fill = GridBagConstraints.BOTH;
            sliders[i - LABEL_SIZE] = new ScrollbarPanel(((Float)vect.elementAt(i)).floatValue(), defaultLabels[i - LABEL_SIZE]);
            gridbag4.setConstraints(sliders[i - LABEL_SIZE], c4);
            if ((i == LABEL_SIZE+2) || (i == LABEL_SIZE+8) ||(i == LABEL_SIZE+15) ) {
                c4.gridwidth = GridBagConstraints.REMAINDER;
                gridbag4.setConstraints(sliders[i - LABEL_SIZE], c4);
                p.add(sliders[i - LABEL_SIZE]);
                end = true;
            } else if ((i == LABEL_SIZE+11) || (i == LABEL_SIZE+5)) {
                c4.gridwidth = GridBagConstraints.REMAINDER;
                gridbag4.setConstraints(sliders[i - LABEL_SIZE], c4);
                p.add(sliders[i - LABEL_SIZE]);
                c4.weightx = 0.0;  
                c4.weighty = 0.0;
                c4.fill = GridBagConstraints.CENTER;
                c4.gridwidth = 1;
                int min =0, max = 3;
                if (i == LABEL_SIZE+11) {
                    min = 3;
                    max = LABEL_SIZE;
                }
                for (int j = min ; j < max ; j++) {
                    values[j] = new TextField(""+((Float)vect.elementAt(j)).intValue(), 3);
                    if (j == 2) 
                        c4.gridwidth = GridBagConstraints.REMAINDER;
                    gridbag4.setConstraints(values[j], c4);
                    p.add(values[j]);
                }
                end = true;
            }
            else {
                gridbag4.setConstraints(sliders[i - LABEL_SIZE], c4);
                p.add(sliders[i - LABEL_SIZE]);
            }
	
            if (end) {
                if (i == CRITERIA_SIZE-1)
                    c.gridwidth = GridBagConstraints.REMAINDER;
                gridbag.setConstraints(p, c);
                controls.add(p);
                end = false;
            }
        }

        c.weightx = 1.0;  
        c.weighty = 0.0;
        c.gridwidth = 1;
        c.fill = GridBagConstraints.NONE;

    
        Panel buttonPanel = new Panel();
        c4.fill = GridBagConstraints.CENTER;
        c4.weightx = 0.0;  
        c4.weighty = 0.0;
        c4.gridwidth = 1;
        c4.insets = new Insets(3,10,3,10);
        buttonPanel.setLayout(gridbag4);

        Button ApplyButton = new Button ("Apply");
        ApplyButton.addActionListener(new java.awt.event.ActionListener() {
                public void actionPerformed(java.awt.event.ActionEvent e) {
                    applyCriteriaFrame();
                }
            });
        gridbag4.setConstraints(ApplyButton, c4);
        buttonPanel.add(ApplyButton);
        Button CloseButton = new Button ("Close");
        CloseButton.addActionListener(new java.awt.event.ActionListener() {
                public void actionPerformed(java.awt.event.ActionEvent e) {
                    end();
                }
            });
        c4.gridwidth = GridBagConstraints.REMAINDER; //end row
        gridbag4.setConstraints(CloseButton, c4);
        buttonPanel.add(CloseButton);
        c.weighty= 0.0;
        c.weightx = 0.0;
        c.gridheight = 4;
        c.gridwidth = GridBagConstraints.REMAINDER;
        c.gridheight = GridBagConstraints.REMAINDER;
        gridbag.setConstraints(buttonPanel, c);
        controls.add(buttonPanel);
        controls.setVisible(false);
    }

  /**
   * Apply the criteria the control frame and applies the
   * data which was set in it.
   */
  public void applyCriteriaFrame() {
    Vector c = new Vector(CRITERIA_SIZE);
    override = true;

    for(int i = 0 ; i < LABEL_SIZE  ; i++) 
      criteria.insertElementAt(new Float(values[i].getText()), i);
    for(int i = LABEL_SIZE ; i < CRITERIA_SIZE  ; i++) 
      criteria.insertElementAt(new Float(sliders[i - LABEL_SIZE].getValue()), i);
    
    controls.setVisible(false);
  }

  /**
   * Cancels the criteria control frame - basically just
   * closes it without applying the settings.
   */
  public void cancelCriteriaFrame() {
      if (controls != null)
          controls.setVisible(false);
  }

  /**
   * Accessor function for the scheduler criteria.
   * @return Vector
   */ 
  public Criteria getCriteria() {
    return criteria;
  }

  /**
   * Set a serie of criteria used by the scheduler, this function is called 
   * on propertyChanged or when the TTAEMS structure has its own criteria.
   * @param v The vector criteria
   */
  public void setCriteria(Criteria v) {
      if ((controls != null) && controls.isVisible())
          openCriteriaFrame(v);
      criteria = v;
  }

    /**
     * updateResourceModeler, this function change any resource Usage 
     * to a future slot to put the resource modeler in the correct
     * position.
     * This method is a functionality of the Partial Order Scheduler
     * @param ScheduleElement se
     * @param offset 
     */
     public synchronized void updateResourceModeler(ScheduleElement se, int offset) {
	 log.log("Warning This method does nothing it's an abstract for the PartialOrderScheduler", Log.LOG_WARNING, FAC_SCHED);
     }

    /**
     * delayFinishTime delays the finish time of a schedule element
     * and readjust all the next schedule elements that were waiting for
     * the completion of this method.
     * @param Schedule mlcs
     * @param ScheduleElement se
     * @param offset 
     * @param set if true it updates the resource modeler
     * @return Schedule 
     */
    
    public Schedule delayFinishTime(Schedule s, ScheduleElement se, int offset)  throws agent.mass.ScheduleErrorException {
        s.delayFinishTime(se,offset);
        return(s);
    }


    /**
     * Delay the start time of a schedule.
     * @param schedule s
     * @param ScheduleElement se
     * @param int offset
     */
    public Schedule delayStartTime(Schedule s, ScheduleElement se, int offset) throws ScheduleErrorException {
        s.delayStartTime(se,offset);
        return(s);
    }

   /**
     * setStartTime sets the start time of the whole Schedule
     * all other Schedule Elements that depends on it
     * @param Schedule s
     * @param offset 
     * @return Schedule 
     */
    public Schedule setStartTime(Schedule s, int offset)  throws ScheduleErrorException {
	s.setStartTime(offset);
	return(s);
    }

  /**
   * Returns true if the component does something with clicks.
   */
  public boolean hasClickAction() {
      return true;
  }

  /**
   * Handles mouse events
   * @param e The event to process
   */
  public void processMouseEvent(MouseEvent e) {

    if (e.getID() == MouseEvent.MOUSE_CLICKED) {
        if (controls != null) {
            openCriteriaFrame(criteria);
            controls.setVisible(!controls.isVisible());
        }
    }
  }

  public void mouseClicked(MouseEvent e) { processMouseEvent(e); }
  public void mousePressed(MouseEvent e) { processMouseEvent(e); }
  public void mouseReleased(MouseEvent e) { processMouseEvent(e); }
  public void mouseDragged(MouseEvent e) { processMouseEvent(e); }
  public void mouseEntered(MouseEvent e) { }
  public void mouseExited(MouseEvent e) { }
  public void mouseMoved(MouseEvent e) { }

  /**
   * Watch for interesting property changes.  Monitors Task changes.
   * @param e The event to look at
   */
    public void propertyChanged(PropertyEvent e) { 
        String key = e.getKey().toString();
    
        if (key.equals(AUTOSCHEDULE)) {
            autoschedule = ((Boolean)e.getProperty()).booleanValue();

        } else if (key.equals(CACHESCHEDULES)) {
            cacheschedules = ((Boolean)e.getProperty()).booleanValue();

        } else if (key.equals(CACHELIMIT)) {
            cachelimit = ((Integer)e.getProperty()).intValue();

        } else if (key.equals(NUMSCHEDULES)) {
            numschedules = ((Integer)e.getProperty()).intValue();

        } else if (key.equals(MAXREADSCHEDULES)) {
            maxschedules = ((Integer)e.getProperty()).intValue();

        } else if (key.equals(AUTOSETSTARTTIME)) {
            autosetstarttime = ((Boolean)e.getProperty()).booleanValue();

        // Automatically (re)schedule if we are a simpleton
        } else if (key.equals(SCHEDULERCRITERIA)) {
            Vector v = (Vector)state.getProperty(SCHEDULERCRITERIA);
            setCriteria((Criteria)v);
	
        } else if (key.equals(SCHEDULE)) {
            Schedule s = (Schedule)e.getProperty();
            Object t = state.getProperty(SimpleTaemsReader.CONTAEMS);
            if (t instanceof Hashtable) {
                Enumeration en = ((Hashtable)t).elements();
                while (en.hasMoreElements()) {
                    t = en.nextElement();
                    ((Taems)t).setDisplayedSchedule(s);
                }
            
            } else if (t instanceof Taems) {
                ((Taems)t).setDisplayedSchedule(s);
            }
	
        } else if (key.equals(SimpleTaemsReader.CONTAEMS)) {
            if (autoschedule) {
                log.log("Autoscheduling...", Log.LOG_DEBUG, FAC_SCHED);
                Hashtable h = null;
                if (e.getProperty() instanceof taems.Taems) {
                    Taems t = (Taems)e.getProperty();
                    h = new Hashtable();
                    h.put(t.getLabel(),t);
                } else
                    h = (Hashtable)e.getProperty();
                for (Enumeration enumr = h.elements(); enumr.hasMoreElements(); ) {
                    Taems t = (Taems)enumr.nextElement();
                    if (t.numSchedules() == 0) {
                        Vector v = scheduleTaems(t,false,true);
                        if (v == null) {
                            log.log("Autoschedule failed, running evaluateTaems", Log.LOG_WARNING, FAC_SCHED);
                            evaluateTaems(t,false,true);
                        }
                    }
                }
            }
        }
    }

    public void propertyRemoved(PropertyEvent e) { propertyChanged(e); }
    public void propertyAdded(PropertyEvent e) { propertyChanged(e); }

    public Vector evaluateTaems(Taems taems) {
	return evaluateTaems(taems, false);
    }
    public Vector evaluateTaems(Taems taems, boolean rescheduling) {
	return evaluateTaems(taems, false, true);
    }

    /**
     * Takes a Taems structure and modifies it
     * to give to the scheduler in order to evaluate possible
     * schedules.  This function also performed several transformations
     * on the resulting clone to correct potential incompatibilities
     * between DTC and current Taems implementations.
     * <P>
     * Currently performed transformations:
     * <UL>
     * <LI> Schedules are removed
     * <LI> Nonlocal methods and tasks are made local, assigned to the owning agent
     * <LI> Consumes interrelationships belonging to remote agents are removed
     * <LI> Produces interrelationships belonging to remote agents are removed
     * <LI> Limits interrelationships are removed
     * <LI> q_last and q_seq_last QAFs are changed to q_seq
     * <LI> "SpecialType" Commitments are removed
     * <LI> Multiple task groups are merged under a single MasterNode with q_sum
     * </UL>
     * @param Taems structrue
     * @param boolean rescheduling flag
     * @param set Set the schedule to the Taems object or not
     * @return A Vector of Schedule
     * @see #scheduleTaems(Taems, boolean, boolean)
     */
    public Vector evaluateTaems(Taems taems, boolean rescheduling, boolean set) {
        Taems local = (Taems)taems.clone();

        // Remove schedules
        local.removeAllSchedules();

        // Remove nonlocal methods and tasks
        TaskBase tb = new TaskBase();
        tb.setNonLocal(true);
        Enumeration NLM = local.findNodes(tb);
        Agent agent = local.getAgent();
        while (NLM.hasMoreElements()) {
            tb = (TaskBase)NLM.nextElement();
            log.log("Un-nonlocl'ing " + tb.getLabel(), Log.LOG_DEBUG, FAC_SCHED);
            tb.setNonLocal(false);
            tb.setAgent(agent);
        }

        // Remove remove resource NLEs
        Enumeration enumr = local.findNodes(new ConsumesInterrelationship());
        while (enumr.hasMoreElements()) {
            ConsumesInterrelationship NLE = 
                (ConsumesInterrelationship)enumr.nextElement();
            if ((NLE.getTo().getAgent() == null) || 
                (!NLE.getTo().getAgent().getLabel().equals(agent.getLabel()))) {
                log.log("Excising " + NLE.getLabel(), Log.LOG_DEBUG, FAC_SCHED);
                local.exciseNode(NLE);
            }
        }
        enumr = local.findNodes(new ProducesInterrelationship());
        while (enumr.hasMoreElements()) {
            ProducesInterrelationship NLEProduces = 
                (ProducesInterrelationship)enumr.nextElement();
            if ((NLEProduces.getTo().getAgent() == null) ||
                (!NLEProduces.getTo().getAgent().getLabel().equals(agent.getLabel()))) {
                log.log("Excising " + NLEProduces.getLabel(), Log.LOG_DEBUG, FAC_SCHED);
                local.exciseNode(NLEProduces);
            }
        }
        enumr = local.findNodes(new LimitsInterrelationship());
        while (enumr.hasMoreElements()) {
            LimitsInterrelationship NLELimits = 
                (LimitsInterrelationship)enumr.nextElement();
            //if ((NLELimits.getFrom().getAgent() == null) ||
            //    (!NLELimits.getFrom().getAgent().getLabel().equals(agent.getLabel()))) {
                log.log("Excising " + NLELimits.getLabel(), Log.LOG_DEBUG, FAC_SCHED);
                local.exciseNode(NLELimits);
            //}
        }

        // Change labels on q_last and q_seq_last
        enumr = local.findNodes(new Task());
        while (enumr.hasMoreElements()) {
            Task t = (Task)enumr.nextElement();
            if (t.getQAF() instanceof LastQAF) {
                log.log("Switching QAF of " + t.getLabel(), Log.LOG_DEBUG, FAC_SCHED);
                t.getQAF().setLabel("q_seq");
            } else if (t.getQAF() instanceof SeqLastQAF) {
                log.log("Switching QAF of " + t.getLabel(), Log.LOG_DEBUG, FAC_SCHED);
                t.getQAF().setLabel("q_seq");
            }
        }

        // Remove special Commitments
        enumr = new utilities.SafeEnumeration(local.getCommitments());
        while (enumr.hasMoreElements()) {
            Commitment c = (Commitment)enumr.nextElement();
            if (c.isSpecialType()) {
                log.log("Removing commitment " + c.getLabel(), Log.LOG_DEBUG, FAC_SCHED);
                local.removeCommitment(c);
                continue;
            } else {
                // ok
            }
        }

        // Combine multiple task groups under a single MasterNode
        enumr = local.getNodes();
        int num = 0;
        while (enumr.hasMoreElements()) {
            Node n = (Node)enumr.nextElement();
            if (!(n instanceof Resource))
                num++;
            if ((n instanceof TaskBase) && ((TaskBase)n).hasSupertasks())
                num--;
        }
        if (num > 1) {
            enumr = new utilities.SafeEnumeration(local.getNodes());
            Task mn = new Task("MasterNode", agent, new SumQAF());
            Task task = (Task)local.findNode(mn);
            if (task == null) {
                log.log("Adding master node task group", Log.LOG_DEBUG, FAC_SCHED);
                task = mn;
                local.addNode(task);
            } else {
                log.log("Using existing master node", Log.LOG_DEBUG, FAC_SCHED);
            }
            while(enumr.hasMoreElements()) {
                Object o = enumr.nextElement();
                if ((o instanceof TaskBase) && (!((TaskBase)o).hasSupertasks())) {
                    tb = (TaskBase)o;
                    task.addSubtask(tb);
                    if (tb.getDeadline() != Integer.MIN_VALUE) {
                        if (task.getDeadline() == Integer.MIN_VALUE) {
                            log.log("Migrating deadline on task " + tb.getLabel() + " to " + task.getLabel(), Log.LOG_DEBUG, FAC_SCHED);
                            task.setDeadline(tb.getDeadline());
                        } else {
                            log.log("Warning, ignoring deadline on task " + tb.getLabel(), Log.LOG_WARNING, FAC_SCHED);
                        }
                        tb.unsetDeadline();
                    }
                    if (tb.getEarliestStartTime() != Integer.MIN_VALUE) {
                        if (task.getEarliestStartTime() == Integer.MIN_VALUE) {
                            log.log("Migrating earliest start time on task " + tb.getLabel() + " to " + task.getLabel(), Log.LOG_DEBUG, FAC_SCHED);
                            task.setEarliestStartTime(tb.getEarliestStartTime());
                        } else {
                            log.log("Warning, ignoring earliest start time on task " + tb.getLabel(), Log.LOG_WARNING, FAC_SCHED);
                        }
                        tb.setEarliestStartTime(Integer.MIN_VALUE);
                    }
                    local.removeNode(tb);
                }
            }
        }

        // Look if there is anything left to run
        enumr = local.findNodes(new Method());
        boolean found = false;
        while (enumr.hasMoreElements()) {
            Method m = (Method)enumr.nextElement();
            if (!m.hasFinished()) {
                found = true;

                // Have they started?
                if (m.hasStarted()) {
                    log.log("Adding finish time to running method " + m.getLabel(), Log.LOG_DEBUG, FAC_SCHED);
                    m.setFinishTime(m.getStartTime() + 1);
                }
            }
        }
        Vector sv;
        if (!found) {
            log.log("Warning, no unexecuted methods found in task structure (no schedule produced)", Log.LOG_WARNING, FAC_SCHED);
            sv = new Vector();
        } else {
            sv = scheduleTaems(local, rescheduling, false);
        }

        if (set) {
            if ((sv != null) && (sv.size() > 0)) {
                taems.setScheduleVector(sv);
                startFireEvent(new ScheduleEvent(this, 
                                                 (Schedule)sv.firstElement(), taems, 
                                                 ScheduleEvent.ADDED));
            } else {
                startFireEvent(new ScheduleEvent(this, null, taems, 
                                                 new ScheduleErrorException("No schedule generated"), 
                                                 ScheduleEvent.EXCEPTION));
            }
        }

        return sv;
    }

  /**
   * See scheduleTaems below.  Assumes the set parameter
   * is true and rescheduling is false.
   * @see #scheduleTaems(Taems, boolean, boolean)
   */
    public Vector scheduleTaems(Taems taems) {
	return scheduleTaems(taems, false);
    }

  /**
   * See scheduleTaems below.  Assumes the set parameter
   * is true.
   * @see #scheduleTaems(Taems, boolean, boolean)
   */
    public Vector scheduleTaems(Taems taems, boolean rescheduling) {
        return(scheduleTaems(taems, rescheduling, true));
    }

    /**
     * Invokes the scheduler to make a Taems schedule.  Uses Tom Wagner's
     * DTC scheduler.  You should not call this function unless you actually
     * have this scheduler.
     * <P>
     * <B>Note</B>:
     * The Design-to-Criteria scheduler is not currently available for general 
     * distribution.  Parties seriously interested in obtaining the scheduler 
     * for well defined research purposes should contact Tom Wagner at 
     * wagner@cs.umass.edu.
     * <P>
     * If you wish to set the time that the schedule schedules for, set
     * the <TT>time_to_schedule_for</TT> attribute in your Taems 
     * structure to an appropriate time (Float).  Note that the
     * rescheduling flag will override this value.
     *
     * @param taems The Taems structure to work on
     * @param rescheduling flag used for rescheduling.  If true,
     *  the time_to_schedule_for and cost_accrued attributes
     *  are set with the current time and the structures current
     *  total cost are specified.
     * @param set If true, will assign the Schedule vector to
     *  the Taems object, and fire an appropriate ScheduleEvent.  If this
     *  is false, note that you will likely need to call setSchedule
     *  with the new schedule on a Taems object before itcan really be 
     *  used.
     * @see taems.Taems#setSchedule
     * @see <a href="http://mas.cs.umass.edu/research/dtc.html">DTC</a>
     */
    public synchronized Vector scheduleTaems(Taems taems, 
            boolean rescheduling, boolean set) 
    {
        int exit_value;
        Vector sv = null;
        Object cachekey = null;
        if (obsched != null) obsched.updateCount(1);

         log.log("Schedule taems start on " + taems.getLabel(), Log.LOG_SPAM, 
                FAC_SCHED);

        // Caching
        if (cacheschedules) 
        {
            cachekey = generateCacheKey(taems);
            if (cache.containsKey(cachekey)) 
            {
                log.log("Using cached copy of schedule with key " + cachekey, 
                        Log.LOG_DEBUG, FAC_SCHED);
                if (obcache != null) obcache.updateCount(1);

                StringBuffer buf = (StringBuffer)cache.get(cachekey);
                sv = readSchedules(new StringReader(buf.toString()));

                if (set) 
                {
                    taems.setScheduleVector(sv);
                    startFireEvent(new ScheduleEvent(this, 
                                (Schedule)sv.firstElement(), taems, 
                                ScheduleEvent.ADDED));
                }

                log.log("Schedule taems cache end on " + taems.getLabel(), 
                        Log.LOG_SPAM, FAC_SCHED);
                return sv;	
            } else {
                log.log("No cached copy found for key " + cachekey, 
                        Log.LOG_DEBUG, FAC_SCHED);
            }
        }

        if (!override)
            if (taems.getScheduleCriteria() != null)
                setCriteria(taems.getScheduleCriteria());
            else 
                taems.setScheduleCriteria(criteria);
        else 
            taems.setScheduleCriteria(criteria);
	
        String name = (String)state.getProperty(State.NAME); 
                // + "_" + taems.getLabel();
        String DTCparameters="";
	
        if (rescheduling) 
        {
            int time = ((Integer)state.getProperty(State.TIME)).intValue();
            float cost = taems.getNode().getCurrentCost();
            DTCparameters += "(spec_attributes\n";
            DTCparameters += "   (time_to_schedule_for Float " + time + ")\n"; 
            DTCparameters += "   (cost_accrued Float " + cost + ")\n";
            DTCparameters += ")\n";
        }

        if (DTCNative.isNativeAvailable())
        {
            /////////////////
            // native call
            /////////////////
            String sched_input;
            String result;

            if (rescheduling)
            {
                sched_input = DTCparameters + taems.toTTaems(Taems.V1_0A);
            } else {
                sched_input = taems.toTTaems(Taems.V1_0A);
            }
            long start = System.currentTimeMillis();
            result = dtc_native.call_dtc(numschedules, "p", 50, 100, 0,
                    sched_input, false);
            long finish = System.currentTimeMillis();
            if (obschedtime != null) obschedtime.updateCount(finish - start);
            if (result == null)
            {
                exit_value = 1;
            } else {
                exit_value = 0;
                if (cacheschedules && (cachekey != null)) 
                {
                    cacheSchedules(cachekey, result);
                    StringBuffer buf = (StringBuffer)cache.get(cachekey);
                    if (buf != null) 
                        sv = parseSchedules(result);
                    else {
                        log.log("Warning: Apparently cached taems structure " +
                                "isn't actually cached", Log.LOG_WARNING, 
                                FAC_SCHED);
                        sv = parseSchedules(result);
                    }
                } else {
                    sv = parseSchedules(result);
                }
            }
                
        } else {
            /////////////////
            // exec call
            /////////////////
        
            // Check
            if (!state.hasProperty(DTC)) 
            {
                log.log("Cannot schedule task structure because DTC location "+ 
                        "has not been specified", Log.LOG_ERR, FAC_SCHED);
                return null;
            }

            log.log("Generating schedule", Log.LOG_INFO, FAC_SCHED);

            // Write the task file
            try {
                String str = SCHED_DIR + name + ".ttaems";
                log.log("Writing task to " + str, Log.LOG_DEBUG, FAC_SCHED);

                FileWriter taskfile = new FileWriter(str);
                if (rescheduling) 
                    taskfile.write(DTCparameters);
                taskfile.write(taems.toTTaems(Taems.V1_0A));
                taskfile.close();

            }  catch (IOException e) {
                log.log("Error:  write TTAEMS " + e, Log.LOG_ERR, FAC_SCHED);
                return null;
            }

            // Run the scheduler
            long start = System.currentTimeMillis();
            Process sched;
            try {
                String cmd;
                if (state.hasProperty(DTCOVERRIDE)) {
                    cmd = (String)state.getProperty(DTCOVERRIDE);
                } else {
                    cmd = SCHEDULER + " " + numschedules + " p 50 100 0";
                }
                cmd += " " + SCHED_DIR + name + ".ttaems" + " " + SCHED_DIR + name + " 0";
                log.log("Running " + cmd, Log.LOG_DEBUG, FAC_SCHED);
                sched = Runtime.getRuntime().exec(cmd);

                BufferedReader schedout = new BufferedReader(new InputStreamReader(sched.getInputStream()));
                BufferedReader schederr = new BufferedReader(new InputStreamReader(sched.getErrorStream()));
                //sched.waitFor();

                // Cheezy waitFor replacement
                while(true) 
                {
                    try {
                        // Ask the process for its exitValue. If the process
                        // is not finished, an IllegalThreadStateException
                        // is thrown. If it is finished, we fall through and
                        // the variable finished is set to true.
                        while (schederr.ready()) {
                            log.log("! " + schederr.readLine(), Log.LOG_DEBUG, FAC_SCHED);
                        }
                        while (schedout.ready()) {
                            log.log("> " + schedout.readLine(), Log.LOG_DEBUG, FAC_SCHED);
                        }
                        //System.err.println("Checking...");
                        exit_value = sched.exitValue();
                        break;
                    }
                    catch (IllegalThreadStateException e) {
                        Thread.currentThread().sleep(100);
                    }
                }
                log.log("Scheduler finished with status " + exit_value,
                        Log.LOG_DEBUG2, FAC_SCHED);

            } catch (InterruptedException e) {
                log.log("Error: " + e, Log.LOG_ERR, FAC_SCHED);
                return null;

            }  catch (IOException e) {
                log.log("Error: Running DTC " + e, Log.LOG_ERR, FAC_SCHED);
                e.printStackTrace();
                return null;
            }
            long finish = System.currentTimeMillis();
            if (obschedtime != null) obschedtime.updateCount(finish - start);

            if (exit_value == 0) 
            {
                if (cacheschedules && (cachekey != null)) 
                {
                    cacheSchedules(cachekey, SCHED_DIR,  name + 
                            ".ttaems_schedules");
                    StringBuffer buf = (StringBuffer)cache.get(cachekey);
                    if (buf != null) 
                        sv = readSchedules(new StringReader(buf.toString()));
                    else {
                        log.log("Warning: Apparently cached taems structure " +
                                "isn't actually cached", Log.LOG_WARNING, 
                                FAC_SCHED);
                        sv = readSchedules(SCHED_DIR,  name + ".ttaems_schedules");
                    }

                } else {
                    sv = readSchedules(SCHED_DIR,  name + ".ttaems_schedules");
                }
            }
        }

        ///////////////////////////////////////////////////
        // processing of results for native or exec call
        ///////////////////////////////////////////////////
        if (exit_value == 0)
        {
            if (sv == null)
               log.log("Error: null schedule was returned", Log.LOG_ERR, 
                       FAC_SCHED);
            else if (sv.size() == 0)
               log.log("Warning: Empty schedule was returned", Log.LOG_WARNING, 
                       FAC_SCHED);
            else
               log.log(sv.size() + " schedule(s) were returned", Log.LOG_INFO, 
                       FAC_SCHED);

            if (set) 
            {
                if ((sv != null) && (sv.size() > 0)) 
                {
                    taems.setScheduleVector(sv);
                    startFireEvent(new ScheduleEvent(this, 
                                (Schedule)sv.firstElement(), taems, 
                                ScheduleEvent.ADDED));
                } else {
                    startFireEvent(new ScheduleEvent(this, null, taems, 
                                new ScheduleErrorException(
                                    "No schedule generated"), 
                                ScheduleEvent.EXCEPTION));
                }
            }

            log.log("Schedule taems end on " + taems.getLabel(), Log.LOG_SPAM, 
                    FAC_SCHED);
            return sv;

        } else {
            log.log("Error: in execution of DTC, exit value was " + 
                    exit_value, Log.LOG_ERR, FAC_SCHED);
            startFireEvent(new ScheduleEvent(this, null, taems, 
                        new ScheduleErrorException("Error in DTC execution"), 
                        ScheduleEvent.EXCEPTION));
            return null;
        }
    }

    /**
     * Given a string of schedules, seperate them into a vector.
     * @param schedules a set of schedules in ttaems format.
     * @return a Vector of schedules.
     */
    protected Vector parseSchedules(String schedules)
    {
        Reader r;

        if (parser == null) 
        {
            parser = new ReadTTaems(utilities.Log.getDefault());
        }
        StringBuffer buf = new StringBuffer();
        int count = 0;
        int loc = 0;
        if (maxschedules > 0) {
            try {
                for (count = 0; count < maxschedules; count++) 
                    {
                        while (true) {
                            if (loc >= schedules.length())
                                throw new EOFException();
                            char c = schedules.charAt(loc++);
                            buf.append(c);
                            if (c == '(') 
                                break;
                        }
                        for (int parens = 1; parens > 0; ) 
                            {
                                if (loc >= schedules.length())
                                    throw new EOFException();
                                char c = schedules.charAt(loc++);
                                buf.append(c);
                                if (c == ')') 
                                    parens--;
                                else if (c == '(') 
                                    parens++;
                            }
                    }
            } catch (EOFException e) {
                log.log("End of schedule string reached.", Log.LOG_DEBUG2, 
                        FAC_SCHED);
            }
            schedules = buf.toString();
        }

        // Make a reader
        r = new StringReader(schedules);

        // Then parse it
        Vector sv = null;
        try {
            Taems task = parser.readTTaems(r); 
            sv = task.getScheduleVector();
            log.log("" + sv.size() + " schedules(s) read in.", Log.LOG_INFO, 
                    FAC_SCHED);
            
            if (sv != null) {
                for (Enumeration e = sv.elements() ; e.hasMoreElements() ; ) { 
                    Schedule schedule = (Schedule)e.nextElement();
                    int time = ((Integer)state.getProperty(State.TIME)).intValue();
                    if (autosetstarttime && (time != 0)) {
                        schedule.setStartTime(time);
                    }
                    for (Enumeration e1 = schedule.getElements(); e1.hasMoreElements() ; ) {
                        ScheduleElement se = (ScheduleElement)e1.nextElement();
                        if (se.getLabel().startsWith("Slack_MyTime"))
                            schedule.removeScheduleElement(se);
                    }
                }
            }

        } catch (Error e) {
            log.log("Error parsing Taems schedule: " + e, Log.LOG_ERR, FAC_SCHED);
            log.log("Schedules: " + schedules, Log.LOG_ERR, FAC_SCHED);

        } catch (Exception e) {
            log.log("Error parsing Taems schedule: " + e, Log.LOG_ERR, FAC_SCHED);
            log.log("Schedules: " + schedules, Log.LOG_ERR, FAC_SCHED);
        }

        return sv;
        
    }
    
    /**
     * Call the schedule parser that read Tom's output and 
     * rank them in a Vector.
     * @return Vector of schedules
     */
    protected Vector readSchedules(String path, String file) {
        String filename;
        File fd = new File(path, file);

        if (fd.exists()) {
            try {
                filename = path + file;
                FileInputStream is = new FileInputStream(filename);
                Vector sv = readSchedules(new InputStreamReader(is));
                return sv;
                
            } catch (FileNotFoundException e) {
                log.log("Error: " + e, Log.LOG_ERR, FAC_SCHED);
                return null;
            }
        }

        log.log("Error: " + path + file + " was not created by the scheduler", Log.LOG_ERR, FAC_SCHED);

        return null;
    }

    /**
     * Call the schedule parser that read Tom's output and 
     * rank them in a Vector.
     * @return Vector of schedules
     */
    protected Vector readSchedules(Reader r) {
        
        if (parser == null) {
            parser = new ReadTTaems(utilities.Log.getDefault());
        }

        // Handle MaxReadSchedules
        if (r != null) {
            lastreader = r;
        } else {
            r = lastreader;
        }

        if (r == null) {
            log.log("Warning: readSchedules called with null Reader.", Log.LOG_WARNING, FAC_SCHED);
            return new Vector();
        }

        if (maxschedules > 0) {
            StringBuffer buf = new StringBuffer();
            int count = 0;
            try {
                for (count = 0; count < maxschedules; count++) {
                    while (true) {
                        int c = r.read();
                        if (c < 0) throw new EOFException();
                        buf.append((char)c);
                        if (c == '(') break;
                    }
                    for (int parens = 1; parens > 0; ) {                       
                        int c = r.read();
                        if (c < 0) throw new EOFException();
                        buf.append((char)c);
                        if (c == ')') parens--;
                        else if (c == '(') parens++;
                    }
                }
            } catch (EOFException e) {
                log.log("End of schedule file reached.", Log.LOG_DEBUG2, FAC_SCHED);
            } catch (IOException e) {
                log.log("Error reading schedules: " + e.toString(), Log.LOG_ERR, FAC_SCHED);
            }
            if (count == 0) {
                return new Vector();
            }
            r = new StringReader(buf.toString());
        }

        try {
            if (!r.ready()) {
                log.log("Reader in readSchedules is not ready", Log.LOG_WARNING, FAC_SCHED);
                return new Vector();
            }
        } catch (IOException e) {
            log.log("Reader in readSchedules is not ready: " + e, Log.LOG_WARNING, FAC_SCHED);
            return new Vector();
        }

        Taems task = parser.readTTaems(r); 
        Vector sv = task.getScheduleVector();
        log.log("" + sv.size() + " schedules(s) read in.", Log.LOG_INFO, FAC_SCHED);
            
        if (sv != null) {
            for (Enumeration e = sv.elements() ; e.hasMoreElements() ; ) {
                Schedule schedule = (Schedule)e.nextElement();
                int time = ((Integer)state.getProperty(State.TIME)).intValue();
                if (autosetstarttime && (time != 0)) {
                    schedule.setStartTime(time);
                }
                for (Enumeration e1 = schedule.getElements();
                     e1.hasMoreElements() ; ) {
                    ScheduleElement se = (ScheduleElement)e1.nextElement();
                    if (se.getLabel().startsWith("Slack_MyTime"))
                        schedule.removeScheduleElement(se);
                }
            }
        }
        
        return sv;
    }

    /**
     * If MaxReadSchedules is > 0, this function will return the next
     * (MaxReadSchedules number) of schedules from the previous call
     * to DTC, provided the file containing those schedules hasn't
     * been removed.  Repeated calls to this function will continue
     * to return subsequent batches of schedules, until no more
     * are avaiable.
     * @return Vector A vector of Schedules 
     */
    public Vector readMoreSchedules() {
        return readSchedules((Reader)null);
    }

    /**
     * Generates a key that should hopefully uniquely identify the
     * taems structure, or at least identify the properties that
     * control its scheduling properties.
     */
    protected Object generateCacheKey(Taems taems) {
        StringBuffer buf = new StringBuffer("");

        Enumeration e = taems.getNodes();
        while (e.hasMoreElements()) {
            Node n = (Node)e.nextElement();
            if (n instanceof Task) {
                Task t = (Task)n;
                buf.append(t.getLabel());
                buf.append(String.valueOf(t.getEarliestStartTime()));
                buf.append(String.valueOf(t.getDeadline()));
            }
        }
        e = taems.findNodes(new Method());
        while (e.hasMoreElements()) {
            Method m = (Method)e.nextElement();
            buf.append(m.getLabel());
            buf.append(String.valueOf(m.getEarliestStartTime()));
            buf.append(String.valueOf(m.getDeadline()));
            buf.append(m.getGlobalOutcome().getDuration().calculateMax());
            buf.append(m.getCurrentQuality());
        }
        e = taems.findNodes(new Interrelationship());
        while (e.hasMoreElements()) {
            Interrelationship i = (Interrelationship)e.nextElement();
            buf.append(i.getLabel());
            if (i.getQuality() != null)
                buf.append(i.getQuality().calculateMax());
            if (i.getCost() != null)
                buf.append(i.getCost().calculateMax());
            if (i.getDuration() != null)
                buf.append(i.getDuration().calculateMax());
        }

        return buf.toString();
    }

    protected void cacheSchedules(Object key, String schedules)
    {
        // Test cache limit
        while ((cachelimit > 0) && (cache.size() > cachelimit)) 
        {
            // I just pick a "random" one here, a better system would
            // evaluate entries based on use
            Object k = cache.keys().nextElement();
            log.log("Cache limit reached, removing key " + k.toString(), 
                    Log.LOG_DEBUG2, FAC_SCHED);
            if (cache.remove(k) == null) 
            {
                // yikes, just bail
                break;
            }
        }

        StringBuffer buf = new StringBuffer(schedules);

        // Save it
        cache.put(key, buf);

        if (log.getLogLevel(FAC_SCHED) >= Log.LOG_SPAM) 
        {
            log.log("Schedule stored with key " + key, Log.LOG_SPAM, FAC_SCHED);
            log.log("Number of schedules now cached: " + cache.size(), 
                    Log.LOG_SPAM, FAC_SCHED);
        }
    }
    
  /**
   * Cache the data from the output file
	 */
    protected void cacheSchedules(Object key, String path, String file) {
        String filename;
        File fd = new File(path, file);

        if (fd.exists()) {
            try {
                filename = path + file;
                FileInputStream is = new FileInputStream(filename);
                InputStreamReader ir = new InputStreamReader(is);
                BufferedReader br = new BufferedReader(ir);

                StringBuffer buf = new StringBuffer();
                String line;
                try {
                    while ((line = br.readLine()) != null) {
                        buf.append(line);
                    }

                } catch (IOException e) {
                    log.log("Error reading from schedule: " + e, Log.LOG_ERR, FAC_SCHED);
                    return;
                }

                // Test cache limit
                while ((cachelimit > 0) && (cache.size() > cachelimit)) {
                    // I just pick a "random" one here, a better system would
                    // evaluate entries based on use
                    Object k = cache.keys().nextElement();
                    log.log("Cache limit reached, removing key " + k.toString(), Log.LOG_DEBUG2, FAC_SCHED);
                    if (cache.remove(k) == null) {
                        // yikes, just bail
                        break;
                    }
                }

                // Save it
                cache.put(key, buf);

                if (log.getLogLevel(FAC_SCHED) >= Log.LOG_SPAM) {
                    log.log("Schedule stored with key " + key, Log.LOG_SPAM, FAC_SCHED);
                    log.log("Number of schedules now cached: " + cache.size(), Log.LOG_SPAM, FAC_SCHED);
                }

            } catch (FileNotFoundException e) {
                log.log("Error: " + e, Log.LOG_ERR, FAC_SCHED);
                return;
            }
        }

        return;
    }

  /**
	 * Inline  class for ScrollBar that build in the value, their type
	 * (Cost, Duration, Quality).
	 */
    class ScrollbarPanel extends Panel {
	private Label display,l;
	private Scrollbar scroll;
	public ScrollbarPanel(float val, String label) {
	    this((int)(val*100), label);
	}

	/** 
	 * Creates a panel with two labels and one ScrollBar.
	 * @param val is the initial value of the scrollbar,
	 * @param label is show in the upper right corner.
	 */
	public ScrollbarPanel(int val, String label) {
	    GridBagLayout gridbag = new GridBagLayout(); 
	    GridBagConstraints c = new GridBagConstraints();
	    setLayout( gridbag );
      
	    display = new Label("0000");
	    l = new Label(label);
	    scroll = new Scrollbar(Scrollbar.VERTICAL, val, 1, 0, 101);
	    scroll.addAdjustmentListener(new AdjustmentListener() {
		    public void adjustmentValueChanged(AdjustmentEvent e) {
			display.setText(String.valueOf(scroll.getValue()));
		    }
		});
      
	    c.fill = GridBagConstraints.BOTH;
	    c.weightx = 1.0;  
	    c.weighty = 1.0;
	    c.gridheight = 2;
	    gridbag.setConstraints(scroll, c);
	    add(scroll);
	    c.weighty = 0.0;                  //reset to the default
	    c.gridwidth = GridBagConstraints.REMAINDER; //end row
	    c.gridheight = 1;               //reset to the default
	    gridbag.setConstraints(l, c);
	    add(l);
	    gridbag.setConstraints(display, c);
	    add(display);
      
	    display.setText(String.valueOf(scroll.getValue()));
	}

	/**
	 * Get the scrollbar value, useful for reading it.
	 * @return an int 
	 */
	public int getValue() { return scroll.getValue(); }
    }

}
