/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * Scripter.java
 *
 * manifest: Name: agent/mass/Scripter.class
 * manifest: Java-Bean: True
 ***************************************************************************************/

package agent.mass;

import java.io.*;
import java.awt.*;
import java.util.*;

import agent.base.*;
import agent.simplest.*;
import utilities.*;
import agent.simplest.Log;

/**
 * The sole purpose of this class is to enable the agent designer
 * to package scripts with the agent, which are transferred to the
 * simulator when the agent connects.
 */
public class Scripter extends AgentComponent {
  static final long serialVersionUID = 1234567890L; // kludge static ser_id.  Bad bad Bryan!

  protected Log log;
  protected State state;
  protected Communicate communicate;

  protected int FAC_SCRIPT;

  public static final String SCRIPTDATA = "ScriptData";

  /**
   * Default constructor.
   */
  public Scripter() {
    super();

    addDependency("Log");
    addDependency("State");
    addDependency("Communicate");

    State.addParameterInfo(SCRIPTDATA, "String", "Name of the extended configuration entry that contains the script data.  This is read in and sent to the Mass simulator");
  }

  /**
   * Init
   */
  public void init() {

    log = (Log)State.findComponent("Log");
    state = (State)State.findComponent("State");
    communicate = (Communicate)State.findComponent("Communicate");

    FAC_SCRIPT = log.getFacilityID(this);

    super.init();
  }

  /**
   * Begin
   */
  public void begin() {

    readScriptData();
  }

  /**
   * Reads in the script data from the extended entry file, if any.
   * The whole file is squished into one line and sent along to the
   * simulator.
   */
  public void readScriptData() {

    if (!state.hasProperty(SCRIPTDATA)) {
      return;
    }

    log.log("Reading script data", Log.LOG_INFO, FAC_SCRIPT);
    
    try {
      BufferedReader file = state.getExtendedEntryReader((String)state.getProperty(SCRIPTDATA));

      if (file == null) {
	log.log("Error reading from script data: " + state.getProperty(SCRIPTDATA), Log.LOG_ERR, FAC_SCRIPT);
	return;
      }

      StringBuffer buf = new StringBuffer("");
      while (file.ready()) {
	String str = file.readLine();

	// Trim
	if (str.indexOf(";") != -1)
	  str = str.substring(0, str.indexOf(";"));
	str = str.trim();
	if (str.length() <= 0) continue;

	// Append it
	buf.append(str + "\n");
      }

      file.close();

      // Send it along
      addScript(buf);

    } catch(IOException e) {
      log.log("Error: Script data file error: " + e, Log.LOG_ERR, FAC_SCRIPT);
    }

    log.log("Finished reading script data", Log.LOG_DEBUG, FAC_SCRIPT);
  }

  /**
   * Simple mechanism for remotely pausing the simulator.
   */
  public void pauseSimulator() {
    StringBuffer buf = new StringBuffer("");

    buf.append("Script, AndScript, Agent Pause, Fire:1\n");
    buf.append("  Assertion, True\n");
    buf.append("  Reaction, StopSim\n");

    addScript(buf);
  }

  /**
   * Adds a script to the system.  Right now, this just sends the
   * information along to the simulator
   * @param buf The buffer containing the script to send along
   */
  public void addScript(StringBuffer buf) {

    log.log("Adding script:\n" + buf.toString(), Log.LOG_SPAM, FAC_SCRIPT);

    KQMLMessage message = new KQMLMessage("tell",
					  "(NewScript " +
					  buf.toString() +
					  ")",
					  "simulator");
    message.addField("type", "control");
    communicate.sendMessage(message);
  }

  /**
   * Removes a script from the system.  Right now, this just sends the
   * request along to the simulator
   * @param s The name of the script to remove
   */
  public void removeScript(String s) {

    log.log("Removing script: " + s, Log.LOG_SPAM, FAC_SCRIPT);

    KQMLMessage message = new KQMLMessage("tell",
					  "(RemoveScript " +
					  s +
					  ")",
					  "simulator");
    message.addField("type", "control");
    communicate.sendMessage(message);
  }
}
