/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * Sensor.java
 *
 * manifest: Name: agent/mass/Sensor.class
 * manifest: Java-Bean: True
 ***************************************************************************************/

package agent.mass;

import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

import agent.base.*;
import agent.simplest.*;
import utilities.*;
import agent.simplest.Log;

/**
 * This class serves as an interface to the sensors instantiated by the
 * simulator.  Use this to make blocking, in-pulse queries for simple
 * sensor data.  The method you'll likely be most interested in here
 * is readSensorData.
 * <P>
 * Here's an example:
 * <PRE>
 *   Sensor sensor = (Sensor)state.findComponent("Sensor");
 *   if (sensor != null) {
 *     Float result = (Float)sensor.readSensorData("My Sensor Query", "Float");
 *     if (result != null) {
 *       // do something with it
 *     } else {
 *       // handle error
 *     }
 *   }
 * </PRE>
 * <P>
 * You should also be able to right-click on this comonent at runtime 
 * to see a list of all supported sensors (same as what would be returned
 * by getSensors())
 */
public class Sensor extends agent.simplest.Sensor implements MessageEventListener {
  static final long serialVersionUID = 1234567890L; // kludge static ser_id.  Bad bad Bryan!
  public static final int TIMEOUT = 20;

  protected agent.simplest.Communicate communicate;

  /**
   * Default constructor.
   */
  public Sensor() {
    super();

    addDependency("Communicate");
  }

  /**
   * Init
   */
  public void init() {
    super.init();

    communicate = (agent.simplest.Communicate)State.findComponent("Communicate");
    communicate.addMessageEventListener(this);
  }

  /**
   * Creates the sensor query string. (this does <B>not</B> actually
   * query the sensor - it literally just generates the string used
   * to do so).
   * @param sensor The sensory to query
   * @param data Any other data to add to the query, ignored if null
   * @return The string
   * @see #readSensorData
   */
    protected String makeSensorQuery(String sensor, Hashtable data) {
        Enumeration e;
        StringBuffer buf = new StringBuffer(sensor);

        // Break out
        if (data == null)
            return buf.toString();

        // Insertion sort the keys
        Vector v = new Vector();
        e = data.keys();
        while (e.hasMoreElements()) {
            String s = (String)e.nextElement();
            for (int i = 0; i < v.size(); i++) {
                String cmp = (String)v.elementAt(i);
                if (s.compareTo(cmp) < 0) {
                    v.insertElementAt(s, i);
                    s = null;
                    break;
                }
            }
            if (s != null)
                v.addElement(s);
        }

        // Make the string
        for (int i = 0; i < v.size(); i++) {
            String s = (String)v.elementAt(i);
            buf.append(", " + s + ":" + Converter.unTypeProperty(data.get(s), ";"));
        }

        // Done
        return buf.toString();
    }

  /**
   * Queries the sensor and returns the supplied sensor data, if any.
   * <br>
   * Note that the type that the sensor emits and the type
   * you specify here have to match!
   * <br>
   * This function uses makeSensorQuery internally to generate the
   * actual string used to query the sensor.
   * @param sensor The sensor to query
   * @param data Any other data to add to the query, ignored if null
   * @param type The data type, String if null
   * @return The data, or null if not found
   */
  public Object readSensorData(String sensor, Hashtable data, String type) {

    sensor = makeSensorQuery(sensor, data);

    return super.readSensorData(sensor, data, type);
  }

  /**
   * Returns the last recorded sensor data, if any, without
   * actually querying the sensor.
   * @param sensor The sensor to query
   * @param data Any other data to add to the query, ignored if null
   * @param type The data type, String if null
   * @return The data, or null if not found
   * @see #readSensorData
   */
  public Object readOldSensorData(String sensor, Hashtable data, String type) {

    sensor = makeSensorQuery(sensor, data);

    return super.readOldSensorData(sensor, data, type);
  }

  /**
   * Trys to fetch the current sensor data from the simulator
   * <P>
   * This is the current model for pseudo-blocking queries to the simulator.
   * Basically, you send a message, then actively call Communicate's
   * retrieveMessage function, with the expectation that a messageEvent
   * arising from this polling will cause a detectable change in local
   * state (here, we wait for a variable to change value).  Once detected,
   * execution can resume, or on timeout an error can be noted.
   * <P>
   * This clearly isn't the greatest model, but it should work until a 
   * better one is thought up.
   * @param sensor The sensor name
   * @return True if successfull, false otherwise
   */
    protected boolean querySensor(String sensor) {

        // Can we even do this yet?
        if (!communicate.ready()) {
            return false;
        }

        // Create the message
        log.log("Sending sensor query for " + sensor, Log.LOG_INFO, FAC_SENSOR);
        KQMLMessage message = new KQMLMessage("ask", "(Sensor " + sensor + ")", "simulator");
        Long id = new Long(State.getRandom().nextLong());
        message.addField("reply-with", id.toString());

        // Clear the old data
        sensors.remove(sensor);
        replies.remove(id);

        // Send the message
        communicate.sendMessage(message);
    
        // Wait for a reply
        int timeout = TIMEOUT * 10;
        while (! replies.containsKey(id)) {
            timeout--;
            if (timeout <= 0) {
                log.log("Timeout waiting for sensor data from simulator (" + sensor + ")", Log.LOG_ERR, FAC_SENSOR);
                break;
            }
            Message m = communicate.receiveMessage(true);
            if (m != null) {
                log.log("Intercepted message on connection " + m.getConnection() + ": " + m, Log.LOG_DEBUG, FAC_SENSOR);
                messageReceived(new MessageEvent(communicate, m, m.getConnection(), MessageEvent.RECEIVED));
            }
            try { Thread.currentThread().sleep(100); } catch (InterruptedException e) { }
        }

        // Return
        if (replies.containsKey(id)) {
            sensors.put(sensor, replies.get(id));
            replies.remove(id);
            return true;

        } else {
            return false;
        }
    }

  /**
   * Adds a sensor description to the list of sensors known to be available
   * to this agent.  The format is the same as in the config file, with
   * the initial Sensor chopped off.
   * @param s The string sensor definition
   */
    public void addSensor(String s) {
        Hashtable hash = new Hashtable();

        log.log("Adding sensor => " + s, Log.LOG_INFO, FAC_SENSOR);

        // Parse it to get the name
        StringTokenizer tok = new StringTokenizer(s, ",");
        tok.nextToken(); // Lose the "Sensor,"
        tok.nextToken(); // Lose the Sensor type
        String name = tok.nextToken();
        name = name.trim();

        // Tokenizer it
        while (tok.hasMoreTokens()) {
            // Get the next value pair
            String str = tok.nextToken().trim();
            if (str.indexOf(":") == -1)
                hash.put(str, "");
            else {
                String n = str.substring(0, str.indexOf(":")).trim();
                String v = str.substring(str.indexOf(":")+1).trim();
                hash.put(n, v);
            }
        }

        // Add it
        String ts = findSensor(name);
        if (ts == null) {
            // Send it along
            KQMLMessage message = new KQMLMessage("tell",
                                                  "(NewSensor " +
                                                  s +
                                                  ")",
                                                  "simulator");
            message.addField("type", "control");
            communicate.sendMessage(message);

            known.put(name, hash);
            menu.add(new MenuItem(name));
            log.log("Sensor added (" + name + ") " + hash, Log.LOG_DEBUG, FAC_SENSOR);
            startFireEvent(new SensorEvent(this, name, SensorEvent.ADDED));

        } else
            log.log("Duplicate sensor name (" + name + ") detected, not added", Log.LOG_WARNING, FAC_SENSOR);
    }

  /**
   * Watch for a message.  Handles property changes that arrive over
   * the network.
   * <P>
   * Currently, we will only accept state variable changes to known
   * variables.  This means that the variable must either previously
   * exist in the state table, or be a known parameter for the
   * data to be incorporated.
   * @param m The event to look at
   */
  public void messageSent(MessageEvent me) {}
  public void messageReceived(MessageEvent me) {

    // Check for other message types
    if (me.getConnection().getType() != Connection.KQML) {
      return;
    }

    KQMLMessage m = (KQMLMessage)me.getMessage();
    String perf = m.getPerformative();

    if (perf.equalsIgnoreCase("reply")) {
      // Giving us a variable?
      String key = m.contentWord();
      if ((key != null) && key.equals("SensorData")) {
	String data = m.contentData();
	Long id = Long.valueOf(m.getField("in-reply-to"));
	replies.put(id, data);
      }

    } else if (perf.equalsIgnoreCase("tell")) {
      // Giving us a new sensor entry?
      String key = m.contentWord();
      if ((key != null) && key.equals("NewSensor")) {
	String data = m.contentData();
	addSensor(data);
      }
    } 

  }
}
