/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * SimpleTaemsReader.java
 *
 * manifest: Name: agent/mass/SimpleTaemsReader.class
 * manifest: Java-Bean: True
 ***************************************************************************************/

package agent.mass;

import java.io.*;
import java.util.*;
import java.awt.*;
import javax.swing.*;
import javax.swing.event.*;

import agent.base.*;
import agent.simplest.*;
import taems.parser.*;
import taems.*;

import utilities.*;
import agent.simplest.Log;

/**
 * This class acts as an interface between textual Taems files and the binary
 * Taems representation.  Using the three properties, ObjTaems, SubTaems and
 * ConTaems, the user may specify the source for the different TAEMS views
 * used by the agent and simulator.  On startup, the three view specifiers
 * are searched for, and if found, are used to retrieve the corresponding
 * .ttaems files from the extended config file.  The subjective and conditioned
 * are parsed and set locally in State as SUBTAEMS and CONTAEMS (string constants),
 * respectively.  The objective view is parsed and automatically sent to the
 * simulator.  If the StoreObjTask parameter is set, the objective view
 * will also be stored locally as OBJTAEMS - but keep in mind you should
 * <B>not</B> use this view for normal agent operations.  If you do not need
 * the structures to be automatically generated, do not specify these properties,
 * and instead use readTTaems method.
 * <P>
 * So, to use this, you should just include this component in your agent, 
 * specify the three view locations and ensure to add them to your extended
 * config file.  After startup, you can gain access to the derived structures
 * through State.  If you need to, you can also reset the [Obj|Sub|Con]Taems
 * properties and call the make[Objective|Subjective|Conditional]View
 * functions manually to generate one of the views by hand.
 * <P>
 * As an example, in State you will have a ConTaems parameter, typically 
 * defined in your config file like this:
 * <PRE>
 * ConTaems: configs/myagent/myagent.ttaems
 * </PRE>
 * SimpleTaemsReader will automatically note this on startup, and will use
 * this to find the ConTaems entry in the extended config file.  Currently, 
 * this generation will be triggered by the addition of RandomSeed to 
 * state (which in Mass is done automatically by the simulator).  After
 * generation, you can get access to the conditional view like this:
 * <PRE>
 * Taems conditional = (Taems)state.getProperty(CONTAEMS);
 * </PRE>
 * ...and then do whatever you want with it.  Note that this structure
 * may not be available in the init or begin phases of control, so you should
 * use state.hasProperty() or check for a null return if you need to try to
 * get this information at those times. See the Taems documentation
 * for more details on working with Taems task structures.
 * <P>
 * So the take home message here is that when you use this component, all
 * the reading, parsing and placing of the various Taems task structure views
 * is handled for you.  So long as you define the various parameters and include
 * your textual taems file(s) in the extended entry file it should be placed in
 * the indicated spots when the agent starts up.
 * <P>
 * Here's a simple three node Taems task structure in textual form:
 * <PRE>
 * ; The owner agent name
 * (spec_agent
 *   (label agent_a)
 * )
 * <BR>
 * ; Start new base structure
 * (spec_task_group
 *    (label Root)
 *    (agent agent_a)
 *    (deadline 18)
 *    (subtasks Method1 Method2)
 *    (qaf q_min)
 * )
 * <BR>
 * ; Method 1
 * (spec_method
 *    (label Method1)
 *    (agent agent_a)
 *    (supertasks Root)
 *    (outcomes
 *       (Outcome_1
 *          (density 1.0)
 *          (quality_distribution 8.0 0.2 9.0 0.8)
 *          (duration_distribution 6.0 0.6 8.0 0.4)
 *          (cost_distribution 5.0 1.0)
 *       )
 *    )
 * )
 * <BR>
 * ; Method 2
 * (spec_method
 *    (label Method2)
 *    (agent agent_a)
 *    (supertasks Root)
 *    (outcomes
 *       (Outcome_1
 *          (density 1.0)
 *          (quality_distribution 7.0 1.0)
 *          (duration_distribution 5.0 0.3 6.0 0.7)
 *          (cost_distribution 6.0 1.0)
 *       )
 *    )
 * )
 * </PRE>
 * <P>
 * The MultiTaems property allows the agent to easily manipulate several
 * tasks structures simultaneously.  When this property is set, structures
 * are stored in hashtables in State, instead of just being single objects.
 * The structures themselves are keyed by their labels (e.g. Taems.get/setLabel()).
 * So, if you used SimpleTaemsReader to read in a structure with the label
 * "FooBar", you would use the following commands to get access to it:
 * <PRE>
 * Hashtable condstructs = (Hashtable)state.getProperty(CONTAEMS);
 * Taems conditional = (Taems)condstructs.get("FooBar");
 * </PRE>
 * If you then read in another structure, with the label "BarFoo", it would
 * be stored in that same hashtable, with a different key.  Note that
 * structures will not be removed in this case, so you must manually
 * remove them, from both the subjective and conditioned hashtables,
 * when you are done with them.  The labels should be consistent, so both
 * your subjective and conditioned (and objective) views should have the
 * same labels.  This allows SimpleTaemsReader to automatically update the
 * remote objective views correctly (e.g. objective FooBar will be removed
 * when you remove subjective FooBar locally).
 * <P>
 * Here's a brief summary of the four different cases you might run into:
 * <UL>
 * <LI><I>Single + no labels provided</I> - Each new structure which is
 * read in will be given a unique label.  Existing structures will be
 * removed.
 * <LI><I>Single + labels provided</I> - New structures read in will
 * keep their labels.  New labeled structure will overwrite old ones,
 * structures with the same label will be added to existing ones (see
 * Taems.importTaems()).
 * <LI><I>Multi + no labels provided</I> - You shouldn't really do this,
 * as your agent will wind up with a lot of structures hanging about
 * that it doesn't need (plus you won't know the key needed to access
 * them in the hashtables).
 * <LI><I>Multi + labels provided</I> - New structures will keep their
 * labels.  Structures with unique labels are added to the hashtable.
 * Structures with labels that already exist are added to the matching
 * structure already in the hashtable (see also Taems.importTaems()).  
 * Objective views will be automatically removed when their subjective
 * counterparts are removed.
 * </UL>
 * @see #makeAllStructures
 */
public class SimpleTaemsReader extends agent.base.AgentComponent implements PropertyEventListener, MessageEventListener {
  static final long serialVersionUID = 1234567890L; // kludge static ser_id.  Bad bad Bryan!

  // Used to tag un-labeled structures
  protected static int NUM = 0;
  private Hashtable olabels = new Hashtable();

  /* Structure Type Constants - these can change value at runtime */
  public static String CONTAEMS = "ConTaemsTask";
  public static String SUBTAEMS = "SubTaemsTask";
  public static String OBJTAEMS = "ObjTaemsTask";

  protected agent.simplest.State state;
  protected agent.simplest.Log log;
  protected agent.simplest.Communicate communicate;

  protected int FAC_ORGDES;
  private JFrame sdisplay, cdisplay, odisplay;
  private int nextUpdate;
  protected ReadTTaems reader = new ReadTTaems(utilities.Log.getDefault());
  private boolean setname, commits, multitaems, retattrib;
  public Vector allstructs = new Vector();

  public static final String SETTAEMSNAME = "SetTaemsName";
  public static final String STOREOBJTASK = "StoreObjTask";
  public static final String MULTITAEMS = "MultiTaems";
  public static final String KEEPCOMMITS = "KeepCommits";
  public static final String DISPLAYTAEMS = "DisplayTaems";
  public static final String RETARGETATTRIBUTES = "RetargetAttributes";

  /**
   * Default constructor
   */
  public SimpleTaemsReader() {
    super();

    addDependency("State");
    addDependency("Log");
    addDependency("Communicate");

    State.addParameterInfo("ObjTaems", "String", "Objective Taems definition file.  If specified, the file will be used to automatically generate and disseminate the objective view.");
    State.addParameterInfo("SubTaems", "String", "Subjective Taems definition file.  If specified, the file will be used to create a local subjective view.");
    State.addParameterInfo("ConTaems", "String", "Conditioned Taems definition file.  If specified, the file will be used to create a local conditioned view.");
    State.addParameterInfo(SETTAEMSNAME, "Boolean", "If true, re-sets Taems agent name for all local nodes of a task structure marked with the original primary agent name after its generation", Boolean.TRUE);
    State.addParameterInfo(STOREOBJTASK, "Boolean", "If true, a copy of the objective view is stored locally (normally the agent would not have access to this)", Boolean.FALSE);
    State.addParameterInfo(MULTITAEMS, "Boolean", "If true, the agent is assumed to be able to handle multiple independant structures concurrently.  Taems structure types are stored in Hashtables.  See the storeTaems procedure for more info.", Boolean.FALSE);
    State.addParameterInfo(KEEPCOMMITS, "Boolean", "If true, commitments are appended onto new Taems structures from a previously existing one (if their labels match) if is it found", Boolean.TRUE);
    State.addParameterInfo(DISPLAYTAEMS, "Boolean", "If true, Taems task structures are automatically displayed in a window when they are created", Boolean.TRUE);
    State.addParameterInfo(RETARGETATTRIBUTES, "Boolean", "If true, calls retargetAttributes on all new Taems task structures", Boolean.FALSE);
  }

  /**
   * Sets up component helpers
   */
  public void init() {
    state = (agent.simplest.State)State.findComponent("State");
    CondRule.log = log = (agent.simplest.Log)State.findComponent("Log");
    communicate = (agent.simplest.Communicate)State.findComponent("Communicate");

    FAC_ORGDES = log.getFacilityID("TaemsReader");

    state.addPropertyEventListener(this);
    communicate.addMessageEventListener(this);

    super.init();
  }

  /**
   * Doesn't do anything right now
   */
  public void begin() { }

  /**
   * Reset
   */
  public void reset() {

    makeAllStructures();
  }

  /**
   * Pulse based control
   */
  boolean first = true;
  public void pulse() {
      if (first && (NUM == 0)) {
          first = false;
          String obj = (String)state.getProperty("ObjTaems");
          String con = (String)state.getProperty("ConTaems");
          String sub = (String)state.getProperty("SubTaems");
          if ((obj != null) || (con != null) || (sub != null))
              makeAllStructures();
      }
  }

    /**
     * Accessors
     */
    public void setKeepCommits(boolean s) { commits = s; }
    public boolean getKeepCommits() { return commits; }
    public void setRetargetAttributes(boolean r) { retattrib = r; }
    public boolean getRetargetAttributes() { return retattrib; }
    public void setSetName(boolean s) { setname = s; }
    public boolean getSetName() { return setname; }
    public void setMultiTaems(boolean s) { 
        multitaems = s;

        if (getMultiTaems()) {
            CONTAEMS = "ConTaemsTasks";
            SUBTAEMS = "SubTaemsTasks";
            OBJTAEMS = "ObjTaemsTasks";
        } else {
            CONTAEMS = "ConTaemsTask";
            SUBTAEMS = "SubTaemsTask";
            OBJTAEMS = "ObjTaemsTask";
        }
    }
    public boolean getMultiTaems() { return multitaems; }
    public ReadTTaems getReader() { return reader; }
 
  /**
   * Reads a textual taems file from State's external configuration
   * data file and returns a Taems object.  It will also update the
   * object's agent name if the SetTaemsName parameter is true.
   * @param name The external config entry to read from
   * @return a Taems reference, or null if the entry was not found
   * or it was not parsable.
   * @see State#getExtendedEntryReader
   */
  public synchronized Taems readTTaems(String name) {
    Taems taems = null;

    log.log("Reading Taems file entry " + name, Log.LOG_DEBUG, FAC_ORGDES);
    
    try {
      BufferedReader file = state.getExtendedEntryReader(name);

      if (file == null) {
          log.log("Error reading from ttaems file: " + name, Log.LOG_ERR, FAC_ORGDES);
          return null;
      }

      // Parse that baby
      log.log("Starting Taems parse", Log.LOG_SPAM, FAC_ORGDES);
      taems = reader.readTTaems(file);
      log.log("Finished Taems parse", Log.LOG_SPAM, FAC_ORGDES);

      file.close();

      if (taems != null) {
          if (getRetargetAttributes())
              taems.retargetAttributes();
          if (getSetName())
              updateTaemsAgentName(taems);
      }

    } catch(IOException e) {
      log.log("Error: Taems file error: " + e, Log.LOG_ERR, FAC_ORGDES);
    }

    log.log("Finished reading Taems file", Log.LOG_SPAM, FAC_ORGDES);
    return taems;
  } 

    /**
     * Reads a Taems structure from a StringReader.
     * @param read The reader to read from
     * @see #readTTaems(String)
     */
    public synchronized Taems readTTaems(StringReader read) {
        Taems taems = null;
        log.log("Starting Taems parse", Log.LOG_SPAM, FAC_ORGDES);
        taems = reader.readTTaems(read);
        log.log("Finished Taems parse", Log.LOG_SPAM, FAC_ORGDES);
	
        if (taems != null) {
            if (getRetargetAttributes())
                taems.retargetAttributes();
            if (getSetName())
                updateTaemsAgentName(taems);
        }

        log.log("Finished reading Taems file", Log.LOG_SPAM, FAC_ORGDES);
        return taems;
    } 

    /**
     * Updates the agent names in the respective taems nodes, if they
     * are local and match the original name of the taems structure itself
     * @param taems the structure to update
     */
    public void updateTaemsAgentName(Taems taems) {
        if (taems == null) {
            log.log("Warning: updateTaemsAgentName given null task structure", Log.LOG_ERR, FAC_ORGDES);
            return;

        }
        String agentname = (String)state.getProperty(State.NAME);
        Agent agent = new Agent(agentname);
        Agent oldagent = taems.getAgent();

        if (oldagent != null) {
            if (oldagent.matches(agent)) {
                log.log("New name " + oldagent.getLabel() + " matches old, no update performed.", Log.LOG_SPAM, FAC_ORGDES);
                return;
            } else
                log.log("Updating taems name from " + oldagent.getLabel() + " to " + agent.getLabel(), Log.LOG_SPAM, FAC_ORGDES);
        } else {
            log.log("Warning, original agent name is not set in Taems object " + taems.getLabel() + ", aborting SetTaemsName", Log.LOG_INFO, FAC_ORGDES);
            return;
        }

        taems.setAgent(agent);
        
        Enumeration e = taems.getAllNodes();
        while (e.hasMoreElements()) {
            Node n = (Node)e.nextElement();
            if (n.getAgent() == null) {
                if (!(n instanceof Resource))
                    log.log("Warning, node " + n.getLabel() + " has no agent specified.", Log.LOG_WARNING, FAC_ORGDES);
                continue;
            }
            if ((n instanceof TaskBase) && ((TaskBase)n).isNonLocal()) {
                continue;
            }
            if (! n.getAgent().matches(oldagent)) continue;
            n.setAgent(agent);
        }
    }

    /**
     * Copies old commitments from a previous structure into a new one.  If
     * an old commitment matches a new one (see Commitment.matches()), it will
     * not be added.
     * @param ot The old task structure
     * @param nt The new task structure, which will be modified
     */
    public void copyOldCommitments(Taems ot, Taems nt) {

        // Check
        if (ot == null) return;
        if (nt == null) return;

        Enumeration oe = ot.getCommitments();
        while (oe.hasMoreElements()) {
            Commitment oc = (Commitment)oe.nextElement();

            // New?
            boolean found = false;
            Enumeration ne = nt.getCommitments();
            while (ne.hasMoreElements()) {
                Commitment nc = (Commitment)ne.nextElement();
                if (oc.matches(nc)) {
                    found = true;
                    break;
                }
            }

            // No? Add it.
            if (!found) {
                Commitment clone = (Commitment)oc.clone();
                Vector v = new Vector();
                Enumeration cv = clone.getTasks();
                while (cv.hasMoreElements()) {
                    Node n = (Node)cv.nextElement();
                    n = new VirtualTaskBase(n.getLabel(), n.getAgent());
                    v.addElement(n);
                }
                clone.setTaskVector(v);
                nt.addCommitment(clone);
            }
        }
    }

  /**
   * Generates the objective, subjective and conditional
   * structures.  These are just the default ones, and 
   * each is also sent or set to its designtated location.
   * This will typically only be called when the agent
   * is created or reset.
   * <P>
   * Also increments the un-labeled structure id.
   * @see #makeConditionalStructure
   * @see #makeSubjectiveStructure
   * @see #makeObjectiveStructure
   */
  public void makeAllStructures() {
    NUM++;

    String obj = (String)state.getProperty("ObjTaems");
    String con = (String)state.getProperty("ConTaems");
    String sub = (String)state.getProperty("SubTaems");

    Taems t = null;
    if (obj != null)
        t = makeObjectiveStructure(true);
    if (t != null) allstructs.addElement(t);
    if (sub != null)
        t = makeSubjectiveStructure(true);
    if (t != null) allstructs.addElement(t);
    if (con != null)
        t = makeConditionalStructure(true);
    if (t != null) allstructs.addElement(t);
  }

    /**
     * This is a convenience function to remove the (up to) three
     * structures generated by the makeAllStructures function.  If
     * you don't use that function, you shouldn't use this one.
     * @see #makeAllStructures();
     */
    public void removeAllStructures() {

        Enumeration e = allstructs.elements();
        while (e.hasMoreElements()) {
            Taems t = (Taems)e.nextElement();
            removeTaems(t);
        }
        allstructs.removeAllElements();
    }

    /**
     * Set all the view using the Taems Task Structure passed in
     * arguments. This is very useful if you generate the TAEMS structure
     * by another way (like a Problem Solver or a Planner) and
     * want to assign it to the agent.  This uses the storeTaems and 
     * handleNewObjectiveStructure  methods to store the conditional, subjective
     * and objective views.
     * <P>
     * Also increments the un-labeled structure id.
     * @param t The new Taems structure
     * @see #storeTaems
     * @see #handleNewObjectiveStructure
     */
    public void setTaems(Taems t) {
        NUM++;

        if (t != null) {
            if (getRetargetAttributes())
                t.retargetAttributes();
            if (getSetName())
                updateTaemsAgentName(t);
            if ((t.getLabel() == null) || t.getLabel().equals(""))
                t.setLabel("Structure_" + NUM);

            Taems cond = (Taems)t.clone();
            storeTaems(CONTAEMS, cond);

            Taems subj = (Taems)t.clone();
            storeTaems(SUBTAEMS, cond);

            handleNewObjectiveStructure(t, true);

        } else {
            log.log("setTaems called with null task structure", Log.LOG_ERR, FAC_ORGDES);
        }
    }

  /**
   * Creates the conditioned version of the task structure.
   * @param set Set this to true if you want the conditional
   *    CONTAEMS object to be updated with this structure.
   * @return The structure, or null if none generated.
   * @see #makeSubjectiveStructure
   */
  public Taems makeConditionalStructure(boolean set) {

    log.log("Generating conditional task structure", Log.LOG_INFO, FAC_ORGDES);

    String name = (String)state.getProperty("ConTaems");

    if (name != null) {

      Taems task = readTTaems(name);
      if ((task.getLabel() == null) || task.getLabel().equals(""))
          task.setLabel("Structure_" + NUM);

      // Set it?
      if (set) {
          task = storeTaems(CONTAEMS, task);
      }

      log.log("Generated conditioned task structure", Log.LOG_SPAM, FAC_ORGDES);
      return task;

    } else {
      log.log("Warning: no conditioned view specified.", Log.LOG_WARNING, FAC_ORGDES);
    }

    log.log("Conditioned generation failed", Log.LOG_WARNING, FAC_ORGDES);
    return null;
  }

  /**
   * Generates the subjective Taems task structure
   * @param set Set this to true if you want the subjective
   *    SUBTAEMS object to be updated with this structure.
   * @return The structure, or null if none generated.
   * @see #makeObjectiveStructure
   */
  public Taems makeSubjectiveStructure(boolean set) {

    log.log("Generating subjective task structure", Log.LOG_INFO, FAC_ORGDES);

    String name = (String)state.getProperty("SubTaems");

    if (name != null) {

      Taems task = readTTaems(name);
      if ((task.getLabel() == null) || task.getLabel().equals(""))
          task.setLabel("Structure_" + NUM);

      // Set it?
      if (set) {
          task = storeTaems(SUBTAEMS, task);
      }

      log.log("Generated subjective task structure", Log.LOG_SPAM, FAC_ORGDES);
      return task;

    } else {
      log.log("Warning: no subjective view specified.", Log.LOG_WARNING, FAC_ORGDES);
    }

    log.log("Subjective generation failed", Log.LOG_WARNING, FAC_ORGDES);
    return null;
  }

  /**
   * Generates the objective Taems task structure.
   * @param send Set this to true if you want the structure
   *   sent to the simulator
   * @return The structure.  Note that you shouldn't really be
   *   looking at the objective view on the agent side.
   */
  public Taems makeObjectiveStructure(boolean send) {

    log.log("Generating objective task structure", Log.LOG_INFO, FAC_ORGDES);

    String name = (String)state.getProperty("ObjTaems");

    if (name != null) {

      Taems task = readTTaems(name);
      if ((task.getLabel() == null) || task.getLabel().equals(""))
          task.setLabel("Structure_" + NUM);
 
      // Handle it
      task = handleNewObjectiveStructure(task, send);

      log.log("Generated objective task structure", Log.LOG_SPAM, FAC_ORGDES);
      return task;
    
    } else {
      log.log("Warning: no objective view specified.", Log.LOG_WARNING, FAC_ORGDES);
    }

    log.log("Objective generation failed", Log.LOG_WARNING, FAC_ORGDES);
    return null;
  }

  /**
   * Handles the addition of a new objective task structure,
   * optionally sending to the Mass simulator it if necessary.
   * @param taems The new task structure
   * @param send Send it to Mass?
   * @return The taems object which was stored
   */
  public Taems handleNewObjectiveStructure(Taems obj, boolean send) {

      // Send it?
      if (send) {
          KQMLMessage m;

          // Check
          if (!(getMultiTaems() || olabels.containsKey(obj.getLabel()))) {
              handleRemoveObjectiveStructure(null);
          }

          // Send new
          m = new KQMLMessage("tell", "(AddTaems " + obj.toString() + ")", "simulator");
          log.log("Sending objective task structure to simulator", Log.LOG_DEBUG, FAC_ORGDES);
          communicate.sendMessage(m);
          
          // Note the label
          olabels.put(obj.getLabel(), new Boolean(true));
      }
      
      // Some people may want this
      if (state.hasProperty(STOREOBJTASK) && (((Boolean)state.getProperty(STOREOBJTASK)).booleanValue() == true)) {
          obj = storeTaems(OBJTAEMS, obj);
      }

      return obj;
  }

  /**
   * Handles the removal of a objective task structure,
   * @param l The label of the task structure to remove. If null,
   *   all old objective structures are removed.
   */
  public void handleRemoveObjectiveStructure(String l) {
      KQMLMessage m;

      log.log("Removing objective task structure " + l + " from simulator", Log.LOG_DEBUG, FAC_ORGDES);

      if (l == null) {
          m = new KQMLMessage("tell", "(RemoveTaems)", "simulator");
      } else {
          m = new KQMLMessage("tell", "(RemoveTaems " + l + ")", "simulator");
          olabels.remove(l);
      }

      communicate.sendMessage(m);
  }

    /**
     * Stores the Taems object locally to the agent, using the object's label
     * as the hashtable key if necessary (when MutliTaems is true).  This is
     * a convenience function that respects the MultiTaems property.
     * <P>
     * Agents generally come in two varieties, with respect to Taems structures.
     * Either they combine all their knowledge into a single comprehensive 
     * structure, or multiple independant structures are mantained and potentially
     * active concurrently.  If the former, Taems structures are stored by this method
     * as-is in State.  If the MultiTaems parameter is true, Taems structures are
     * stored in Hashtables in state.  Each type will have it's own hashtable, containing
     * the structures keyed by their label.
     * <P>
     * If a structure exists with the same label in the location, the two
     * will be merged (see also Taems.importTaems()), and the resulting
     * structure returned.
     * @param type The type of structure (e.g. SUBTAEMS, CONTAEMS, OBJTAEMS)
     * @param taems The task structure
     * @return The actual structure which was added.
     */
    public Taems storeTaems(String type, Taems taems) {
        Taems exist = null;
        log.log("Storing task structure " + type, Log.LOG_DEBUG, FAC_ORGDES);

        if (getMultiTaems()) {
            if (taems.getLabel() == null) {
                log.log("Warning: Taems structure has no label", Log.LOG_WARNING, FAC_ORGDES);
                taems.setLabel("Unknown");
            }
            Hashtable ht = (Hashtable)state.getProperty(type);
            if (ht == null)
                ht = new Hashtable();

            exist = (Taems)ht.get(taems.getLabel());
            if ((exist != null) && (exist instanceof Taems)) {
                if (getKeepCommits())
                    copyOldCommitments(exist, taems);
                if ((exist.getLabel() != null) && exist.getLabel().equals(taems.getLabel())) {
                    log.log("Importing new task structure into existing table entry for " + exist.getLabel(), Log.LOG_DEBUG2, FAC_ORGDES);
                    exist.importTaems(taems);
                } else {
                    log.log("Adding new task structure into hashtable", Log.LOG_DEBUG2, FAC_ORGDES);
                    exist = taems;
                }
            } else {
                log.log("Adding new task structure into hashtable", Log.LOG_DEBUG2, FAC_ORGDES);
                exist = taems;
            }

            ht.put(exist.getLabel(), exist);
            state.setProperty(type, ht, exist.getLabel());

        } else {
            exist = (Taems)state.getProperty(type);
            if ((exist != null) && (exist instanceof Taems)) {
                if (getKeepCommits())
                    copyOldCommitments(exist, taems);
                if ((exist.getLabel() != null) && exist.getLabel().equals(taems.getLabel())) {
                    log.log("Importing new task structure into existing " + exist.getLabel(), Log.LOG_DEBUG2, FAC_ORGDES);
                    exist.importTaems(taems);
                } else {
                    log.log("Adding new task structure to state", Log.LOG_DEBUG2, FAC_ORGDES);
                    exist = taems;
                }
            } else {
                log.log("Adding new task structure to state", Log.LOG_DEBUG2, FAC_ORGDES);
                exist = taems;
            }

            state.setProperty(type, exist, exist.getLabel());
        }

        return exist;
    }

    /**
     * Removes the Taems structure from the local state component.
     * @param taems The task structure
     */
    public void removeTaems(Taems taems) {
        String types[] = { CONTAEMS, SUBTAEMS, OBJTAEMS };

        log.log("Removing task structure " + taems.getLabel(), Log.LOG_DEBUG, FAC_ORGDES);

        // Generate the list
        if (getMultiTaems()) {
            for (int i = 0; i < types.length; i++) {
                Hashtable ht = (Hashtable)state.getProperty(types[i]);
                if (ht == null) continue;
                for (Enumeration e = ht.keys(); e.hasMoreElements(); ) {
                    String key = (String)e.nextElement();
                    Taems t = (Taems)ht.get(key);
                    if (t == taems) {
                        ht.remove(key);
                        state.setProperty(types[i], ht);
                        break;
                    }
                }
            }

        } else {
            for (int i = 0; i < types.length; i++) {
                Taems t = (Taems)state.getProperty(types[i]);
                if (t == null) continue;
                if (t == taems) {
                    state.removeProperty(types[i]);
                    break;
                }
            }
        }
    }

    /**
     * Given a node, searches through the known Taems structures to find
     * the structure that contains it  This is a convenience function that
     * respects the MultiTaems property.
     * @param n The node to search for
     * @return The structure containing it, or null if none found
     */
    public Taems findTaems(Node n) {
        String types[] = { CONTAEMS, SUBTAEMS, OBJTAEMS };
        Vector v = new Vector();
        Taems taems;

        log.log("Looking for the structure containing node " + n.getLabel(), Log.LOG_DEBUG, FAC_ORGDES);

        // Generate the list
        if (getMultiTaems()) {
            for (int i = 0; i < types.length; i++) {
                Hashtable ht = (Hashtable)state.getProperty(types[i]);
                if (ht == null) continue;
                for (Enumeration e = ht.elements(); e.hasMoreElements(); ) {
                    v.addElement(e.nextElement());
                }
            }

        } else {
            for (int i = 0; i < types.length; i++) {
                taems = (Taems)state.getProperty(types[i]);
                if (taems == null) continue;
                v.addElement(taems);
            }
        }

        // Search through each of them
        for (Enumeration e = v.elements(); e.hasMoreElements(); ) {
            taems = (Taems)e.nextElement();
            log.log("Looking at " + taems.getLabel(), Log.LOG_DEBUG2, FAC_ORGDES);
            for (Enumeration en = taems.findNodes(n); en.hasMoreElements(); ) {
                if (n == en.nextElement()) {
                    log.log("Found node in " + taems.getLabel(), Log.LOG_DEBUG, FAC_ORGDES);
                    return taems;
                } else {
                    log.log("Found matching but not equal node in " + taems.getLabel() + ": " + n, Log.LOG_DEBUG2, FAC_ORGDES);
                }
            }
        }

        log.log("Could not find node.", Log.LOG_DEBUG, FAC_ORGDES);
        return null;
    }

  /**
   * Displays a taems data structure in a window.  Repeated calls
   * re-use this same window.
   * @param f The frame to store it in, or nil if none created yet
   * @param Taems The TTaems structure to display
   * @param title The title of the window
   * @return The frame the structure is displayed in
   */
    public JFrame displayTaems(JFrame f, Object taems, String title) {

        if ((!hasGUI()) || (!((Boolean)state.getProperty(DISPLAYTAEMS)).booleanValue())) {
            return f;
        }

        log.log("Displaying Taems: " + title, Log.LOG_DEBUG, FAC_ORGDES);

        // Make the window
        if (f == null) {
            log.log("Creating new window for " + title, Log.LOG_DEBUG, FAC_ORGDES);
            if (State.hasComponent("WindowManager")) {
                WindowManager wm = (WindowManager)state.findComponent("WindowManager");
                f = wm.createWindow(title + " Taems Views");
            } else {
                f = new JFrame(title + " Taems Views");
            }

            f.setJMenuBar(new JMenuBar());
            f.getContentPane().setLayout(new BorderLayout());

            f.setSize(new Dimension(640, 460));

            JPanel p = new JPanel();
            p.setLayout(new CardLayout());
            f.getContentPane().add(p, "Center", 0);

            if (!agent.simplest.State.hasComponent("WindowManager"))
                f.setVisible(true);
        }

        // Get the structures
        Hashtable structs = null;
        if (taems instanceof Taems) {
            structs = new Hashtable(1);
            structs.put(title, taems);
        } else if (taems instanceof Hashtable) {
            structs = (Hashtable)taems;
        } else {
            log.log("Warning, attempt to add unknown object to Taems Views: " + taems, Log.LOG_WARNING, FAC_ORGDES);
            return f;
        }

        // Find the menu
        JMenu menu = null;
        for (int i = 0; i < f.getJMenuBar().getMenuCount(); i++) {
            menu = f.getJMenuBar().getMenu(i);
            if (menu.getText().equals(title)) {
                break;
            } else {
                menu = null;
            }
        }
        if (menu == null) {
            menu = new JMenu(title);
            f.getJMenuBar().add(menu);
        }
        menu.removeAll();

        // Do it
        JPanel p = (JPanel)f.getContentPane().getComponent(0);
        p.removeAll();

        Enumeration e = structs.keys();
        while (e.hasMoreElements()) {
            String key = (String)e.nextElement();
            Taems t = (Taems)structs.get(key);
            p.add(title + key, t);

            JMenuItem item = new JMenuItem(key);
            item.addMenuDragMouseListener(new myMenuDragMouseListener(p, title + key));
            menu.add(item);
        }

        f.validate();

        return f;
    }

class myMenuDragMouseListener implements MenuDragMouseListener {
    JPanel panel;
    String key;

    public myMenuDragMouseListener(JPanel p, String k) {
        panel = p;
        key = k;
    }
    public void menuDragMouseEntered(MenuDragMouseEvent e) { };
    public void menuDragMouseExited(MenuDragMouseEvent e) { };
    public void menuDragMouseDragged(MenuDragMouseEvent e) { };
    public void menuDragMouseReleased(MenuDragMouseEvent e) {
        ((CardLayout)panel.getLayout()).show(panel, key);
    }
}

  /**
   * Watch for messages
   * @param m The message event to look at
   */
  public void messageSent(MessageEvent e) { }
  public void messageReceived(MessageEvent e) {

    // Check for other message types
    if (e.getConnection().getType() != Connection.KQML) {
      return;
    }

    KQMLMessage m = (KQMLMessage)e.getMessage(); 
    String perf = m.getPerformative();

    if (perf.equalsIgnoreCase("tell")) {
        String key = m.contentWord();
        if (key.equalsIgnoreCase("GenerateTaskStructure")) {
            String data = m.contentData();

            if (data.equalsIgnoreCase("All"))
                makeAllStructures();
            else if (data.equalsIgnoreCase("Objective"))
                makeObjectiveStructure(true);
            else if (data.equalsIgnoreCase("Subjective"))
                makeSubjectiveStructure(true);
            else if (data.equalsIgnoreCase("Conditional"))
                makeConditionalStructure(true);
            else
                log.log("Unknown view " + data + " cannot be generated", Log.LOG_WARNING, FAC_ORGDES);
        }
    }
  }

  /**
   * Watch for interesting property changes.  Monitors Task changes.
   * @param e The event to look at
   */
  public void propertyChanged(PropertyEvent e) { 
    String key = e.getKey().toString();

    if (key.equalsIgnoreCase(SUBTAEMS)) {
      // Subjective view
      if (getMultiTaems()) {
          Hashtable h = (Hashtable)e.getProperty();
          Enumeration en = olabels.keys();
          while (en.hasMoreElements()) {
              Object k = en.nextElement();
              if (! h.containsKey(k)) {
                  handleRemoveObjectiveStructure((String)k);
              }
          }
      }

      sdisplay = displayTaems(sdisplay, e.getProperty(), "Subjective");

    } if (key.equalsIgnoreCase(CONTAEMS)) {
      // Conditional view
      cdisplay = displayTaems(cdisplay, e.getProperty(), "Conditioned");

    } if (key.equalsIgnoreCase(OBJTAEMS)) {
      odisplay = displayTaems(odisplay, e.getProperty(), "Objective");

    } if (key.equalsIgnoreCase(SETTAEMSNAME)) {
      setSetName(((Boolean)e.getProperty()).booleanValue());

    } if (key.equalsIgnoreCase(MULTITAEMS)) {
      setMultiTaems(((Boolean)e.getProperty()).booleanValue());

    } if (key.equalsIgnoreCase(KEEPCOMMITS)) {
      setKeepCommits(((Boolean)e.getProperty()).booleanValue());

    } if (key.equalsIgnoreCase(RETARGETATTRIBUTES)) {
      setRetargetAttributes((((Boolean)e.getProperty()).booleanValue()));
    }
  }
  public void propertyRemoved(PropertyEvent e) { propertyChanged(e); }
  public void propertyAdded(PropertyEvent e) { propertyChanged(e); }
}


