/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * State.java
 *
 * manifest: Name: agent/mass/State.class
 * manifest: Depends-On: agent/simplest/State.class
 * manifest: Java-Bean: True
 ***************************************************************************************/

package agent.mass;

import java.io.*;
import java.awt.*;
import java.util.*;

import agent.base.*;
import agent.simplest.*;
import utilities.Message;
import utilities.KQMLMessage;
import utilities.Connection;
import utilities.Distribution;

/**
 * Serves as a public data repository.  Builds upon the simplest
 * State component by watching the message stream for data requests
 * and submissions from external programs.
 * <P>
 * See the parent class for more information.
 */
public class State extends agent.simplest.State implements MessageEventListener {
  static final long serialVersionUID = 1234567890L; // kludge static ser_id.  Bad bad Bryan!
  protected Communicate communicate;

  public static final String LOCATION = "Location";

  /**
   * Default constructor.
   */
  public State() {
    super();
    addDependency("Communicate");

    State.addParameterInfo(LOCATION, "String", "The pysical location of the agent, changing this value will move the agent within the local provided by Mass");
    State.addParameterInfo("Constraints", "Vector", "The environmental constraints on the agent");
  }

  /**
   * Init
   */
  public void init() {

    communicate = (Communicate)State.findComponent("Communicate");
    communicate.addMessageEventListener(this);

    super.init();
  }

  /**
   * Add support for further parsing of the constraint list.  It basically
   * converts it from a vector of strings to a vector of Constraints.
   * @param property The key and data, in a pointer
   */
  public void propertyPreprocess(StatePointer property) {
    Object key = property.key;

    super.propertyPreprocess(property);

    if (key instanceof String) {
        String skey = (String)key;
        if (skey.equalsIgnoreCase("Constraints") && (property.ptr instanceof Vector)) {
            if (!(((Vector)property.ptr).lastElement() instanceof Constraint))
                if (getRandom() != null) {
                    Vector temp = (Vector)property.ptr;
                    for (int i = 0; i < temp.size(); i++)
                        temp.setElementAt(Constraint.parseConstraint((String)temp.elementAt(i)), i);
                } else {
                    property.key = "TempConstraints";
                }

        } else if ((skey.equalsIgnoreCase("RandomSeed")) &&
                   hasProperty("TempConstraints")) {
            // We need this garbage because Random may not be set
            // when the constraints are found
            Vector temp = (Vector)getProperty("TempConstraints");
            for (int i = 0; i < temp.size(); i++)
                temp.setElementAt(Constraint.parseConstraint((String)temp.elementAt(i)), i);
            removeProperty("TempConstraints");
            setProperty("Constraints", temp);

        } else if (skey.equalsIgnoreCase("Location") ||
                   skey.equalsIgnoreCase("Display")) {
            if (property.ptr != null) {
                communicate.sendMessage(new KQMLMessage("(" + key + " " + (String)property.ptr + ")", "simulator"));	
            }
        } else if (skey.startsWith("DisplayImage")) {
            StringTokenizer tok = new StringTokenizer(skey, "-");
            if (tok.countTokens() == 2) {
                String cmd = tok.nextToken();
                String id = tok.nextToken();
                String data;
                if (property.ptr != null) {
                    data = property.ptr.toString();
                } else {
                    data = "null";
                }
                communicate.sendMessage(new KQMLMessage("(" + cmd + " " + id + " " + (String)property.ptr + ")", "simulator"));	
            }
        }
    }
  }

  /**
   * Watch for a message.  Handles property changes that arrive over
   * the network.
   * <P>
   * Currently, we will only accept state variable changes to known
   * variables.  This means that the variable must either previously
   * exist in the state table, or be a known parameter for the
   * data to be incorporated.
   * @param m The event to look at
   */
  public void messageSent(MessageEvent me) {}
  public void messageReceived(MessageEvent me) {

    // Check for other message types
    if (me.getConnection().getType() != Connection.KQML) {
      return;
    }

    KQMLMessage reply, m = (KQMLMessage)me.getMessage();
    String perf = m.getPerformative();

    if (perf.equalsIgnoreCase("tell")) {
      // Giving us a variable?
      String key = m.contentWord();
      String data = m.contentData();
      if ((key != null) && (key.length() > 0) && (hasProperty(key) || usesParameter(key))) {
	setProperty(key, data);
	log.log("Changed \"" + key + "\" to \"" + data + "\"", Log.LOG_INFO, FAC_STATE);
      }
    } else if (perf.equalsIgnoreCase("ask")) {
      // Asking us about a variable?
      String key = m.contentWord();
      String ID = "ask";
      if (hasProperty(key)) {
	reply = new KQMLMessage("reply", "(" + key + " " + (String)getProperty(key) + ")", m.getSourceAddr());
	if (m.getField("reply-with") != null ) 
                ID = m.getField("reply-with");

	reply.addField("in-reply-to", ID);
	//reply.addField(key, (String)getProperty(key));
	communicate.sendMessage(reply);
      }
    }
  }
}





