/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * TaemsAction.java
 ***************************************************************************************/

package agent.mass;

import java.io.*;
import java.util.*;
import java.awt.*;

import agent.simplest.*;
import taems.*;

/**
 * A Taems action class, basically a Method wrapper.  Use this when
 * creating Actions that are in fact derived from Taems methods.
 */
public class TaemsAction extends agent.simplest.Action implements Serializable, Cloneable {
  protected Method method = null;
  protected ScheduleElement schedelem = null;
  protected Schedule sched = null;

  /**
   * Generic constructor
   * @param m The method to base the action on
   */
  public TaemsAction(Method m) {
    super(m.getLabel());
    setMethod(m);
  }

  /**
   * Generic constructor, generally shouldn't be used
   */
  public TaemsAction(agent.simplest.Action a) {
    super(a.getName());
    setID(a.getID());
    setMethod(null);
  }

  /**
   * Schedule based constructor.  The method is obtained
   * from the ScheduleElement.
   * @param se The schedule element to perform
   * @param s The schedule it belongs to
   */
  public TaemsAction(ScheduleElement se, Schedule s) {
    this(se.getMethod());

    if (se.hasStarted())
        setStart(se.getActualStart());
    if (se.hasFinished()) {
        setFinish(se.getFinish());
        setQuality(se.getQuality());
        setTime(se.getDuration());
    }
    setScheduleElement(se);
    setSchedule(s);
  }

  /**
   * Returns the current known start time of the action.
   * @return The start time, or -1 if none specified
   */
  public int getStart() {
    if (getMethod() == null)
      return super.getStart();

    return getMethod().getStartTime();
  }

  /**
   * Sets the start time of the action
   * @param s The new start time
   */
  public void setStart(int s) {
    if (getMethod() != null)
      getMethod().setStartTime(s);

    if (getScheduleElement() != null)
      getScheduleElement().setActualStart(s);

    super.setStart(s);
  }

  /**
   * Returns the current known finish time of the action.
   * @return The finish time, or -1 if none specified
   */
  public int getFinish() {
    if (getMethod() == null)
      return super.getFinish();

    return getMethod().getFinishTime();
  }

  /**
   * Sets the finish time of the action
   * @param f The new finish time
   */
  public void setFinish(int f) {
    if (getMethod() != null)
      getMethod().setFinishTime(f);

    if (getScheduleElement() != null)
      getScheduleElement().setActualFinish(f);

    super.setFinish(f);
  }

  /**
   * Returns the current known accrued time of the action.
   * @return The accrued time, or -1 if none specified
   */
  public int getTime() {
    if (getMethod() == null)
      return super.getTime();

    return getMethod().getAccruedTime();
  }

  /**
   * Sets the accrued time of the action
   * @param t The new accrued time
   */
  public void setTime(int t) {
    if (getMethod() != null) {
      getMethod().setAccruedTime(t);
      getMethod().setCurrentDuration(t);
    }

    super.setTime(t);
  }

  /**
   * Returns the current known quality of the action.
   * @return The quality, or -1 if none specified
   */
  public float getQuality() {
    if (getMethod() == null)
      return -1;

    return getMethod().getCurrentQuality();
  }

  /**
   * Sets the quality of the action
   * @param q The new quality
   */
  public void setQuality(float q) {
    if (getMethod() != null)
      getMethod().setCurrentQuality(q);

    super.setQuality(q);
  }

  /**
   * Returns the current known quality of the action.
   * @return The quality, or -1 if none specified
   */
  public float getCost() {
    if (getMethod() == null)
      return -1;

    return getMethod().getCurrentCost();
  }

  /**
   * Sets the cost of the action
   * @param c The new cost
   */
  public void setCost(float c) {
    if (getMethod() != null)
      getMethod().setCurrentCost(c);

    super.setCost(c);
  }

  /**
   * Clone.  Does not clone method or schedule element, just copies in
   * references to them.
   */
  public Object clone() {
    TaemsAction cloned = null;

    try {
      cloned = (TaemsAction)super.clone();
    } catch (Exception e) {
      System.err.println("Clone Error: " + e);
    }

    cloned.setMethod(getMethod());
    cloned.setScheduleElement(getScheduleElement());
    cloned.setSchedule(getSchedule());

    return cloned;
  }

    /**
     * Looks first at the Action data object, if it isn't found there,
     * it looks in the Method's attribute list.
     * @param key The data item's key, typically a string
     * @return The data item, or null if not found
     */
    public synchronized Object getData(Object key) {
        Object o = super.getData(key);

        if ((o == null) && (getMethod() != null))
            return getMethod().getAttribute(key);
        else
            return o;
    }

    /**
     * Sets a specific data item in the data table. If a method
     * is available it also sets it there.
     * @param key The data item's key, typically a string
     * @return The data item, or null if not found
     */
    public synchronized void setData(Object key, Object d) {
        super.setData(key, d);

        if (getMethod() != null)
            getMethod().setAttribute(key, d);
    }

  /**
   * Get/set variable accessor functions
   */
  public Method getMethod() {
      //return getScheduleElement().getMethod();
      return method;
  }
  public void setMethod(Method m) { method = m; }

  public ScheduleElement getScheduleElement() { return schedelem; }
  public void setScheduleElement(ScheduleElement s) { schedelem = s; }
  public Schedule getSchedule() { return sched; }
  public void setSchedule(Schedule s) { sched = s; }

  public String toString() {
    //if (sched != null)
    //  return super.toString() + "\n" + sched.toString();
    //else
      return super.toString();
  }
}
