/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * Action.java
 ***************************************************************************************/

package agent.simplest;

import java.io.*;
import java.util.*;
import java.awt.*;

/**
 * Action
 */
public class Action implements Serializable, Cloneable {
  private long id;
  protected boolean done = false;
  protected String name;
  protected Hashtable data = new Hashtable(25);
  
  /**
   * Generic constructor
   */
  public Action() {
    this("");
  }

  /**
   * Another constructor
   * @param n The name of the action to perform
   */
  public Action(String n) {
    setID(State.getRandom().nextLong());
    name = n;
  }

  /**
   * Returns the name of the action
   * @return The name
   */
  public String getName() { return name; }

  /**
   * This supposedly would start the action, but doesn't do 
   * anything yet.  This will be called once when the action is to start.
   * Derived instances of the action class which actually do things
   * should override this function.
   */
  public void start() {
    done = false;
  }

  /**
   * This supposedly would execute the action, but doesn't do 
   * anything yet.  This will be called periodically after the action is
   * started. Derived instances of the action class which actually do
   * things should override this function.
   */
  public void execute() {
  }

  /**
   * This supposedly would cancel the action, but doesn't do 
   * anything yet. This will be called when an action is aborted.
   * Derived instances of the action class which actually do things
   * should override this function.
   */
  public void cancel() {
    done = true;
  }

  /**
   * This supposedly would complete the action, but doesn't do 
   * anything yet. It should be called when the Execute component
   * determines the action has completed.  Derived instances of the
   * action class which actually do things should override this function.
   */
  public void complete() { 
    done = true;
  }

  /**
   * Returns the hashtable data object.  Use this to store action
   * specific information about either the input or results of the
   * action
   * @return The data table
   */
  public Hashtable getDataTable() { return data; }

  /**
   * Returns a specific data item from the data table.
   * @param key The data item's key, typically a string
   * @return The data item, or null if not found
   */
  public synchronized Object getData(Object key) {
    return data.get(key);
  }

  /**
   * Determines if the action contains data of a particular
   * type.
   * @param key The data item's key, typically a string
   * @return True if the actions contains data corresponding
   *  to the given hey
   */
  public synchronized boolean hasData(Object key) {
    return data.containsKey(key);
  }

  /**
   * Sets a specific data item in the data table.
   * Commonly (but not always) used data items:
   * <UL>
   *   <LI>Status - status of the action
   *   <LI>Start - the start time of the action
   *   <LI>Finish - the finish time of the action
   *   <LI>Cost - the cost of the action
   *   <LI>Quality - the current quality of the action
   *   <LI>Time - the remaining time 
   *   <LI>Update - time of the last status update
   * </UL>
   * @param key The data item's key, typically a string
   * @param d The data item to be stored
   */
  public synchronized void setData(Object key, Object d) {
    data.put(key, d);
  }

  /**
   * Returns the current known start time of the action.
   * This value is normally stored in the data table.
   * @return The start time, or -1 if none specified
   */
  public int getStart() {
    Integer i = (Integer)getData("Start");

    if (i == null)
      return -1;
    else
      return i.intValue();
  }

  /**
   * Sets the start time of the action
   * @param s The new start time
   */
  public void setStart(int s) {
    setData("Start", new Integer(s));
  }

  /**
   * Returns the current known finish time of the action.
   * This value is normally stored in the data table.
   * @return The finish time, or -1 if none specified
   */
  public int getFinish() {
    Integer i = (Integer)getData("Finish");

    if (i == null)
      return -1;
    else
      return i.intValue();
  }

  /**
   * Sets the finish time of the action
   * @param f The new finish time
   */
  public void setFinish(int f) {
    setData("Finish", new Integer(f));
  }

  /**
   * Returns the current known accrued time of the action.
   * This value is normally stored in the data table.
   * @return The accrued time, or -1 if none specified
   */
  public int getTime() {
    Integer i = (Integer)getData("Time");

    if (i == null)
      return -1;
    else
      return i.intValue();
  }

  /**
   * Sets the accrued time of the action
   * @param t The new accrued time
   */
  public void setTime(int t) {
    setData("Time", new Integer(t));
  }

  /**
   * Convenience method for getting duration.  To set duration,
   * set the start and finish times.
   * @return The duration, or -1 if not known
   */
  public int getDuration() {
    int start = getStart();
    int finish = getFinish();

    if ((start != -1) && (finish != -1)) {
      return finish - start;
    }

    return -1;
  }

  /**
   * Returns the current known quality of the action.
   * This value is normally stored in the data table.
   * @return The quality, or -1 if none specified
   */
  public float getQuality() {
    Float f = (Float)getData("Quality");

    if (f == null)
      return 0;
    else
      return f.floatValue();
  }

  /**
   * Sets the quality of the action
   * @param q The new quality
   */
  public void setQuality(float q) {
    setData("Quality", new Float(q));
  }

  /**
   * Returns the current known cost of the action
   * This value is normally stored in the data table.
   * @return The cost
   */
  public float getCost() {
    Float f = (Float)getData("Cost");

    if (f == null)
      return 0;
    else
      return f.floatValue();
  }

  /**
   * Sets the cost of the action
   * @param c The new cost
   */
  public void setCost(float c) {
    setData("Cost", new Float(c));
  }

  /**
   * Returns weather or not this action is still executing
   * @return true if it is executing
   */
  public boolean isDone() { return done; }

  /**
   * Returns the unique action id number.
   * @return The id number
   */
  public long getID() { return id; }

  /**
   * Sets the action id number.
   * @param The new id number
   */
  public void setID(long i) { id = i; }

  /**
   * Clone
   */
  public Object clone() {
    Action cloned = null;

    try {
      cloned = (Action)super.clone();
    } catch (Exception e) {
      System.err.println("Clone Error: " + e);
    }

    if (getName() != null)
        cloned.name = new String(getName());
    if (getDataTable() != null)
        cloned.data = ((Hashtable)getDataTable().clone());

    return cloned;
  }

  /**
   * Returns a string representation of the action
   * @return A string
   */
  public String toString() {
    return "[Action-" + id + "] " + name + " (S:" + getStart() + " F:" + getFinish() + ")";
  }
}
