/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * ActionMonitor.java
 *
 * manifest: Name: agent/simplest/ActionMonitor.class
 * manifest: Java-Bean: True
 ***************************************************************************************/

package agent.simplest;

/* Global includes */
import java.awt.*;
import java.io.*;
import java.util.*;
import java.awt.event.*;

import agent.base.*;
import utilities.*;

/**
 * ActionMonitor is sort of a trivial learning component, which watches
 * execution performance data and attempts to divine its underlying
 * characteristics.  Currently, only duration data is evaluated.  For each
 * action performed, the component notes its duration in a running queue
 * of maximum length DataSize.  Later, other components can query for
 * this information on a per-action basis.  Two modes are currently supported,
 * average and max.  In both cases the high and low value from the data list
 * are discarded, after which either the average or maximum value is used
 * to create a Distribution reflecting the performance.  This data can then
 * be used to condition action expectations wherever necessary (e.g. changing
 * the duration fields in a Taems method outcome.)
 * <P>
 * Other data modes, and other examined data should be easy to add to this
 * component as needs dictate.
 * <P>
 * The component also supports persistant state.  The internal data structure
 * used by the component can be obtains with dumpDataCache.  Later, this
 * information can be re-read from a file, specified with the DataCache
 * property.
 */
public class ActionMonitor extends agent.base.AgentComponent implements ActionEventListener,PropertyEventListener, MouseListener  {
  static final long serialVersionUID = 1234567890L; // kludge static ser_id.  Bad bad Bryan!

  protected static final int AVERAGE = 0;
  protected static final int MAXIMUM = 1;

  protected Log log;
  protected Execute execute;
  protected State state;
  protected int FAC_AM;
  private static PopupMenu menu = new PopupMenu("Actions");
  private Hashtable durations = new Hashtable();
  private int datasize = 5;
  private int mode = AVERAGE;

  public static final String DATASIZE = "DataSize";
  public static final String DATACACHE = "DataCache";
  public static final String DATAMODE = "DataMode";

  /**
   * Default constructor
   */
  public ActionMonitor() {
    super();
    
    addDependency("Log");
    addDependency("Execute");
    addDependency("State");

    State.addParameterInfo(DATASIZE, "Integer", "Size of the data list to be used for learning", new Integer(5));
    State.addParameterInfo(DATACACHE, "String", "Location of the data cache file for historical results");
    State.addParameterInfo(DATAMODE, "String", "Data calculation mode (average,max)", new String("average"));

    addMouseListener(this);
    add(menu);
  }

  /**
   * Sets up component helpers
   */
  public void init() {
    log = (Log)agent.simplest.State.findComponent("Log");

    FAC_AM = log.getFacilityID(this);

    execute = (Execute)agent.simplest.State.findComponent("Execute");
    execute.addActionEventListener(this);
    state = (State)agent.simplest.State.findComponent("State");
    state.addPropertyEventListener(this);
    
    readConfigFile();

    super.init();
  }

  /**
   * pulsing function, use to update the display
   */
  public void pulse() {
  }

  /**
   * Resets the component
   */
  public void reset() {
      log.log("Clearing action table", Log.LOG_INFO, FAC_AM);
      durations.clear();
  }

  /**
   * 
   */
  public void end() {
  }

  /**
   * Reads in and appends data from the DataCache extended config entry
   */
    public void readConfigFile() {
        String filename = (String)state.getProperty(DATACACHE);

        if (filename == null) {
            return;
        }

        log.log("Reading data cache entries", Log.LOG_INFO, FAC_AM);
    
        try {
            BufferedReader file = state.getExtendedEntryReader(filename);

            if (file == null) {
                log.log("Error reading from directory file: " + filename, Log.LOG_ERR, FAC_AM);
                return;
            }

            int type = 0;
            while (file.ready()) {
                String str = file.readLine().trim();
	
                if (str.startsWith("#"))
                    continue;
                if (str.equals(""))
                    continue;
                if (str.equals("Duration")) {
                    type = 0;
                    continue;
                }

                String action = str.substring(0, str.indexOf(":"));
                String data = str.substring(str.indexOf(":") + 1);
                Vector v = (Vector)Converter.reTypeProperty(data, "Vector");

                if (v == null) {
                    log.log("Error parsing data cache line: " + str, Log.LOG_WARNING, FAC_AM);
                    continue;
                }

                try {
                    Enumeration e = v.elements();
                    while (e.hasMoreElements()) {
                        if (type == 0) {
                            float value = Float.parseFloat((String)e.nextElement());
                            addDurationPoint(action, value);
                        } else {
                            log.log("Unknown data cache type " + type, Log.LOG_ERR, FAC_AM);
                        }
                    }
                } catch (NumberFormatException ex) {
                    log.log("Error parsing data cache line: " + str + "\n" + ex, Log.LOG_WARNING, FAC_AM);
                    continue;
                }
            }

            file.close();

        } catch (IOException e) {
            log.log("Error: Data cache file error: " + e, Log.LOG_ERR, FAC_AM);
        }
    }

  /**
   * Adds a duration point to the list.
   * @param action The action to add data to
   * @param p The data point to add
   */
  public void addDurationPoint(String action, float p) {

      if (!durations.containsKey(action))
          durations.put(action, new Vector());

      Vector v = getDurationData(action);
      while ((v.size() >= datasize) && (!v.isEmpty()))
          v.remove(v.firstElement());
      v.add(new Float(p));

      if (log.getLogLevel(FAC_AM) >= Log.LOG_DEBUG2) {
          log.log("Learned duration for " + action + " is now: " + getDuration(action), Log.LOG_DEBUG2, FAC_AM);
      }
  }

  /**
   * Gets the observed duration distribution.
   * @param action The action to query
   * @return The duration distribution, or null if not available
   */
  public Vector getDurationData(String action) {
      return (Vector)durations.get(action);
  }

  /**
   * Gets the observed duration distribution.
   * @param action The action to query
   * @return The duration distribution, or null if not computable
   */
  public Distribution getDuration(String action) {
      return calculate(getDurationData(action));
  }

  /**
   * Gets the data
   * @param v The vector of data
   * @return The duration distribution, or null if not computable
   */
    protected Distribution calculate(Vector v) {

      switch (mode) {
      case AVERAGE:
          return calculateMiddleAverage(v);
      case MAXIMUM:
          return calculateMaximum(v);
      default:
      }

      return null;
  }

  /**
   * Calculates the distribution of a vector of Float values.
   * <P>
   * Right now this throws out the highest and lowest values, 
   * averages the remainer, and returns a single value distribution
   * of that average.
   * @param v The vector of data
   * @return The duration distribution, or null if not computable
   */
    protected Distribution calculateMiddleAverage(Vector v) {
        if (v == null) return null;
        if (v.isEmpty()) return null;

        TreeSet set = new TreeSet(v);
        set.remove(set.first());
        if (set.isEmpty()) return null;
        set.remove(set.last());

        if (!set.isEmpty()) {
            float sum = 0;
            Iterator i = set.iterator();
            while (i.hasNext())
                sum += ((Float)i.next()).floatValue();
            Distribution d = new Distribution(sum / (float)set.size(), 1.0f);
            return d;
        }
        
        return null;
    }

  /**
   * Calculates the distribution of a vector of Float values.
   * <P>
   * Right now this throws out the highest and lowest values, 
   * and returns the maximum value after that.
   * @param v The vector of data
   * @return The duration distribution, or null if not computable
   */
    protected Distribution calculateMaximum(Vector v) {
        if (v == null) return null;
        if (v.isEmpty()) return null;

        TreeSet set = new TreeSet(v);
        set.remove(set.first());
        if (set.isEmpty()) return null;
        set.remove(set.last());

        if (!set.isEmpty()) {
            float max = ((Float)set.last()).floatValue();
            Distribution d = new Distribution(max, 1.0f);
            return d;
        }
        
        return null;
    }

  /**
   * Returns an enumeration of the action (names) that we have been tracking.
   * @return An enumeration
   */
  public Enumeration getActions() {

    return durations.keys();
  }

    public void dumpDataCache() {
        Enumeration e;

        System.err.println("# Duration values");
        System.err.println("Duration");
        e = getActions();
        while (e.hasMoreElements()) {
            StringBuffer buf = new StringBuffer("");

            String action = (String)e.nextElement();
            Vector v = getDurationData(action);

            System.err.println(action + ":" + Converter.unTypeProperty(v, null));
        }
    }

  /**
   * Called when an action begins executing
   */
    public void actionStarted(ActionEvent e) { }
    public void actionAborted(ActionEvent e) { }
    public void actionCompleted(ActionEvent e) {
        Action a = e.getAction();
        addDurationPoint(a.getName(), a.getDuration());
    }

  /**
   * Watch for interesting property changes.  Monitors time changes.
   * @param e The event to look at 
   */
  public void propertyRemoved(PropertyEvent e) { }
  public void propertyAdded(PropertyEvent e) { propertyChanged(e); }
  public void propertyChanged(PropertyEvent e) { 
    String key = e.getKey().toString();

    if (key.equalsIgnoreCase(DATASIZE)) {
        datasize = ((Integer)e.getProperty()).intValue();
    } else if (key.equalsIgnoreCase(DATAMODE)) {
        String str = ((String)e.getProperty());
        if (str.equalsIgnoreCase("average"))
            mode = AVERAGE;
        else if (str.equalsIgnoreCase("max"))
            mode = MAXIMUM;
        else
            log.log("Error, unknown DataMode " + str + " was ignored", Log.LOG_ERR, FAC_AM);
    }
  }

  /**
   * Returns true if the component has a popupmenu.
   */
  public boolean hasPopupMenu() {
      return true;
  }

  /**
   * Responds to mouse clicks
   */
    public void processMouseEvent(MouseEvent e) {
        MenuItem item;

        if ((e.getID() == MouseEvent.MOUSE_PRESSED) || e.isPopupTrigger()) {
            menu.removeAll();

            // Dumper
            menu.add(item = new MenuItem("Dump Data Cache"));
            item.addActionListener(new ActionListener() {
                    public void actionPerformed(java.awt.event.ActionEvent event) {
                        dumpDataCache();
                    }
                });

            // Durations
            menu.add(item = new MenuItem("- Durations -"));
            item.setEnabled(false);
            Enumeration en = new SortedIterator(getActions());
            while (en.hasMoreElements()) {
                String action = (String)en.nextElement();
                menu.add(item = new MenuItem(action + ": " + getDuration(action)));
                item.addActionListener(new MyActionMonitorMenuListener(action));
            }
            menu.show(this, e.getX(), e.getY());
        }
    }
    public void mouseClicked(MouseEvent e) { processMouseEvent(e); }
    public void mousePressed(MouseEvent e) { processMouseEvent(e); }
    public void mouseReleased(MouseEvent e) { }
    public void mouseDragged(MouseEvent e) { }
    public void mouseEntered(MouseEvent e) { }
    public void mouseExited(MouseEvent e) { }
    public void mouseMoved(MouseEvent e) { }

    /**
     * Little class to handle data printouts
     */
    class MyActionMonitorMenuListener implements ActionListener {
        String action;
        public MyActionMonitorMenuListener(String action) {
            this.action = action;
        }
        public void actionPerformed(java.awt.event.ActionEvent event) {
            System.err.println(action + " duration: " + getDuration(action));
            if (getDurationData(action) != null)
                System.err.println(" " + getDurationData(action).toString());
            else
                System.err.println(" [no data available]");
        }
    }
}
