/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * Communicate.java
 *
 * manifest: Name: agent/simplest/Communicate.class
 * manifest: Java-Bean: True
 * manifest: Depends-On: agent/simplest/Message.class
 * manifest: Depends-On: agent/simplest/MessageEvent.class
 * manifest: Depends-On: agent/simplest/MessageListener.class
 ***************************************************************************************/

package agent.simplest;

import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.net.*;

import agent.base.*;
import utilities.*;

/**
 * This simple communications class is capable of listing on a
 * ports and maintaining several TCP connections, either accepted from
 * that server port or opened explicitly.
 * <P>
 * To use this class, you generate a Message object (from utilties)
 * and pass it to sendMessage.  You can also listen for incoming messages
 * by attaching your component as a listener to Communicate and them
 * monitoring the messageRecieved event stream.
 * <P>
 * Here's a simple example of how to do this:
 * <PRE>
 * public class MyProblemSolver extends AgentComponent implements MessageEventListener {
 *   Communicate communicate;
 * <BR>
 *   public MyProblemSolver() {
 * 
 *     // Mark MyProblemSolver as being dependent on the Communicate component
 *     addDependency("Communicate");
 *   }
 * <BR>
 *   public void init() {
 * <BR>
 *     // Find the Communicate component
 *     communicate = (Communicate)State.findComponent("Communicate");
 * <BR>
 *     // And start listening to it's events
 *     communicate.addMessageEventListener(this);
 *   }
 * <BR>
 *   public void pulse() {
 * <BR>
 *     // Create a message to sent
 *     Message m = new Message("Hello World", "joe.foo.net", 25000);
 * <I>or, to send a KQML message...</I>
 *     KQMLMessage m = new KQMLMessage("tell", "(Hello World"), "joe.foo.net", 25000);
 * <BR>
 *     // Send it
 *     communicate.sendMessage(m);
 *   }
 * <BR>
 *   public void messageReceived(MessageEvent e) {
 * <BR>
 *     // Look at incoming messages for potential responses
 *     // or new communications
 *     Message m = e.getMessage();
 * <BR>
 *     // Use m.getContent(), getSourceAddr(), etc
 *     // to view message characteristics
 *   }
 *   public void messageSent(MessageEvent e) {}
 * }
 * </PRE>
 * <P>
 * The ServerPort property specifies the port number the agent should listen
 * on for new connections.
 * <P>
 * Communicate also allows you to specify destination agents "symbolically" with the
 * RemoteHosts property.  RemoteHosts is a hashtable, where the key is a destination's
 * symbol (an agent's name, for instance), and the value is the real destination address
 * in "host:port" form.  So if you had a RemoteHosts that looked like this: 
 * "Agent1,host.foo.net:1234,Agent2,host.foo.net:2345", you could simply specify "Agent1"
 * as the destination address and Communicate will re-map it to host.foo.net:1234.
 * This allows you to get around hard coding destinations in your source or data files.
 *
 * @see utilities.Message
 * @see MessageEventListener
 */
public class Communicate extends AgentComponent implements PropertyEventListener, Runnable, MouseListener {
  static final long serialVersionUID = 1234567890L; // kludge static ser_id.  Bad bad Bryan!

  protected static final int RECIEVE_DELAY = 50;
  public static final String SEQUENCE = "sequence";

  protected ListenerVector listeners = new ListenerVector();
  protected transient ServerSocket servsocket;
  protected transient Vector oqueue;
  protected transient Vector iqueue;
  protected transient Vector connections = new Vector();
  protected transient int thread = 0;
  protected int FAC_COM;
  protected Hashtable remotes = new Hashtable();
  protected Thread recv, acpt;
  private static PopupMenu menu = new PopupMenu("Communicate");

  // Components we use
  protected Log log;
  protected State state;
  protected Control control;

  // Default values
  private String remoteaddr = null;
  private int remoteport = -1, serverport = -1;
  protected int send = 0, receive = 0;
  protected boolean queue = false, ready = false;
  protected int deftype = Connection.KQML;

  public static final String HOST = "Host";
  public static final String PORT = "Port";
  public static final String SERVERPORT = "ServerPort";
  public static final String REMOTEHOSTS = "RemoteHosts";
  public static final String CONNECTIONS = "Connections";
  public static final String QUEUEMESSAGES = "QueueMessages";

  /**
   * Constructor
   */
  public Communicate() { 
    oqueue = new Vector(10);
    iqueue = new Vector(10);

    State.addParameterInfo(HOST, "String", "If specified, the agent will automatically connect to this host address when it starts up");
    State.addParameterInfo(PORT, "Integer", "The port number of the Host to connect to (-1 = don't connect)", new Integer(-1));
    State.addParameterInfo(SERVERPORT, "Integer", "The port number this agent should listen on for new connections (-1 = don't listen)", new Integer(-1));
    State.addParameterInfo(REMOTEHOSTS, "Hashtable", "Listing of remote agent's host addresses and ports.  If a message's destination agent is in this list, the respective host address and port are used for delivery. (e.g. AgentA,foo.bar.org:5000,AgentB,foo.bar.org:5001...)");
    State.addParameterInfo(CONNECTIONS, "Hashtable", "Connection type specifications (see addConnection)");
    State.addParameterInfo(QUEUEMESSAGES, "Boolean", "If true, incoming messages will be queued and locally delivered only at pulse time in the primary control thread.  Otherwise they are delived directly from the message receiving thread.", Boolean.TRUE);

    addDependency("Log");
    addDependency("State");
    addDependency("Control");

    // Listen for mouses
    addMouseListener(this);
    add(menu);
  }

  /**
   * Closes the communication streams.
   */
  public void end() {
    flushOQueue();

    closeAllConnections();
    closeserver();

    if (log == null) return;

    log.log("Communication Summary:", Log.LOG_INFO, FAC_COM);
    log.log("           Sent: " + send, Log.LOG_INFO, FAC_COM);
    log.log("       Received: " + receive, Log.LOG_INFO, FAC_COM);
  }

  /**
   * Get/Set functions for parameters
   */
  public String getRemoteHostAddr() { return remoteaddr; }
  public void setRemoteHostAddr(String ra) {
    remoteaddr = ra;
    if (state != null) state.setProperty(HOST, ra);
  }
  public int getRemoteHostPort() { return remoteport; }
  public void setRemoteHostPort(int rp) {
    remoteport = rp;
    if (state != null) state.setProperty(PORT, new Integer(rp));
  }
  public int getServerPort() { return serverport; }
  public void setServerPort(int sp) {
    serverport = sp;
    if (state != null) state.setProperty(SERVERPORT, new Integer(sp));
  }

  /**
   * Sets up component helpers
   */
  public void init() {
    log = (Log)State.findComponent("Log");
    state = (State)State.findComponent("State");
    control = (Control)State.findComponent("Control");

    FAC_COM = log.getFacilityID(this);

    state.addPropertyEventListener(this);

    // Observation stuff
    if (State.hasComponent("Observe")) {
      Observe observe = (Observe)State.findComponent("Observe");

      observe.addObserver(new Observer("Messages Sent") {
	public boolean checkEvent(AgentEvent e) {
	  if ((e instanceof MessageEvent) && (e.getID() == MessageEvent.SENT))
	      return true;
	  return false;
	}
      });
      observe.addObserver(new Observer("Messages Received") {
	public boolean checkEvent(AgentEvent e) {
	  if ((e instanceof MessageEvent) && (e.getID() == MessageEvent.RECEIVED))
	      return true;
	  return false;
	}
      });
    }

    log.log("Communicate component initialized", Log.LOG_DEBUG2, FAC_COM);
  }

  /**
   * Starts our activity.  Opens up a local server and/or a connection to
   * a remote host (if applicable).  Starts up a thread for reading from
   * either of these conenctions.
   */
  public void begin() {
    log.log("Communicate component starting", Log.LOG_DEBUG2, FAC_COM);

    if (State.hasComponent("LogViewer")) {
        LogViewer viewer = (LogViewer)State.findComponent("LogViewer");

        viewer.addFilter(new LogViewerFilter("Msg Sent", false) {
                public boolean isMyLogEntry(String s) {
                    return s.indexOf("Message Sent: ") != -1;
                }
                public String formatMyLogEntry(String s) {
                    return "From: " + getAgent(s) + "\n" +
                        "To: " + getTo(s) + "\n" +
                        "Mesg: " + s.substring(s.indexOf('"'));
                }                
                public void drawToken(Graphics g, int x, int y, LogViewerEntrySlot v) {
                    Color c = g.getColor();
                    if (v.size() == 1) {
                        LogViewerEntry e = (LogViewerEntry)v.firstElement();
                        String to = getTo(e.getData());
                        g.setColor(LogViewer.getColor(to));
                    }
                    drawTokenType(g, x, y, 0);
                    g.setColor(c);
                }
                public void drawTokenType(Graphics g, int x, int y, int type) {
                    super.drawTokenType(g, x, y, 2);
                }
                public String getTokenType(int type) {
                    switch (type) {
                    case 0:
                        return "Message Sent (color is destination)";
                    }
                    return null;
                }
                public String getTo(String s) {
                    int i = s.indexOf("(to ") + 4;
                    return s.substring(i, s.indexOf(')', i));
                }
                public String getUID(String s) {
                    int i = s.indexOf(":uid ") + 5;
                    if (i < 0) return null;
                    return s.substring(i, s.indexOf(' ', i));
                }
                public void select(LogViewerEntrySlot v, LogViewer viewer) {
                    Enumeration e = v.elements();
                    LogViewerFilter msgrecv = findLogViewerFilter("Msg Recv");
                    if (msgrecv == null || !msgrecv.isActive()) return;
                    while (e.hasMoreElements()) {
                        LogViewerEntry entry = (LogViewerEntry)e.nextElement();
                        String uid = getUID(entry.getData());
                        if (uid != null) {
                            String agent = getTo(entry.getData());
                            TreeMap t = (TreeMap)msgrecv.getEntries(agent);
                            if (t == null) continue;

                            Iterator i = t.values().iterator();
                            while (i.hasNext()) {
                                LogViewerEntrySlot s = (LogViewerEntrySlot)i.next();
                                Enumeration se = s.elements();

                                while (se.hasMoreElements()) {
                                    LogViewerEntry sentry = (LogViewerEntry)se.nextElement();
                                    String euid = getUID(sentry.getData());
                                    if (euid != null && uid.equals(euid)) {
                                        s.setColor(Color.pink);
                                        myhighlights.addElement(s);
                                        break;
                                    }
                                }
                            }
                        }
                    }
                }
            });
        viewer.addFilter(new LogViewerFilter("Msg Recv") {
                public boolean isMyLogEntry(String s) {
                    return s.indexOf("Message Received: ") != -1;
                }
                public String formatMyLogEntry(String s) {
                    return "From: " + getFrom(s) + "\n" +
                        "To: " + getAgent(s) + "\n" +
                        "Mesg: " + s.substring(s.indexOf('"'));
                }
                public void drawToken(Graphics g, int x, int y, LogViewerEntrySlot v) {
                    Color c = g.getColor();
                    if (v.size() == 1) {
                        LogViewerEntry e = (LogViewerEntry)v.firstElement();
                        String from = getFrom(e.getData());
                        g.setColor(LogViewer.getColor(from));
                    }
                    drawTokenType(g, x, y, 1);
                    g.setColor(c);
                }
                public void drawTokenType(Graphics g, int x, int y, int type) {
                    super.drawTokenType(g, x, y, 1);
                }
                public String getTokenType(int type) {
                    switch (type) {
                    case 0:
                        return "Message Received (color is source)";
                    }
                    return null;
                }
                public String getFrom(String s) {
                    int i = s.indexOf("(from ") + 6;
                    return s.substring(i, s.indexOf(')', i));
                }
                public String getUID(String s) {
                    int i = s.indexOf(":uid ") + 5;
                    if (i < 0) return null;
                    return s.substring(i, s.indexOf(' ', i));
                }
                public void select(LogViewerEntrySlot v, LogViewer viewer) {
                    Enumeration e = v.elements();
                    LogViewerFilter msgrecv = findLogViewerFilter("Msg Sent");
                    if (msgrecv == null || !msgrecv.isActive()) return;
                    while (e.hasMoreElements()) {
                        LogViewerEntry entry = (LogViewerEntry)e.nextElement();
                        String uid = getUID(entry.getData());
                        if (uid != null) {
                            String agent = getFrom(entry.getData());
                            TreeMap t = (TreeMap)msgrecv.getEntries(agent);
                            if (t == null) continue;

                            Iterator i = t.values().iterator();
                            while (i.hasNext()) {
                                LogViewerEntrySlot s = (LogViewerEntrySlot)i.next();
                                Enumeration se = s.elements();

                                while (se.hasMoreElements()) {
                                    LogViewerEntry sentry = (LogViewerEntry)se.nextElement();
                                    String euid = getUID(sentry.getData());
                                    if (euid != null && uid.equals(euid)) {
                                        s.setColor(Color.green);
                                        myhighlights.addElement(s);
                                        break;
                                    }
                                }
                            }
                        }
                    }
                }
            });
    }

    // Connect?
    if (serverport > -1)
        openserver(serverport);
    if ((remoteaddr != null) && (remoteport > -1))
        open(remoteaddr, remoteport, deftype);
   
    // Thread 1
    recv = new Thread(this);
    recv.setName("Msg Receiver");
    recv.setDaemon(true);
    control.registerThread(recv);
    recv.start();
    Thread.yield();

    // Thread 2
    acpt = new Thread(this);
    acpt.setName("Con Listener");
    acpt.setDaemon(true);
    acpt.setPriority(acpt.getPriority() - 1);
    control.registerThread(acpt);
    acpt.start();

    log.log("Communicate component started", Log.LOG_SPAM, FAC_COM);

    ready = true;
  }

  /**
   * Resets me.  Closes both server and connection streams.
   */
  public void reset() {
    closeAllConnections();
    closeserver();
    synchronized (oqueue) {
      oqueue.removeAllElements();
    }
  }

  /**
   * The function called by the thread created in begin.  It continuously
   * tries to accept clients on a port or receive messages from a remote
   * connection.
   * @see #begin
   */
  private static final Integer tlock = new Integer(0);
  public void run() {
    boolean mess = false;

    synchronized (tlock) {
        if (thread++ == 0) mess = true;
    }

    if (mess) {
        log.log("Starting message receive loop", Log.LOG_DEBUG, FAC_COM);
        while (true) {
            try {
                if (!control.isPaused()) {
                    flushOQueue();
                    receiveMessage();
                }
            } catch (Exception e) {
                log.log("Error: " + e, Log.LOG_ERR, FAC_COM);
                e.printStackTrace();
            }
            try {
                Thread.yield();
                Thread.sleep(100);
            } catch (InterruptedException e) {
                //log.log(e.toString(), Log.LOG_ERR, FAC_COM);
            }
        }

    } else {
        log.log("Starting connection acceptance loop", Log.LOG_DEBUG, FAC_COM);
        while (true) {
            try {
                if (!control.isPaused()) {
                    accept();
                }
            } catch (Exception e) {
                log.log("Error: " + e, Log.LOG_ERR, FAC_COM);
                e.printStackTrace();
            }
            try {
                Thread.yield();
                Thread.sleep(500);
            } catch (InterruptedException e) { log.log(e.toString(), Log.LOG_ERR, FAC_COM); }
        }
    }
  }

    /**
     * Pulses me
     */
    public void pulse() {

        // Flush the iqueue
        flushIQueue();
    }

  /**
   * Adds a listener to the list
   */
  public void addMessageEventListener(MessageEventListener l) {
    //listeners.add(l, l.getMessageEventPriority());
    listeners.add(l);
  }

  /**
   * Removes a listener
   */
  public void removeMessageEventListener(MessageEventListener l) {
    listeners.remove(l);
  }

  /**
   * Sends an event
   * @param e The event to send
   */
  protected void fireEvent(AgentEvent event) {
    MessageEvent e = (MessageEvent)event;
    MessageEventListener l; 

    log.log("Firing a message event", Log.LOG_DEBUG2, FAC_COM);
    
    Enumeration enumr = listeners.elements();
    while (enumr.hasMoreElements()) {
      l = (MessageEventListener)enumr.nextElement();
      
      if (!e.isActive()) break;
      
      switch (e.getID()) {
      case MessageEvent.SENT:
	l.messageSent(e);
	break;
      case MessageEvent.RECEIVED:
	l.messageReceived(e);
	break;
      default:
	break;
      }
    }
    
    if (log != null) log.log(e, Log.LOG_INFO, FAC_COM);
    
    log.log("Fired a message event", Log.LOG_SPAM, FAC_COM);
  }

  /**
   * Opens a TCP connection to another socket.  This will connect
   * to the remote host, and add a new Connection to that host
   * to the connection list.
   * @param host The address of the remote host to connect to
   * @param port The port number to connect to
   * @param type The Connection type
   * @see utilities.Connection
   */
  public void open(String host, int port, int type) {

    if (port < 0) return;

    log.log("Trying to connect to " + host + " at port " + port, Log.LOG_INFO, FAC_COM);

    try {
      // Check to see if we have an alternate addr for this guy
      String alth = getAlternateHostAddr(host);
      int altp = getAlternateHostPort(host);
      if (alth != null)
          host = alth;
      if (altp > -1)
          port = altp;

      // Now open it
      Socket socket = new Socket(InetAddress.getByName(host), port);
      if (socket != null) {
          addConnection(new Connection(socket, type, false));

          log.log("Connected to " + host + " at port " + port, Log.LOG_DEBUG, FAC_COM);
      } else
        log.log("Error connecting to host " + host + " at port " + port, Log.LOG_ERR, FAC_COM);

    } catch (Exception e) {
        log.log("Error connecting to host " + host + " at port " + port + ": " + e, Log.LOG_ERR, FAC_COM);
        e.printStackTrace();
    }
  }

    /**
     * Returns the alternate host ip address (as specified in RemoteHosts),
     * if any.
     * @param host The host address or agent name
     * @return The alternate address, or null if none
     */
    public String getAlternateHostAddr(String host) {

      if (remotes.containsKey(host)) {
          host = (String)remotes.get(host);
          if (host.indexOf(":") != -1) {
              return host.substring(0, host.indexOf(":"));
          }
      }

      return null;
    }

    /**
     * Returns the alternate host port number (as specified in RemoteHosts),
     * if any.
     * @param host The host address or agent name
     * @return The alternate port address, or -1 if none
     */
    public int getAlternateHostPort(String host) {

      if (remotes.containsKey(host)) {
          host = (String)remotes.get(host);
          if (host.indexOf(":") != -1) {
              return Integer.parseInt(host.substring(host.indexOf(":") + 1));
          }
      }

      return -1;
    }

  /**
   * Closes a currently open socket
   * @param c The connection to close
   */
  public void closeAllConnections() {
    Enumeration e = connections.elements();

    while (e.hasMoreElements()) {
      close((Connection)e.nextElement());
    }

    connections.removeAllElements();
  }

  /**
   * Closes a currently open socket
   * @param c The connection to close
   */
  public synchronized void close(Connection c) {

    log.log("Closing the socket ", Log.LOG_DEBUG, FAC_COM);

    if (c != null)
      try {
          c.close();

      } catch (IOException e) {
          log.log(e, Log.LOG_ERR, FAC_COM);
      }
  }

  /**
   * Opens a server socket, connected to a specific port number.
   * @port The port number to listen on.  If less than 1, the function
   *   returns immediately.
   */
  public void openserver(int port) {

    if (port <= 0) return;

    log.log("Trying to open server at port " + port, Log.LOG_INFO, FAC_COM);

    if (servsocket == null)
      try {
          servsocket = new ServerSocket(port);
          log.log("Opened a server at port " + port, Log.LOG_DEBUG, FAC_COM);
      } catch (IOException e) {
          log.log("Error opening server port at " + port + ": " + e, Log.LOG_ERR, FAC_COM);
          e.printStackTrace();
      }
    else
        log.log("Error: Server socket already established", Log.LOG_ERR, FAC_COM);
  }

  /**
   * Accepts a connection on the server socket.  Stores the client
   * stream in the data structure used for connecting to remote hosts.
   * (this means that you can't talk to two hosts at once).
   */
    public void accept() {

        if (servsocket == null) return;

        log.log("Trying to accept a connection", Log.LOG_DEBUG2, FAC_COM);

        if (servsocket != null) 
            try {
                Socket socket = servsocket.accept();

                Connection c = new Connection(socket, deftype, true);
                addConnection(c);

                log.log("Got a connection", Log.LOG_DEBUG, FAC_COM);
            } catch (SocketException e) {
                log.log("Got a socket exception, exiting connection accept: " + e, Log.LOG_INFO, FAC_COM);

            } catch (IOException e) {
                log.log("Error accepting connections: " + e, Log.LOG_ERR, FAC_COM);
                e.printStackTrace();
            }
        else
            log.log("Error: attempt to accept with no server listening", Log.LOG_ERR, FAC_COM);
    }

  /**
   * Closes the server socket
   */
  public synchronized void closeserver() {
    if (servsocket != null)
      try {
          servsocket.close();
          log.log("Server (listen) connection closed.", Log.LOG_INFO, FAC_COM);
          servsocket = null;
      } catch (IOException e) {
          log.log("Error closing the server: " + e, Log.LOG_ERR, FAC_COM);
          e.printStackTrace();
      }
  }

  /**
   * Returns an enumeration of the connections the component has
   * @return The connections
   */
  public Enumeration getConnections() {
      return connections.elements();
  }

  /**
   * Finds a connection in the connection list
   * @param host The host's connection to find
   * @return The connection, or null if not found
   */
  public Connection findConnection(String host) {
    Enumeration e = connections.elements();

    while (e.hasMoreElements()) {
      Connection c = (Connection)e.nextElement();
      String name = c.getRemoteHostName();
      int port = c.getRemotePort();
      if ((name != null) && name.equalsIgnoreCase(host))
          return c;
      String altn = getAlternateHostAddr(host);
      int altp = getAlternateHostPort(host);
      if ((name != null) && (altn != null) && name.equalsIgnoreCase(altn) && (port == altp))
          return c;
    }

    return null;
  }

  /**
   * Finds the default connection in the connection list
   * @return The connection, or null if not found
   */
  public Connection findDefaultConnection() {
    Enumeration e = connections.elements();

    while (e.hasMoreElements()) {
      Connection c = (Connection)e.nextElement();
      if (c.isDefault())
          return c;
    }

    return null;
  }

  /**
   * Sets the default connection in the connection list
   * @param c The connection to make default
   */
  public void setDefaultConnection(Connection c) {
    Enumeration e = connections.elements();

    while (e.hasMoreElements()) {
      ((Connection)e.nextElement()).setDefault(false);
    }

    c.setDefault(true);

    log.log("Connection [" + c + "] set to default.", Log.LOG_INFO, FAC_COM);
  }

  /**
   * Adds a connection to Communicate's known connection list.
   * This will also use the Connections parameter to modify
   * the connection type to/from recognized hosts.
   * <P>
   * The Connections parameter, a hashtable is composed of a set of keys
   * (connection specs) and their respective object (connection pseudo-
   * protocol values).  The recognized connection specs are as follows:
   * <UL>
   * <LI>For accepted connections: "remotehost:serverport", "serverport", "remotehost"
   * <LI>For established connections: "remotehost:remoteport", "remotehost"
   * </UL>
   * If the key is a single star (*), then the associated value will be used
   * as the default connection type for any connections which are otherwise
   * unspecified.
   * <P>
   * See Connection for a list of recognized pseudo-protocol types.
   * @param c The connection to add
   * @see utilities.Connection
   * @see State#propertyPreprocess
   */
  public void addConnection(Connection c) {

      if (state.hasProperty(CONNECTIONS)) {
          Hashtable connections = (Hashtable)state.getProperty(CONNECTIONS);
          String host = c.getRemoteHostName();
          int port;
          int type = -1;

          if (host == null)
              host = "???";

          if (c.wasAccepted()) // Use local server port if accepted connection
              port = c.getLocalPort();
          else
              port = c.getRemotePort();

          if (connections.containsKey(host + ":" + port)) {
              // host:port
              type = (Integer.parseInt((String)connections.get(host + ":" + port)));
 
          } else if (c.wasAccepted() && (connections.containsKey("" + port))) {
              // accepted port
              type = (Integer.parseInt((String)connections.get("" + port)));
 
          } else if (connections.containsKey(host)) {
              // host
              type = (Integer.parseInt((String)connections.get(host)));

          } else if (connections.containsKey("*")) {
              // default '*'
              type = (Integer.parseInt((String)connections.get("*")));
          }

          if (type >= 0) {
              c.setType(type);
              log.log("Reset type of connection [" + c + "]", Log.LOG_DEBUG2, FAC_COM);
          } else
              log.log("Unspecified type for connection [" + c + "]", Log.LOG_WARNING, FAC_COM);
      }

    connections.addElement(c);

    log.log("Added connection " + c, Log.LOG_INFO, FAC_COM);
  }

  /**
   * Removes a connection from Communicate's known connection list
   * @param c The connection to add
   * @param close If true, closes the connection before removal
   */
  public void removeConnection(Connection c, boolean close) {

    if (connections.removeElement(c) && close)
      close(c);

    log.log("Removed connection " + c, Log.LOG_INFO, FAC_COM);
  }

  /**
   * Determines if it is ok to send a message yet
   * @return true if messages can be sent
   */
  public boolean ready() {

      return ready;
  }

  /**
   * Sends a message.  It first attempts to use findConnection to determine
   * which connection to send the message out on.  If no connection is
   * found, it calls open to create a Connection.  If no output stream is
   * able to be created, the  message is sent out on the first default
   * connection.
   * @param m The message you want to send
   * @return True if the message was successfully sent
   * @see utilities.Message
   * @see #findConnection
   * @see #open
   * @see #findDefaultConnection
   */
  public boolean sendMessage(Message m) {
    boolean success = false;

    log.log("Trying to send a message to " + m.getDestAddr(), Log.LOG_DEBUG, FAC_COM);

    if (ready()) {
      log.log("Sending message \"" + m.getData() + "\"", Log.LOG_DEBUG2, FAC_COM);

      // Find the right connection to send out on
      Connection c;
      c = findConnection(m.getDestAddr());
      if (c == null) {
          open(m.getDestAddr(), m.getDestPort(), deftype);
          c = findConnection(m.getDestAddr());
      }
      if (c == null)
          c = findDefaultConnection();

      // Can we send it?
      if (c != null) {
          log.log("Using connection [" + c + "] to send message.", Log.LOG_SPAM, FAC_COM);
          if ((m.getSourceAddr() == null) || (m.getSourceAddr().equals("")))
              m.setSourceAddr(c.getLocalHostName());
          if (m.getSourcePort() == 0)
              m.setSourcePort(c.getLocalPort());
          
          synchronized (c.getOutput()) {
              success = m.send(c.getOutput(), c.getDelim());
          }
      } else {
          log.log("Error, could not find appropriate connection to send message on.", Log.LOG_ERR, FAC_COM);
      }
      
      // Did we send it?
      if (success) {
          send++;
          m.setSendTime(((Integer)state.getProperty(State.TIME)).intValue());
          startFireEvent(new MessageEvent(this, m, c, MessageEvent.SENT));
      } else {
          log.log("Failed sending message: " + m, Log.LOG_ERR, FAC_COM);
      }

    } else {
        log.log("Attempted to send message before connection was established, queueing", Log.LOG_WARNING, FAC_COM);
        synchronized (oqueue) {
            oqueue.addElement(m);
        }
    }

    return success;
  }

  /**
   * This fires all the messages we've queued up while waiting for
   * a pulse
   */
  protected void flushIQueue() {
      log.log("Flushing incoming message queue (size" + iqueue.size() + ")", Log.LOG_DEBUG, FAC_COM);

      while (! iqueue.isEmpty()) {
          MessageEvent e = null;
          synchronized(iqueue) {
              e = (MessageEvent)iqueue.firstElement();
              if (!iqueue.removeElement(e))  {
                  log.log("Error removing message event from input queue", Log.LOG_ERR, FAC_COM);
              }
          }
          if (e != null)
              startFireEvent(e);
      }

      log.log("Done flushing incoming message queue", Log.LOG_SPAM, FAC_COM);
  }

  /**
   * This fires all the messages we've queued up while waiting for
   * a connection
   */
  protected void flushOQueue() {

    if (!ready()) return;

    log.log("Flushing outgoing message queue (size " + oqueue.size() + ")", Log.LOG_DEBUG, FAC_COM);

    synchronized (oqueue) {
        Enumeration e = oqueue.elements();
        Vector remove = new Vector();
        while (e.hasMoreElements()) {
            Message m = (Message)e.nextElement();
            if (!sendMessage(m))
                log.log("Warning, could not send message", Log.LOG_ERR, FAC_COM);
            else
                remove.addElement(m);
        }
        
        e = remove.elements();
        while (e.hasMoreElements()) {
            oqueue.removeElement(e.nextElement());
        }
    }

    log.log("Done flushing outgoing message queue", Log.LOG_SPAM, FAC_COM);
  }

  /**
   * Tries to receive a messagae on all available connections,
   * firing a message event if successful.
   */
  public void receiveMessage() {
    receiveMessage(false);
  }

  /**
   * Same as above, but if ret is set, the first discovered message
   * will be returned immediately after it is processed.
   * @param ret Set to true if your want the first message returned
   * @return A message (if ret), or null if none.
   */
  public Message receiveMessage(boolean ret) {
      boolean gotone = false;

      log.log("Looking for incoming messages...", Log.LOG_SPAM, FAC_COM);

    try {
        Enumeration e = connections.elements();
        while (e.hasMoreElements()) {
            Connection c = (Connection)e.nextElement();
            BufferedReader input = c.getInput();
            
            if ((input != null) && input.ready()) {
                Message m;
                
                log.log("Trying to receive a message on connection " + c, Log.LOG_DEBUG, FAC_COM);
                
                synchronized(input) {
                    //if (! input.ready()) continue;

                    switch (c.getType()) {
                    case Connection.KQML:
                        m = new KQMLMessage(KQMLMessage.receive(input, c.getDelim()));
                        break;
                        
                    case Connection.LENGTHPREFIX:
                        m = new LengthPrefixMessage(LengthPrefixMessage.receive(input));
                        break;
                        
                    case Connection.RAW:
                    default:
                        m = new Message(Message.receive(input, c.getDelim()));
                        break;
                    }
                }
                
                log.log("Finished getting message, " + m.getData().length() + " bytes read", Log.LOG_SPAM, FAC_COM);
                
                receive++;
                gotone = true;
                
                messagePreprocess(m, c);
                messagePostprocess(m, c);
                
                if (ret)
                    return m;
            }
        }
        
    } catch (IOException e) {
        log.log("Error receiving message: " + e, Log.LOG_ERR, FAC_COM);
        e.printStackTrace();
    }

    if (!gotone) {
        try { Thread.currentThread().sleep(RECIEVE_DELAY); } catch (InterruptedException e) { }
    }
    
    return null;
  }

  /**
   * Check messages that arrive before they are fired (this also fires
   * the event).
   * @param m The message event to look at
   * @param c The connection is was received on
   * @return true if the Message was fired
   */
  public boolean messagePreprocess(Message m, Connection c) {

    // Fill in the time
    m.setReceiveTime(((Integer)state.getProperty(State.TIME)).intValue());

    // Fill in remote's information
    if ((m.getSourceAddr() == null) || (m.getSourceAddr().equals("")))
      m.setSourceAddr(c.getRemoteHostName());
    if (m.getSourcePort() == 0)
      m.setSourcePort(c.getRemotePort());

    // Fill in local information
    if ((m.getDestAddr() == null) || (m.getDestAddr().equals("")))
      m.setDestAddr(c.getLocalHostName());
    if (m.getDestPort() == 0)
      m.setDestPort(c.getLocalPort());

    // Yea, set that too
    if (c == null) {
        log.log("Warning: No connection specified for message", Log.LOG_WARNING, FAC_COM);
    } else {
        m.setConnection(c);
    }

    // Fire it
    if (!queue) {
        startFireEvent(new MessageEvent(this, m, c, MessageEvent.RECEIVED));
        return true;

    } else {
    // Queue it
        log.log("Queing incoming message", Log.LOG_SPAM, FAC_COM);
        synchronized (iqueue) {
            iqueue.addElement(new MessageEvent(this, m, c, MessageEvent.RECEIVED));
        }
    }

    return false;
  }

  /**
   * Postprocess (after event firing) the message.
   * @param m The message event to look at
   * @param c The connection is was received on
   */
  public void messagePostprocess(Message m, Connection c) { }

  /**
   * Watch for interesting property changes.  Basically, this is to
   * see if anyone changes the host or port variables.
   * @param e The event to look at
   */
  public void propertyRemoved(PropertyEvent e) { }
  public void propertyChanged(PropertyEvent e) { propertyAdded(e); }
  public void propertyAdded(PropertyEvent e) {
    String key = e.getKey().toString();

    if (key.equals(PORT)) {
      // Port change
      remoteport = ((Integer)e.getProperty()).intValue();

    } else if (key.equals(HOST)) {
      // Host change
      remoteaddr = ((String)e.getProperty());

    } else if (key.equals(SERVERPORT)) {
      // ServerPort change
      serverport = ((Integer)e.getProperty()).intValue();

    } else if (key.equals(QUEUEMESSAGES)) {
      // QueueMessages change
      queue = ((Boolean)e.getProperty()).booleanValue();

    } else if (key.equals(REMOTEHOSTS)) {
      // RemoteHosts change
      remotes = (Hashtable)e.getProperty();
    } 
  }

  /**
   * Returns true if the component has a popupmenu.
   */
  public boolean hasPopupMenu() {
      return true;
  }

  /**
   * Responds to mouse clicks
   */
  public void processMouseEvent(MouseEvent e) {
    if ((e.getID() == MouseEvent.MOUSE_PRESSED) || e.isPopupTrigger()) {
        MenuItem item;

        if ((e.getID() == MouseEvent.MOUSE_PRESSED) || e.isPopupTrigger()) {
            menu.removeAll();

            item = new MenuItem("Print Incoming Messages");
            item.addActionListener(new ActionListener() {
                    public void actionPerformed(java.awt.event.ActionEvent event) {
                        Enumeration e = iqueue.elements();
                        while (e.hasMoreElements()) {
                            Message m = (Message)e.nextElement();
                            System.err.println(m);
                        }
                    }
                });
            menu.add(item);

            item = new MenuItem("Print Outgoing Messages");
            item.addActionListener(new ActionListener() {
                    public void actionPerformed(java.awt.event.ActionEvent event) {
                        Enumeration e = oqueue.elements();
                        while (e.hasMoreElements()) {
                            Message m = (Message)e.nextElement();
                            System.err.println(m);
                        }
                    }
                });
            menu.add(item);

            Enumeration en = getConnections();
            if (en.hasMoreElements()) {
                while (en.hasMoreElements()) {
                    Connection c = (Connection)en.nextElement();
                    item = new MenuItem(c.getRemoteHostName());
                    item.addActionListener(new MyConnectionMenuListener(c));
                    menu.add(item);
                }
            } else {
                item = new MenuItem("None");
                menu.add(item);
                item.setEnabled(false);
            }
            menu.show(this, e.getX(), e.getY());
        }
    }
  }
  public void mouseClicked(MouseEvent e) { processMouseEvent(e); }
  public void mousePressed(MouseEvent e) { processMouseEvent(e); }
  public void mouseReleased(MouseEvent e) { processMouseEvent(e); }
  public void mouseDragged(MouseEvent e) { processMouseEvent(e); }
  public void mouseEntered(MouseEvent e) { }
  public void mouseExited(MouseEvent e) { }
  public void mouseMoved(MouseEvent e) { }

    /**
     * Little class to handle observer printouts
     */
    class MyConnectionMenuListener implements ActionListener {
        Connection c;
        public MyConnectionMenuListener(Connection c) {
            this.c = c;
        }
        public void actionPerformed(java.awt.event.ActionEvent event) {
            System.err.println(c.toString());
        }
    }
}
