/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * Control.java
 *
 * manifest: Name: agent/simplest/Control.class
 * manifest: Java-Bean: True
 ***************************************************************************************/

package agent.simplest;

import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import java.util.zip.*;

import agent.base.*;
import utilities.Converter;

/**
 * The control class is responsible for initializing and starting the
 * components in the agent.  It also controls the thread registry and
 * can pause, restart and halt execution.  The main() function in 
 * Control is the method which actually starts the agent's execution.
 * <P>
 * This Control component provides several phases of execution.  It
 * is important to understand the ordering of the phases, and what
 * behavior limitations are assumed, to make effective and correct
 * use of them.  Details on the various phases are covered in the
 * AgentComponent class, and a brief listing appears below:
 * <P><OL>
 * <LI> <B>Construction</B> - The component constructors are called.
 *  Component dependencies and state properties should be declared here,
 *  in addition to any low level setup your component might need.
 * <LI> <B>Initialization</B> - The component init() methods are called.
 *  Components typically obtain handles to other components they
 *  will need, although they should not be used in this phase (with
 *  a few exceptions).  Any setup that can be performed independantly
 *  of other components should be finished here.  Your component
 *  may be used by others in the agent after this phase is complete.
 * <LI> <B>Begin</B> - The component begin() methods are called.  
 *  Other components may now be used.  The component's setup should
 *  be finished.
 * <LI> <B>Pulse</B> - The component's pulse() method is called.  This
 *  function will usually be called repeatedly during the agent's
 *  lifetime.  Along with event reactions, activities placed here
 *  should implement the bulk of the agent's runtime behavior.
 * <LI> <B>End</B> - The component's end() method is called.  This
 *  will be performed after the agent has decided it should complete
 *  its execution (for instance, Control.quit() is called).  Any
 *  needed completion activities should be performed here (i.e.
 *  closing file handles or network connections, consolidating
 *  statistics, etc.).
 * </OL><P>
 * You will typically not need to interact with this component, instead
 * it will be responsible for managing your component's activity behind
 * the scenes.  For more details on what your component should do and when,
 * use the AgentComponent references below.
 * <P>
 * <B>** Mandatory class</B>
 * @see agent.base.AgentComponent
 * @see agent.base.AgentComponent#init
 * @see agent.base.AgentComponent#begin
 * @see agent.base.AgentComponent#pulse
 * @see agent.base.AgentComponent#end
 * @see agent.base.AgentComponent#reset
 */
public class Control extends AgentComponent implements Runnable, ActionListener, MouseListener, PropertyEventListener {
  static final long serialVersionUID = 4712010406850428679L; // kludge static ser_id.  Bad bad Bryan!
  protected boolean dopulsing = true;
  protected boolean dopulse = false;
  protected int pulsedelay = 50;
  protected Log log;
  protected State state;
  protected int FAC_CTRL = -1;
  private transient Vector threads = new Vector(20);
  private transient Thread setup = null;
  private transient Thread controlthread = null;
  private transient static boolean started = false;
  private transient boolean hasQuit = false;
  private boolean paused = false;
  private int quittime = -1;
  private int pulsewarninglength = -1;
  private long lastpulseduration = -1;
  private PopupMenu menu = new PopupMenu("Control Functions");
  private MenuItem pauseitem;
  private static Frame agentframe;
  private static Thread shutdownHook;

  public static final String AUTOPULSE = "AutoPulse";
  public static final String PULSEWARNINGLENGTH = "PulseWarningLength";
  public static final String AUTOPULSEDELAY = "AutoPulseDelay";
  public static final String QUITTIME = "QuitTime";
  public static final String COMPONENTS = "Components";

    /**
     * Constructor.  Starts up the primary control thread which does
     * stuff with the other installed components.
     */
    public Control() {
        this(true);
    }

    /**
     * If run is set to true, then when the thread creating the Control
     * object has completed, a new Control thread will activate, taking
     * care of the remaining setup of the components and running of the
     * agent.  If run is false, then none of this will take place 
     * (presumably someone else will handle it)
     * @param run To run or not
     */
    public Control(boolean run) {
        addDependency("Log");
        addDependency("State");

        State.addParameterInfo(AUTOPULSE, "Boolean", "If true, Control will automatically pulse components.  This should be turned off if the agent's pulse cycle is driven by an external entity", Boolean.TRUE);
        State.addParameterInfo(AUTOPULSEDELAY, "Integer", "The amount of time to pause after a pulse cycle is complete", new Integer(50));
        State.addParameterInfo(QUITTIME, "Integer", "If set, the agent will quit when this time is reached");
        State.addParameterInfo(PULSEWARNINGLENGTH, "Integer", "Control will flag pulse log messages who's duration is greater that this value (-1 for no flagging)", new Integer(pulsewarninglength));

        // Handle quitting gracefully
        //System.runFinalizersOnExit(true);
        try {
            Runtime.getRuntime().addShutdownHook(shutdownHook = new Thread() {
                    public void run() {
                        Control control = (Control)agent.simplest.State.findComponent("Control");
                        log("Shutdownhook is quitting the agent...", Log.LOG_INFO, FAC_CTRL);
                        control.quit();
                    }
                });
        } catch (NoSuchMethodError ex) {
            log("Failed to addShutdownHook, you may be running an older version of java", Log.LOG_WARNING, FAC_CTRL);
        }

        // Who's running us?
        if (run) {
            setup = Thread.currentThread();
            setup.setName("Setup");
        } else {
            setup = null;
        }

        // Listen for mouses
        addMouseListener(this);

        // Make default menu
        MenuItem m;

        m = new MenuItem("Pulse");
        m.addActionListener(this);
        menu.add(m);

        pauseitem = m = new MenuItem("Pause");
        m.addActionListener(this);
        menu.add(m);

        //m = new MenuItem("Restart");
        //m.addActionListener(this);
        //menu.add(m);

        m = new MenuItem("Threads");
        m.addActionListener(this);
        menu.add(m);

        m = new MenuItem("Quit");
        m.addActionListener(this);
        menu.add(m);

        menu.addActionListener(this);
        add(menu);

        if (agentframe != null) {
            agentframe.addWindowListener(new WindowListener() {
                    public void windowActivated(WindowEvent e) {}
                    public void windowClosed(WindowEvent e) {}
                    public void windowClosing(WindowEvent e) {
                        quit();
                    } 
                    public void windowDeactivated(WindowEvent e) {}
                    public void windowDeiconified(WindowEvent e) {}
                    public void windowIconified(WindowEvent e) {}
                    public void windowOpened(WindowEvent e) {}
                });
      }

        // Finish
        if (run) {
            controlthread = new Thread(this);
            controlthread.setName("Control");
            controlthread.start();
        } else {
            started = true;
        }
    }

  /**
   * Performs dependency checks, initializes and starts up each of the
   * other components in the process after constructors have completed
   * and bean connections have been re-established.  It then
   * pulses each component forever (or at least until the quit method
   * is called).
   */
  public void run() {
    int wait = 0;

    // Wait for the init thread to complete before starting
    log("Waiting for setup thread " + setup.getName() + " to die...", Log.LOG_DEBUG, FAC_CTRL);
    try {
        setup.join();
    } catch (InterruptedException e) {
        log("Got exception while waiting for setup to die." + e, Log.LOG_ERR, FAC_CTRL);
        e.printStackTrace();
    }
    setup = null;

    // Weeee doggie!
    started = true;

    // Do something useful
    try { 
      setupComponents();

      // Auto execute
      if (dopulsing) {
          while (dopulsing) {
              if (!isPaused()) {
                  if (quittime > 0) {
                      int time = ((Integer)state.getProperty(State.TIME)).intValue();
                      if (time > quittime) {
                          log.log("Quit time of " + quittime + " reached, quitting", Log.LOG_INFO, FAC_CTRL);
                          quit();
                      }
                  }
                  pulseComponents();
              } else {
                  try { Thread.currentThread().sleep(100); } catch (InterruptedException e) { }
              }
              Thread.currentThread().yield();
          }
          
      // Directected execute
      } else {
          while (!hasQuit) {
              if (quittime > 0) {
                  int time = ((Integer)state.getProperty(State.TIME)).intValue();
                  if (time > quittime) {
                      log.log("Quit time of " + quittime + " reached, quitting", Log.LOG_INFO, FAC_CTRL);
                      quit();
                  }
              }
              if (dopulse) {
                  pulseComponents();
                  dopulse = false;
              } else {
                  try { Thread.currentThread().sleep(100); } catch (InterruptedException e) { }
              }
              Thread.currentThread().yield();
          }
      }

    } catch (DependencyFailedException e) {
        log("Error: " + e.toString(), Log.LOG_ERR, FAC_CTRL);

    } catch (Exception e) {
        log("Error: " + e, Log.LOG_ERR, FAC_CTRL);
        e.printStackTrace();
    }

    // Done?
    log("Control.run has ended.", Log.LOG_INFO, FAC_CTRL);
  }

  /**
   * Returns a reference to the control thread.  You shouldn't
   * need this.
   */
  protected Thread getControlThread() { return controlthread; }

  /**
   * Should trigger the Control thread to perform a pulse.  This
   * should only be used if AutoExecute is false.
   */
  public void doPulse() {
      if (dopulsing) {
          log.log("Warning doPulse called while auto-executing", Log.LOG_WARNING, FAC_CTRL);
      }
      dopulse = true;
  }

  /**
   * Sets up component helpers.
   */
  public void init() {
    log = (Log)State.findComponent("Log");
    state = (State)State.findComponent("State");

    state.addPropertyEventListener(this);

    FAC_CTRL = log.getFacilityID(this);
  }

    /**
     * Begin
     */
    public void begin() {

        if (State.hasComponent("LogViewer")) {
            LogViewer viewer = (LogViewer)State.findComponent("LogViewer");

            viewer.addFilter(new LogViewerFilter("Heartbeat", false) {
                    public boolean isMyLogEntry(String s) {
                        return (s.indexOf("Pulsing components") > 0);
                    }
                    public void drawToken(Graphics g, int x, int y, LogViewerEntrySlot v) {
                        Color c = g.getColor();
                        int type = 0;
                        LogViewerEntry e = (LogViewerEntry)v.firstElement();
			if (e.getData().indexOf("excessive") > 0) type = 1;
                        drawTokenType(g, x, y, type);
                    }
                    public void drawTokenType(Graphics g, int x, int y, int type) {
                        Color c = g.getColor();
                        switch (type) {
                        default:
                        case 0:
                            g.setColor(Color.black);
                            super.drawTokenType(g, x, y, 17);
                            break;
                        case 1:
                            g.setColor(Color.red);
                            super.drawTokenType(g, x, y, 19);
                            break;
                        }
                        g.setColor(c);
                    }
                    public String getTokenType(int type) {
                        switch (type) {
                        case 0:
                            return "Pulse";
                        case 1:
                            return "Long Pulse";
                        }
                        return null;
                    }
                });
        }
    }

    /**
     * Pulse
     */
    public void pulse() {
    }

  /**
   * Reset.  Kills off all the threads.
   * @see #killThreads
   */
  public void reset() {
    killThreads();
  }

  /**
   * This will probably never be called, but...
   */
  protected void finalize() {
    quit();
  }

    /**
     * Performs the initial setup phase of the agent.  This
     * first checks for dependency failures, then inits each
     * component, and finally begins each component.  After this
     * has completed, components may be pulsed as normal.
     */
    public void setupComponents() throws DependencyFailedException {
        
        checkDependencies();
        
        initComponents();
        startComponents();
    }

  /**
   * Checks for the existance of all required components and the
   * depedencies for all of the registered components. If a check fails,
   * an exception is thrown and execution is terminated.
   * @exception DependencyFailedException Thrown when a dependency
   * check fails
   */
  public void checkDependencies() throws DependencyFailedException {
    Enumeration c = State.components();
    AgentComponent component;
    String failed = "";
    String[] dependencies;
    String[] required = { "Control", "State" };

    // Check required
    for (int i = 0; i < required.length; i++) {
      if (!State.hasComponent(required[i]))
	throw new DependencyFailedException("Couldn't find required component " + required[i]);
    }

    // Check dependent
    while (c.hasMoreElements()) {
      component = (AgentComponent)c.nextElement();
      dependencies = component.getDependencies();

      for (int i = 0; i < dependencies.length; i++) {
	if (!State.hasComponent(dependencies[i]))
	  failed += dependencies[i] + " ";
      }
    }

    // Check em
    if (!failed.equals("")) {
      throw new DependencyFailedException("Couldn't find dependency target(s) " + failed);
    }
  }

  /**
   * Initializes all of the registered components.  Should only be
   * called once during agent execution, typically by the control.
   * <P>
   * Don't call this unless you know what you're doing.
   * @see agent.base.AgentComponent#init
   */
  public void initComponents() {
    Enumeration c = State.components();
    AgentComponent comp;
    String components = "";

    if (log == null)
        log = (Log)State.findComponent("Log");

    log("Initing components", Log.LOG_INFO, FAC_CTRL);

    while (c.hasMoreElements()) {
      comp = (AgentComponent)c.nextElement();
      try {
          log("Initializing " + comp.getDescriptor(), Log.LOG_INFO, FAC_CTRL);
          comp.init(); 
          log(comp.getDescriptor() + " initialized", Log.LOG_INFO, FAC_CTRL);
      } catch (Exception e) {
          log("Exception thrown while initializing " + comp.getDescriptor() + ": " + e, Log.LOG_ERR, FAC_CTRL);
          e.printStackTrace();
      }
    }
  }

  /**
   * Starts the components. (i.e., calls the begin
   * method in each of them.
   * <P>
   * I know by my conventions here that this should be called,
   * "beginComponents", but that sounds stupid.
   * <P>
   * Don't call this unless you know what you're doing.
   * @see agent.base.AgentComponent#begin
   */
  public void startComponents() {
    Enumeration c = State.components();
    AgentComponent comp;

    log("Starting components", Log.LOG_INFO, FAC_CTRL);

    while (c.hasMoreElements()) {
      comp = (AgentComponent)c.nextElement();
      try {
          log("Starting " + comp.getDescriptor(), Log.LOG_INFO, FAC_CTRL);
          comp.begin();
          log(comp.getDescriptor() + " started", Log.LOG_DEBUG, FAC_CTRL);
      } catch (Exception e) {
          log("Exception thrown while starting " + comp.getDescriptor() + ": " + e, Log.LOG_ERR, FAC_CTRL);
          e.printStackTrace();
      }
    }
  }

  /**
   * Stops the components. (i.e., calls the end
   * method in each of them.
   * <P>
   * Don't call this unless you know what you're doing.
   * @see agent.base.AgentComponent#end
   */
  public void stopComponents() {
    Enumeration c = State.components();
    AgentComponent comp;

    log("Stopping components", Log.LOG_INFO, FAC_CTRL);

    while (c.hasMoreElements()) {
      comp = (AgentComponent)c.nextElement();
      try {
          log("Stopping " + comp.getDescriptor(), Log.LOG_INFO, FAC_CTRL);
          comp.end();
          log(comp.getDescriptor() + " stopped", Log.LOG_DEBUG, FAC_CTRL);
      } catch (Exception e) {
          log("Exception thrown while stopping " + comp.getDescriptor() + ": " + e, Log.LOG_ERR, FAC_CTRL);
          e.printStackTrace();
      }
    }
  }

  /**
   * Resets each of the registered components (i.e., calls the reset
   * method in each of them
   * @see agent.base.AgentComponent#reset
   */
  public void resetComponents() {
    Enumeration c = State.components();
    AgentComponent comp;

    log("Resetting components", Log.LOG_INFO, FAC_CTRL);

    while (c.hasMoreElements()) {
      comp = (AgentComponent)c.nextElement();
      try {
          log("Resetting " + comp.getDescriptor(), Log.LOG_INFO, FAC_CTRL);
          comp.reset();
          log(comp.getDescriptor() + " reset", Log.LOG_DEBUG, FAC_CTRL);
      } catch (Exception e) {
          log("Exception thrown while resetting " + comp.getDescriptor() + ": " + e, Log.LOG_ERR, FAC_CTRL);
          e.printStackTrace();
      }
    }
  }

  /**
   * Pulses each of the registered components (i.e., calls the pulse
   * method in each of them
   * @see agent.base.AgentComponent#pulse
   */
  public void pulseComponents() {
    Enumeration c = State.components();
    AgentComponent comp;

    log("Pulsing components (last pulse " + lastpulseduration + (lastpulseduration > 0 && pulsewarninglength > 0 && lastpulseduration > pulsewarninglength ? " was excessive" : "") + ")", Log.LOG_DEBUG, FAC_CTRL);

    long start = System.currentTimeMillis();
    if (pulsePreprocess()) {
        while (c.hasMoreElements()) {
            comp = (AgentComponent)c.nextElement();
            comp.pulse();
            log(comp.getDescriptor() + " pulsed", Log.LOG_SPAM, FAC_CTRL);
        }
        pulsePostprocess();
    }
    lastpulseduration = System.currentTimeMillis() - start;
  }

  /**
   * This is called before each pulse.  If this function returns
   * false, the pulse will not be delievered to the components, nor
   * will the pulsePostProcess be called.
   * @return true if the pulse should be continued
   */
  public boolean pulsePreprocess() {
      return true;
  }

  /**
   * This is called after each pulse
   */
  public void pulsePostprocess() {
      if (pulsedelay > 0) {
	 try { Thread.currentThread().sleep(pulsedelay); } catch (InterruptedException e) { }
      }
  }

  /**
   * Resets and then starts the components.
   * @see #resetComponents
   * @see #startComponents
   */
  public void restartComponents() {

      log("Restarting components", Log.LOG_INFO, FAC_CTRL);

      resetComponents();
      startComponents();
  }

  /**
   * Components creating a thread should register it here
   * @param t The thread you create
   */
  public void registerThread(Thread t) {
    threads.addElement(t);
  }

  /**
   * Components creating a thread should unregister it here
   * @param t The thread you own
   * @return True if the thread was found and removed
   */
  public boolean unregisterThread(Thread t) {
    return threads.removeElement(t);
  }

  /**
   * Stops all non-essential threads (that have registered)
   */
  public void suspendThreads() {
    Enumeration e = threads.elements();

    log("Not suspending threads", Log.LOG_WARNING, FAC_CTRL);
    /*
    log("Suspending threads", Log.LOG_INFO, FAC_CTRL);

    while (e.hasMoreElements()) {
        ((Thread)e.nextElement()).suspend();
    }
    */
  }

  /**
   * Restarts all non-essential threads (that have registered)
   */
  public void resumeThreads() {
    Enumeration e = threads.elements();

    log("Not resuming threads", Log.LOG_WARNING, FAC_CTRL);
    /*
    log("Resuming threads", Log.LOG_INFO, FAC_CTRL);

    while (e.hasMoreElements()) {
        ((Thread)e.nextElement()).resume();
    }
    */
  }

  /**
   * Kills all non-essential threads (that have registered)
   */
  public void killThreads() {
    Enumeration e = threads.elements();
    Thread t;

    log("Not killing threads", Log.LOG_WARNING, FAC_CTRL);

    while (e.hasMoreElements()) {
      t = (Thread)e.nextElement();
      /*t.stop();
      t.destroy();
      threads.removeElement(t);
      */
    }
  }

  /**
   * Prints stack traces from all existing threads.
   */
  public void dumpThreads() {
      int count = Thread.activeCount();
      Thread threads[] = new Thread[Thread.activeCount()];
      Thread.enumerate(threads);

      for (int i = 0; i < threads.length; i++) {
          Thread t = (Thread)threads[i];
          if (t != null) {
              System.err.println(t.toString());
          }
      }
  }

  /**
   * Quits the agent program
   * @see #stopComponents
   */
  public void quit() {

    if ( ((setup == null) || (!setup.isAlive())) && (!hasQuit) ) {
      hasQuit = true;
      log("Quitting", Log.LOG_INFO, FAC_CTRL);
      dopulsing = false;
      try {
          stopComponents();
      } catch (Exception e) {
          log("Exception thrown while stopping components: " + e, Log.LOG_ERR, FAC_CTRL);
          e.printStackTrace();
      }
      if (Thread.currentThread() != shutdownHook) {
          System.runFinalization();
          System.exit(0);
      }

    } else {
        if ((setup != null) && (setup.isAlive())) {
            log("Error, cannot quit while setup thread is still active", Log.LOG_ERR, FAC_CTRL);
        } else {
            // Just someone getting overexcited about quitting...
        }
    }
  }

  /**
   * Returns weather or not the agent has actually begun the primary
   * execution stage.  Component construction is not in this stage.
   * @return true if we've started
   */
  public static boolean hasStarted() { return started; }

  /**
   * Returns weather or not the component is paused, that is, 
   * weather or not it will send pulses to individual components.
   * @return true if we're paused
   */
  public boolean isPaused() { return paused; }

  /**
   * Sets the pause status of the control component.
   * @param p Set to true to pause it
   */
  public void setPaused(boolean p) {
      if (p != isPaused()) {
          paused = p;
          if (isPaused()) {
              if (pauseitem != null)
                  pauseitem.setLabel("Unpause");
              log("Pausing...", Log.LOG_INFO, FAC_CTRL);
          } else {
              if (pauseitem != null)
                  pauseitem.setLabel("Pause");
          }
          /*
          if (isPaused()) {
              suspendThreads();
          } else {
              resumeThreads();
          }
          */
      }
  }

  /**
   * Reponds to normal menu selections
   * @param e The event to look at
   */
  public void actionPerformed(java.awt.event.ActionEvent e) {
    String command = e.getActionCommand();

    if (command.equals("Quit")) {
      quit();
    } else if (command.equals("Restart")) {
      restartComponents();
    } else if (command.equals("Pulse")) {
      pulseComponents();
    } else if (command.equals("Pause")) {
      setPaused(true);
    } else if (command.equals("Threads")) {
      dumpThreads();
    } else if (command.equals("Unpause")) {
      setPaused(false);
    }
  }

  /**
   * Returns true if the component has a popupmenu.
   */
  public boolean hasPopupMenu() {
      return true;
  }

  /**
   * Responds to mouse clicks
   */
  public void processMouseEvent(MouseEvent e) {
    if ((e.getID() == MouseEvent.MOUSE_PRESSED) || e.isPopupTrigger()) {
        menu.show(this, e.getX(), e.getY());
    }
  }
  public void mouseClicked(MouseEvent e) { processMouseEvent(e); }
  public void mousePressed(MouseEvent e) { processMouseEvent(e); }
  public void mouseReleased(MouseEvent e) { processMouseEvent(e); }
  public void mouseDragged(MouseEvent e) { processMouseEvent(e); }
  public void mouseEntered(MouseEvent e) { }
  public void mouseExited(MouseEvent e) { }
  public void mouseMoved(MouseEvent e) { }

  /**
   * Watch for interesting property changes.  Basically, this is to
   * see if anyone changes the host or port variables.
   * @param e The event to look at
   */
  public void propertyRemoved(PropertyEvent e) { }
  public void propertyChanged(PropertyEvent e) { }
  public void propertyAdded(PropertyEvent e) {
    String key = e.getKey().toString();

    if (key.equalsIgnoreCase(AUTOPULSE)) {
      // Auto pulse?
      dopulsing = ((Boolean)e.getProperty()).booleanValue();

    } else if (key.equalsIgnoreCase(AUTOPULSEDELAY)) {
      // Auto pulse delay
      pulsedelay = ((Integer)e.getProperty()).intValue();

    } else if (key.equalsIgnoreCase(PULSEWARNINGLENGTH)) {
      // Pulse warning length
      pulsewarninglength = ((Integer)e.getProperty()).intValue();

    } else if (key.equalsIgnoreCase(QUITTIME)) {
      // Quit time
      quittime = ((Integer)e.getProperty()).intValue();
    } 
  }

    /**
     * Local log function, since we frequently may not have access to
     * the normal one.
     * @param str The string to log
     * @param level The level to log at
     * @param fac The facility code to log at
     */
    protected void log(String str, int level, int fac) {

        if ((this.log != null) && (fac >= 0))
            this.log.log(str, level, fac);
        else
            utilities.Log.getDefault().log(str, level);
    }

    /**
     * Parses the contents of a configuration file (.gnt) file
     * and stores the parameters it contains in the Java System
     * property list.
     * @param r The reader to read the config file from
     */
    protected static void parseConfigFile(BufferedReader r) throws IOException {

        while (true) {
            String s = r.readLine();
            if (s == null) break;
            s = s.trim();
            if (s.startsWith("#")) continue;
            if (s.startsWith(";")) continue;
            if (s.length() == 0) continue;
                      
            if (s.indexOf(":") > 0) {
                String name = s.substring(0, s.indexOf(":")).trim();
                String value = s.substring(s.indexOf(":")+1).trim();
                if (System.getProperty(name) == null) {
                    utilities.Log.getDefault().log("Setting " + name + ":" + value, Log.LOG_DEBUG2);
                    System.setProperty(name, value);
                } else {
                    utilities.Log.getDefault().log("Keeping existing " + name + ":" + System.getProperty(name), Log.LOG_DEBUG2);              
                }
            } else {
                utilities.Log.getDefault().log("Unknown configuration line: " + s, Log.LOG_WARNING);              
            }
        }
    }

  /**
   * Starts the applet in a new thread so it will die after 
   * construction, allowing  the primary control thread to 
   * continue initialization.
   */
    public static void main(String args[]) {
        Vector v = null;
      
        // Read the configuration file from somewhere
        for (int i = 0; i < args.length; i++) {
            if (args[i].equals("--config") && (i+1 < args.length)) {
                try {
                    Reader file = new FileReader(args[i+1]);
                    parseConfigFile(new BufferedReader(file));
                  
                } catch (IOException e) {
                    utilities.Log.getDefault().log("Error reading from file " + args[i+1] + "\n" + e, Log.LOG_ERR);
                }
                i++;

            } else if (args[i].equals("--jarconfig") && (i+2 < args.length)) {
                try {
                    ZipFile jarfile = new ZipFile(args[i+1]);
                    ZipEntry jarentry = jarfile.getEntry(args[i+2]);
                    parseConfigFile(new BufferedReader(new InputStreamReader(jarfile.getInputStream(jarentry))));
                  
                } catch (IOException e) {
                    utilities.Log.getDefault().log("Error reading from jar file " + args[i+1] + "\n" + e, Log.LOG_ERR);
                }
                i+=2;

            } else if (args[i].equals("--help") || args[i].equals("-h")) {
                usage();
                
            } else {
                utilities.Log.getDefault().log("Unknown argument " + args[i], Log.LOG_ERR);
                usage();
            }
        }

        // Find the component list
        if ((System.getProperty(COMPONENTS) != null) && (v == null)) {
            String s = System.getProperty(COMPONENTS);
            v = (Vector)Converter.reTypeProperty(s, "Vector");
        }

        // Make an agent frame
        if ((v != null) && !v.isEmpty()) {
            AgentFrame f = new AgentFrame("Agent", v);
            agentframe = f;
            Thread t = new Thread(f);
            t.setName("Setup");
            t.start();

        } else {
            utilities.Log.getDefault().log("Error: no components found to load", Log.LOG_ERR);
            usage();
        }

        // This thread is dead
        utilities.Log.getDefault().log("Control main ending", Log.LOG_DEBUG);

        utilities.Log.getDefault().log("Threads I know about:", Log.LOG_DEBUG);
        Thread list[] = new Thread[20];
        int i = Thread.enumerate(list);
        for (i--; i >= 0; i--) {
            utilities.Log.getDefault().log("  " + list[i].getName(), Log.LOG_DEBUG);
        }
    }

    /**
     * Prints usage information
     */
    public static void usage() {
        System.err.println("usage: java Control [--config config.file] [--jarconfig file.jar config.file]");
        System.err.println(" --config    : Read configuration from config.file");
        System.err.println(" --jarconfig : Read configuration from config.file in file.jar");
        System.err.println(" --help      : Print this help message");
        System.err.println("The Components line in the config file specifies the components to load.");
        System.exit(0);
    }
}
