/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * Execute.java
 *
 * manifest: Name: agent/simplest/Execute.class
 * manifest: Java-Bean: True
 * manifest: Depends-On: agent/simplest/Action.class
 * manifest: Depends-On: agent/simplest/ActionEvent.class
 * manifest: Depends-On: agent/simplest/ActionListener.class
 ***************************************************************************************/

package agent.simplest;

import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;

import agent.base.*;
import utilities.*;
import agent.simplest.Log;


/**
 * Responsible for executing things.  Whenever you want your agent
 * to perform some sort of action, you should use this class, or some
 * derivation of it, to perform that action.
 * <P>
 * To use Execute, you make an Action object and ask Execute to perform
 * it for you.  Execute then does something with the action and fires off
 * an ActionEvent when something happens.  For instance, when the action
 * is finished, an actionCompleted event will be fired, which you can
 * listen for and gather execution characteristics from.
 * <P>
 * Here's a simple example of how to do this:
 * <PRE>
 * public class MyProblemSolver extends AgentComponent implements ActionEventListener {
 *   Execute execute;
 * <BR>
 *   public MyProblemSolver() {
 * <BR>
 *     // Mark MyProblemSolver as being dependent on the Execute component
 *     addDependency("Execute");
 *   }
 * <BR>
 *   public void init() {
 * <BR> 
 *     // Find the Execute component
 *     execute = (Execute)State.findComponent("Execute");
 * <BR> 
 *     // And start listening to it's events
 *     execute.addActionEventListener(this);
 *   }
 * <BR>
 *   public void pulse() {
 * <BR>
 *     // Create an action to execute
 *     Action a = new Action("My-Action-Name");
 * <BR>
 *     // Execute it
 *     execute.executeAction(a);
 *   }
 * <BR>
 *   public void actionCompleted(ActionEvent e) {
 * <BR>
 *     // Get the action and do something with it
 *     Action a = (Action)e.getAction();
 * <BR>
 *     // Use a.getDuration(), getQuality(), getCost(), etc
 *     // to view execution characteristics
 *   }
 *   public void actionStarted(ActionEvent e) {}
 *   public void actionAborted(ActionEvent e) {}
 * }
 * </PRE>
 *
 * @see Action
 * @see ActionEventListener
 */
public class Execute extends AgentComponent implements MouseListener, ActionListener {
  static final long serialVersionUID = 1234567890L; // kludge static ser_id.  Bad bad Bryan!
  private static final int NUM_ACTIONS = 10;

  protected int FAC_EXECUTE;
  protected ListenerVector listeners = new ListenerVector();
  protected Log log;
  protected State state;
  protected Vector actions;
  private static PopupMenu menu = new PopupMenu("Current Actions");

  /**
   * Default constructor
   */
  public Execute() {
    addDependency("Log");
    addDependency("State");

    // Listen for mouses
    addMouseListener(this);
    add(menu);
  }

  /**
   * Sets up component helpers
   */
    public void init() {
        log = (Log)State.findComponent("Log");
        state = (State)State.findComponent("State");
    
        FAC_EXECUTE = log.getFacilityID(this);

        // Observation stuff
        if (State.hasComponent("Observe")) {
            Observe observe = (Observe)State.findComponent("Observe");

            observe.addObserver(new Observer("Methods Started") {
                    public boolean checkEvent(AgentEvent e) {
                        if ((e instanceof ActionEvent) && (e.getID() == ActionEvent.START))
                            return true;
                        return false;
                    }
                });
            observe.addObserver(new Observer("Methods Completed") {
                    public boolean checkEvent(AgentEvent e) {
                        if ((e instanceof ActionEvent) && (e.getID() == ActionEvent.COMPLETE))
                            return true;
                        return false;
                    }
                });
            observe.addObserver(new Observer("Methods Aborted") {
                    public boolean checkEvent(AgentEvent e) {
                        if ((e instanceof ActionEvent) && (e.getID() == ActionEvent.ABORT))
                            return true;
                        return false;
                    }
                });
            /*
            observe.addObserver(new Observer("Total Method Quality") {
                    public boolean checkEvent(AgentEvent e) {
                        if ((e instanceof ActionEvent) && (e.getID() == ActionEvent.COMPLETE))
                            return true;
                        return false;
                    }
                    public void addEvent(AgentEvent e) {	  
                        if (checkEvent(e)) {
                            Action a = ((ActionEvent)e).getAction();
                            updateCount(a.getQuality());
                        }
                    }
                });
            observe.addObserver(new Observer("Total Method Cost") {
                    public boolean checkEvent(AgentEvent e) {
                        if ((e instanceof ActionEvent) && (e.getID() == ActionEvent.COMPLETE))
                            return true;
                        return false;
                    }
                    public void addEvent(AgentEvent e) {	  
                        if (checkEvent(e)) {
                            Action a = ((ActionEvent)e).getAction();
                            updateCount(a.getCost());
                        }
                    }
                });
            observe.addObserver(new Observer("Total Method Duration") {
                    public boolean checkEvent(AgentEvent e) {
                        if ((e instanceof ActionEvent) && (e.getID() == ActionEvent.COMPLETE))
                            return true;
                        return false;
                    }
                    public void addEvent(AgentEvent e) {	  
                        if (checkEvent(e)) {
                            Action a = ((ActionEvent)e).getAction();
                            updateCount(a.getDuration());
                        }
                    }
                });
            */
        } else
            log.log("Observe not found, long term observations not instantiated.", Log.LOG_WARNING, FAC_EXECUTE);

        actions = new Vector(NUM_ACTIONS);
    }

    public void begin() {

        if (State.hasComponent("LogViewer")) {
            LogViewer viewer = (LogViewer)State.findComponent("LogViewer");

            viewer.addFilter(new LogViewerFilter("Actions", false) {
                    public boolean isMyLogEntry(String s) {
                        return ((s.indexOf("Action Start: ") != -1) ||
                                (s.indexOf("Action Complete: ") != -1) ||
                                (s.indexOf("Action Abort: ") != -1));
                    }
                    public void drawToken(Graphics g, int x, int y, LogViewerEntrySlot v) {
                        if (v.size() == 1) {
                            LogViewerEntry e = (LogViewerEntry)v.firstElement();
                            int type = getType(e.getData());
                            if (type == ActionEvent.COMPLETE)
                                drawTokenType(g, x, y, 1);
                            else if (type == ActionEvent.START)
                                drawTokenType(g, x, y, 0);
                            else if (type == ActionEvent.ABORT)
                                drawTokenType(g, x, y, 2);
                        } else {
                            drawTokenType(g, x, y, 3);
                        }
                    }
                    public void drawTokenType(Graphics g, int x, int y, int type) {
                        Color c = g.getColor();
                        switch (type) {
                        case 0:
                            g.setColor(LogViewer.getColor(1));
                            super.drawTokenType(g, x, y, 0);
                            break;
                        case 1:
                            g.setColor(LogViewer.getColor(0));
                            super.drawTokenType(g, x, y, 0);
                            break;
                        case 2:
                            g.setColor(LogViewer.getColor(0));
                            super.drawTokenType(g, x, y, 4);
                            break;
                        case 3:
                            super.drawTokenType(g, x, y, 0);
                            break;
                        }
                        g.setColor(c);
                    }
                    public String getTokenType(int type) {
                        switch (type) {
                        case 0:
                            return "Action Started";
                        case 1:
                            return "Action Completed";
                        case 2:
                            return "Action Aborted";
                        }
                        return null;
                    }
                    public void select(LogViewerEntrySlot v, LogViewer viewer) {
                        Enumeration e = v.elements();
                        while (e.hasMoreElements()) {
                            LogViewerEntry entry = (LogViewerEntry)e.nextElement();
                            int type = getType(entry.getData());
                            String agent = getAgent(entry.getData());
                            if ((type == ActionEvent.COMPLETE) ||
                                (type == ActionEvent.ABORT)) {
                                String stime = getStart(entry.getData());
                                Integer time = Integer.valueOf(stime);
                                LogViewerEntrySlot s = findAction(agent, getAction(entry.getData()), ActionEvent.START, time);
                                if ((s != null) && (s.getColor() == null)) {
                                    s.setColor(Color.green);
                                    myhighlights.addElement(s);
                                }
                            } else if (type == ActionEvent.START) {
                                String stime = getTime(entry.getData());
                                Integer time = Integer.valueOf(stime);
                                LogViewerEntrySlot s = findAction(agent, getAction(entry.getData()), ActionEvent.COMPLETE, time);
                                if ((s != null) && (s.getColor() == null)) {
                                    s.setColor(Color.pink);
                                    myhighlights.addElement(s);

                                } else if (s == null) {
                                    s = findAction(agent, getAction(entry.getData()), ActionEvent.ABORT, time);
                                    if ((s != null) && (s.getColor() == null)) {
                                        s.setColor(Color.red);
                                        myhighlights.addElement(s);
                                    }
                                }
                            }
                        }
                    }
                    public LogViewerEntrySlot findAction(String agent, String action, int type, Integer time) {
                        TreeMap t = (TreeMap)getEntries(agent);
                        SortedMap map;
                        if ((type == ActionEvent.COMPLETE) ||
                            (type == ActionEvent.ABORT)) {
                            map = t.tailMap(time);
                        } else {
                            map = t.subMap(time, new Integer(time.intValue() + 1));
                        }
                        Iterator i = map.values().iterator();
                        while (i.hasNext()) {
                            LogViewerEntrySlot v = (LogViewerEntrySlot)i.next();
                            Enumeration e = v.elements();
                            while (e.hasMoreElements()) {
                                LogViewerEntry entry = (LogViewerEntry)e.nextElement();
                                if ((getAction(entry.getData()).equals(action)) &&
                                    (getType(entry.getData()) == type))
                                    return v;
                            }
                        }
                        return null;
                    }
                    public String getStart(String s) {
                        int i = s.indexOf("(S:") + 3;
                        return s.substring(i, s.indexOf(' ', i));
                    }
                    public String getAction(String s) {
                        int i = s.indexOf("[Action") + 1;
                        return s.substring(i, s.indexOf(']', i));
                    }
                    public int getType(String s) {
                        int i = s.indexOf("Action ") + 7;
                        s = s.substring(i, s.indexOf(':', i));
                        if (s.equals("Complete")) return ActionEvent.COMPLETE;
                        if (s.equals("Start")) return ActionEvent.START;
                        if (s.equals("Abort")) return ActionEvent.ABORT;
                        return -1;
                    }
                });
        }
    }

  /**
   * Resets me
   */
  public void reset() {
    cancelAllActions("Reset");
  }

  /**
   * Destructor
   */
  public void end() {
  }

  /**
   * Prints out currently executing actions
   */
  public void pulse() {

    if (isExecuting()) {
      log.log("Executing current actions:", Log.LOG_DEBUG2, FAC_EXECUTE);
      Enumeration e = actions();
      while (e.hasMoreElements()) {
          Action a = (Action)e.nextElement();
          log.log("Executing " + a, Log.LOG_DEBUG2, FAC_EXECUTE);
          a.execute();
      }
    } else {
      log.log("I'm not executing anything right now.", Log.LOG_SPAM, FAC_EXECUTE);
    }
  }

  /**
   * Adds a listener to the list
   */
  public void addActionEventListener(ActionEventListener l) {
    //listeners.add(l, l.getActionEventPriority());
    listeners.add(l);
  }

  /**
   * Removes a listener
   */
  public void removeActionEventListener(ActionEventListener l) {
    listeners.remove(l);
  }

  /**
   * Sends an action event to each of the registered listeners
   * @param e The event to send
   */
  protected void fireEvent(AgentEvent event) {
    ActionEvent e = (ActionEvent)event;
    ActionEventListener l;
    Enumeration enumr = listeners.elements();

    log.log("Firing action event", Log.LOG_DEBUG2, FAC_EXECUTE);

    while (enumr.hasMoreElements()) {
      l = (ActionEventListener)enumr.nextElement();

      if (!e.isActive()) break;

      switch (e.getID()) {
      case ActionEvent.START:
	l.actionStarted(e);
	break;
      case ActionEvent.COMPLETE:
	l.actionCompleted(e);
	break;
      case ActionEvent.ABORT:
	l.actionAborted(e);
	break;
      default:
	break;
      }
    }
    
    if (log != null)
        log.log(e, Log.LOG_INFO, FAC_EXECUTE);

    log.log("Fired an action event", Log.LOG_SPAM, FAC_EXECUTE);
  }

  /**
   * Starts to execute the given action. Does not wait for
   * action completion.
   * <P>
   * (This doesn't really do anything in this class, but
   * in a derived class it would be responsible for initiating
   * executing in some way)
   */
  public void executeAction(Action a) {

    log.log("Trying to execute action " + a, Log.LOG_DEBUG, FAC_EXECUTE);

    actions.insertElementAt(a, 0);
    a.setData("Status", "Running");
    a.setData("State", state);
    if (a.getStart() < 0)
        a.setStart(((Integer)state.getProperty(State.TIME)).intValue());
    a.start();
    startFireEvent(new ActionEvent(this, a, ActionEvent.START));

    log.log("Action " + a + " started", Log.LOG_DEBUG, FAC_EXECUTE);
  }

  /**
   * This is called when the action has been completed
   * @param a The action
   */
  public void completeAction(Action a) {
    boolean status = false;

    log.log("Trying to complete action " + a, Log.LOG_DEBUG, FAC_EXECUTE);

    status = actions.removeElement(a);
    if (status) {
      a.complete();
      a.setData("Status", "Completed");
      if (a.getFinish() < 0)
          a.setFinish(((Integer)state.getProperty(State.TIME)).intValue());
      a.setTime(a.getDuration());
      startFireEvent(new ActionEvent(this, a, ActionEvent.COMPLETE));
    } else
      log.log("Error: Tried to complete an action I don't know about", Log.LOG_ERR, FAC_EXECUTE);

    log.log("Action " + a + " completed", Log.LOG_DEBUG, FAC_EXECUTE);
  }

  /**
   * Returns a status string about an action
   * @param a The action to check up on
   * @return the action, with pertinant data items filled in
   */
  public Action statusAction(Action a) {

    a.setData("Status", "Unknown");
    a.setData("Update", state.getProperty(State.TIME));
    log.log("statusAction does not do anything right now", Log.LOG_ERR, FAC_EXECUTE);

    return a;
  }

  /**
   * Cancels the specified action.
   * <BR>
   * Note that cancelAction will actually result in an ActionEvent.ABORT
   * event, so the terminology is a bit inconsistant.
   * @param a The action
   * @param reason Why the action was cancelled
   * @return true if the action was found and cancelled
   */
  public boolean cancelAction(Action a, String reason) {
    boolean status = false;
    
    if (a == null) return false;

    log.log("Trying to cancel action " + a + " (" + reason + ")", Log.LOG_WARNING, FAC_EXECUTE);

    status = actions.removeElement(a);
    if (status) {
        status = finishCancelAction(a, reason);
        if (status)
            log.log("Action " + a + " cancelled", Log.LOG_INFO, FAC_EXECUTE);
        else
            log.log("Action " + a + " not cancelled", Log.LOG_WARNING, FAC_EXECUTE);

    } else
      log.log("Error: Tried to cancel an action I don't know about", Log.LOG_ERR, FAC_EXECUTE);

    return status;
  }

  /**
   * Finishes cancelling the specified action.
   * @param a The action
   * @param reason Why the action was cancelled
   * @return true if the action was found and cancelled
   */
  protected boolean finishCancelAction(Action a, String reason) {
    boolean status = false;
    
    if (a == null) return false;

    a.cancel();
    a.setData("Status", "Aborted");
    a.setData("AbortReason", reason);
    a.setFinish(((Integer)state.getProperty(State.TIME)).intValue());
    a.setTime(a.getDuration());
    
    startFireEvent(new ActionEvent(this, a, ActionEvent.ABORT));

    return true;
  }

  /**
   * Cancels all the currently executing actions
   * @param reason Why you are doing this
   */
  public void cancelAllActions(String reason) {
    Enumeration e = actions.elements();

    log.log("Cancelling all actions", Log.LOG_WARNING, FAC_EXECUTE);

    while(e.hasMoreElements())
      cancelAction((Action)e.nextElement(), reason);

    log.log("All actions should be cancelled now", Log.LOG_INFO, FAC_EXECUTE);
  }

  /**
   * Finds an action by name
   * @param name The name
   * @return The action, or null if not found
   */
  public Action findAction(String name) {
    Enumeration e = actions.elements();

    while (e.hasMoreElements()) {
      Action a = (Action)e.nextElement();
      if (a.getName().equalsIgnoreCase(name))
	return a;
    }

    return null;
  }

  /**
   * Finds an action by ID
   * @param id the action's id
   * @return The action, or null if not found
   */
  public Action findAction(long id) {
    Enumeration e = actions.elements();

    while (e.hasMoreElements()) {
      Action a = (Action)e.nextElement();
      if (a.getID() == id)
	return a;
    }

    return null;
  }

  /**
   * Returns an enumeration of the current actions
   * @return The enumration
   */
  public Enumeration actions() { return actions.elements(); }

  /**
   * Returns true if we're executing something
   * @return The execution state
   */
  public boolean isExecuting() { return (!actions.isEmpty()); }

    /**
     * Determines if the Execute component is "done", that is, all
     * known actions or schedules have completed.
     * @return true if Execute is done (as far as it can tell)
     */
    public boolean isDone() {
        return (!isExecuting());
    }


  /**
   * Reponds to normal menu selections
   * @param e The event to look at
   */
  public void actionPerformed(java.awt.event.ActionEvent e) {
    Object item = e.getSource();
    Object command = e.getActionCommand();

    Action a = null;
    if (item instanceof ActionMenuItem) {
        a = ((ActionMenuItem)item).getAction();
    } else {
        a = findAction(command.toString());
    }

    if (a == null)
        System.err.println("Whoops, couldn't find action " + command.toString());
    else
        System.err.println(a.toString());
  }

  /**
   * Returns true if the component has a popupmenu.
   */
  public boolean hasPopupMenu() {
      return true;
  }

  /**
   * Responds to mouse clicks
   */
  public void processMouseEvent(MouseEvent e) {
    if ((e.getID() == MouseEvent.MOUSE_PRESSED) || e.isPopupTrigger()) {
        MenuItem m;

        try {
            for (int i = 0; true; i++) {
                m = menu.getItem(i);
                m.removeActionListener(this);
            }
        } catch (ArrayIndexOutOfBoundsException ex) {
        }
        menu.removeAll();

        if (isExecuting()) {
            Enumeration en = actions();
            while (en.hasMoreElements()) {
                Action a = (Action)en.nextElement();
                m = new ActionMenuItem(a);
                m.addActionListener(this);
                menu.add(m);
            }
        } else {
            m = new MenuItem("None");
            menu.add(m);
        }

       menu.show(this, e.getX(), e.getY());
    }
  }
  public void mouseClicked(MouseEvent e) { processMouseEvent(e); }
  public void mousePressed(MouseEvent e) { processMouseEvent(e); }
  public void mouseReleased(MouseEvent e) { processMouseEvent(e); }
  public void mouseDragged(MouseEvent e) { processMouseEvent(e); }
  public void mouseEntered(MouseEvent e) { }
  public void mouseExited(MouseEvent e) { }
  public void mouseMoved(MouseEvent e) { }

    /**
     * Used to hold the data needed for the action menu
     */
    class ActionMenuItem extends MenuItem {
        Action action;
        public ActionMenuItem(Action a) {
            super(a.getName());
            action = a;
            setLabel(getLabel() + " (" + action.getStart() + ")"); 
        }
        public Action getAction() { return action; }
    }

}






