/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * Log.java
 *
 * manifest: Name: agent/simplest/Log.class
 * manifest: Java-Bean: True
 ***************************************************************************************/

package agent.simplest;

import java.io.*;
import java.awt.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;

import agent.base.*;

import utilities.VFlowLayout;

/**
 * Use this class as a pass-through for logging messages.  The Log class
 * (and all subclasses) should be designed in such a manner that you can
 * always send log messages to them, once you have a handle to a Log
 * instance. Log with impunity!
 * <P>
 * Elements to be logged should be assigned an appropriate level to be
 * logged at.  The general idea is that the lower the level, the more
 * important the message is.  Then, at runtime, you can specify a global
 * LogLevel which will determine which messages are actually recorded
 * (i.e. a message is recorded if its level is <= the global level).
 * <P>
 * Some guidelines for assigning log levels:
 * <UL>
 *  <DD>LOG_NONE (-1) - nothin, nada, zilch (log level only)
 *  <DD>LOG_ERR (0) - unexpected runtime errors
 *  <DD>LOG_WARNING (1) - essential operation notices
 *  <DD>LOG_INFO (2) - runtime activity messages
 *  <DD>LOG_DEBUG (3) - less important activity messages
 *  <DD>LOG_DEBUG2 (4) - periodic or trace messages
 *  <DD>LOG_SPAM (5) - spam<BR>
 * </UL>
 * <P>
 * Log items may also be given a log "facility" code, which is essentially
 * a subject flag describing what sort of topic the log item pertains to.
 * <P>
 * By default, logging is done with the "General" facility, but you can
 * define your own in your source code.  To do this, pick a descriptive
 * string indicative of the data you want to flag.  Use this string
 * as a parameter to getFacilityID, which will return a numerical id
 * for you to use as a parameter to the log functions.  Later, you
 * can define a per-facility logging scheme allowing you to define
 * different log levels for each facility (using the LogFacility property
 * in your config file or the Log Facilities slider control panel).
 * <P>
 * Here's an example:
 * <PRE>
 * protected Log log;
 * protected int FAC_MYFAC;
 *
 * public void init() {
 *   // First find Log
 *   log = (Log)State.findComponent("Log");
 *
 *   // Then get your code
 *   FAC_MYFAC = log.getFacilityID(this);
 * <I>or</I>
 *   FAC_MYFAC = log.getFacilityID("MyAlternateName");
 *
 *   // Then use it
 *   log.log("I'm logging at the error level", Log.LOG_ERR, FAC_MYFAC);
 *   log.log("I'm logging at the information level", Log.LOG_INFO, FAC_MYFAC);
 *   log.log("I'm spamming up the airwaves", Log.LOG_SPAM, FAC_MYFAC);
 * }
 * </PRE>
 * <P>
 * Facility codes may be specified in the LogFacilities parameter.  For example:
 * <BR>
 * &nbsp;<TT>LogFacilities: Communicate,2,Execute,3</TT>
 * <BR>
 * would set the log level for Communicate to 2 and Execute to 3.
 * <P>
 * All events using the log(AgentEvent) method are logged with the
 * "Event" facility code.
 */
public class Log extends AgentComponent implements PropertyEventListener {
  static final long serialVersionUID = 1234567890L; // kludge static ser_id.  Bad bad Bryan!
  protected ListenerVector listeners = new ListenerVector();

  /**
   * Logging constants, use them if you want
   */
  public static final int LOG_NONE    = -1;
  public static final int LOG_EMERG   = 0;
  public static final int LOG_ALERT   = 0;
  public static final int LOG_CRIT    = 0;
  public static final int LOG_ERR     = 0;
  public static final int LOG_WARNING = 1;
  public static final int LOG_INFO    = 2;
  public static final int LOG_DEBUG   = 3;
  public static final int LOG_DEBUG2  = 4;
  public static final int LOG_SPAM    = 5;

  /**
   * Error threshold
   */
  public static final int ERROR_THRESH = LOG_ERR;

  /**
   * Flush threshold
   */
  public static final int FLUSH_THRESH = ERROR_THRESH;

  /**
   * Facility/ log level stuff
   */
  private static int ID = 0;
  private int FAC_GENERAL = 0, FAC_EVENT = 0;
  protected int FAC_LOG = 0;
  private static final int MAX_FAC = 1000;
  public static int LEV_EVENT = LOG_DEBUG;
  public static int LEV_GENERAL = LOG_INFO;
  protected static Hashtable facilities = new Hashtable();
  protected static String facnames[] = new String[MAX_FAC];
  protected static int levels[] = new int[MAX_FAC];

  private Container faccon = null;
  Vector factemp = new Vector();
  Hashtable slides = new Hashtable();

  private Label label = new Label("                                                    ", Label.LEFT);
  private utilities.Log utilitieslog = utilities.Log.getDefault();
  protected transient PrintWriter output;
  private int length;
  protected boolean complex = true;
  protected boolean audible;
  private boolean flush;
  private ScreenLog screenlog = null;
  protected String name = "";
  protected State state = null;
  protected long time = Calendar.getInstance().getTime().getTime();
  protected int clock;
  protected boolean logevents = false;
  protected boolean logmemory = false;
  protected boolean logtime = false;
  protected boolean logcontrolonly = false;
  protected int logdelaythresh = -1;
  protected LogItem lastlogitem = null;
  protected boolean linebreaks = true;
  protected boolean gui = false;
  private Runtime runtime = Runtime.getRuntime();
  private Calendar cal = Calendar.getInstance();
  private long lasttime = 0, thistime;
  protected Hashtable files = new Hashtable();
  protected Control control;

  public static final String LOGLEVEL = "LogLevel";
  public static final String LOGFLUSH = "LogFlush";
  public static final String LOGLENGTH = "LogLength";
  public static final String LOGFACILITIES = "LogFacilities";
  public static final String LOGCOMPLEX = "LogComplex";
  public static final String LOGTIME = "LogTime";
  public static final String LOGMEMORY = "LogMemory";
  public static final String LOGEVENTS = "LogEvents";
  public static final String LOGGUI = "LogGUI";
  public static final String AUDIBLEERRORS = "AudibleErrors";
  public static final String USEANSI = "UseANSI";
  public static final String LOGLINEBREAKS = "LogLinebreaks";
  public static final String LOGDELAYTHRESH = "LogDelayThresh";
  public static final String LOGCONTROLONLY = "LogControlOnly";

  /**
   * Default constructor
   */
  public Log() {
    setLogStream(System.err);

    State.addParameterInfo(LOGLEVEL, "Integer", "The default log level value, used for the general log level and facilities who's levels' are unspecified", new Integer(LOG_INFO));
    State.addParameterInfo(LOGFLUSH, "Boolean", "If true, the log stream will be flushed after each logging action", Boolean.TRUE);
    State.addParameterInfo(LOGLENGTH, "Integer", "Maximum permitted length for log lines (0 for no limit)", new Integer(500));
    State.addParameterInfo(LOGFACILITIES, "Hashtable", "Per-component facility log specifications (e.g. Execute,2,State,3...)");
    State.addParameterInfo(LOGCOMPLEX, "Boolean", "If true, extra information is prepended to each log message", Boolean.TRUE);
    State.addParameterInfo(LOGTIME, "Boolean", "If true, includes wall clock time with each log message", Boolean.FALSE);
    State.addParameterInfo(LOGMEMORY, "Boolean", "If true, the agent's current memory information is logged with each log message", Boolean.FALSE);
    State.addParameterInfo(LOGEVENTS, "Boolean", "Generate LogEvents for each logging action.", Boolean.FALSE);
    State.addParameterInfo(LOGGUI, "Boolean", "Update Log component GUI with log messages", Boolean.FALSE);
    State.addParameterInfo(AUDIBLEERRORS, "Boolean", "Plays a system beep when an error occurs", Boolean.TRUE);
    State.addParameterInfo(USEANSI, "Boolean", "ANSI-colorize textual output.", Boolean.TRUE);
    State.addParameterInfo(LOGLINEBREAKS, "Boolean", "If true, linebreak are permitted in log events, otherwise logged data will be condensed onto one line", Boolean.TRUE);
    State.addParameterInfo(LOGDELAYTHRESH, "Integer", "If LogTime is true, log items will always be shown if their delay is above this threshold (if > -1)", new Integer(logdelaythresh));
    State.addParameterInfo(LOGCONTROLONLY, "Boolean", "If true, only log events from the control thread will be processed", new Boolean(logcontrolonly));

    addDependency("State");

    if (utilitieslog != null)
	utilitieslog.setName("JAF");

    // Visual display junk
    add(new Label("Status:", Label.RIGHT));
    add(label);
    setLabel("Un-updated");

    // Set up event facilities
    if (this.getClass().getName().equals("agent.simplest.Log")) {
      FAC_LOG = getFacilityID(this);
      FAC_GENERAL = getFacilityID("General");
      FAC_EVENT = getFacilityID("Event");
      updateAllLogLevels();
    }
  }

  /**
   * Init
   */
  public void init() {
    state = (State)State.findComponent("State");
    if (State.hasComponent("ScreenLog"))
      screenlog = (ScreenLog)State.findComponent("ScreenLog");
    control = (Control)State.findComponent("Control");

    state.addPropertyEventListener(this);
    updateAllLogLevels();

    if (hasGUI()) {
        // Facility window
        JFrame window;
        if (State.hasComponent("WindowManager")) {
            WindowManager wm = (WindowManager)state.findComponent("WindowManager");
            window = wm.createWindow("Log Facilities");
        } else
            window = new JFrame("Log Facilities");
        JScrollPane scroll = new JScrollPane();
        faccon = new JPanel();
        window.getContentPane().add(scroll);
        scroll.setViewportView(faccon);
        faccon.setLayout(new VFlowLayout(true));
        Enumeration e = factemp.elements();
        while (e.hasMoreElements()) {
            faccon.add((Component)e.nextElement());
        }
        //window.pack();
        window.setSize(new Dimension(450, 300));
    }
  }

  /**
   * Begin
   */
  public void begin() {
    printLogLevels();

        if (State.hasComponent("LogViewer")) {
            LogViewer viewer = (LogViewer)State.findComponent("LogViewer");

            viewer.addFilter(new LogViewerFilter("Error/Warn", true) {
                    public boolean isMyLogEntry(String s) {
                        int level = getLogLevel(s);
                        return ((level == Log.LOG_ERR) ||
                                (level == Log.LOG_WARNING));
                    }
                    public void drawToken(Graphics g, int x, int y, LogViewerEntrySlot v) {
                        Color c = g.getColor();
                        Enumeration e = v.elements();
                        int type = 0;
                        while (e.hasMoreElements()) {
                            LogViewerEntry entry = (LogViewerEntry)e.nextElement();
                            if (getLogLevel(entry.getData()) == Log.LOG_ERR) {
                                type = 1;
                                break;
                            }
                        }
                        drawTokenType(g, x, y, type);
                    }
                    public void drawTokenType(Graphics g, int x, int y, int type) {
                        Color c = g.getColor();
                        switch (type) {
                        case 1:
                            g.setColor(LogViewer.getColor(0));
                            g.fillOval(x - 3, y - 11, 7, 7);
                            break;
                        default:
                        case 0:
                            g.setColor(LogViewer.getColor(6));
                            super.drawTokenType(g, x, y, 0);
                        }
                        g.setColor(c);
                    }
                    public String getTokenType(int type) {
                        switch (type) {
                        case 1:
                            return "Error";
                        case 0:
                            return "Warning";
                        }
                        return null;
                    }
                    public int getLogLevel(String s) {
                        int i = s.indexOf("]");
                        i = s.lastIndexOf("(", i);
                        if ((i < 0) || (i+1 >= s.length())) {
                            if (s.indexOf("java.lang.") != -1) {
                                if (s.indexOf("Exception") != -1 ||
                                    s.indexOf("Error") != -1)
                                    return Log.LOG_ERR;
                            }
                            return -1;
                        }
                        if (s.length() <= i+3) return -1;
                        char c = s.charAt(i+1);
                        if (s.charAt(i+2) != ')') return -1;
                        if (s.charAt(i+3) != ']') return -1;
                        if (Character.isDigit(c))
                            return Character.getNumericValue(c);
                        return -1;
                    }
                });
        }
  }

  /**
   * Destructor
   */
  protected void finalize() {
    closeLogStream();
  }

  /**
   * De-serialization function that resets the log stream
   */
  private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
    in.defaultReadObject();
    setLogStream(System.err);
  }

  /**
   * Adds a listener to the list
   */
  public void addLogEventListener(LogEventListener l) {
    //listeners.add(l, l.getLogEventPriority());
    listeners.add(l);
  }

  /**
   * Removes a listener
   */
  public void removeLogEventListener(LogEventListener l) {
    listeners.remove(l);
  }

  /**
   * Sends an event
   * @param e The event to send
   */
  protected void fireEvent(AgentEvent event) {
    LogEvent e = (LogEvent)event;
    LogEventListener l;
    Enumeration enumr = listeners.elements();

    while (enumr.hasMoreElements()) {
      l = (LogEventListener)enumr.nextElement();

      if (!e.isActive()) break;

      switch (e.getID()) {
      case LogEvent.ERROR:
          l.errorLog(e);
          if (screenlog != null)
              screenlog.log(e);
          break;

      case LogEvent.EVENT:
          AgentEvent ae = (AgentEvent)e.getItem().getObject();
          if (ae.getSource() != l) {
              // Avoid spamming the originator
              l.eventReflection(ae);
          }
          l.eventLog(e);
          break;

      case LogEvent.STRING:
          l.stringLog(e);
          break;

      case LogEvent.OBJECT:
          l.objectLog(e);
          break;

      default:
          break;
      }
    }
  }

  /**
   * General log level accessors
   */
  public int getLogLevel() { return LEV_GENERAL; }
  public void setLogLevel(int l) { 
    if (LEV_GENERAL != l) {
      setLogLevel(FAC_GENERAL, l);
      updateAllLogLevels();
      if (state != null) state.setProperty(LOGLEVEL, new Integer(LEV_GENERAL));
    }
  }

  /**
   * Updates all the levels, using the general level and the faciilities
   * table, if any
   */
    public void updateAllLogLevels() {

        // Get the facility level table
        Hashtable faclevels = null;
        if (state == null)
            state = (State)State.findComponent("State");
        if (state != null)
            faclevels = (Hashtable)state.getProperty(LOGFACILITIES);

        // Set all the levels to default
        for (int i = 0; i < levels.length; i++) {
            String fac = getFacilityByID(i);
            if (fac != null) {
                if ((faclevels != null) && faclevels.containsKey(fac)) {
                    try {
                        int lev = Integer.parseInt((String)faclevels.get(fac));
                        setLogLevel(i, lev);
                        
                    } catch (NullPointerException ee) {
                        log("Error setting level for facility \"" + fac + "\"", LOG_ERR, FAC_LOG);
                        
                    } catch (NumberFormatException ee) {
                        log("Error: Facility \"" + fac + "\" did not have a parseable number value", LOG_ERR, FAC_LOG);
                    }

                } else {
                    if (i == FAC_EVENT)
                        setLogLevel(i, LEV_EVENT);
                    else
                        setLogLevel(i, LEV_GENERAL);
                }
            } else
                break;
        }

        /*
        // Set the event level
        setLogLevel(FAC_EVENT, LEV_EVENT);

        // Set the facility levels
        if (faclevels != null) {
            Enumeration e = faclevels.keys();
            while (e.hasMoreElements()) {
                String f = (String)e.nextElement();
                try {
                    int lev = Integer.parseInt((String)faclevels.get(f));
                    setLogLevel(getFacilityID(f), lev);

                } catch (NullPointerException ee) {
                    log("Error setting level for facility \"" + f + "\"", LOG_ERR, FAC_LOG);

                } catch (NumberFormatException ee) {
                    log("Error: Facility \"" + f + "\" did not have a parseable number value", LOG_ERR, FAC_LOG);
                }
            }
            }*/

        log("All log levels updated (General is " + levels[FAC_GENERAL] + ")", LOG_INFO, FAC_LOG);
    }

  /**
   * Prints out the log level information
   */
  public void printLogLevels() {

    // Set the facility levels
    log("Registered log levels:", LOG_INFO, FAC_LOG);
    if (facilities != null) {
      Enumeration e = facilities.keys();
      while (e.hasMoreElements()) {
          String f = (String)e.nextElement();
          int id = getFacilityID(f);
          log("  Level[" + f + "(" + id + ")] = " + getLogLevel(id), LOG_DEBUG, FAC_LOG);
      }
    }
  }

  /**
   * Sets a log level for a certain facility id
   * @param id The facility id
   * @param l The new log level
   */
  public void setLogLevel(int id, int l) {

      if (levels[id] != l) {
          levels[id] = l;
          log("Log level for facility " + id + " is now " + l, LOG_DEBUG, FAC_LOG);

          // Known facs
          if ((id == FAC_GENERAL) && (utilitieslog != null)) {
              LEV_GENERAL = l;
              utilitieslog.setLevel(LEV_GENERAL);
          } else if (id == FAC_EVENT) {
              LEV_EVENT = l;
          }
          
          JSlider slide = (JSlider)slides.get(new Integer(id));
          if (slide != null) {
              if (slide.getValue() != l)
                  slide.setValue(l);
          }
      }
  }

  /**
   * Gets a log level for a certain facility id
   * @param id The facility id
   * @return The log level, or -1 if not found
   */
  public int getLogLevel(int id) {

      if ((id >= 0) && (id < levels.length))
          return levels[id];
      else
          return -1;
  }

  /**
   * Sets a log level for a certain facility code.
   * @param sid The facility code text representation
   * @param l The new log level
   */
  public void setLogLevel(String sid, int l) {

    setLogLevel(getFacilityID(sid), l);
  }

  /**
   * Gets a log level for a certain facility code
   * @param sid The facility code text representation
   * @return The log level, or -1 if not found
   */
  public int getLogLevel(String sid) {

      return getLogLevel(getFacilityID(sid));
  }

  /**
   * Log item length accessors.  This limits the character length
   * of the log items.
   */
  public int getLogLength() { return length; }
  public void setLogLength(int l) { 
    length = l;
    if (state != null) state.setProperty(LOGLENGTH, new Integer(length));
  }

  /**
   * Flush accessors
   */
  public boolean getFlush() { return flush; }
  public void setFlush(boolean f) { 
    flush = f;
    if (state != null) state.setProperty(LOGFLUSH, new Boolean(flush));
    if (flush) flush();
  }

  /**
   * Gets a facility ID number for an AgentComponent by removing the
   * java path prefix (e.g. agent.simplest.Log == Log).  Note that if
   * you have two components with the same suffix, their names will
   * clash.
   * @param c The component
   * @return The ID number
   */
  public int getFacilityID(AgentComponent c) {

    return getFacilityID((Object)c);
  }

  /**
   * Gets a facility ID number for an Object by removing the
   * java path prefix (e.g. agent.simplest.Log == Log).  Note that if
   * you have two objects with the same suffix, their names will
   * clash.
   * @param c The component
   * @return The ID number
   */
  public int getFacilityID(Object o) {
    String f = o.getClass().getName();

    f = f.substring(f.lastIndexOf(".") + 1);

    return getFacilityID(f);
  }

  /**
   * Gets a facility ID number, which is used as a parameter to the actual
   * logging functions.  If one does not exist, an ID is created
   * for it. 
   * <P>
   * Do not to call this for every log call you make, as it may
   * slow things down to an unacceptable level - it is preferable
   * to use this function once (from init, for instance) and cache
   * the result in an instance variable within your component
   * for future use.
   * @param f The facility's name
   * @return The ID number
   */
    public synchronized int getFacilityID(String f) {
        Integer key;

        // Find the id
        key = (Integer)facilities.get(f);

        // Or create it if none
        if (key == null) {
            if (ID >= (MAX_FAC - 1)) {
                log("Error: Maximum number of facilities exceeded, using General", LOG_ERR, FAC_LOG);
                return getFacilityID("General");
            }
            key = new Integer(ID++);
            facilities.put(f, key);
            facnames[key.intValue()] = f;
            levels[key.intValue()] = -999;
            updateAllLogLevels();

            if (hasGUI()) {
                // Make selection item
                JPanel item = new JPanel();
                item.setLayout(new FlowLayout(FlowLayout.RIGHT));
                item.setAlignmentX(1);
                JSlider slide = new JSlider(JSlider.HORIZONTAL, -1, 5, levels[key.intValue()]);
                slide.setAlignmentX(1);
                slide.setName(f);
                slide.addChangeListener(new ChangeListener() {
                        public void stateChanged(ChangeEvent e) {
                            setLogLevel(((JSlider)e.getSource()).getName(), ((JSlider)e.getSource()).getValue());
                        }
                    });
                slide.setSnapToTicks(true);
                slide.setMajorTickSpacing(1);
                slide.setPaintTicks(true);
                slide.setPaintLabels(true);
                Dictionary d = new Hashtable();
                d.put(new Integer(LOG_NONE), new JLabel("No"));
                d.put(new Integer(LOG_ERR), new JLabel("Er"));
                d.put(new Integer(LOG_WARNING), new JLabel("Wr"));
                d.put(new Integer(LOG_INFO), new JLabel("In"));
                d.put(new Integer(LOG_DEBUG), new JLabel("D1"));
                d.put(new Integer(LOG_DEBUG2), new JLabel("D2"));
                d.put(new Integer(LOG_SPAM), new JLabel("Sp"));
                slide.setLabelTable(d);
                slides.put(key, slide);
                item.add(new JLabel(f + " (" + key + ")"));
                item.add(slide);
                if (faccon == null)
                    factemp.addElement(item);
                else
                    faccon.add(item);
            }
      
            log("New log facility created: \"" + f + "\", level " + levels[key.intValue()], LOG_INFO, FAC_LOG);
        }

        return key.intValue();
    }

  /**
   * Returns a facility's string name, given an id number
   * @return The name, or null if not found
   */
  public String getFacilityByID(int i) {

    if (facilities != null) {
        if ((i >= 0) && (i < facnames.length))
            return facnames[i];
    }

    return null;
  }

  /**
   * Set the logging stream
   * @param s The stream you want log messages to go to
   */
  public void setLogStream(OutputStream s) {

    closeLogStream();
    output = new PrintWriter(s, false);
  }

  /**
   * Closes the logging stream, if there is one
   */
  public void closeLogStream() {

    if (output != null) {
      flush();
      output.close();
      output = null;
    }
  }

  /**
   * Logs an event, using the default event logging level, or the Event
   * facility's level if re-specified.
   * <P>
   * To facilitate agent construction, we encourage you to automatically
   * log events from your fireEvent method, rather than manually
   * linking the event stream to this method.  Assuming that your component
   * has a local Log reference called "log", this can be done by putting
   * the following line of code at the bottom of your fireEvent method:
   * <CENTER><TT>if (log != null) log.log(e);</TT></CENTER>
   * Where "e" is the AgentEvent reference.  This encourages good logging
   * practices and greatly facilitates agent construction.
   * <P>
   * Events are logged with the "Event" facility, with level LEV_EVENT.
   * @param e The event to log
   */
  public void log(AgentEvent e) {

      log(e, LEV_EVENT, FAC_EVENT);
  }

  /**
   * Logs some other random object
   * @param e The event to log
   * @param l The level to log at
   * @param f The facility ID to use
   */
  public void log(AgentEvent e, int l, int f) {
    if (logcontrolonly && control != null && Thread.currentThread() != control.getControlThread()) return;

      if (logevents || loggable(l, f))
          log(new LogItem(e.toString(), l, f));
      if (logtime && logdelaythresh > 0) {
          lastlogitem = new LogItem(e.toString(), l, f);
      }
  }

  /**
   * Logs some other random object, using the "General" facility.
   * @param o The object, its toString() method is called
   * @param l The level to log at
   */
  public void log(Object o, int l) {

      log(o, l, FAC_GENERAL);
  }

  /**
   * Logs some other random object
   * @param o The object, its toString() method is called
   * @param l The level to log at
   * @param f The facility ID to use
   */
  public void log(Object o, int l, int f) {

      log(o.toString(), l, f);
  }

  /**
   * Logs a string, using the "General" facility
   * @param s The string to log
   * @param l The level to log at
   */
  public void log(String s, int l) {

    log(s, l, FAC_GENERAL);
  }

  /**
   * Logs a string.  This is the primary method that your logging is likely
   * to use.
   * @param s The string to log
   * @param l The level to log at
   * @param f The facility ID to use
   */
  public void log(String s, int l, int f) {
    if (logcontrolonly && control != null && Thread.currentThread() != control.getControlThread()) return;

    if (loggable(l, f))
        log(new LogItem(s, l, f));
      if (logtime && logdelaythresh > 0) {
          lastlogitem = new LogItem(s, l, f);
      }
  }

  /**
   * Updates the label
   * @param l The string to set the label to
   */
  public void setLabel(String l) {
    label.setText(l);
  }

  /**
   * To log or not to log
   * @param l The level of the prospective log line
   * @param f The facility ID to check
   * @return true if the item should be logged
   */
  public boolean loggable(int l, int f) {
    if ((l < Log.LOG_ERR) || (l > Log.LOG_SPAM)) {
        log("Error: Log attempt by " + getFacilityByID(l) + " level " + l + " out of range", Log.LOG_ERR, FAC_LOG);
        new Exception().printStackTrace();
    }

	// LogDelayThresh
	if (logtime && logdelaythresh > 0) {
	    thistime = System.currentTimeMillis();
	    if ((thistime - lasttime) > logdelaythresh) {
		if (lastlogitem != null) {
			lastlogitem.setAux(getFacilityByID(lastlogitem.getFacility()) + "(" + lastlogitem.getLevel() + ")");
			output.println(lastlogitem + " (LogDelayThresh " + (thistime - lasttime) + ")");
			return true;
		}
	    }
	}

    return (l <= levels[f]);
  }

  /**
   * To log or not to log
   * @param i The prospective log item
   * @return true if the item should be logged
   */
  protected boolean loggable(LogItem i) {
    return loggable(i.getLevel(), i.getFacility());
  }

  /**
   * Logs a log item to the designated stream.  It will
   * automatically flush error level (FLUSH_THRESH) log items.
   * @param i The item to log
   * @see #setLogStream
   */
    String aux;
    protected void log(LogItem i) {

        //if (loggable(i)) {
        if (true) {
            i.clip(length);

            if (complex) {
                aux = clock + ":" + name;

                if (logmemory)
                    aux += ":" + runtime.freeMemory() + "/" + runtime.totalMemory(); 
                if (logtime) {
                    //thistime = Calendar.getInstance().getTime().getTime();
                    thistime = System.currentTimeMillis();
                    aux += ":" + thistime + "(" + (thistime - lasttime) + ")"; 
                    lasttime = thistime;
                }

                aux += ":" + getFacilityByID(i.getFacility()) + "(" + i.getLevel() + ")";
                i.setAux(aux);
            }

            // This prints it out
            if (linebreaks) {
                output.println(i);
            } else {
                output.println(i.toString().replace('\n', ' '));
            }

            if (gui)
                setLabel(i.getItem());
            if ((flush) || (i.getLevel() <= FLUSH_THRESH)) flush();

            if (i.getLevel() <= ERROR_THRESH) {
                if (audible && hasGUI())
                    Toolkit.getDefaultToolkit().beep();
                flush();
            }
        }

        if (logevents) {
            if (i.getLevel() <= ERROR_THRESH)
                startFireEvent(new LogEvent(this, i, LogEvent.ERROR));
            if (i.getObject() == null) {
                startFireEvent(new LogEvent(this, i, LogEvent.STRING));
            } else if (i.getObject() instanceof AgentEvent) {
                startFireEvent(new LogEvent(this, i, LogEvent.EVENT));
            } else {
                startFireEvent(new LogEvent(this, i, LogEvent.OBJECT));
            }
        }
    }

    /**
     * Opens an auxilary log file and caches the OutputStream.
     * Subsequent calls to this function will returned the cached
     * stream, rather than opening a new one.
     * @param filename The file to write to
     * @param append Append to the file or not
     */
    protected OutputStream getFileOutputStream(String filename, boolean append) {
        OutputStream stream = (OutputStream)files.get(filename);

        if (stream == null) {
            try {
                stream = new FileOutputStream(filename, append);
                files.put(filename, stream);
            } catch (IOException e) {
                log("Error opening file output stream for " + filename + ":" + e, LOG_ERR, FAC_LOG);
            }
        }

        return stream;
    }

    /**
     * Logs data to an auxilary log file, overwriting the file when
     * it is first opened.
     * @param o The object to log
     * @param filename The file to log to
     * @see #log(Object, String, boolean)
     */
    public void log(Object o, String filename) {
        log(o, filename, false);
    }

    /**
     * A convenience function which logs arbitrary data to an auxilary log
     * file.  Note this does not add line breaks to the data, so if you want
     * one entry per line you must add the linebreaks directly to the string
     * output of the object you are logging.
     * @param o The object to log
     * @param filename The file to log to
     * @param append Append the data to an existing file, or false to overwrite it.
     *    Note that this flag is only used the first time the file is
     *    written to in a session.
     */
    public void log(Object o, String filename, boolean append) {
        OutputStream stream = getFileOutputStream(filename, append);

        if (stream != null) {
            try {
                String s = o.toString();
                stream.write(s.getBytes());
                stream.flush();
            } catch (IOException e) {
                log("Error using output stream for " + filename + ":" + e, LOG_ERR, FAC_LOG);
            }
        }
    }

  /**
   * Flushes the log
   */
  public void flush() {
      output.flush();
  }

  /**
   * Watch for interesting property changes.  Basically, this is to
   * see if anyone changes the agent's name and monitor the simulated
   * clock time.  Also grabs new log level events.
   * @param e The event to look at
   */
  public void propertyChanged(PropertyEvent e) { propertyAdded(e); }
  public void propertyRemoved(PropertyEvent e) {  }
  public void propertyAdded(PropertyEvent e) {
    String key = e.getKey().toString();

    if (key.equals(State.TIME)) {
      clock = ((Integer)e.getProperty()).intValue();

    } else if (key.equals(State.NAME)) {
      name = ((String)e.getProperty());
      utilitieslog.setName(name);

    } else if (key.equals(AUDIBLEERRORS)) {
      audible = ((Boolean)e.getProperty()).booleanValue();

    } else if (key.equals(USEANSI)) {
      LogItem.setAnsi(((Boolean)e.getProperty()).booleanValue());

    } else if (key.startsWith("Log")) {
        // Just to avoid excessive str checks
        if (key.equals(LOGLEVEL)) {
            LEV_GENERAL = ((Integer)e.getProperty()).intValue();
            utilitieslog.setLevel(LEV_GENERAL);
            updateAllLogLevels();

        } else if (key.equals(LOGFLUSH)) {
            flush = ((Boolean)e.getProperty()).booleanValue();
            if (flush) flush();

        } else if (key.equals(LOGDELAYTHRESH)) {
            logdelaythresh = ((Integer)e.getProperty()).intValue();

        } else if (key.equals(LOGLENGTH)) {
            length = ((Integer)e.getProperty()).intValue();

        } else if (key.equals(LOGEVENTS)) {
            logevents = ((Boolean)e.getProperty()).booleanValue();

        } else if (key.equals(LOGLINEBREAKS)) {
            linebreaks = ((Boolean)e.getProperty()).booleanValue();

        } else if (key.equals(LOGCONTROLONLY)) {
            logcontrolonly = ((Boolean)e.getProperty()).booleanValue();

        } else if (key.equals(LOGFACILITIES)) {
            updateAllLogLevels();

        } else if (key.equals(LOGCOMPLEX)) {
            complex = ((Boolean)e.getProperty()).booleanValue();

        } else if (key.equals(LOGMEMORY)) {
            logmemory = ((Boolean)e.getProperty()).booleanValue();

        } else if (key.equals(LOGTIME)) {
            logtime = ((Boolean)e.getProperty()).booleanValue();

        } else if (key.equals(LOGGUI)) {
            gui = ((Boolean)e.getProperty()).booleanValue();
        }
    }
  }
}
