/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * LogViewer.java
 *
 * manifest: Name: agent/simplest/LogViewer.class
 * manifest: Java-Bean: True
 ***************************************************************************************/

package agent.simplest;

/* Global includes */
import java.awt.*;
import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;

import agent.base.*;
import utilities.*;

/**
 * This utility component provides a generic way of viewing
 * a collection of time-based logfiles, such as those generated
 * by the JAF Log components, in real time.  Developer-defined
 * log "filters" are added to the LogViewer, along with one or
 * more logfiles, which define the types of events you are
 * interested in and the data to be parsed, respectively.
 * <P>
 * Each filter is used to evaluate each line in each logfile,
 * to determine if that log line represents a certain type
 * of activity.  For instance, you might define a "Sent Messages"
 * filter, which looks for instances of when messages were sent 
 * by the agent.  As each log line is parsed, the filter must
 * determine if the line indicates a message has been sent or not.
 * This information is then used to generate a graphical timeline,
 * where tokens drawn on the timeline indicate that a filtered
 * activity has taken place.  So each log line found by the
 * "Sent Messages" filter will eventually be represented by a
 * token on that timeline.
 * <P>
 * Since you may design your own filter, you can search for
 * any type of pattern, provided the necessary data fits on 
 * a single log line.  Graphically viewing the combined logs
 * of a number of agents, with arbitrary filters controlling
 * the view of the data gives you a powerful way of interpreting
 * and debugging agent activity.
 * <P>
 * The LogViewer interface allows you to hilight interesting
 * events on the timeline, scale the time timeline width,
 * and dynamically search all existing filtered events.
 * <P>
 * Note that because LogViewer operates on log file data, you
 * must both have a sufficiently high Log level for the interesting
 * data to be logged in the first place, and have the log data
 * in a file.  This latter requirement is typically satisfied by
 * redirecting the output of the agent to a file.
 * <P>
 * The LogViewer can be run independantly of a JAF agent:<BR>
 * &nbsp; <TT>java agent.simplest.LogViewer [-f filter] [-l filter] [logfiles]</TT><BR>
 * Make sure you have an appropriate classpath containing
 * the JAF and utilities libraries.  The stand along LogViewer
 * takes in as arguments both logfiles and filters.  The
 * filters should be specified by their full class names, using
 * either a <TT>-f</TT> or <TT>-l</TT> flag.  Those specified
 * with <TT>-f</TT> are loaded normally, while those with <TT>-l</TT>
 * are loaded and deactivated.  A example execution might look
 * like this:
 * &nbsp; <TT>java agent.simplest.LogViewer -f agent.example.MyFilter agent1.out agent2.out</TT>
 *
 * @see agent.simplest.LogViewerFilter
 */
public class LogViewer extends agent.base.AgentComponent implements PropertyEventListener, MouseListener {
    static final long serialVersionUID = 1234567890L; // kludge static ser_id.  Bad bad Bryan!

    protected static int FAC_LOGVIEW;

    protected Log log;
    protected State state;
    protected ViewerFrame window;

    protected Vector logfiles = new Vector();
    protected Hashtable logfilereaders = new Hashtable();
    protected Hashtable logfilelines = new Hashtable();
    protected Vector filters = new Vector();
    protected float timemultiplier = 1;
    protected int start = 0;
    protected int finish = 0;

    protected static Color[] colors;
    protected static Hashtable colorset = new Hashtable();
    protected boolean standalone = false;
    protected LogfileReaderThread reader = null;

    public static final String LOGFILES = "Logfiles";
    public static final String LOGFILTERS = "LogFilters";
    public static final String LOGVIEWERUPDATE = "LogViewerUpdate";
    public static final String LOGVIEWERREAD = "LogViewerRead";
    public static final String LOGSCALE = "LogScale";

    static {
        colors = new Color[8];
        colors[0] = new Color(255, 0, 0);  // red
        colors[1] = new Color(0, 153, 0);  // green
        colors[2] = new Color(0, 0, 255);  // blue
        colors[3] = new Color(153, 0, 204); // purple
        colors[4] = new Color(102, 102, 0); // brown
        colors[5] = new Color(204, 102, 0); // orange
        colors[6] = new Color(255, 51, 255); // magenta
        colors[7] = new Color(102, 102, 153); // slate
    }

    /**
     * Default constructor
     */
    public LogViewer() {
        super();
    
        addDependency("Log");
        addDependency("State");

        State.addParameterInfo(LOGFILES, "Vector", "List of the logfiles to be viewed, these are automatically opened when the component starts", new Vector());
        State.addParameterInfo(LOGFILTERS, "Vector", "List of the log filters to be added", new Vector());
        State.addParameterInfo(LOGVIEWERUPDATE, "Integer", "Time delay between log viewer continuous read updates", new Integer(2000));
        State.addParameterInfo(LOGVIEWERREAD, "Boolean", "If true, continuous reading of log files is enabled", Boolean.TRUE);

        addMouseListener(this);
    }

    /**
     * Sets up component helpers
     */
    public void init() {
        log = (Log)agent.simplest.State.findComponent("Log");
        FAC_LOGVIEW = log.getFacilityID(this);

        state = (State)agent.simplest.State.findComponent("State");
        state.addPropertyEventListener(this);

        boolean read = ((Boolean)state.getProperty(LOGVIEWERREAD)).booleanValue();
        if (read) startReader();
    
        super.init();
    }

    /**
     * pulsing function, use to update the display
     */
    public void pulse() {
    }

    /**
     * Resets the component
     */
    public void reset() {
    }

    /**
     * Start and finish time accessors.
     */
    public int getStartTime() {
        int low = Integer.MAX_VALUE;
        for (Enumeration e = getFilters(); e.hasMoreElements();) {
            LogViewerFilter f = (LogViewerFilter)e.nextElement();
            if (f.getNumAgents() > 0) {
                for (Iterator ie = f.getEntryAgents(); ie.hasNext();) {
                    TreeMap t = f.getEntries((String)ie.next());
                    if (t.isEmpty()) continue;
                    int i = ((Integer)t.firstKey()).intValue();
                    if (i < low)
                        low = i;
                }
            }
        }
        return low;
    }
    public int getFinishTime() {
        int hi = Integer.MIN_VALUE;
        for (Enumeration e = getFilters(); e.hasMoreElements();) {
            LogViewerFilter f = (LogViewerFilter)e.nextElement();
            if (f.getNumAgents() > 0) {
                for (Iterator ie = f.getEntryAgents(); ie.hasNext();) {
                    TreeMap t = f.getEntries((String)ie.next());
                    if (t.isEmpty()) continue;
                    int i = ((Integer)t.lastKey()).intValue();
                    if (i > hi)
                        hi = i;
                }
            }
        }
        return hi;
    }

    /**
     * Starts the auto-reader
     */
    public synchronized void startReader() {
        if (readerIsActive()) return;

        reader = new LogfileReaderThread();
        reader.start();

        if (window != null) window.updateMenus();
    }

    /**
     * Stops the auto-reader
     */
    public synchronized void stopReader() {
        if (!readerIsActive()) return;

        reader.alive = false;
        reader = null;

        if (window != null) window.updateMenus();
    }

    /**
     * The reader
     */
    class LogfileReaderThread extends Thread {
        public boolean alive = true;
        public void run() {
            int sleep = ((Integer)state.getProperty(LOGVIEWERUPDATE)).intValue();
            while (alive) {
                try { Thread.currentThread().sleep(sleep); } catch (InterruptedException ex) { }
                if ((window == null) || (!window.isVisible())) continue;
                Enumeration e = new SafeEnumeration(logfilereaders.keys());
                while (e.hasMoreElements()) {
                    int total = 0;
                    String filename = (String)e.nextElement();
                    if (logfilereaders.containsKey(filename)) {
                        total += parseLogfile(filename, null, false);
                    }
                    if ((total > 0) && (window != null)) {
                        window.updateDisplay();
                    }
                }
            }
        }
    };

    /**
     * Is the reader reading?
     */
    public boolean readerIsActive() {
        return (reader != null);
    }

    /**
     * Parses all the known logfiles
     */
    public synchronized void parseAllLogfiles(LogViewerFilter filter) {

        logfilereaders.clear();
        logfilelines.clear();

        Enumeration en = getLogfiles();
        while (en.hasMoreElements()) {
            parseLogfile((String)en.nextElement(), filter, true);
        }
    }

    /**
     * Parses a logfile and tests each line against one or more filters.
     * @param filename The filename to parse
     * @param filter The filter to test against.  If null, the logfile is
     * tested against all active filters.
     * @param verbose If true, status log messages are printed
     * @return The number of new lines read from the file
     */
    public int parseLogfile(String filename, LogViewerFilter filter, boolean verbose) {
        try {
            String line;
            int linenum = 0;
            LogViewerFilter fs[] = null;

            if (filter == null) {
                fs = new LogViewerFilter[filters.size()];
                for (int i = 0; i < fs.length; i++) {
                    LogViewerFilter f = (LogViewerFilter)filters.elementAt(i);
                    fs[i] = (f.isActive() ? f : null);
                }
            } else {
                fs = new LogViewerFilter[1];
                fs[0] = filter;
            }

            BufferedReader br;
            if (logfilereaders.containsKey(filename)) {
                br = (BufferedReader)logfilereaders.get(filename);
                linenum = ((Integer)logfilelines.get(filename)).intValue();
            } else {
                FileReader fr = new FileReader(filename);
                br = new BufferedReader(fr);
                logfilereaders.put(filename, br);
            }

            int start = linenum;
            String lasttime = "0", lastagent = "???";
            for (; (line = br.readLine()) != null; linenum++) {
                if ((linenum > 0) && (linenum % 500 == 0) && verbose)
                    log.log("Read line " + linenum, Log.LOG_DEBUG, FAC_LOGVIEW);
                for (int i = 0; i < fs.length; i++) {
                    if (fs[i] == null) continue;

                    if (fs[i].isMyLogEntry(line)) {
                        String time = fs[i].getTime(line);
                        if (time == null) time = lasttime;
                        lasttime = time;
                        String agent = fs[i].getAgent(line);
                        if (agent == null) agent = lastagent;
                        lastagent = agent;
                        fs[i].addEntry(Integer.valueOf(time), agent, new LogViewerEntry(filename, linenum, line));
                    }
                }
            }

            if (start != linenum) {
                linenum--;
                logfilelines.put(filename, new Integer(linenum));
                if (verbose)
                    log.log("Done parsing " + filename + " - " + (linenum - start) + " lines read", Log.LOG_INFO, FAC_LOGVIEW);
            }

            return linenum - start;

        } catch (IOException e) {
            log.log("Logfile " + filename + " not found.", Log.LOG_WARNING, FAC_LOGVIEW);
            return 0;
        }
    }

    /**
     * Gets an enumeration of the logfiles
     */
    public Enumeration getLogfiles() {
        return logfiles.elements();
    }

    /**
     * Determines if the logfile is being tracked.
     */
    public boolean hasLogfile(String filename) {
        for (Enumeration e = logfiles.elements(); e.hasMoreElements();) {
            if (filename.equals((String)e.nextElement()))
                return true;
        }
        return false;
    }
   
    /**
     * Adds a file to the logfile list.  The file will be processed
     * for viewing
     */
    public void addLogfile(String filename) {
        log.log("Adding logfile " + filename, Log.LOG_INFO, FAC_LOGVIEW);
        if (!hasLogfile(filename)) {
            if (parseLogfile(filename, null, true) > 0) {
                logfiles.add(filename);
                if (window != null) window.updateMenus();
            } else {
                clearLogfileData(filename);
            }

        } else
            log.log("Logfile " + filename + " not added, it is already being tracked", Log.LOG_WARNING, FAC_LOGVIEW);
    }

    /**
     * Removes a file from the logfile list.
     */
    public void removeLogfile(String filename) {
        log.log("Removing logfile " + filename, Log.LOG_INFO, FAC_LOGVIEW);
        if (hasLogfile(filename)) {
            clearLogfileData(filename);
            logfiles.remove(filename);
            if (window != null) window.updateMenus();

        } else
            log.log("Logfile " + filename + " not removed, it is not being tracked", Log.LOG_WARNING, FAC_LOGVIEW);
    }

    /**
     */
    public void removeAllLogfiles() {
        Enumeration en = getLogfiles();
        while (en.hasMoreElements()) {
            removeLogfile((String)en.nextElement());
        }
    }

    /**
     */
    public synchronized void clearAllLogfileData() {
        Enumeration e = getFilters();
        while (e.hasMoreElements()) {
            LogViewerFilter f = (LogViewerFilter)e.nextElement();
            f.clearAllEntries();
            logfilereaders.clear();
            logfilelines.clear();
        }
        if (window != null) window.updateDisplay();
    }

    /**
     */
    public synchronized void clearLogfileData(String filename) {
        log.log("Clearing data from " + filename, Log.LOG_INFO, FAC_LOGVIEW);

        logfilereaders.remove(filename);
        logfilelines.remove(filename);

        Enumeration e = getFilters();
        while (e.hasMoreElements()) {
            LogViewerFilter f = (LogViewerFilter)e.nextElement();
            synchronized (f) {
                for (Iterator ie = f.getEntryAgents(); ie.hasNext();) {
                    String agent = (String)ie.next();
                    TreeMap t = f.getEntries(agent);
                    Iterator i = t.keySet().iterator();
                    while (i.hasNext()) {
                        Integer time = (Integer)i.next();
                        Vector v = f.getEntry(time, agent);
                        for (Enumeration en = new SafeEnumeration(v.elements()); en.hasMoreElements();) {
                            LogViewerEntry le = (LogViewerEntry)en.nextElement();
                            if (le.getSource().equals(filename)) {
                                if (f.clearEntry(time, agent, le)) {
                                    i.remove();
                                    break;
                                }
                            }
                        }
                    }
                }
            }
        }
        if (window != null) window.updateDisplay();
    }

    /**
     * Gets an enumeration of the filters
     */
    public Enumeration getFilters() {
        return filters.elements();
    }

    /**
     * Determines if the filter is present
     */
    public boolean hasFilter(LogViewerFilter f) {
        for (Enumeration e = getFilters(); e.hasMoreElements();) {
            if (f.getLabel().equals(((LogViewerFilter)e.nextElement()).getLabel()))
                return true;
        }
        return false;
    }
   
    /**
     * Adds a filter to the filter list, using reflection.
     * @param fname The class name of the filter to add
     * @return The filter instance, or null if none created
     */
    public LogViewerFilter addFilter(String fname) {

        try {
            Class c = Class.forName(fname);
            Object o = c.newInstance();
            if (o instanceof LogViewerFilter) {
                addFilter((LogViewerFilter)o);
                return (LogViewerFilter)o;
            } else {
                log.log("Error: Class " + fname + " is not a LogViewerFilter", Log.LOG_ERR, FAC_LOGVIEW);
            }

        } catch (Exception ex) {
            log.log("Error: Could not instantiate class " + fname + "\n" + ex, Log.LOG_ERR, FAC_LOGVIEW);
        }

        return null;
    }
   
    /**
     * Adds a filter to the filter list.
     */
    public void addFilter(LogViewerFilter f) {
        log.log("Adding filter " + f.getLabel(), Log.LOG_INFO, FAC_LOGVIEW);
        if (!hasFilter(f)) {
            filters.add(f);
            f.setLogViewer(this);
            parseAllLogfiles(f);
            if (window != null) window.updateMenus();

        } else
            log.log("Filter " + f.getLabel() + " not added, it is already present", Log.LOG_WARNING, FAC_LOGVIEW);
    }

    /**
     * Removes a filter from the filter list.
     */
    public void removeFilter(LogViewerFilter f) {
        log.log("Removing filter " + f.getLabel(), Log.LOG_INFO, FAC_LOGVIEW);
        if (hasFilter(f)) {
            filters.remove(f);
            f.setLogViewer(null);
            deactivateFilter(f);
            if (window != null) window.updateMenus();

        } else
            log.log("Filter " + f.getLabel() + " not removed, it is not present", Log.LOG_WARNING, FAC_LOGVIEW);
    }

    /**
     * Finds a filter, given its name
     */
    public LogViewerFilter findFilter(String n) {
        Enumeration e = getFilters();
        while (e.hasMoreElements()) {
            LogViewerFilter f = (LogViewerFilter)e.nextElement();
            if (n.equals(f.getLabel()))
                return f;
        }
        return null;
    }

    /**
     * Activates a filter
     */
    public void activateFilter(LogViewerFilter f) {
        if (f.isActive()) return;
        f.setActive(true);
        parseAllLogfiles(f);
        if (window != null) window.updateMenus();
    }

    /**
     * Deactivates a filter
     */
    public void deactivateFilter(LogViewerFilter f) {
        if (!f.isActive()) return;
        f.setActive(false);
        f.clearAllEntries();
        if (window != null) window.updateMenus();
    }

    /**
     * Gets a specific perty color, for use by the visualization part.
     * In general, only numbers 0-2 (red, green, blue) will remain
     * constant, any past that may be changed in later revisions..
     */
    public static Color getColor(int i) {
        return colors[i];
    }

    /**
     * Sets the time multiplier
     */
    public void setTimeMultiplier(float t) {
	timemultiplier = t;
    }
    public float getTimeMultiplier() {
	return timemultiplier;
    }

    /**
     * Gets a perty color, for use by the visualization part.
     */
    public static Color getColor(String s) {
        Integer color = (Integer)colorset.get(s);
        if (color != null) return colors[color.intValue()];

        // Pick a color
        Vector v;
        color = new Integer(0);
        for (int i = 0; i < colors.length - 1; i++) {
            Integer num1 = (Integer)colorset.get(colors[i]);
            if (num1 == null) {
                num1 = new Integer(0);
                colorset.put(colors[i], num1);
            }
            Integer num2 = (Integer)colorset.get(colors[i+1]);
            if (num2 == null) {
                num2 = new Integer(0);
                colorset.put(colors[i+1], num2);
            }
            if (num2.intValue() < num1.intValue()) {
                color = new Integer(i+1);
                break;
            }
        }
        Integer num = (Integer)colorset.get(colors[color.intValue()]);
        colorset.put(colors[color.intValue()], new Integer(num.intValue() + 1));
        colorset.put(s, color);

        return colors[color.intValue()];
    }

    /**
     * Watch for interesting property changes.  Monitors time changes.
     * @param e The event to look at 
     */
    public void propertyRemoved(PropertyEvent e) { }
    public void propertyAdded(PropertyEvent e) { propertyChanged(e); }
    public void propertyChanged(PropertyEvent e) { 
        String key = e.getKey().toString();

        if (key.equalsIgnoreCase(LOGFILES)) {
            removeAllLogfiles();
            Vector v = (Vector)e.getProperty();
            if (v != null) {
                Enumeration en = v.elements();
                while (en.hasMoreElements()) {
                    addLogfile((String)en.nextElement());
                }
            }

        } else if (key.equalsIgnoreCase(LOGFILTERS)) {
            Vector v = (Vector)e.getProperty();
            if (v != null) {
                Enumeration en = v.elements();
                while (en.hasMoreElements()) {
                    addFilter((String)en.nextElement());
                }
            }
        }
    }

    /**
     * Returns true if the component does something when you click on it.
     */
    public boolean hasClickAction() {
        return true;
    }

    /**
     * Responds to mouse clicks
     */
    public void handleMouseEvent(MouseEvent e) {

        if (window == null) {
            window = new ViewerFrame(this);
            window.setSize(640,400);
            //window.pack();
            if (State.hasComponent("WindowManager")) {
                WindowManager wm = (WindowManager)state.findComponent("WindowManager");
                wm.addWindow(window);
            }
        }

        window.setVisible(true);
    }
    public void mouseClicked(MouseEvent e) { handleMouseEvent(e); }
    public void mousePressed(MouseEvent e) { handleMouseEvent(e); }
    public void mouseReleased(MouseEvent e) { }
    public void mouseDragged(MouseEvent e) { }
    public void mouseEntered(MouseEvent e) { }
    public void mouseExited(MouseEvent e) { }
    public void mouseMoved(MouseEvent e) { }

    /**
     * Allows this component to be run outside of the agent.
     */
    public static void main(String []argv) {

        // Create the necessary components
        agent.simplest.Control control = new agent.simplest.Control(false);
        agent.simplest.Log log = new agent.simplest.Log();
        agent.simplest.State state = new agent.simplest.State();
        agent.simplest.LogViewer viewer = new agent.simplest.LogViewer();

        // ...and a couple to add some Filters
        agent.simplest.Communicate communicate = new agent.simplest.Communicate();
        agent.simplest.Execute execute = new agent.simplest.Execute();

        viewer.standalone = true;
        state.setProperty("Name", "LogView");

        // Then init them
        try {
            control.setupComponents();
        } catch (DependencyFailedException ex) {
            log.log(ex, Log.LOG_ERR, FAC_LOGVIEW);
        }
        log.setLogLevel(FAC_LOGVIEW, Log.LOG_DEBUG);
        log.setLogLevel(2);

        // Add the specified logfiles
        for (int i = 0; i < argv.length; i++) {
            if (argv[i].equals("-h")) {
                System.err.println("usage: LogViewer [-t <time multiplier>] [-f class.LogFilter] [-l class.LogFilter] [logfiles]");

            } else if (argv[i].equals("-t")) {
                i++;
                if (argv.length > i)
                    viewer.setTimeMultiplier(Float.parseFloat(argv[i]));
                else
                    log.log("Error: Filter time multiplier must follow -t", Log.LOG_ERR, FAC_LOGVIEW);

            } else if (argv[i].equals("-f")) {
                i++;
                if (argv.length > i)
                    viewer.addFilter(argv[i]);
                else
                    log.log("Error: Filter class name must follow -f", Log.LOG_ERR, FAC_LOGVIEW);

            } else if (argv[i].equals("-l")) {
                i++;
                if (argv.length > i) {
                    LogViewerFilter f = viewer.addFilter(argv[i]);
                    if (f == null) {
			f = viewer.findFilter(argv[i]);
		    }
                    if (f != null) {
                        log.log("Deactivating " + f.getLabel(), Log.LOG_DEBUG, FAC_LOGVIEW);
                        viewer.deactivateFilter(f);
                    }
                } else
                    log.log("Error: Filter class name must follow -l", Log.LOG_ERR, FAC_LOGVIEW);

            } else {
                viewer.addLogfile(argv[i]);
            }
        }

        // Now do something
        viewer.mouseClicked(null);
    }
}

/**
 * Window for the log viewer
 */
class ViewerFrame extends JFrame {
    LogViewer viewer;
    JMenu logfiles, filters, help;
    ViewerPanel panel;

    public ViewerFrame(LogViewer v) {
        super("Log Viewer");
        viewer = v;

        JMenuBar mbar = new JMenuBar();

        logfiles = new JMenu("Logfiles");
        mbar.add(logfiles);

        filters = new JMenu("Filters");
        mbar.add(filters);

        help = new JMenu("Help");
        mbar.add(help);

        setJMenuBar(mbar);

        getContentPane().add(panel = new ViewerPanel(viewer));

        updateMenus();

        if (viewer.standalone) {
            addWindowListener(new WindowAdapter() {
                    public void windowClosing(WindowEvent e) {
                        Control control = (Control)State.findComponent("Control");
                        if (control != null) control.quit();
                    }
                    public void windowClosed(WindowEvent e) {
                        Control control = (Control)State.findComponent("Control");
                        if (control != null) control.quit();
                    }
            });
        }
    }

    public void updateMenus() {

        // Logfile menu
        logfiles.removeAll();
        JMenuItem item = new JMenuItem("Add Logfiles...");
        item.addActionListener(new ActionListener() {
                public void actionPerformed(java.awt.event.ActionEvent e) {
                    String dir = System.getProperty("user.dir");
                    if (dir == null) dir = ".";
                    JFileChooser choose = new JFileChooser(dir);
                    choose.setMultiSelectionEnabled(true);
                    choose.setDialogTitle("Select one or more JAF logfiles...");
                    int ret = choose.showOpenDialog(panel);
                    if (ret == JFileChooser.APPROVE_OPTION) {
                        File[] files = choose.getSelectedFiles();
                        for (int i = 0; i < files.length; i++) {
                            viewer.addLogfile(files[i].getAbsolutePath());
                        }
                    }
                }
            });
        logfiles.add(item);
        if (viewer.getLogfiles().hasMoreElements()) {
            item = new JMenuItem("Refresh All");
            item.addActionListener(new ActionListener() {
                    public void actionPerformed(java.awt.event.ActionEvent e) {
                        viewer.clearAllLogfileData();
                        viewer.parseAllLogfiles(null);
                    }
                });
            logfiles.add(item);
            logfiles.addSeparator();
            item = new JCheckBoxMenuItem("Continuous Read");
            item.addActionListener(new ActionListener() {
                    public void actionPerformed(java.awt.event.ActionEvent e) {
                        JCheckBoxMenuItem i = (JCheckBoxMenuItem)e.getSource();
                        if (i.getState())
                            viewer.startReader();
                        else
                            viewer.stopReader();
                    }
                });
            if (viewer.readerIsActive())
                ((JCheckBoxMenuItem)item).setState(true);
            logfiles.add(item);
            item = new JCheckBoxMenuItem("Scroll On Update");
            item.addActionListener(new ActionListener() {
                    public void actionPerformed(java.awt.event.ActionEvent e) {
                        JCheckBoxMenuItem i = (JCheckBoxMenuItem)e.getSource();
                        if (i.getState())
                            panel.resizescroll = true;
                        else
                            panel.resizescroll = false;
                    }
                });
            if ((panel != null) && (panel.resizescroll))
                ((JCheckBoxMenuItem)item).setState(true);
            logfiles.add(item);
            logfiles.addSeparator();
            for (Enumeration en = viewer.getLogfiles(); en.hasMoreElements();) {
                String lf = (String)en.nextElement();
                item = new JMenuItem(lf);
                item.addActionListener(new ActionListener() {
                        public void actionPerformed(java.awt.event.ActionEvent e) {
                            viewer.clearLogfileData(((JMenuItem)e.getSource()).getText());
                            viewer.parseLogfile(((JMenuItem)e.getSource()).getText(), null, true);
                        }
                    });
                logfiles.add(item);
            }
            logfiles.addSeparator();
            JMenu menu = new JMenu("Remove");
            for (Enumeration en = viewer.getLogfiles(); en.hasMoreElements();) {
                String lf = (String)en.nextElement();
                item = new JMenuItem(lf);
                item.addActionListener(new ActionListener() {
                        public void actionPerformed(java.awt.event.ActionEvent e) {
                            viewer.removeLogfile(((JMenuItem)e.getSource()).getText());
                        }
                    });
                menu.add(item);
            }
            logfiles.add(menu);
        }

        if (viewer.standalone) {
            logfiles.addSeparator();
            item = new JMenuItem("Quit");
            item.addActionListener(new ActionListener() {
                    public void actionPerformed(java.awt.event.ActionEvent e) {
                        dispose();
                    }
                });
            logfiles.add(item);
        }

        // Filter menu
        filters.removeAll();

        item = new JMenuItem("Enable All");
        item.addActionListener(new ActionListener() {
                public void actionPerformed(java.awt.event.ActionEvent e) {
                    Enumeration en = viewer.getFilters();
                    while (en.hasMoreElements()) {
                        viewer.activateFilter((LogViewerFilter)en.nextElement());
                    }
                    updateMenus();
                }
            });
        filters.add(item);
        item = new JMenuItem("Disable All");
        item.addActionListener(new ActionListener() {
                public void actionPerformed(java.awt.event.ActionEvent e) {
                    Enumeration en = viewer.getFilters();
                    while (en.hasMoreElements()) {
                        viewer.deactivateFilter((LogViewerFilter)en.nextElement());
                    }
                    updateMenus();
                }
            });
        filters.add(item);
        filters.addSeparator();
        item = new JMenuItem("Show Legend");
        item.addActionListener(new ActionListener() {
                public void actionPerformed(java.awt.event.ActionEvent e) {
                    JDialog tokens = new JDialog(ViewerFrame.this, "Token Legend");
                    tokens.getContentPane().setLayout(new BorderLayout());
                    JScrollPane scroller = new JScrollPane();
                    scroller.getViewport().add(panel.makeTokenLegend());
                    tokens.getContentPane().add(scroller);
                    tokens.pack();
                    tokens.setLocation(ViewerFrame.this.getLocation().x + 20,
                                       ViewerFrame.this.getLocation().y + 20);
                    tokens.setVisible(true);
                }
            });
        filters.add(item);
        filters.addSeparator();

        for (Enumeration en = viewer.getFilters(); en.hasMoreElements();) {
            LogViewerFilter lf = (LogViewerFilter)en.nextElement();
            item = new JCheckBoxMenuItem(lf.getLabel());
            if (lf.isActive())
                ((JCheckBoxMenuItem)item).setState(true);
            item.addActionListener(new ActionListener() {
                    public void actionPerformed(java.awt.event.ActionEvent e) {
                        JCheckBoxMenuItem i = (JCheckBoxMenuItem)e.getSource();
                        LogViewerFilter f = viewer.findFilter(i.getText());
                        if (f != null) {
                            if (i.getState())
                                viewer.activateFilter(f);
                            else
                                viewer.deactivateFilter(f);
                        }
                    }
                });
            filters.add(item);
        }

        // Help
        item = new JMenuItem("About Log Viewer...");
        item.addActionListener(new ActionListener() {
                public void actionPerformed(java.awt.event.ActionEvent e) {
                    JOptionPane.showMessageDialog(null,
                                                  "Log Viewer\na Java Agent Framework component\n\nby Bryan Horling\nbhorling@cs.umass.edu\n\nMulti-Agent Systems Lab\nUniversity of Massachusetts\nhttp://mas.cs.umass.edu/",
                                                  "About the Log Viewer",
                                                  JOptionPane.INFORMATION_MESSAGE);
                }
            });
        help.add(item);

        help.addSeparator();

        item = new JMenuItem("Help");
        item.addActionListener(new ActionListener() {
                public void actionPerformed(java.awt.event.ActionEvent e) {
                    //panel.text2.setText(panel.helpText);
                    JOptionPane.showMessageDialog(null,
                                                  panel.helpText,
                                                  "Log Viewer Help",
                                                  JOptionPane.INFORMATION_MESSAGE);
                }
            });
        help.add(item);

        updateDisplay();
    }

    public void updateDisplay() {
        if (panel != null) {
            panel.tc.invalidate();
            panel.lc.invalidate();
            panel.revalidate();
            panel.repaint();
        }
    }
}

/**
 * The panel which contains the viewer subpanels.  This acts
 * as a container, which sets up the visible layout
 */
class ViewerPanel extends JPanel implements MouseMotionListener, MouseListener, ChangeListener {
    public int TICKWIDTH = 3;
    public int TICKINC = 50 / TICKWIDTH;
    public final int PAD = 5;
    public final int LINE = 10 + PAD;

    LogViewer viewer;
    LegendCanvas lc;
    TimelineCanvas tc;
    JLabel foot;
    JTextArea text1, text2;
    JScrollPane tcscroll;
    JScrollPane textscroll1, textscroll2;
    Point spot;
    JSlider widthslider;
    JSplitPane textsplit;
    String blank = " ";
    Color hilightcolor = new Color(225, 225, 225);
    Color althilightcolor = new Color(255, 255, 150);
    Color searchcolor = new Color(175, 255, 255);
    JTextField search;
    boolean resizescroll = true;

    LogViewerEntrySlot highlighted = null;
    LogViewerEntrySlot althighlighted = null;
    protected Vector searched = new Vector();

    String helpText = "Brief Instructions:\n\n- Click on a token to view its details.\n- Right-click a token to display its details in the alternate (yellow) text area.\n\n- Enter a term in the blue box followed by a return to search all visible tokens.\n\n- Drag the divider bar to change the text areas' viewing sizes.\n- Right-click a text area to swap their arrangement.\n\n- Use the slider bar above to change the viewing scale.\n- See the menus for more options.";

    public ViewerPanel(LogViewer v) {
        super();

        viewer = v;

        // Create the stuff
        JPanel lp = new JPanel();
        lp.setLayout(new BorderLayout());
        lc = new LegendCanvas(viewer);
        search = new JTextField();
        search.setBackground(searchcolor);
        search.setText("Search Text");
        lp.add(lc);
        lp.add(search, BorderLayout.SOUTH);

        tcscroll = new JScrollPane();
        tcscroll.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
        tcscroll.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_NEVER);
        tc = new TimelineCanvas(viewer);
        tcscroll.getViewport().add(tc);

        widthslider = new JSlider(1, 10, TICKWIDTH);
        widthslider.addChangeListener(this);
        foot = new JLabel(blank);

        JPanel tp = new JPanel();
        tp.setLayout(new GridLayout(1, 2));
        tp.add(foot);
        tp.add(widthslider);

        JPanel p = new JPanel();
        p.setLayout(new BorderLayout());
        p.add(lp, BorderLayout.WEST);
        p.add(tcscroll);
        p.add(tp, BorderLayout.SOUTH);

        textscroll1 = new JScrollPane();
        text1 = new JTextArea("\n\n\n");
        text1.setLineWrap(true);
        text1.setWrapStyleWord(true);
        text1.setEditable(false);
        text1.setBackground(althilightcolor);
        textscroll1.getViewport().add(text1);

        textscroll2 = new JScrollPane();
        text2 = new JTextArea("\n\n\n");
        text2.setLineWrap(true);
        text2.setWrapStyleWord(true);
        text2.setEditable(false);
        text2.setBackground(hilightcolor);
        textscroll2.getViewport().add(text2);

        text2.setText(helpText);

        textsplit = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, textscroll1, textscroll2);

        MouseListener switcher = new MouseAdapter() {
                public void mouseClicked(MouseEvent e) {
                    if ((e.getModifiers() & MouseEvent.BUTTON3_MASK) != 0) {
                        int o = textsplit.getOrientation();
                        textsplit.setOrientation(o == JSplitPane.VERTICAL_SPLIT ? JSplitPane.HORIZONTAL_SPLIT : JSplitPane.VERTICAL_SPLIT);
                        textsplit.setDividerLocation((textsplit.getMinimumDividerLocation() + textsplit.getMaximumDividerLocation()) / 2);
                    }
                }
            };
        textsplit.addMouseListener(switcher);
        text1.addMouseListener(switcher);
        text2.addMouseListener(switcher);

        //JSplitPane sp = new JSplitPane(JSplitPane.VERTICAL_SPLIT, p, textsplit);
        JPanel sp = new JPanel();
        sp.setLayout(new BorderLayout());
        sp.add(p, BorderLayout.NORTH);
        sp.add(textsplit);

        // Lay it out
        setLayout(new BorderLayout());
        add(sp);
        //add(foot, BorderLayout.SOUTH);

        // Who are we listening to
        tc.addMouseMotionListener(this);
        tc.addMouseListener(this);
        tc.addKeyListener(new KeyAdapter() {
                public void keyTyped(KeyEvent e) {
                    //System.err.println(e);
                }
            });
        search.addKeyListener(new KeyAdapter() {
                public void keyPressed(KeyEvent e) {
                    if (e.getKeyCode() == KeyEvent.VK_ENTER) {
                        int count;
                        if (search.getText().length() > 0) {
                            count = hilightMatches(search.getText());
                        } else
                            count = hilightMatches(null);
                    }
                }
            });
        tc.addComponentListener(new ComponentAdapter() {
                public void componentResized(ComponentEvent e) {
                    if (resizescroll) {
                        JScrollBar scroll = tcscroll.getHorizontalScrollBar();
                        scroll.setValue(scroll.getMaximum());
                    }
                }
            });

        // This erases the initial search string the first time focus is gained
        search.addFocusListener(new FocusAdapter() {
                boolean first = true;
                public void focusGained(FocusEvent e) {
                    if (first) search.setText("");
                    first = false;
                }
            });

        // Set the size
        setSize(300,100);
        revalidate();

        // Quotes, woo hoo!
        Thread t = new Thread() {
                public void run() {
                    try {
                        java.net.URL foo = new java.net.URL("http://shakti.trincoll.edu/~bhorling/quotes");
                        java.net.URLConnection con = foo.openConnection();
                        con.connect();
                        Object o = con.getContent();
                        if (o instanceof InputStream) {
                            BufferedReader r = new BufferedReader(new InputStreamReader((InputStream)o));
                            String line;
                            Vector lines = new Vector();
                            while ((line = r.readLine()) != null) {
                                lines.addElement(line);
                            }
                            if (lines.size() != 0) {
                                blank = (String)lines.elementAt(new Random().nextInt(lines.size()));
                                blank = blank.trim();
                            }
                        }
                    } catch (IOException e) {
                        //System.err.println(e);
                    }
                }
            };
        t.start();
    }

    public JPanel makeTokenLegend() {
        return new TokenCanvas(viewer);
    }

    public void stateChanged(ChangeEvent e) {
        if ((e.getSource() == widthslider) && (TICKWIDTH != widthslider.getValue())) {
            Rectangle r = tcscroll.getViewport().getViewRect();
            r.x = (r.x + r.width) / 2;
            r.width = 1;
            TICKWIDTH = widthslider.getValue();
            TICKINC = 50 / TICKWIDTH;
            tc.revalidate();
            tc.repaint();
            tcscroll.getViewport().scrollRectToVisible(r);
        }
    }

    public void mouseDragged(MouseEvent e) {
        /*
        Point np = tcscroll.getViewport().getViewPosition();
        np.x += e.getX() - spot.x;
        System.err.println(spot + " " + e.getPoint() + " " + (e.getX() - spot.x));
        spot = e.getPoint();
        spot.x += e.getX() - spot.x;
        tcscroll.getViewport().setViewPosition(np);
        */
    }
    public void mouseMoved(MouseEvent e) {
        int time = tc.calculateTime(e.getX());
        LogViewerFilter filter = tc.calculateFilter(e.getY());
        String agent = tc.calculateAgent(e.getY());
        int num = -1;
        if (filter != null) {
            Vector v = filter.getEntry(new Integer(time), agent);
            if (v != null) num = v.size();
        }

        foot.setText(" Time: " + ((time >= 0) ? String.valueOf(time) : "-") +
                     "  Filter: " + ((filter != null) ? filter.getLabel() : "-") +
                     "  Agent: " + ((agent != null) ? agent : "-") + 
                     (num > 0 ? "  Entries: " + num : ""));
    }
    public void mouseClicked(MouseEvent e) {
        int time = tc.calculateTime(e.getX());
        LogViewerFilter filter = tc.calculateFilter(e.getY());
        String agent = tc.calculateAgent(e.getY());
        boolean alt = false;

        if (((e.getModifiers() & MouseEvent.BUTTON2_MASK) != 0) ||
            ((e.getModifiers() & MouseEvent.BUTTON3_MASK) != 0)) {
            if (!selectAltEntry(filter, time, agent)) {
                for (int i = 1; i < 1 + Math.ceil(10 / TICKWIDTH); i++) {
                    if (selectAltEntry(filter, time - i, agent))
                        return;
                    if (selectAltEntry(filter, time + i, agent))
                        return;
                }
            }

        } else {

            if (!selectEntry(filter, time, agent)) {
                for (int i = 1; i < 1 + Math.ceil(10 / TICKWIDTH); i++) {
                    if (selectEntry(filter, time + i, agent))
                        return;
                    if (selectEntry(filter, time - i, agent))
                        return;
                }
            }
        }
    }
    public void mouseEntered(MouseEvent e) { }
    public void mouseExited(MouseEvent e) {
        foot.setText(blank);
    }
    public void mousePressed(MouseEvent e) {
        tc.requestFocus();
        spot = e.getPoint();
    }
    public void mouseReleased(MouseEvent e) { }

    /**
     */
    public int hilightMatches(String text) {
        Enumeration e = searched.elements();

        while (e.hasMoreElements()) {
            LogViewerEntrySlot v = (LogViewerEntrySlot)e.nextElement();
            v.setColor(null);
        }
        searched.clear();

        if (text != null) {
            e = viewer.getFilters();
            while (e.hasMoreElements()) {
                LogViewerFilter f = (LogViewerFilter)e.nextElement();
                if (!f.isActive()) continue;
                synchronized (f) {
                    for (Iterator ie = f.getEntryAgents(); ie.hasNext();) {
                        String agent = (String)ie.next();
                        TreeMap t = f.getEntries(agent);
                        Iterator i = t.keySet().iterator();
                        while (i.hasNext()) {
                            Integer time = (Integer)i.next();
                            LogViewerEntrySlot v = f.getEntry(time, agent);
                            for (Enumeration en = v.elements(); en.hasMoreElements();) {
                                LogViewerEntry le = (LogViewerEntry)en.nextElement();
                                if (le.getData().indexOf(text) >= 0) {
                                    searched.addElement(v);
                                    break;
                                }
                            }
                        }
                    }
                }
            }
        }
        tc.repaint();

        foot.setText(" " + searched.size() + " matches found");

        return searched.size();
    }

    public boolean selectAltEntry(LogViewerFilter filter, int time, String agent) {
        LogViewerEntrySlot old = althighlighted;
        JTextArea text = text1;
        JScrollPane scroll = textscroll1;

        if (old != null) {
            old.setColor(null);
        }
        althighlighted = null;
        if ((time >= 0) && (filter != null) && (agent != null)) {
            LogViewerEntrySlot v = (LogViewerEntrySlot)filter.getEntry(new Integer(time), agent);
            if (v != null) {
                althighlighted = v;
                if (althighlighted != old) {
                    text.setText(v.toString());
                }
                //scroll.getViewport().setViewPosition(new Point(0,0));
            } else {
                text.setText("");
            }
        }
        if (old != althighlighted)
            tc.repaint();

        return (althighlighted != null);
    }

    public boolean selectEntry(LogViewerFilter filter, int time, String agent) {
        LogViewerEntrySlot old = highlighted;

        if (old != null) {
            old.getFilter().deselect(old, viewer);
            old.setColor(null);
        }
        highlighted = null;
        if ((time >= 0) && (filter != null) && (agent != null)) {
            LogViewerEntrySlot v = (LogViewerEntrySlot)filter.getEntry(new Integer(time), agent);
            JTextArea text = text2;
            JScrollPane scroll = textscroll2;
            if (v != null) {
                highlighted = v;
                if (highlighted != old) {
                    text.setText(v.toString());
                }
                filter.select(v, viewer);
                //scroll.getViewport().setViewPosition(new Point(0,0));
            } else {
                text.setText("");
            }
        }
        if (old != highlighted)
            tc.repaint();

        return (highlighted != null);
    }

    public void drawString(Graphics g, String s, int x, int y, int align) {
        FontMetrics metrics = g.getFontMetrics();
        int width = metrics.stringWidth(s);
        
        y -= metrics.getDescent();
        
        switch (align) {
        case 0: // Left
            x += PAD;
            break;
        case 1: // Center
            x -= width / 2;
            break;
        case 2: // Right
            x -= width + PAD;
            break;
        }
        
        g.drawString(s, x, y);
    }

    public int calculateDataHeight() {
        int h = LINE * 2;
        
        for (Enumeration e = viewer.getFilters(); e.hasMoreElements();) {
            LogViewerFilter f = (LogViewerFilter)e.nextElement();
            if (f.isActive())
                h += LINE * Math.max(f.getNumAgents(), 1);
        }

        return h;
    }
    
    /**
     * This draws the token legend, indicating
     * the various token types
     */
    class TokenCanvas extends JPanel {
        LogViewer viewer;

        public TokenCanvas(LogViewer v) {
            viewer = v;
        }

        public boolean isOpaque() { return true; }

        public void paintComponent(Graphics g) {
            int y = 0;
            int x = 100;
            Color c;

            c = g.getColor();
            g.setColor(Color.white);
            g.fillRect(0, 0, getWidth(), getHeight());
            g.setColor(c);

            Enumeration e = viewer.getFilters();
            for (y += LINE; e.hasMoreElements(); y += LINE) {
                LogViewerFilter f = (LogViewerFilter)e.nextElement();

                drawString(g, f.getLabel(), 0, y, 0);

                c = g.getColor();
                for (int type = 0; f.getTokenType(type) != null; type++) {
                    f.drawTokenType(g, x, y, type);
                    drawString(g, f.getTokenType(type), x + PAD * 2, y, 0);
                    if (f.getTokenType(type+1) != null) y += LINE;
                }
                g.setColor(c);

                if (e.hasMoreElements()) {
                    c = g.getColor();
                    g.setColor(Color.lightGray);
                    g.drawLine(0, y, getWidth(), y);
                    g.setColor(c);
                }
            }
        }

        public Dimension getPreferredSize() {
            int w = 150;
            int h = 0;
            FontMetrics metrics = getFontMetrics(getFont());

            Enumeration e = viewer.getFilters();
            while (e.hasMoreElements()) {
                LogViewerFilter f = (LogViewerFilter)e.nextElement();
                if (f.getTokenType(0) == null) h += LINE;
                for (int type = 0; f.getTokenType(type) != null; type++) {
                    h += LINE;
                    w = Math.max(w, 100 + PAD * 4 + metrics.stringWidth(f.getTokenType(type)));
                }
            }

            return new Dimension(w, h);
        }
        public Dimension getMinimumSize() { return getPreferredSize(); }
        public Dimension getMaximumSize() { return getPreferredSize(); }
    }
    
    /**
     * This draws the "legend" on the left hand side, indicating
     * the row labels
     */
    class LegendCanvas extends JPanel {
        LogViewer viewer;
        Graphics lastg = null;

        public LegendCanvas(LogViewer v) {
            viewer = v;
        }

        public boolean isOpaque() { return true; }

        public void paintComponent(Graphics g) {
            int y = LINE + tcscroll.getInsets().top;
            Color c;

            lastg = g;

            c = g.getColor();
            g.setColor(Color.white);
            g.fillRect(0, 0, getWidth(), getHeight());
            g.setColor(c);

            drawString(g, "Time", getWidth(), y, 2);
            y += LINE;
            
            Enumeration e = viewer.getFilters();
            for (y += LINE; e.hasMoreElements(); y += LINE) {
                LogViewerFilter f = (LogViewerFilter)e.nextElement();
                if (!f.isActive()) {
                    y -= LINE;
                    continue;
                }

                drawString(g, f.getLabel(), 0, y, 0);

                c = g.getColor();
                synchronized (f) {
                    for (Iterator ie = f.getEntryAgents(); ie.hasNext();) {
                        String agent = (String)ie.next();
                        /*
                          g.setColor(Color.black);
                          drawString(g, agent, getWidth(), y + 1, 2);
                        */
                        g.setColor(LogViewer.getColor(agent));
                        drawString(g, agent, getWidth() - 1, y, 2);
                        if (ie.hasNext()) y += LINE;
                    }
                }
                g.setColor(c);

                if (e.hasMoreElements()) {
                    c = g.getColor();
                    g.setColor(Color.lightGray);
                    g.drawLine(0, y, getWidth(), y);
                    g.setColor(c);
                }
            }
        }

        public Dimension getPreferredSize() {
            int w = 125;
            FontMetrics metrics = getFontMetrics(getFont());

            Enumeration e = viewer.getFilters();
            while (e.hasMoreElements()) {
                LogViewerFilter f = (LogViewerFilter)e.nextElement();
                for (Iterator ie = f.getEntryAgents(); ie.hasNext();) {
                    String agent = (String)ie.next();
                    w = Math.max(w, PAD * 4 + metrics.stringWidth(f.getLabel()) + metrics.stringWidth(agent));
                }
            }

            return new Dimension(w, calculateDataHeight());
        }
        public Dimension getMinimumSize() { return getPreferredSize(); }
        public Dimension getMaximumSize() { return getPreferredSize(); }
    }

    /**
     * This draws the "timeline" on the right hand side, containing
     * most of the interesting data
     */
    class TimelineCanvas extends JPanel {
        private int HINSET = PAD * 4;

        LogViewer viewer;

        public TimelineCanvas(LogViewer v) {
            viewer = v;
            setAutoscrolls(true);
        }

        public boolean isOpaque() { return true; }
        public boolean isFocusable() { return true; }

        public void paintComponent(Graphics g) {
            int y = LINE * 2;
            int start = getStartTime();
            int finish = getFinishTime();
            Color c;
            Rectangle bounds = tcscroll.getViewport().getViewRect();
            
            // Limit what we actually draw
            int vstart = calculateTime(bounds.x);
            if (vstart < 0) vstart = start;
            vstart -= PAD;
            int vfinish = calculateTime(bounds.x + bounds.width);
            if (vfinish < 0) vfinish = finish;
            vfinish += PAD;

            c = g.getColor();
            g.setColor(Color.white);
            //g.fillRect(0, 0, getWidth(), getHeight());
            g.fillRect(bounds.x, bounds.y, bounds.width, bounds.height);
            g.setColor(c);

            if (getFinishTime() < 0) {
                drawString(g, "No Data Found", getWidth() / 2, getHeight() / 2, 1);
                return;
            }

            g.drawLine(HINSET, y - PAD * 2,
                       (finish - start) * TICKWIDTH + HINSET, y - PAD * 2);
            for (int i = finish - start; i >= 0; i -= TICKINC) {
                if (i + start < vstart - TICKWIDTH) continue;
                if (i + start > vfinish + TICKWIDTH) continue;
                g.drawLine(HINSET + i * TICKWIDTH, y - PAD,
                           HINSET + i * TICKWIDTH, y - PAD * 3 + 2);
                drawString(g, String.valueOf(i + start),
                           HINSET + i * TICKWIDTH,  y - LINE, 1);
            }
            c = g.getColor();
            g.setColor(Color.lightGray);
            for (int i = (int)(Math.floor(finish / 100) * 100) - start; i >= 0; i -= 100) {
                g.drawLine(HINSET + i * TICKWIDTH, y,
                           HINSET + i * TICKWIDTH, y + getHeight());
            }
            g.setColor(c);

            Enumeration e = viewer.getFilters();
            for (y += LINE; e.hasMoreElements(); y += LINE) {
                LogViewerFilter f = (LogViewerFilter)e.nextElement();
                if (!f.isActive()) {
                    y -= LINE;
                    continue;
                }

                synchronized (f) {
                    for (Iterator ie = f.getEntryAgents(); ie.hasNext();) {
                        String agent = (String)ie.next();
                        TreeMap t = f.getEntries(agent);
                        Iterator i = t.keySet().iterator();
                        while (i.hasNext()) {
                            Integer time = (Integer)i.next();
                            if (time.intValue() < vstart) continue;
                            if (time.intValue() > vfinish) continue;
                            LogViewerEntrySlot v = (LogViewerEntrySlot)t.get(time);
                            Color back = null;
                            if (v == highlighted) {
                                back = hilightcolor;
                            } else if (v == althighlighted) {
                                back = althilightcolor;
                            } else if (v.getColor() != null) {
                                back = v.getColor();
                            } else if (searched.contains(v)) {
                                back = searchcolor;
                            }
                            if (back != null) {
                                c = g.getColor();
                                g.setColor(Color.lightGray);
                                //g.drawRect(HINSET - 1 + (time.intValue() - start) * TICKWIDTH - TICKWIDTH / 2, y - LINE - 1, TICKWIDTH + 1, LINE + 1);
                                g.drawRect(HINSET - 3 + (time.intValue() - start) * TICKWIDTH - TICKWIDTH / 2, y - LINE - 1, TICKWIDTH + 5, LINE + 1);
                                g.setColor(back);
                                g.fillRect(HINSET - 2 + (time.intValue() - start) * TICKWIDTH - TICKWIDTH / 2, y - LINE, TICKWIDTH + 4, LINE);
                                g.setColor(c);
                            }
                            f.drawToken(g, HINSET + (time.intValue() - start) * TICKWIDTH, y, f.getEntry(time, agent));
                        }
                        if (ie.hasNext()) y += LINE;
                    }
                }

                if (e.hasMoreElements()) {
                    c = g.getColor();
                    g.setColor(Color.lightGray);
                    g.drawLine(0, y, getWidth(), y);
                    g.setColor(c);
                }
            }
        }

        public int calculateTime(int x) {
            int time = (x - HINSET + (TICKWIDTH / 2)) / TICKWIDTH + getStartTime();
            if (time < getStartTime() || time > getFinishTime())
                return -1;
            else
                return time;
        }

        public LogViewerFilter calculateFilter(int y) {
            int h = LINE * 2;
        
            for (Enumeration e = viewer.getFilters(); y > h && e.hasMoreElements();) {
                LogViewerFilter f = (LogViewerFilter)e.nextElement();
                if (!f.isActive()) continue;
                h += LINE * Math.max(f.getNumAgents(), 1);
                if (y < h) return f;
            }

            return null;
        }

        public String calculateAgent(int y) {
            int h = LINE * 2;
        
            for (Enumeration e = viewer.getFilters(); y > h && e.hasMoreElements();) {
                LogViewerFilter f = (LogViewerFilter)e.nextElement();
                if (!f.isActive()) continue;
                if (y < h + LINE * Math.max(f.getNumAgents(), 1)) {
                    if (f.getNumAgents() <= 0) return null;
                    h += LINE;
                    for (Iterator ie = f.getEntryAgents(); ie.hasNext(); h += LINE) {
                        String s = (String)ie.next();
                        if (y < h) return s;
                    }
                } else {
                    h += LINE * Math.max(f.getNumAgents(), 1);
                }
            }

            return null;
        }

        public int getStartTime() {
            int time = viewer.getStartTime();
            return time - (time % TICKINC);
        }

        public int getFinishTime() {
            int time = viewer.getFinishTime();
            return time + TICKINC - (time % TICKINC);
        }

        public Dimension getPreferredSize() {
            int w = Math.max((getFinishTime() - getStartTime()) * TICKWIDTH + HINSET * 2, 100);
            return new Dimension(w, calculateDataHeight());
        }
        public Dimension getMinimumSize() { return getPreferredSize(); }
        public Dimension getMaximumSize() { return getPreferredSize(); }
    }
}
