/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * LogViewerFilter.java
 ***************************************************************************************/

package agent.simplest;

import java.util.*;
import java.awt.*;

/**
 * This class performing the filtering for the LogViewer component.  One
 * LogViewerFilter will exist for each type of log activity you wish to
 * show up in the viewer - thus if you want to add a line to the viewer
 * for activity generated by your component you need to create one of
 * these filters for it.  The following code, which should go in your
 * component's begin() method, is a trivial example of how to do that.
 * <P>
 * <PRE>
 *   if (State.hasComponent("LogViewer")) {
 *       LogViewer viewer = (LogViewer)State.findComponent("LogViewer");
 *
 *       viewer.addFilter(new LogViewerFilter("Msg Sent") {
 *               public boolean isMyLogEntry(String s) {
 *                   return s.indexOf("Message Sent: ") != -1;
 *               }
 *           });
 *   }
 * </PRE>
 * <P>
 * The filter created above checks for log lines which contain the
 * phrase "Message Sent: ", which corresponds to a message being sent
 * by the Communicate component.
 * <P>
 * Each LogViewerFilter must define, at a minimum, the isMyLogEntry
 * method.  This method will be called once for each line in each
 * logfile the LogViewer is parsing, and should return true when
 * a line is seen that represents the activity the filter is looking
 * for.  In the above case, isMyLogEntry returns true whenever
 * it sees "Message Sent: " in the log line.  Note that because this
 * methods gets called so freqently, you should ensure that it runs
 * as fast as possible.
 * <P>
 * Overriding only this one method will give you a usable LogViewerFilter,
 * which places bullet-style tokens on the timeline to represent activity, and
 * prints the simple log line out in the text field when selected.  More
 * advanced filters may optionally choose to control the appearance token
 * (by overriding drawToken), or the data written to the text field (by
 * overriding formatMyLogEntry).  You may also control, to a certain
 * extent, the behavior when an entry is selected by overriding the
 * select and deselect methods.
 * <P>
 * If you wish to use the LogViewer on log files other than those 
 * formatted in the JAF log file style, you may also need to override
 * (at least) the getTime and getAgent methods.
 */
public abstract class LogViewerFilter {
    public String label;
    protected boolean active;
    public TreeMap entries = new TreeMap();
    public Vector myhighlights = new Vector();
    protected LogViewer viewer = null;

    /**
     * Creates an active filter.
     * @param l The filter's label
     */
    public LogViewerFilter(String l) {
        this (l, true);
    }

    /**
     * @param l The filter's label
     * @param a Whether the filter should be active or not
     */
    public LogViewerFilter(String l, boolean a) {
        label = l;
        active = a;
    }

    /**
     * Returns the filter's label
     */
    public String getLabel() { return label; }

    /**
     * Returns whether the filter is active or not
     */
    public boolean isActive() { return active; }

    /**
     * A reference to the containing LogViewer, if any
     */
    public LogViewer getLogViewer() { 
        return viewer;
    }
    public void setLogViewer(LogViewer v) { 
        viewer = v;
    }

    /**
     * An inactive filter will not be drawn on the timeline,
     * and will not be called during logfile parsing
     */
    public void setActive(boolean a) { active = a; }

    public Iterator getEntryAgents() {
        return entries.keySet().iterator();
    }
    public synchronized int getNumAgents() { return entries.size(); }
    public TreeMap getEntries(String agent) { return (TreeMap)entries.get(agent); }

    protected LogViewerFilter findLogViewerFilter(String name) {
        return viewer.findFilter(name);
    }

    public void clearAllEntries() {
        entries.clear();
    }

    /**
     * Called when the given LogViewerEntrySlot is selected
     */
    public void select(LogViewerEntrySlot v, LogViewer viewer) { }

    /**
     * Called when the given LogViewerEntrySlot is deselected
     */
    public void deselect(LogViewerEntrySlot v, LogViewer viewer) {
        Enumeration e = myhighlights.elements();
        while (e.hasMoreElements()) {
            ((LogViewerEntrySlot)e.nextElement()).setColor(null);
        }
    }

    /**
     * @return true if the time slot should be removed once the
     * clear has completed
     */
    public synchronized boolean clearEntry(Integer time, String agent, Object data) {
        if ((time == null) || (agent == null) || (data == null)) return false;
        if (entries.containsKey(agent)) {
            TreeMap t = (TreeMap)entries.get(agent);
            if (t.containsKey(time)) {
                LogViewerEntrySlot v = (LogViewerEntrySlot)t.get(time);
                v.removeElement(data);
                if (v.isEmpty()) return true;
            }
        }
        return false;
    }

    /**
     * Gets a specific entry for the log filter.  This is generally
     * just for the LogViewer component to call during display event
     * handling.
     */
    public synchronized LogViewerEntrySlot getEntry(Integer time, String agent) {
        if ((time == null) || (agent == null)) return null;
        if (entries.containsKey(agent)) {
            TreeMap t = (TreeMap)entries.get(agent);
            if (t.containsKey(time)) {
                LogViewerEntrySlot v = (LogViewerEntrySlot)t.get(time);
                return v;
            }
        }

        return null;
    }
    
    /**
     * Adds an entry to the log filter.  This is generally just for the
     * LogViewer component to call during processing.
     */
    public synchronized void addEntry(Integer time, String agent, Object data) {
        TreeMap t;
        LogViewerEntrySlot v;

	if (viewer.getTimeMultiplier() != 1) {
		time = new Integer((int)Math.round(time.floatValue() * viewer.getTimeMultiplier()));
	}

        if (!entries.containsKey(agent)) {
            t = new TreeMap();
            entries.put(agent, t);
        } else {
            t = (TreeMap)entries.get(agent);
        }

        if (!t.containsKey(time)) {
            v = new LogViewerEntrySlot(this, time, agent);
            t.put(time, v);
        } else {
            v = (LogViewerEntrySlot)t.get(time);
        }

        v.addElement(data);
    }

    /**
     * Override this function to return true or false, depending
     * on whether or not the provided log entry line represents
     * the type of activity your filter is detecting.
     * @param s A log file line
     * @return True if that line is recognized by the filter,
     * false otherwise
     */
    public abstract boolean isMyLogEntry(String s);

    /**
     * Override this function to control the data which is
     * written to the textual display when the log entry is
     * selected.
     * @param s The log file string to format
     * @return The "formatted" version of that log line,
     * presumably easier for humans to read.
     */
    public String formatMyLogEntry(String s) {
        StringBuffer buf = new StringBuffer(s);
        int index = -1;

        while ((index = buf.toString().indexOf(LogItem.ANSI_S)) >= 0) {
            int end = buf.toString().indexOf('m', index);
            if (end < 0) break;
            buf.replace(index, end+1, "");
        }

        return buf.toString();
    }

    /**
     * Override this function to control the shape and color
     * of the token representation of your log activity.  By
     * default a simple bullet will be drawn.
     * <P>
     * Although you can draw tokens to be whatever size you
     * want, we recommend at this time that you fit them
     * as best possible within a 5x15 pixel area.  The bottom
     * of this area should be the y coordinate given, while
     * it should be centered along the x coordinate.  Like
     * this:
     * <PRE>
     *    #####
     *    #####
     *    #####
     *    #####
     *    #####
     *  Y #####
     *      X
     * </PRE>
     */
    public void drawToken(Graphics g, int x, int y, LogViewerEntrySlot v) {
        drawTokenType(g, x, y, 0);
    }

    /**
     * Individual token types should be drawn within this function,
     * where the type parameter indicates the token type to draw.
     * Type numbers must start at 0 and increment continuously.
     * <P>
     * You should override this function to draw your tokens, as
     * returned by the getTokenType function.
     * <P>
     * You may also make calls from your drawTokenType to this
     * drawTokenType, to take advantage of some of the "default"
     * tokens which this can draw.  The following type numbers
     * draw the specified symbols:<BR>
     * <BLOCKQUOTE><PRE>
     * 0 - Dot
     * 1 - Up arrow
     * 2 - Down arrow
     * 3 - X
     * 4 - x (little x)
     * 5 - Right arrow
     * 6 - Left arrow
     * 7 - Exclamation point
     * 8 - Hat (^)
     * 9 - Vee (v)
     * 10 - Plus
     * 11 - Minus
     * 12 - Front slash
     * 13 - Back slash
     * 14 - Empty dot
     * 15 - Filled square
     * 16 - Empty square
     * 17 - Little dot
     * 18 - Question mark
     * 19 - Big dot
     * </PRE></BLOCKQUOTE>
     */
    public void drawTokenType(Graphics g, int x, int y, int type) {
        switch (type) {
        case 0: // dot
            g.fillOval(x - 2, y - 10, 5, 5);
            break;
        case 1: // down arrow
            g.drawLine(x, y - 2, x + 0, y - 13);
            g.drawLine(x, y - 2, x - 2, y - 4);
            g.drawLine(x, y - 2, x + 2, y - 4);
            break;
        case 2: // up arrow
            g.drawLine(x, y - 13, x + 0, y - 2);
            g.drawLine(x, y - 13, x - 2, y - 11);
            g.drawLine(x, y - 13, x + 2, y - 11);
            break;
        case 3: // X
            g.drawLine(x + 2, y - 11, x - 2, y - 4);
            g.drawLine(x - 2, y - 11, x + 2, y - 4);
            break;
        case 4: // x
            g.drawLine(x - 2, y - 10, x + 2, y - 5);
            g.drawLine(x - 2, y - 5, x + 2, y - 10);
            break;
        case 5: // right arrow
            g.drawLine(x - 2, y - 7, x + 2, y - 7);
            g.drawLine(x, y - 9, x + 2, y - 7);
            g.drawLine(x, y - 5, x + 2, y - 7);
            break;
        case 6: // left arrow
            g.drawLine(x - 2, y - 7, x + 2, y - 7);
            g.drawLine(x - 2, y - 7, x, y - 9);
            g.drawLine(x - 2, y - 7, x, y - 5);
            break;
        case 7: // exclamation
            g.drawLine(x, y - 13, x, y - 6);
            g.fillOval(x - 1, y - 4, 3, 3);
            break;
        case 8: // hat
            g.drawLine(x - 2, y - 7, x, y - 9);
            g.drawLine(x + 2, y - 7, x, y - 9);
            break;
        case 9: // vee
            g.drawLine(x - 2, y - 9, x, y - 7);
            g.drawLine(x + 2, y - 9, x, y - 7);
            break;
        case 10: // plus
            g.drawLine(x - 2, y - 8, x + 2, y - 8);
            g.drawLine(x, y - 10, x, y - 6);
            break;
        case 11: // minus
            g.drawLine(x - 2, y - 8, x + 2, y - 8);
            break;
        case 12: // front slash
            g.drawLine(x + 2, y - 11, x - 2, y - 4);
            break;
        case 13: // back slash
            g.drawLine(x - 2, y - 11, x + 2, y - 4);
            break;
        case 14: // empty dot
            g.drawOval(x - 2, y - 10, 4, 4);
            break;
        case 15: // filled square
            g.fillRect(x - 2, y - 10, 4, 4);
            break;
        case 16: // empty square
            g.drawRect(x - 2, y - 10, 4, 4);
            break;
        case 17: // little dot
            g.fillOval(x - 1, y - 9, 3, 3);
            break;
        case 18: // question mark
            g.drawArc(x - 2, y - 13, 4, 4, -90, 270);
            g.drawLine(x, y - 8, x, y - 6);
            g.fillOval(x - 1, y - 4, 3, 3);
            break;
        case 19: // big dot
            g.fillOval(x - 3, y - 11, 7, 7);
            break;
        }
    }

    /**
     * Individual token types should have string representations,
     * for the legend.  This function should return the string
     * representation for a given type number, or null if no
     * type number exists.
     */
    public String getTokenType(int type) {
        return null;
    }

    /**
     * Given a log line, returns the time indicated by that
     * log line, or null if not found.
     * <P>
     * This specific implementation will extract the time
     * from the first a block of information that matches
     * "\[(\d+):" in the log line. (e.g. "[123:")
     * <P>
     * If you have an alternate log file format you wish to
     * use, it should override this function as necessary.
     * @param a The log file line
     * @return The time as a numeric string, or null if not found
     */
    public String getTime(String s) {
        char c[] = s.toCharArray();

        for (int i = 0; i < c.length; i++) {
            if (c[i] == '[') {
                for (int j = i+1; j < c.length; j++) {
                    if (c[j] == ':') {
                        if (i+1 < j)
                            return s.substring(i+1, j);
                        else
                            break;
                    } else if (!Character.isDigit(c[j])) {
                        break;
                    }
                }
            }
        }

        return null;
    }

    /**
     * Given a log line, returns the agent name indicated by that
     * log line, or null if not found.
     * <P>
     * This specific implementation will extract the time
     * from the first a block of information that looks more or less like
     * "\[[^:]:(\w+)[:|\]" in the log line (e.g. "[123:AgentA:")
     * <P>
     * If you have an alternate log file format you wish to
     * use, it should override this function as necessary.
     * @param a The log file line
     * @return The agent name, or null if not found
     */
    public String getAgent(String s) {
        char c[] = s.toCharArray();

        for (int i = 0; i < c.length; i++) {
            if (c[i] == '[') {
                for (int j = i+1; j < c.length; j++) {
                    if (c[j] == ':') {
                        for (int k = j+1; k < c.length; k++) {
                            if ((c[k] == ':') || (c[k] == ']')) {
                                if (j+1 < k)
                                    return s.substring(j+1, k);
                                else
                                    break;
                            }
                        }
                        break;
                    } else if (!Character.isDigit(c[j])) {
                        break;
                    }
                }
            }
        }

        return null;
    }
}
