/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * Observe.java
 *
 * manifest: Name: agent/simplest/Observe.class
 * manifest: Java-Bean: True
 ***************************************************************************************/

package agent.simplest;

/* Global includes */
import java.awt.*;               // ScrollPane, PopupMenu, MenuShortcut, etc.
import java.io.*;                // Object serialization streams.
import java.util.*;
import java.awt.event.*;
import javax.swing.*;

import agent.base.*;


/**
 * An observation engine.  This component accepts Observers, to which it
 * will pass events (to be observed) and keep up to date as far as 
 * the current time is concerned.  To use this, just send in an Observer
 * (addObserver).  Keep the Observer's reference to obtain the counts as
 * they are produced.  Call removeObserver when you are done with it.
 * <P>
 * Here's an example:
 * <PRE>
 * Observe observe = (Observe)State.findComponent("Observe");
 * Observer observer;
 * <BR>
 * observe.addObserver(observer = new Observer("Messages Sent") {
 *   public boolean checkEvent(AgentEvent e) {
 *     if ((e instanceof MessageEvent) && (e.getID() == MessageEvent.SENT))
 *       return true;
 *     return false;
 *   }
 * });
 * <P>
 * <I>And to access the data...</I>
 * <P>
 * float count = observer.getCount();
 * </PRE>
 * @see Observer
 */
public class Observe extends agent.base.AgentComponent implements LogEventListener, PropertyEventListener, MouseListener  {
  static final long serialVersionUID = 1234567890L; // kludge static ser_id.  Bad bad Bryan!
  protected Vector observers = new Vector(25);
  protected Log log;
  protected int FAC_OBSERVE;
  private String title = "Observe"; // Title of the windows
  private transient JFrame window;   // Windows 
  private HistogramComponent hc;    // Histogram Component
  private static PopupMenu menu = new PopupMenu("Current Actions");

  /**
   * Default constructor
   */
  public Observe() {
    super();
    
    addDependency("Log");
    addDependency("State");

    addMouseListener(this);
    add(menu);
  }

  /**
   * Sets up component helpers
   */
  public void init() {
    log = (Log)agent.simplest.State.findComponent("Log");
    log.addLogEventListener(this);

    FAC_OBSERVE = log.getFacilityID(this);

    State state = (State)agent.simplest.State.findComponent("State");
    state.addPropertyEventListener(this);
    
    if (hasGUI()) {
        if (State.hasComponent("WindowManager")) {
            WindowManager wm = (WindowManager)state.findComponent("WindowManager");
            window = wm.createWindow(title);
        } else
            window = new JFrame(title);
        hc = new HistogramComponent(null, 800, 500);
        //    window.setLayout(null);       
        window.setSize(window.getInsets().left + window.getInsets().right + 800,
                       window.getInsets().top + window.getInsets().bottom + 500); 
    
        window.getContentPane().add(hc);
        window.pack();
        if (!agent.simplest.State.hasComponent("WindowManager"))
            window.setVisible(true);
    }

    super.init();
  }

  /**
   * Resets the component
   */
  public void reset() {

    log.log("Removing all registered observers", Log.LOG_INFO, FAC_OBSERVE);
    observers.removeAllElements();
  }

  /**
   * Prints out a final summary
   */
  public void end() {
    if (window != null)
      window.dispose();

    if (log == null) return;

    log.log("Observation summary:", Log.LOG_INFO, FAC_OBSERVE);

    Enumeration e = getObservers();
    while (e.hasMoreElements()) {
      Observer o = (Observer)e.nextElement();
      StringBuffer pad = new StringBuffer("");

      for (int j = 0, len = o.getDescription().length(); j < (25 - len); j++)
          pad.insert(0, " ");

      log.log(pad + o.toString(),  Log.LOG_INFO, FAC_OBSERVE);
    }
  }

  /**
   * Adds an observer to the pool of active observers.  You can
   * call this method from init.
   * @param o The observer to add
   */
  public void addObserver(Observer o) {

    observers.addElement(o);

    if (log == null)
        log = (Log)State.findComponent("Log");

    if (log != null)
      log.log("Observer activated: " + o, Log.LOG_DEBUG, FAC_OBSERVE);
    else
      System.err.println("Observer activated: " + o);
  }

  /**
   * Removes an observer from the pool
   * @param o The observer to remove
   */
  public void removeObserver(Observer o) {

    observers.removeElement(o);
    log.log("Observer removed: " + o, Log.LOG_DEBUG, FAC_OBSERVE);
  }

  /**
   * Finds an observer, based on its description.
   * @param d The description to search for
   * @return A reference to the first observer matching the
   *  description, or null if not found
   */
  public Observer findObserver(String d) {

    Enumeration e = getObservers();
    while (e.hasMoreElements()) {
      Observer o = (Observer)e.nextElement();

      if (o.getDescription().equalsIgnoreCase(d))
	return o;
    }

    return null;
  }

  /**
   * Returns an enumeration of the active observers
   * @return An enumeration
   */
  public Enumeration getObservers() {

    return observers.elements();
  }

  /**
   * Relays an event to the pertinant observers.
   * @param e the event to add
   */
  public void addEvent(AgentEvent ev) {

    Enumeration e = getObservers();
    while (e.hasMoreElements()) {
      Observer o = (Observer)e.nextElement();

      if (o.usesEvents())
	o.addEvent(ev);
    }
  }

  /**
   * pulsing function, use to update the display
   */
  public void pulse() {
      if ((window != null) && window.isShowing()) {
          Enumeration e = getObservers();
          int count = 0;
          Observer o;
          Vector v = new Vector();
          while (e.hasMoreElements()) {
              o = (Observer)e.nextElement();
              if (o.getDisplay()) {
                  v.insertElementAt(o,count);
                  count = count + 1;
              }
          }
      
          int[] data = new int[count];
          String[]     labels = new String[count];

          int max = observers.size();
          for (int i=0 ; i < count ; i++) {
              o = (Observer)v.elementAt(i);
              data[i] = (int)o.getCount();
              labels[i] = o.getDescription();
          }
      
          hc.setHistogramData(data);
          hc.setHistogramLabels(labels);
          validate();
          hc.drawHisto();
          //      hc.repaint();
      }
  }

  /**
   * Gets called whenever the time changes.  This in turns 
   * notifies each of the observers, allowing them to update their
   * windows or perform whatever per-pulse observations they need
   * to do.
   * @param time The current time
   */
  public void newTime(int time) {

    Enumeration e = getObservers();
    while (e.hasMoreElements()) {
      Observer o = (Observer)e.nextElement();

      o.newTime(time);
    }
  }

  /**
   * Watch for interesting property changes.  Monitors time changes.
   * @param e The event to look at 
   */
  public void propertyRemoved(PropertyEvent e) { }
  public void propertyAdded(PropertyEvent e) { propertyChanged(e); }
  public void propertyChanged(PropertyEvent e) { 
    String key = e.getKey().toString();

    if (key.equalsIgnoreCase(State.TIME)) {
      // Get count
      int time = ((Integer)e.getProperty()).intValue();
      if (time >= 0) {
	newTime(time);
      }
    }
  }

  /**
   * Called something is logged
   */
  public void errorLog(LogEvent e) { }
  public void eventLog(LogEvent e) { }
  public void stringLog(LogEvent e) { }
  public void objectLog(LogEvent e) { }
  public void eventReflection(AgentEvent e) {
    addEvent(e);
  }

  /**
   * Returns true if the component has a popupmenu.
   */
  public boolean hasPopupMenu() {
      return true;
  }

  /**
   * Responds to mouse clicks
   */
    public void processMouseEvent(MouseEvent e) {
        MenuItem item;

        if ((e.getID() == MouseEvent.MOUSE_PRESSED) || e.isPopupTrigger()) {
            menu.removeAll();
            Enumeration en = getObservers();
            while (en.hasMoreElements()) {
                Observer obs = (Observer)en.nextElement();
                item = new MenuItem(obs.toString());
                item.addActionListener(new MyObserverMenuListener(obs));
                menu.add(item);
            }
            menu.show(this, e.getX(), e.getY());
        }
    }
  public void mouseClicked(MouseEvent e) { processMouseEvent(e); }
  public void mousePressed(MouseEvent e) { processMouseEvent(e); }
  public void mouseReleased(MouseEvent e) { processMouseEvent(e); }
  public void mouseDragged(MouseEvent e) { processMouseEvent(e); }
  public void mouseEntered(MouseEvent e) { }
  public void mouseExited(MouseEvent e) { }
  public void mouseMoved(MouseEvent e) { }

    /**
     * Little class to handle observer printouts
     */
    class MyObserverMenuListener implements ActionListener {
        Observer obs;
        public MyObserverMenuListener(Observer obs) {
            this.obs = obs;
        }
        public void actionPerformed(java.awt.event.ActionEvent event) {
            System.err.println(obs.dumpData());
        }
    }
}
