/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * Observer.java
 ***************************************************************************************/

package agent.simplest;

import java.util.*;

import agent.base.*;

/**
 * Generic observer class.  Use this to monitor a particular event type
 * (typically), or make per-pulse observations of other non standard
 * data.  To do so, create a subclass of Observer, overriding the
 * necessary functions (usually checkEvent), and add the new Observer
 * to your Observe component.  Keep a reference to the Observer and
 * use its getCount function to look at the current statistics.
 * You can display your observation in the window by adding an optional
 * argument to set the display
 * <P>
 * Here's an example:
 * <PRE>
 * Observe observe = (Observe)State.findComponent("Observe");
 * Observer observer;
 * <BR>
 * observe.addObserver(observer = new Observer("Messages Sent", <B>true</B>) {
 *   public boolean checkEvent(AgentEvent e) {
 *     if ((e instanceof MessageEvent) && (e.getID() == MessageEvent.SENT))
 *       return true;
 *     return false;
 *   }
 * });
 * <P>
 * @see agent.simplest.Observe
 */
public class Observer {
  public static final int NO_WINDOW = 0;
  protected int count, window, index;
  protected String description = "Data";
  protected Hashtable exdata = new Hashtable(5);
  protected double data[];
  protected boolean display;

  /**
   * Constructor, uses NO_WINDOW
   * @param d The description of the data observed by this class
   */
  public Observer(String d) {
    this(d, NO_WINDOW,true);
  }

   /**
   * Constructor, uses NO_WINDOW
   * @param d The description of the data observed by this class
   * @param display Set the display to true or false
   */
  public Observer(String d, boolean display) {
    this(d, NO_WINDOW, display);
  }

  /**
   * Constructor
   * @param d The description of the data observed by this class
   * @param w The event window to use
   */
  public Observer(String d, int w) {
    this(d,w,true);
  }
  
   /**
   * Constructor
   * @param d The description of the data observed by this class
   * @param w The event window to use
   * @param dis Set the display to true or false
   */
  public Observer(String d, int w, boolean dis) {
    
    display = dis;
    description = d;
    window = w;
    index = 0;

    if (window > 0)
      data = new double[window];
    else
      data = new double[1];

    for (int i = 0; i < data.length; i++)
      data[i] = Double.MIN_VALUE;
  }

  /**
   * Resets the internal count structure used by the Observer, but not
   * the auxillary data structure.
   */
  public void reset() {

    for (int i = 0; i < data.length; i++)
      data[i] = Double.MIN_VALUE;
  }

  /**
   * Observes an event, updating local data if necessary.  Under normal 
   * circumstances, this function calls checkEvent for each event
   * passed to it.  If checkEvent returns true, the local count is
   * incremented by 1. Override this if you are counting different things
   * (e.g. call updateCount(x) where x is method quality, if you want
   * to keep track of total quality over some period of time).
   * @param e the event to observe
   * @see #checkEvent
   */
  public void addEvent(AgentEvent e) {

    if (checkEvent(e))
      updateCount(1);
  }

  /**
   * Determines if an event is to be counted or not.  This is the
   * method you are likely to override if you are monitoring events.
   * @param e The event to check out
   * @return True if the event should be counted
   */
  public boolean checkEvent(AgentEvent e) {

    return false;
  }

  /**
   * Updates the local data array, using the supplied value.
   * @return The current count (for this time slot, if windowed)
   */
  public double updateCount(double c) {

    if (usesWindow())
      data[index] = c;
    else {
      if (data[index] == Double.MIN_VALUE)
        data[index] = 0;
      data[index] += c;
    }

    return data[index];
  }

  /**
   * Gets called whenever the time changes.  This updates the window,
   * if any is being used.  Override this method if you want to perform
   * non-event based observation (but make sure you also make a super-call
   * back to this method).
   * @param time The current time
   */
  public void newTime(int time) {

    if (usesWindow()) {
      index++;

      if (index >= window)
  	    index = 0;

      data[index] = 0;
    }
  }

  /**
   * Override this if you want your observer to not process events
   * @return True if the event should be sent events for observation
   */
  public boolean usesEvents() { return true; }

  /**
   * Returns the description of the data this observer is responsible
   * for detecting. (e.g. "Sent Messages" or "Aborted Actions")
   * @return The description
   */
  public String getDescription() { return description; }

  /**
   * Returns the current count, by processing the gathered data.
   * Currently calls getSum().
   * @return The count
   */
  public double getCount() {
    return getSum();
  }

  /**
   * Returns the current count, by summing the gathered data
   * @return The count
   */
  public double getSum() {
    double count = 0;

    for(int i = 0; i < data.length; i++) {
        if (data[i] == Double.MIN_VALUE) continue;
        count += data[i];
    }

    return count;
  }

  /**
   * Returns the current count, by summing the gathered data and
   * dividing by the total number of data points.
   * @return The count
   */
  public double getAverage() {
      return (getSum() / getCountTotal());
  }

  /**
   * Returns the current count, by finding the difference between
   * sucessive data points, and then dividing by the total number
   * of data points - 1.
   * @return The count
   */
  public double getAverageDifference() {
    double count = 0;
    double last = Double.MIN_VALUE;
    int stop = ((index != 0) ? index - 1 : (data.length - 1));

    if (getCountTotal() <= 1) return 0;

    for(int i = index; true; i++) {
        if (i >= data.length) i = 0;
        if (data[i] != Double.MIN_VALUE) {
            if (last != Double.MIN_VALUE) {
                count += (data[i] - last);
            }
            last = data[i];
        }
        if (i == stop) break;
    }

    return (count / (getCountTotal() - 1));
  }

  /**
   * Returns the current count, by processing the gathered data starting
   * at the current time, back through window slots.  If window is larger
   * than the total amount of data, all the known data is used.
   * @param w The size of the window to check
   * @return The count
   */
  public double getCount(int w) {
    double count = 0;

    if (w > data.length)
      count = getCount();
    else
      for(int i = index, j = 0; j < w; j++, i--) {
          if (i < 0)
              i = data.length - 1;
          if (data[i] == Double.MIN_VALUE) continue;
          count += data[i];
      }

    return count;
  }

  /**
   * Returns a reference to the local data table.  You can use this 
   * to store random junk needed for observations.
   * <P>
   * I added this to get around the no-constructor problem with
   * anonymous classes.
   * @return A reference to the table
   */
  public Hashtable getData() {
    return exdata;
  }

  /**
   * Returns true if a sliding window of the data is used
   * @return True or false
   */
  public boolean usesWindow() { return (window != NO_WINDOW); }

  /**
   * Returns the window size
   * @return The size
   */
  public int getWindowSize() { return window; }

  /**
   * Returns the number of data points currently being stored
   */
  public int getCountTotal() {
      int count = 0;
      for(int i = 0; i < data.length; i++) {
          if (data[i] == Double.MIN_VALUE) continue;
          count++;
      }
      return count;
  }
    
  /**
   * Return if this observer is supposed to be displayed
   * @return A boolean
   */
  public boolean getDisplay() {
    return display;
  }

  /**
   * Toggle the display
   */ 
  public void toggleDisplay() {
    if (display)
      display = false;
    else
      display = true;
  }

  /**
   * Dumps the internal data
   */
  public String dumpData() {
      StringBuffer buf = new StringBuffer(toString() + "\n");
      int stop = ((index != 0) ? index - 1 : (data.length - 1));
      
      for(int i = index; true; i++) {
          if (i >= data.length) i = 0;
          buf.append(String.valueOf(i) + ":");
          if (data[i] == Double.MIN_VALUE)
              buf.append("-");
          else
              buf.append(String.valueOf(data[i]));
          if (i == stop) break;
          buf.append(", ");
      }

      return buf.toString();
  }

  /**
   * String me
   */
  public String toString() {

      if (usesWindow())
          return getDescription() + ": " + getCount() + " [" + getWindowSize() + "]";
      else
          return getDescription() + ": " + getCount();
  }
}
