/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * PropertyEvent.java
 ***************************************************************************************/

package agent.simplest;

import agent.base.AgentEvent;

/**
 * An event signifying the change of a monitored property
 */
public class PropertyEvent extends AgentEvent {
  public static final int ADDED = 0;
  public static final int REMOVED = 1;
  public static final int CHANGED = 2;
  private Object key, oldproperty, newproperty;
  private Object hint = null;

  /**
   * Constructor
   * @param source Who owns this property
   * @param k The property key (e.g. "Name")
   * @param np The new property value
   * @param id The event id
   */
  public PropertyEvent(Object source, Object k, Object np, int id) {
    this(source, k, null, np, id);
  }

  /**
   * Constructor
   * See State.setProperty() for more information about the hint data.
   * @param source Who owns this property
   * @param k The property key (e.g. "Name")
   * @param np The new property value
   * @param id The event id
   * @param h The hint data
   * @see agent.simplest.State#setProperty(Object, Object, Object)
   */
  public PropertyEvent(Object source, Object k, Object np, int id, Object h) {
    this(source, k, null, np, id);
    setHint(h);
  }

  /**
   * Constructor
   * @param source Who owns this property
   * @param k The property key (e.g. "Name")
   * @param op The old proerty value
   * @param np The new property value
   * @param id The event id
   */
  public PropertyEvent(Object source, Object k, Object op, Object np, int id) {
    super(source, id);
    setKey(k);
    setProperty(np);
    setOldProperty(op);
  }

  /**
   * Constructor
   * @param source Who owns this property
   * @param k The property key (e.g. "Name")
   * @param op The old proerty value
   * @param np The new property value
   * @param id The event id
   * @param h The hint data
   * @see agent.simplest.State#setProperty(Object, Object, Object)
   */
  public PropertyEvent(Object source, Object k, Object op, Object np, int id, Object h) {
    this(source, k, op, np, id);
    setHint(h);
  }

  /**
   * Get/set accessor functions
   */
  public Object getProperty() { return newproperty; }
  public void setProperty(Object np) { newproperty = np; }
  public Object getOldProperty() { return oldproperty; }
  public void setOldProperty(Object op) { oldproperty = op; }
  public Object getKey() { return key; }
  public void setKey(Object k) { key = k; }
  public Object getHint() { return hint; }
  public void setHint(Object h) { hint = h; }

  /**
   * What type of data is this?
   * @return The class type of the property
   */
  public String getType() {
    return newproperty.getClass().getName();
  }

  /**
   * Stringify the event
   */
  public String toString() {
    if (getID() == CHANGED)
      return "(PropertyEvent) Type:" + getID() + 
	" Key:" + key + 
	" Hint:" + hint + 
	" Old:" + oldproperty +
	" New:" + newproperty;
    else
      return "(PropertyEvent) Type:" + getID() + 
	" Key:" + key + 
	" Hint:" + hint + 
	" Prop:" + oldproperty;
  }
}
