/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * Sensor.java
 *
 * manifest: Name: agent/simplest/Sensor.class
 * manifest: Java-Bean: True
 ***************************************************************************************/

package agent.simplest;

import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

import agent.base.*;
import utilities.*;
import agent.simplest.Log;

/**
 * A base class for other sensor objects.  This is basically just a
 * shell - no actual sensing capabilities are implemented.
 */
public class Sensor extends AgentComponent implements MouseListener {
  static final long serialVersionUID = 1234567890L; // kludge static ser_id.  Bad bad Bryan!
  public static final int TIMEOUT = 20;
  protected ListenerVector listeners = new ListenerVector();

  protected agent.simplest.Log log;
  protected agent.simplest.State state;

  protected int FAC_SENSOR;
  protected Hashtable sensors = new Hashtable(20);
  protected Hashtable replies = new Hashtable(20);
  protected Hashtable known = new Hashtable(10);
  protected Hashtable age = new Hashtable(10);
  protected PopupMenu menu = new PopupMenu("Known Sensors");

  protected boolean cache = true;

  public static final String SENSORDATA = "SensorData";
  public static final String SENSORCACHE = "SensorCache";

  /**
   * Default constructor.
   */
  public Sensor() {
    super();

    addDependency("Log");
    addDependency("State");

    State.addParameterInfo(SENSORDATA, "String", "Name of the sensor data entry");
    State.addParameterInfo(SENSORCACHE, "Boolean", "Try to cache results or always query?", Boolean.FALSE);

    add(menu);

    // Listen for mouses
    addMouseListener(this);    
  }

  /**
   * Adds a listener to the list
   */
  public void addSensorEventListener(SensorEventListener l) {
    //listeners.add(l, l.getActionEventPriority());
    listeners.add(l);
  }

  /**
   * Removes a listener
   */
  public void removeSensorEventListener(SensorEventListener l) {
    listeners.remove(l);
  }

  /**
   * Sends an event to each of the registered listeners
   * @param e The event to send
   */
  protected void fireEvent(AgentEvent event) {
    SensorEvent e = (SensorEvent)event;
    SensorEventListener l;
    Enumeration enumr = listeners.elements();

    log.log("Firing sensor event", Log.LOG_DEBUG2, FAC_SENSOR);

    while (enumr.hasMoreElements()) {
      l = (SensorEventListener)enumr.nextElement();

      if (!e.isActive()) break;

      switch (e.getID()) {
      case SensorEvent.ADDED:
	l.sensorAdded(e);
	break;
      default:
	break;
      }
    }
    
    if (log != null) log.log(e, Log.LOG_INFO, FAC_SENSOR);

    log.log("Fired a sensor event", Log.LOG_SPAM, FAC_SENSOR);
  }

  /**
   * Init
   */
  public void init() {

    log = (agent.simplest.Log)State.findComponent("Log");
    state = (agent.simplest.State)State.findComponent("State");

    FAC_SENSOR = log.getFacilityID(this);

    super.init();
  }

  /**
   * Begin
   */
  public void begin() {

    if (state.hasProperty(SENSORCACHE))
      cache = ((Boolean)state.getProperty(SENSORCACHE)).booleanValue();

    readSensorFileData();
  }

  /**
   * Reset
   */
  public void reset() {
    sensors = new Hashtable(20);
    replies = new Hashtable(20);
  }

  /**
   * Reads in the sensor data from the extended entry file, if any.
   * Each line is sent to the simulator as a new sensor specification,
   * where it should be interpreted as belonging to the agent.
   */
    protected void readSensorFileData() {

        if (!state.hasProperty(SENSORDATA)) {
            return;
        }

        log.log("Reading sensor data", Log.LOG_INFO, FAC_SENSOR);
    
        try {
            BufferedReader file = state.getExtendedEntryReader((String)state.getProperty(SENSORDATA));

            if (file == null) {
                log.log("Error reading from sensor data: " + state.getProperty(SENSORDATA), Log.LOG_ERR, FAC_SENSOR);
                return;
            }

            while (file.ready()) {
                String str = file.readLine();

                // Trim
                if (str.indexOf(";") != -1)
                    str = str.substring(0, str.indexOf(";"));
                str = str.trim();
                if (str.length() <= 0) continue;

                // Add it
                addSensor(str);
            }

            file.close();

        } catch(IOException e) {
            log.log("Error: Sensor data file error: " + e, Log.LOG_ERR, FAC_SENSOR);
        }

        log.log("Finished reading sensor data", Log.LOG_DEBUG, FAC_SENSOR);
    }

  /**
   * Queries the sensor and returns the supplied sensor data, if any.
   * <br>
   * Note that the type that the sensor emits and the type
   * you specify here have to match!
   * <br>
   * This function uses makeSensorQuery internally to generate the
   * actual string used to query the sensor.
   * @param sensor The sensor to query
   * @param data Any other data to add to the query, ignored if null
   * @param type The data type, String if null
   * @return The data, or null if not found
   */
  public Object readSensorData(String sensor, Hashtable data, String type) {

    return readSensorData(sensor, type);
  }

  /**
   * Queries the sensor and returns the supplied sensor data, if any.
   * <br>
   * Note that the type that the sensor emits and the type
   * you specify here have to match!
   * <br>
   * This function uses makeSensorQuery internally to generate the
   * actual string used to query the sensor.
   * @param sensor The sensor to query
   * @param type The data type, String if null
   * @return The data, or null if not found
   */
    public Object readSensorData(String sensor, String type) {
        Integer time = (Integer)state.getProperty(State.TIME);
    
        if (cache) {
            if (age.containsKey(sensor)) {
                if ((((Integer)age.get(sensor)).equals(time)) && (sensors.get(sensor) != null)) {
                    log.log("Using cached data for sensor query \"" + sensor + "\"", Log.LOG_DEBUG, FAC_SENSOR);
                    return convertSensorData((String)sensors.get(sensor), type);
                }
            }
        }

        if (querySensor(sensor)) {
            if (sensors.containsKey(sensor) && (sensors.get(sensor) != null)) {
                age.put(sensor, time);
                return convertSensorData((String)sensors.get(sensor), type);
            }
        }

        return null;
    }

  /**
   * Returns the last recorded sensor data, if any, without
   * actually querying the sensor.
   * @param sensor The sensor to query
   * @param data Any other data to add to the query, ignored if null
   * @param type The data type, String if null
   * @return The data, or null if not found
   * @see #readSensorData
   */
  public Object readOldSensorData(String sensor, Hashtable data, String type) {

    return readOldSensorData(sensor, type);
  }

  /**
   * Returns the last recorded sensor data, if any, without
   * actually querying the sensor.
   * @param sensor The sensor to query
   * @param type The data type, String if null
   * @return The data, or null if not found
   * @see #readSensorData
   */
  public Object readOldSensorData(String sensor, String type) {

    if (sensors.containsKey(sensor) && (sensors.get(sensor) != null))
      return convertSensorData((String)sensors.get(sensor), type);

    return null;
  }

  /**
   * Converts the data the way we want to, using State for the 
   * majority of data conversion.
   * @param data The data to be converted
   * @param type The type to convert to, or null if String
   * @return The converted data, or null if an error occurs
   * @see utilities.Converter#reTypeProperty
   */
  protected Object convertSensorData(String data, String type) {
    
    // Unexpected Error
    if (data == null)
      return null;

    String s = data.trim();

    // Access error
    if (s.equalsIgnoreCase("AccessDenied"))
      return null;

    // Other error
    if (s.startsWith("Error")) {
      log.log("Error: " + data, Log.LOG_ERR, FAC_SENSOR);
      return null;
    }

    // Type conversion
    return Converter.reTypeProperty(s, type);
  }

  /**
   * Trys to fetch the current sensor data from the simulator
   * <P>
   * Right now this does nothing.
   * @param sensor The sensor name
   * @return True if successfull, false otherwise
   */
  protected boolean querySensor(String sensor) {

      log.log("Warning, cannot make sensor query for " + sensor, Log.LOG_WARNING, FAC_SENSOR);

      return false;
  }

  /**
   * Adds a sensor description to the list of sensors known to be available
   * to this agent.  The format is the same as in the config file, with
   * the initial Sensor chopped off.
   * @param s The string sensor definition
   */
  public void addSensor(String s) {
    Hashtable hash = new Hashtable();

    log.log("Adding sensor => " + s, Log.LOG_INFO, FAC_SENSOR);

    // Parse it to get the name
    StringTokenizer tok = new StringTokenizer(s, ",");
    tok.nextToken(); // Lose the "Sensor,"
    tok.nextToken(); // Lose the Sensor type
    String name = tok.nextToken();
    name = name.trim();

    // Tokenizer it
    while (tok.hasMoreTokens()) {
      // Get the next value pair
      String str = tok.nextToken().trim();
      if (str.indexOf(":") == -1)
	hash.put(str, "");
      else {
	String n = str.substring(0, str.indexOf(":")).trim();
	String v = str.substring(str.indexOf(":")+1).trim();
	hash.put(n, v);
      }
    }

    // Add it
    String ts = findSensor(name);
    if (ts == null) {
      known.put(name, hash);
      menu.add(new MenuItem(name));
      log.log("Sensor added (" + name + ") " + hash, Log.LOG_DEBUG, FAC_SENSOR);
      startFireEvent(new SensorEvent(this, name, SensorEvent.ADDED));
    } else
      log.log("Duplicate sensor name (" + name + ") detected, not added", Log.LOG_WARNING, FAC_SENSOR);
  }

  /**
   * Removes a sensor from the list of sensors known to be available
   * to this agent.
   * @param n The name of the sensor to remove
   * @return True if the sensor is sucessfully removed, false otherwise
   */
  public boolean removeSensor(String n) {

    String s = findSensor(n);
    if (s != null) {
      known.remove(s);
      for (int i = 0; i < menu.getItemCount(); i++) {
          MenuItem m = menu.getItem(i);
          if (m.getLabel().equals(n)) {
              menu.remove(m);
              break;
          }
      }
      log.log("Sensor removed (" + n + ")", Log.LOG_INFO, FAC_SENSOR);
    } else
      return false;

    return true;
  }

  /**
   * Finds a sensor in the known sensor list
   * @param n The name of the sensor to find
   * @return A reference to the sensor (just a string right now), or
   * null if not found.
   */
  public String findSensor(String n) {

    Enumeration e = getSensors();
    while (e.hasMoreElements()) {
      String s = (String)e.nextElement();
      if (s.equals(n)) return s;
    }

    return null;
  }

  /**
   * Returns an enumeration of the available sensors.
   * @return An enumeration
   */
  public Enumeration getSensors() {

    return known.keys();
  }

  /**
   * Returns a hashtable of known sensor data, either real parameters
   * or parameter names and their types.  All entries should be in
   * String format.
   * @return A hashtable if the entry if found, or null if not
   */
  public Hashtable getSensorInfo(String s) {

    return (Hashtable)known.get(s);
  }

  /**
   * Returns true if the component has a popupmenu.
   */
  public boolean hasPopupMenu() {
      return true;
  }

  /**
   * Responds to mouse clicks
   */
  public void processMouseEvent(MouseEvent e) {
    if ((e.getID() == MouseEvent.MOUSE_PRESSED) || e.isPopupTrigger()) {
       menu.show(this, e.getX(), e.getY());
    }
  }
  public void mouseClicked(MouseEvent e) { processMouseEvent(e); }
  public void mousePressed(MouseEvent e) { processMouseEvent(e); }
  public void mouseReleased(MouseEvent e) { processMouseEvent(e); }
  public void mouseDragged(MouseEvent e) { processMouseEvent(e); }
  public void mouseEntered(MouseEvent e) { }
  public void mouseExited(MouseEvent e) { }
  public void mouseMoved(MouseEvent e) { }
}
