/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * State.java
 *
 * manifest: Name: agent/simplest/State.class
 * manifest: Java-Bean: True
 * manifest: Depends-On: agent/simplest/PropertyEvent.class
 * manifest: Depends-On: agent/simplest/PropertyListener.class
 ***************************************************************************************/

package agent.simplest;

import java.io.*;
import java.awt.*;
import java.util.*;
import java.util.zip.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.ResourceBundle;
import java.lang.reflect.*;

import agent.base.*;
import utilities.*;

/**
 * Serves as a public data repository.  Any data structure of a public
 * nature, or one that is to be used by more than one component,
 * should be stored here.  Also maintains the current list of active
 * components in the agent.
 * <P>
 * State is essentially a big hashtable.  You can store anything you
 * want in it, keyed by some other well-known object (typically a
 * String).  When your component is constructed, it will typically
 * call State.addParameterInfo one or more times, listing those
 * parameters which it can use, that should be searched for in any
 * external configuration data State happens to know about.  State
 * then uses this list to determine which values from the configuration
 * to import, and how to convert the data when it does the import.
 * <P>
 * Here's a couple examples, one with a default value, and one with none:
 * <PRE>
 *   State.addParameterInfo("ParamX", "String", "Param X is used for...");
 *   State.addParameterInfo("ParamY", "Long", " ParamY is used for...", new Long(12345));
 * </PRE>
 * So, in this case, when the agent starts, it will look for those two
 * paramters.  As defined here, if no ParamY parameter is found, the default
 * value 12345 will be used instead.  If no ParamX is found during the search,
 * that parameter will receieve a null value.  The complete list of parameters
 * known to State will generally be logged when your agent starts up (dependant
 * on the current log level).  As mentioned earlier, the addParameterInfo
 * calls should be placed in your component's constructor whenever possible.
 * <P>
 * "Parameters" to the State component are those data elements garnered
 * from the search process described above - they are the "parameters" given
 * to the agent when it starts.  Once read in, these values, and all other
 * values assigned by other components in the agent, are considered to be
 * "properties" of the agent.  So parameters are things that are read in
 * on agent start up, while properties are the actual things that are stored
 * in State.
 * <P>
 * You can access the property objects stored in State like this:
 * <P>
 * <PRE>
 *   // Set ParamX to a String
 *   state.setProperty("ParamX", "Behold the power of cheese.");
 * &nbsp;
 *   // Get the Long value stored in ParamY
 *   Long l = (Long)state.getProperty("ParamY");
 * </PRE>
 * ...where setProperty sets the value associated with a given key and
 * getProperty retrieves it.  When setProperty is called, State will
 * also fire a PropertyEvent describing the changes that were made to
 * that entry, so that you may also passively monitor changes to State
 * properties by listening to the event stream which State produces.
 * <P>
 * A variant of the setProperty call also permits the inclusion of "hints":
 * <P>
 * <PRE>
 *   // Get an existing hashtable and modify it's "Cheese" entry
 *   Hashtable table = (Hashtable)state.getProperty("MyTable");
 *   table.put("Cheese", "Bleu");
 * &nbsp;
 *   // Then store it again, with a hint
 *   state.setProperty("MyTable", table, "Cheese");
 * </PRE>
 * <P>
 * In the above example, we retrieved an existing hashtable from state
 * and modified its "Cheese" entry.  Normally, we would simply set
 * it back in state, which would result in a PropertyEvent being
 * sent out that it had been changed.  The problem with this is that
 * it can be hard for other components to determine exactly what changed
 * in the property - if other components need to react to changes in this
 * table they must first determine what those changes are.  To make
 * this process easier, the changing component can supply an arbitrary hint
 * which is passed along with that PropertyEvent for use by other
 * components which are monitoring that property.  In the code above the
 * component has used the key of the field ("Cheese") as its hint.  Any
 * data can be used, however (a Vector index, Taems node reference, or
 * some more complex textual description come to mind - you could even
 * include a whole Vector as a hint, where each entry is itself a hint!).
 * Here's an example of how another component would use the hint data
 * supplied above:
 * <P>
 * <PRE>
 * public void propertyChanged(PropertyEvent e) {
 *   String key = e.getKey().toString();
 * &nbsp;
 *   // Did someone change my table?
 *   if (key.equalsIgnoreCase("MyTable")) {
 *     Hashtable table = (Hashtable)e.getProperty();
 * &nbsp;
 *     // Now determine what field was changed
 *     Object hint = e.getHint();
 *     if ((hint != null) && (hint instanceof String)) {
 *        // Aha!  Now we know what field was changed!
 *        String field = (String)hint;
 * &nbsp;
 *        // ...and can easily get the altered data!
 *        String data = table.get(field);
 * &nbsp;
 *     } else {
 *        // Hmm, no useful hint data was used, what do I do now?
 *     }
 *   }
 * }
 * </PRE>
 * <P>
 * Note how we checked to ensure that the hint was both non-null
 * and of an expected type (String, in this case).  There are no fixed
 * rules associated with hints - components must simply abide by 
 * property-specific convensions when supplying them.  There is some 
 * probability, then, that a component may elect to not supply, or supply
 * a unconvensional hint.  State also does not care what hints are used,
 * so you should perform some extra checking on hints before you 
 * actually use them.
 * <P>
 * State also serves as a repository for component handles.  Whenever
 * an AgentComponent starts up, it automatically registers itself with 
 * the State component available to it.  Later, when another component
 * needs to access that component, it may query State to obtain a handle
 * to it.
 * <P>
 * Here's an example:
 * <PRE>
 * protected Log log;
 * protected Execute execute;
 * &nbsp;
 * public void init() {
 *   log = (Log)State.findComponent("Log");
 *   execute = (State)State.findComponent("Execute");
 * }
 * </PRE>
 * <P>
 * The function call is named "findComponent" because string matching is
 * allowed in addition to an exact match (an exact-match-only variant, 
 * getComponent, is also available).  So in the above, the parameter
 * "Log" would match "agent.simplest.Log" or "agent.mass.Log" or
 * "agent.simplest.ScreenLog" (or if two matching components are found, the
 * closest match "agent.simplest.Log" will be returned).  So, if you are
 * generating a series of components which have similar functions, you should
 * give them the same interface and similar names.  This will allow you to
 * get "plug-and-play" type interactions (as shown above with Log) where 
 * components will automatically use whatever variant you happen to put in 
 * your agent, without needing any code changes.  The findComponent
 * comments give more details on name matching.
 * <P>
 * Note that you should not call findComponent in your constructor, as
 * it is not guarunteed that all components have registered at that time.
 * <P>
 * One other important aspect of State is the external configuration
 * file (ExtConfig).  The property list is generally used for items
 * that can fit in a single line of text, e.g. you would store a reference
 * to the agent's task structure here, but not the entire structure itself.
 * Multi-line elemenets such as this should be stored in the external
 * configuration file, which is a .jar file the agent will have access to
 * at runtime.  State serves as an interface to this file, so that components
 * can simply ask State for a reference to some part of the file, rather than
 * having to dig it out themselves.  The reference to the entry in the
 * .jar file acts just as a regular (read-only) file would, so the actual
 * location of the data will be transparent.
 * <P>
 * Here's an example:
 * <PRE>
 * BufferedReader file = state.getExtendedEntryReader("somefile.txt");
 * if (file == null) {
 *   // Read your data
 *   MyObject myobject = importConfigData(file);
 * &nbsp;
 *   // Do something with it
 *   initMyself(myobject);
 * &nbsp;
 *   // Close the file
 *   file.close();
 * }
 * </PRE>
 * <P>
 * <B>** Mandatory class</B>
 * @see #addParameterInfo
 * @see #getProperty
 * @see #setProperty
 * @see #findComponent
 * @see #getExtendedEntryReader
 */
public class State extends AgentComponent implements MouseListener, ActionListener {
  static final long serialVersionUID = 1234567890L; // kludge static ser_id.  Bad bad Bryan!
  private static boolean collected = false;
  private boolean pauseevents = false;
  private Vector events = new Vector();
  protected int FAC_STATE;
  protected transient static TreeMap components = new TreeMap();
  protected transient static Vector parameters = new Vector(50);
  protected transient static Random random;
  protected ZipFile extFile = null;
  protected static Log log = null;
  protected ListenerVector listeners = new ListenerVector();
  protected static Hashtable defaults = new Hashtable(100);
  protected Hashtable properties = new Hashtable(100);
  protected boolean pulsetime = false;
  protected int menulen = Integer.MAX_VALUE;
  protected int nextmenustart = 0;
  private static PopupMenu menu = new PopupMenu("State Data");

  public static final String NAME = "Name";
  public static final String TIME = "Time";
  public static final String RANDOMSEED = "RandomSeed";
  public static final String PULSETIME = "PulseTime";
  public static final String EXTCONFIG = "ExtConfig";
  public static final String MASSHOME = "MassHome";
  public static final String STATEMENULENGTH = "StateMenuLength";

  /**
   * Default constructor.  Sets up properties hashtable and notes
   * dependencies.
   */
  public State() {
    addParameterInfo(NAME, "String", "The name of the agent (required)", new String("You need to set a name"));
    addParameterInfo(TIME, "Integer", "The agent's local time", new Integer(-1));
    addParameterInfo(RANDOMSEED, "Long", "The random number generator seed.  This is used to seed the agent's random number generator to allow more deterministic performance.  See getRandom().");
    addParameterInfo(PULSETIME, "Boolean", "If set, the agent will automatically set the time to be the current pulse count.", Boolean.FALSE);
    addParameterInfo(EXTCONFIG, "String", "Pathname to the extended configuration file (relative to MassHome, if it exists)");
    addParameterInfo(MASSHOME, "String", "Pathname for the Mass home directory");
    addParameterInfo(STATEMENULENGTH, "Integer", "Maximum allowed length of State's contextual menu.  Repeated clicks on the component will see the next set of entries.", new Integer(50));
    
    properties.put(TIME, new Integer(-1));

    addDependency("Log");

    // Listen for mouses
    addMouseListener(this);
    add(menu);
  }

  /**
   * Init.  Also gets the HTML or environement variables.
   * @see #getOtherParams
   */
  public void init() {
    log = (Log)State.findComponent("Log");

    FAC_STATE = log.getFacilityID(this);

    if (hasProperty(TIME) && (((Integer)getProperty(TIME)).intValue() < 0))
        setProperty(TIME, new Integer(0));

    getOtherParams();
    printParameterInfo();

    log.log("State component initialized", Log.LOG_DEBUG, FAC_STATE);

    if (hasProperty(PULSETIME))
	pulsetime = ((Boolean)getProperty(PULSETIME)).booleanValue();
  }

  /**
   * Just resets the data state, not the list of installed components
   * or listeners.
   */
  public void reset() {
    properties.clear();

    setProperty(TIME, new Integer(0));

    collected = false;
    getOtherParams();
  }

  /**
   * Pulse.
   */
  public void pulse() {

      if (pulsetime) {
	  setProperty(TIME, new Integer(((Integer)getProperty(TIME)).intValue() + 1), PULSETIME);
      }
  }

  /**
   * Gets variables/parameters from external sources of information.
   * The currently supported sources are:
   * <UL>
   *   <LI>the applet context (e.g. &lt;PARAM NAME="Name" VALUE="AgentName"&gt;)
   *   <LI>the system property list (e.g. java -DName="AgentName" javaclass..)
   * </UL>
   * The bad news is that it adds all these parameters as strings,
   * even if they really should be numeric.  One sneaky way of getting
   * around this is to use the propertyPreprocess function, which
   * is called whenever a property is added.  The propertyPreprocess
   * function in this class supports the interpretation of several basic types.
   * <P>
   * To make use of other parameters, update the params string array.
   * @see #getParameterInfo
   * @see #propertyPreprocess
   */
    public void getOtherParams() {
        String[][] params = getParameterInfo();
        String[] paramlist = new String[params.length];
        Object p = this;
        java.applet.Applet a;

        // Check to make sure
        if (!Control.hasStarted()) { return; }

        // So we can collect (just once) before being init'd
        if (collected == true) { return; }
        collected = true;
        pauseevents = true;

        // Just to make sure...
        if (log == null) {
            log = (Log)State.findComponent("Log");
            FAC_STATE = log.getFacilityID(this);
        }

        // Get the params array
        for (int i = 0; i < params.length; i++) {
            paramlist[i] = params[i][0];
        }

        // Look for applet params
        log.log("Checking applet params", Log.LOG_INFO, FAC_STATE);
        while (true) {
            if (p instanceof java.applet.Applet) {
                a = (java.applet.Applet)p;

                // Does the parameter method work?
                try { a.getParameter("_dum_"); }
                catch (NullPointerException e) { break; }

                // Check each known param
                for (int i = 0; i < paramlist.length; i++) {
                    if (a.getParameter(paramlist[i]) != null) {
                        if (!properties.containsKey(paramlist[i])) {
                            setProperty(paramlist[i], a.getParameter(paramlist[i]), "Parameters");
                        } else
                            log.log("Ignoring duplicate property " + paramlist[i], Log.LOG_INFO, FAC_STATE);
                    }
                }
                break;
            }
            if (p == this.getParent()) break;
            p = this.getParent();
            if (p == null) break;
        }

        // Look for system properties
        log.log("Checking system properties", Log.LOG_INFO, FAC_STATE);
        try {
            for (int i = 0; i < paramlist.length; i++) {
                if (System.getProperty(paramlist[i]) != null) {
                    if (!properties.containsKey(paramlist[i])) {
                        setProperty(paramlist[i], System.getProperty(paramlist[i]), "System");
                    } else
                        log.log("Ignoring duplicate property " + paramlist[i], Log.LOG_INFO, FAC_STATE);
                }
            }
        } catch (sun.applet.AppletSecurityException e) {
            log.log("Error: Applet security error, can't check system properties", Log.LOG_ERR, FAC_STATE);
        }

        // Look for defaults
        log.log("Checking default params", Log.LOG_INFO, FAC_STATE);
        Enumeration e = defaults.keys();
        while (e.hasMoreElements()) {
            Object key = e.nextElement();
            if (!properties.containsKey(key))
                setProperty(key, defaults.get(key), "Default");
        }
        defaults.clear();

        // Flush em
        flushEvents();
    }

  /**
   * Prints the known parameter info to stderr to log, or stderr if log
   * is not available.  This function should be automatically called
   * when State is initialized.
   * @see #addParameterInfo
   * @see #getParameterInfo
   */
    public void printParameterInfo() {
        String[] p;
        StringBuffer str = new StringBuffer("");
        TreeSet set = new TreeSet(new Comparator() {
                public int compare(Object o1,
                                   Object o2) {
                    int i = ((String[])o1)[3].compareTo(((String[])o2)[3]);
                    if (i != 0)
                        return i;
                    else
                        return ((String[])o1)[0].compareTo(((String[])o2)[0]);
                }
            });
        set.addAll(parameters);
        Iterator it = set.iterator();

        // Screen widths
        int S_WIDTH = 77;

        // Field widths
        int F_ONE = 20;
        int F_TWO = 10;

        // Ansi?
        boolean ansi = false;
        if (hasProperty(Log.USEANSI))
            ansi = ((Boolean)getProperty(Log.USEANSI)).booleanValue();

        String lcls = "";
        if (it.hasNext()) {
            for (int i = 0; it.hasNext(); i++) {
                p = (String[])it.next();
                if ((p[3] != null) && (!p[3].equals(lcls))) {
                    lcls = p[3];
                    if (ansi)
                        str.append("\n ::: " + LogItem.ANSI_S + LogItem.ANSI_C_CYN + lcls + LogItem.ANSI_F + " :::\n");
                    else
                        str.append("\n ::: " + lcls + " :::\n");
                }
                if (ansi)
                    str.append("  " + LogItem.ANSI_S + LogItem.ANSI_C_GRN + p[0] + LogItem.ANSI_F + " ");
                else
                    str.append("  " + p[0] + " ");
                for (int j = 0; j < F_ONE - p[0].length(); j++) { str.append(" "); }
                str.append(p[1] + " ");
                for (int j = 0; j < F_TWO - p[1].length(); j++) { str.append(" "); }

                String desc = p[2];
                StringBuffer prep = new StringBuffer("");
                while (true && desc.length() > 0) {
                    if (desc.length() > (S_WIDTH - (F_ONE + F_TWO))) {
                        int index = desc.substring(0, (S_WIDTH - (F_ONE + F_TWO))).lastIndexOf(" ");
                        if (index <= 0) index = S_WIDTH - (F_ONE + F_TWO);
 
                        str.append(prep.toString() + desc.substring(0, index) + "\n");
                        if (desc.charAt(index) == ' ')
                            index++;
                        desc = desc.substring(index);

                        prep = new StringBuffer("");
                        for (int j = 0; j < F_ONE + F_TWO + 4; j++) { prep.append(" "); }
                    } else {
                        str.append(prep.toString() + desc);
                        break;
                    }
                }

                str.append("\n");

                if (properties.containsKey(p[0])) {
                    if (ansi)
                        str.append("     == " + LogItem.ANSI_S + LogItem.ANSI_C_PUR + getProperty(p[0]).toString() + LogItem.ANSI_F + "\n");
                    else
                        str.append("     == " + getProperty(p[0]).toString() + "\n");
                }
            }
        }

        if (log != null) {
            int l = log.getLogLength();
            if (l != 0)
                log.setLogLength(0);
            log.log("Known properties:\n" + str, Log.LOG_INFO, FAC_STATE);
            if (l != 0)
                log.setLogLength(l);
        } else
            utilities.Log.getDefault().log("Known parameters:\n" + str, Log.LOG_INFO);
    }

  /**
   * Same as above, just for a single parameter.
   * @see #printParameterInfo
   */
  public void printParameterInfo(String param) {
    String[] p = getParameterInfo(param);
    StringBuffer str;

    // Field widths
    int F_ONE = 12;
    int F_TWO = 8;

    if (p == null) {
      return;
    }

    str = new StringBuffer();
    str.append("  " + p[0] + " ");
    for (int j = 0; j < F_ONE - p[0].length(); j++) { str.append(" "); }
    str.append(p[1] + " ");
    for (int j = 0; j < F_TWO - p[1].length(); j++) { str.append(" "); }
    str.append(p[2]);
    
    if (log != null)
      log.log(str.toString(), Log.LOG_INFO, FAC_STATE);
    else
      utilities.Log.getDefault().log(str.toString(), Log.LOG_INFO);
  }

  /**
   * Adds a parameter to the parameter information list.  If components
   * needs to make use of information residing in the external configuration
   * file or environment variables, they must be declared with this function.
   * <P>
   * It is preferable that this method be called from components' 
   * constructors, and not elsewhere in their code.  Later calls will
   * work, but will cause an interruption in the flow of execution
   * to search for the data (parameters specified in constructors, on the
   * other hand, are queued and searched for en masse).
   * <P>
   * If multiple parameters with the same name are added, the last
   * setting will take precedence, with two details.  If the type
   * and desc parameters are non-null, they will overwite the existing
   * values, otherwise they are ignored.  If the def value is null,
   * it will remove the default value, otherwise the value will overwrite
   * the existing one.  The ability to override a previously defined
   * parameter can be useful for components which need to modify the default
   * behaviors exhibited by their parent component.
   * @param name The name of the parameter to read
   * @param type The data type of the parameter to read (may affect
   * how this data is processed, see below)
   * @param desc The textual description of the parameter to read
   * @param def The default value for the parameter, or null for none
   * @see #propertyPreprocess
   */
  public static void addParameterInfo(String name, String type, String desc, Object def) {
      addParameterInfo(name, type, desc, def, null);
  }

  /**
   * Same as other addParameterInfo, except no default value is given.
   * @param name The name of the parameter to read
   * @param type The data type of the parameter to read (may affect
   * how this data is processed, see below)
   * @param desc The textual description of the parameter to read
   * @see #addParameterInfo(String, String, String, Object)
   */
  public static void addParameterInfo(String name, String type, String desc) {
      addParameterInfo(name, type, desc, null, null);
  }

    /**
     * This is a hack, to allow access to the stack of callers
     * of a particular function.
     */
    static class SecHack extends SecurityManager
    {
        private static SecHack sec = new SecHack();
        public static Class getCallersClass(int i)
        {
            return sec.getClassContext()[i];
        }
    }


  /**
   * This version of addParameterInfo also allows the caller to specify the
   * Class of the component adding the parameter
   * @param name The name of the parameter to read
   * @param type The data type of the parameter to read (may affect
   * how this data is processed, see below)
   * @param desc The textual description of the parameter to read
   * @param def The default value for the parameter, or null for none
   * @param cls The class adding the parameter, if null it will try
   *   to auto-determine it
   * @see #addParameterInfo(String, String, String, Object)
   */
  private static void addParameterInfo(String name, String type, String desc, Object def, String cls) {

      if (cls == null) {
          cls = "";
          try {
              cls = SecHack.getCallersClass(3).toString();
              cls = cls.substring(cls.indexOf(" ") + 1);
          } catch (Exception e) {
              utilities.Log.getDefault().log("Error determining calling class: " + e, Log.LOG_WARNING);
          }
      }

    if (!usesParameter(name))
        parameters.addElement(new String[] { name, type, desc, cls });
    else {
        String []param = getParameterInfo(name);
        if (type != null) param[1] = type;
        if (desc != null) param[2] = desc;
        if (cls != null) param[3] = cls;
    }


    if (def != null)
        defaults.put(name, def);
    else
        defaults.remove(name);

    if (collected == true) {
      if (log != null)
	log.log("Dagnabit, someone added parameter information for " + name + " post-collection...recollecting", Log.LOG_INFO);
      collected = false;
    }
  }

  /**
   * Returns the parameter information list, a la java.applet.Applet.
   * Each property description array has three String elements, 
   * are are organized like this:
   * <UL>
   *  <LI>0 - The property key
   *  <LI>1 - The property type
   *  <LI>2 - The property free form description
   * </UL>
   * @return An array of property description arrays.
   * @see #getOtherParams
   * @see #addParameterInfo
   * @see java.applet.Applet#getParameterInfo
   */
  public static String[][] getParameterInfo() {
    String[] p;
    String[][] plist = new String[parameters.size()][4];
    Enumeration e = parameters.elements();

    if (e.hasMoreElements()) {
      for (int i = 0; e.hasMoreElements(); i++) {
	p = (String[])e.nextElement();
	plist[i][0] = p[0];
	plist[i][1] = p[1];
	plist[i][2] = p[2];
      }
    }

    return plist;
  }

  /**
   * Returns the property information list for a specific parameter.
   * @param key The property key
   * @return The info, or null if not found
   * @see #getParameterInfo
   */
  public static String[] getParameterInfo(String key) {
    Enumeration e = parameters.elements();

    while (e.hasMoreElements()) {
      String info[] = (String [])e.nextElement();
      if (info[0].equalsIgnoreCase(key))
	return info;
    }

    return null;
  }

  /**
   * Determines if the key is a known parameter to some component.
   * @param key The property key
   * @return true if the parameter is known, false otherwise
   * @see #getParameterInfo
   */
  public static boolean usesParameter(String key) {
    Enumeration e = parameters.elements();

    while (e.hasMoreElements()) {
      String info[] = (String [])e.nextElement();
      if (info[0].equalsIgnoreCase(key))
	return true;
    }

    return false;
  }

  /**
   * Returns a reader attached to the correct entry in the extended
   * configuration file attached to the agent.
   * @param n The name of the entry
   * @return A BufferedReader attached to the entry, or null if none
   *   found.
   * @see #getExtendedEntryStream
   */
  public BufferedReader getExtendedEntryReader(String n) {
    InputStream is = getExtendedEntryStream(n);

    if (is != null)
      return new BufferedReader(new InputStreamReader(is));
    else
      log.log("Error creating reader for " + n, Log.LOG_ERR, FAC_STATE);

    return null;
  }

  /**
   * Returns a stream attached to the correct entry in the extended
   * configuration file attached to the agent.  Store data in this
   * file when it is sufficiently detailed to require a file of its 
   * own.
   * @param n The name of the entry
   * @return A stream attached to the entry, or null if none found.
   * @see #getExtendedEntryFile
   */
  public InputStream getExtendedEntryStream(String n) {

    log.log("Getting extended entry " + n, Log.LOG_DEBUG, FAC_STATE);

    try {
      ZipFile ext = getExtendedEntryFile();

      if (ext != null) {
	ZipEntry ent = ext.getEntry(n);
	
	if (ent != null) {
	  return ext.getInputStream(ent);
	} else {
	  log.log("Error finding entry " + n, Log.LOG_ERR, FAC_STATE);
        }
        
      } else {
	log.log("Error finding entry file " + n, Log.LOG_ERR, FAC_STATE);
      }

    } catch (IOException e) {
      log.log("Error reading extended entry : " + e, Log.LOG_ERR, FAC_STATE);
    }

    // Back-up, look for the bare file
    try {
        InputStream is = new FileInputStream(n);
        log.log("Entry \"" + n + "\" was not found in the extended configuration file, but it was found in the local filesystem.  I'll let you use it, but please add it to your ExtConfig file.", Log.LOG_WARNING, FAC_STATE);
        return is;
    } catch (FileNotFoundException e) {
        // ignore this
    }

    return null;
  }

  /**
   * Returns the cached reference to the extended entry file.
   * <BR>
   * We currently cache a ext reference, and re-use it over and over again
   * so don't close it or do anything wierd to it unless you want bad
   * stuff to happen.
   * @return A ZipFile reference
   */
  public ZipFile getExtendedEntryFile() {

    return getExtendedEntryFile(true);
  }

  /**
   * Returns a reference to the extended entry file stored in
   * the ExtConfig property.
   * @param cached Return the relatively static cached reference,
   * or a new one.  If you get a new one, make sure you close it when
   * you are done
   * @param cached Set to true if you want to use the cached entry reference
   * @return A reference to the ext file
   */
  public ZipFile getExtendedEntryFile(boolean cached) {
    String extname;

    if (!hasProperty(EXTCONFIG)) {
      log.log("Error: No ExtConfig file specified", Log.LOG_ERR, FAC_STATE);
      return null;
    }

    if (hasProperty(MASSHOME)) {
      extname = (String)getProperty(MASSHOME) + "/" + (String)getProperty(EXTCONFIG);
    } else {
      extname = (String)getProperty(EXTCONFIG);
    }

    try {
      if (!cached) {
	log.log("Getting new reference to extended entry file " + extname, Log.LOG_DEBUG2, FAC_STATE);
	return new ZipFile(extname);
      }

      if (extFile == null) {
	log.log("Getting new reference to extended entry file " + extname, Log.LOG_DEBUG2, FAC_STATE);
	extFile = new ZipFile(extname);
      } else
	log.log("Getting cached reference to extended entry file " + extname, Log.LOG_DEBUG2, FAC_STATE);

      if (extFile == null) {
	log.log("Whoops, extended entry file ref is null", Log.LOG_WARNING, FAC_STATE);
      }

      return extFile;

    } catch (IOException e) {
      log.log("Error reading extended entry : " + e, Log.LOG_ERR, FAC_STATE);
    }

    return null;
  }

  /**
   * Adds a listener to the list and sends the listener an added event
   * for each active state variable.
   */
  public void addPropertyEventListener(PropertyEventListener l) {
    //listeners.add(l, l.getPropertyEventPriority());
    listeners.add(l);

    // Bring the listener up to date
    if (Control.hasStarted()) {
      Enumeration e = properties();
      Object o;
      while (e.hasMoreElements()) {
	o = e.nextElement();
	l.propertyAdded(new PropertyEvent(this, o, getProperty(o), PropertyEvent.ADDED));
      }
    }
  }

  /**
   * Removes a listener
   */
  public void removePropertyEventListener(PropertyEventListener l) {
    listeners.remove(l);
  }

  /**
   * Sends a property change event to each listener
   * @param event The event to send.
   */
  protected void fireEvent(AgentEvent event) {
    PropertyEvent e = (PropertyEvent)event;
    PropertyEventListener l;
    Enumeration enumr = listeners.elements();

    // Hold it!
    if (pauseevents) {
      events.addElement(e);
      return;
    }

    if (log != null)
        log.log("Firing a property change event", Log.LOG_DEBUG2, FAC_STATE);
    else
        utilities.Log.getDefault().log("Firing a property change event", Log.LOG_DEBUG2);
        

    while (enumr.hasMoreElements()) {
      l = (PropertyEventListener)enumr.nextElement();

      if (!e.isActive()) break;

      switch (e.getID()) {
      case PropertyEvent.ADDED:
	l.propertyAdded(e);
	break;
      case PropertyEvent.REMOVED:
	l.propertyRemoved(e);
	break;
      case PropertyEvent.CHANGED:
	l.propertyChanged(e);
	break;
      default:
	break;
      }
    }

    if ((log != null) && (log.getLogLevel(FAC_STATE) >= Log.LOG_DEBUG))
        log.log(e, Log.LOG_DEBUG, FAC_STATE);

    if (log != null)
        log.log("Fired a property change event", Log.LOG_SPAM, FAC_STATE);
    else
        utilities.Log.getDefault().log("Fired a property change event", Log.LOG_SPAM);
  }

  /**
   * Flushes the events that were held up.
   */
  private void flushEvents() {
    
    pauseevents = false;

    Enumeration e = events.elements();
    while(e.hasMoreElements()) {
      startFireEvent((PropertyEvent)e.nextElement());
    }
    events.removeAllElements();
  }

  /**
   * Adds a component.  Should be called by AgentComponents'
   * constructors. Generally done only at construction time.
   * Note that only one entry can exist for any descriptor
   * tag, so a duplicate entry will override something.
   * @param a The component to add
   */
  public static synchronized void addComponent(AgentComponent a) {
    String []descriptor = null;

    if (a.getDescriptor() != null)
      descriptor = a.getDescriptor().value();

    if (descriptor != null) {
        for (int i = 0; i < descriptor.length; i++) {
            if (components.containsKey(descriptor[i])) {
                utilities.Log.getDefault().log("Warning: Component " + descriptor[i] + " has been overwritten", Log.LOG_WARNING);
            }
            components.put(descriptor[i], a);
            utilities.Log.getDefault().log("Registered component " + descriptor[i], Log.LOG_INFO);
        }
    } else {
        utilities.Log.getDefault().log("Warning: Component found with null descriptor", Log.LOG_WARNING);
    }
  }

  /**
   * Removes a component if it exists in the table.  Should be called
   * by AgentComponents' constructors. Generally done only at
   * construction time when a component resets its descriptor entry.
   * @param a The component to add
   */
  public static synchronized void removeComponent(AgentComponent a) {
    String []descriptor = null;

    if (a.getDescriptor() != null)
      descriptor = a.getDescriptor().value();

    if (descriptor != null) {
      for (int i = 0; i < descriptor.length; i++) {
	if (components.containsKey(descriptor[i])) {
	  utilities.Log.getDefault().log("Removed component " + descriptor[i], Log.LOG_INFO);
	  components.remove(descriptor[i]);
	}
      }
    }
  }

  /**
   * Gets a component.  Don't call this from a constructor, as the
   * complete component hashtable may not have been built yet.
   * @param descriptor The full descriptor name of the component to get
   * @return A reference to that component, or null if none found.
   * @see #findComponent
   */
  public static AgentComponent getComponent(String descriptor) {

    return (AgentComponent)components.get(descriptor);
  }

  /**
   * Returns an enumeration of all active components.
   */
  public static Enumeration components() {

    return new IteratorEnumeration(components.values().iterator());
  }

  /**
   * Finds a component.  Don't call this from a constructor, as the
   * complete component hashtable may not have been built yet.
   * Basically does a substring search on all the registered componenets
   * and returns the first match.  Order of search is exact "STR" match,
   * suffix ".STR" match, suffix "STR" match and substring "STR" match.
   * @param descriptor The name pattern to look for
   * @param A reference to a matching component, or null if none found.
   */
  public static AgentComponent findComponent(String descriptor) {
    Iterator it;
    String key;

    descriptor = descriptor.toLowerCase();

    // Exact match
    it = components.keySet().iterator();
    while (it.hasNext()) {
      key = ((String)it.next());
      if (key.equalsIgnoreCase(descriptor))
	return getComponent(key);
    }

    // End .match
    it = components.keySet().iterator();
    while (it.hasNext()) {
      key = ((String)it.next());
      if ((key.toLowerCase()).endsWith("." + descriptor))
	return getComponent(key);
    }

    // End match
    it = components.keySet().iterator();
    while (it.hasNext()) {
      key = ((String)it.next());
      if ((key.toLowerCase()).endsWith(descriptor))
	return getComponent(key);
    }

    // Normal substring
    it = components.keySet().iterator();
    while (it.hasNext()) {
      key = ((String)it.next());
      if ((key.toLowerCase()).indexOf(descriptor) > -1)
	return getComponent(key);
    }

    // Achtung!
    //utilities.Log.getDefault().log("  Warning, couldn't find a match for \"" + descriptor + "\"", Log.LOG_WARNING);

    return null;
  }

  /**
   * Determines if the component is present or not
   * @param descriptor The name pattern to look for
   * @return true if the component can be found, false otherwise
   */
  public static boolean hasComponent(String descriptor) {
    return (findComponent(descriptor) != null);
  }

  /**
   * Returns an enumeration of all active properties (e.g. the property
   * names or keys)..
   */
  public Enumeration properties() {

    if ((properties == null) || (!collected)) {
      if (log == null)
	utilities.Log.getDefault().log("Grumble...someone's trying to get the property list before I have it", Log.LOG_INFO);
      else
	log.log("Grumble...someone's trying to get the property list before I have it", Log.LOG_INFO, FAC_STATE);
      getProperty("");
    }

    return properties.keys();
  }

  /**
   * Determines if the property is currently stored in the table.
   * @param key The property's identifier key, typically (but not
   * necessarily) a string.
   * @return True, if the property exists
   */
  public boolean hasProperty(Object key) {

    if ((!properties.containsKey(key)) && (!collected)) {
      if (log == null)
	utilities.Log.getDefault().log("Grumble...someone's trying to check property " + key + " before I have it", Log.LOG_INFO);
      else
	log.log("Grumble...someone's trying to check property " + key + " before I have it", Log.LOG_INFO, FAC_STATE);
      getProperty(key);
    }

    return properties.containsKey(key);
  }

  /**
   * Gets a property
   * @param key The property's identifier key, typically (but not
   * necessarily) a string.
   * @return A reference to the property, or null if not found.
   *   This reference should not be stored, as future state accesses
   *   may change the data.
   */
  public Object getProperty(Object key) {

    if ((!properties.containsKey(key)) && (!collected)) {
      if (log == null) {
          utilities.Log.getDefault().log("We interrupt our normally scheduled programming for this brief message...", Log.LOG_DEBUG);
          utilities.Log.getDefault().log("   (some impatient component is looking for " + key + ")", Log.LOG_DEBUG);
      } else {
          log.log("We interrupt our normally scheduled programming for this brief message...", Log.LOG_DEBUG, FAC_STATE);
          log.log("(some impatient component is looking for " + key + ")", Log.LOG_DEBUG, FAC_STATE);
      }
      getOtherParams();
      if (log == null)
          utilities.Log.getDefault().log("And now back to the show", Log.LOG_DEBUG);
      else
          log.log("And now back to the show", Log.LOG_DEBUG, FAC_STATE);
    }

    return properties.get(key);
  }

  /**
   * Sets a property, using no hint data
   * @param key The property's identifier key, typically (but not
   * necessarily) a string.
   * @param property The data
   */
  public void setProperty(Object key, Object property) {
      setProperty(key, property, null);
  }

  /**
   * Sets a property.  The hint parameter allows the component
   * adding or modifying the property to include an arbitrary
   * "hint" describing what they did.  For instance, if the component
   * modified a single field in a hashtable, it might include that
   * key as the hint, so that components monitoring that hashtable
   * can more easily know what change took place.  The actual hint
   * data and type used is entirely up to the developer, so if you
   * intend to supply this data, make sure your conventions are
   * available somewhere for others to refer to.
   * @param key The property's identifier key, typically (but not
   * necessarily) a string.
   * @param property The data
   * @param hint The hint data (can be null)
   */
  public void setProperty(Object key, Object property, Object hint) {
    int id;
    Object oldproperty;

    synchronized(this) {
        StatePointer ptr = new StatePointer(key, property);
        propertyPreprocess(ptr);
        key = ptr.key;
        property = ptr.ptr;
        
        if (properties.containsKey(key)) {
            id = PropertyEvent.CHANGED;
            oldproperty = getProperty(key);
        } else {
            id = PropertyEvent.ADDED;
            oldproperty = property;
        }
        
        properties.put(key, property);
        
        propertyPostprocess(ptr);
    }

    startFireEvent(new PropertyEvent(this, key, oldproperty, property, id, hint));
  }

  /**
   * Deletes a property
   * @param key The property's identifier key, typically (but not
   * necessarily) a string.
   * @retrun True if the property was found
   */
  public boolean removeProperty(Object key) {
      return removeProperty(key, null);
  }

  /**
   * Deletes a property
   * @param key The property's identifier key, typically (but not
   * necessarily) a string.
   * @param hint The hint data (can be null)
   * @return True if the property was found
   */
  public boolean removeProperty(Object key, Object hint) {
    Object property;
    
    if (properties.containsKey(key)) {
      property = properties.get(key);
      properties.remove(key);
      startFireEvent(new PropertyEvent(this, key, property, PropertyEvent.REMOVED, hint));
    } else {
      return false;
    }

    return true;
  }

  /**
   * Override this to add preprocessing to the added data.  By default
   * this function uses Converter.reTypeProperty to convert string versions of
   * known properties to their correct types.
   * @param property The key and data, in a pointer
   * @see #getOtherParams
   */
  public void propertyPreprocess(StatePointer property) {
    Object key = property.key;
    String delim = ",";

    // Retype it
    if ((key instanceof String) && (property.ptr instanceof String)) {
      String info[] = getParameterInfo((String)key);
      if (info != null) {
	String type = info[1];
	property.ptr = Converter.reTypeProperty((String)property.ptr, type);
      }
    }

    // Watch for seed changes
    if (key.equals(NAME) && (property.ptr instanceof String)) {
       Container c = this;
       while (c.getParent() != null) {
           c = c.getParent();
       }
       if ((c != null) && (c instanceof Frame)) {
           ((Frame)c).setTitle("Agent: " + property.ptr);
       }
    }
  }

  /**
   * Override this to add postprocessing to the added data.
   * @param property The key and data, in a pointer
   */
  public void propertyPostprocess(StatePointer property) {
    Object key = property.key;
    String delim = ",";
  
    // Watch for extconfig setting
    if (hasProperty(EXTCONFIG) && hasProperty(MASSHOME) &&
        (key.equals(EXTCONFIG) || key.equals(MASSHOME))) {
        log.log(EXTCONFIG + " entries:", Log.LOG_INFO, FAC_STATE);
        ZipFile ext = getExtendedEntryFile();
        if (ext != null) {
            Enumeration e = ext.entries();
            while(e.hasMoreElements()) {
                log.log("\t" + (ZipEntry)e.nextElement(), Log.LOG_INFO, FAC_STATE);
            }
        }

    } else if (key.equals(STATEMENULENGTH)) {
        menulen = ((Integer)getProperty(STATEMENULENGTH)).intValue();
    }
  }

  /**
   * Use this method to obtain a Random object for all your 
   * random number generating needs.  This is necessary to
   * make the agents deterministic.
   * <P>
   * Note that the random object here will be initialized
   * locally until the seed is set.  This means that you will
   * not be deterministic if you use random before the seed
   * is aquired - watch out for usage of this function in
   * constructors or init!  Note that some classes (such as
   * Action) use this random object to generate their ID numbers
   * so this same warning applies to them as well (e.g. don't
   * create an Action object in your constructor).
   * @return The local random object
   */
  public static Random getRandom() {

    if (random == null) {
        State state = (State)State.findComponent("State");
        if ((state != null) && state.hasProperty(RANDOMSEED)) {
            long seed = ((Long)state.getProperty(RANDOMSEED)).longValue();
            if (state.hasProperty(NAME))
                seed += state.getProperty(NAME).hashCode();
            log.log("Generating random number generator with seed " + seed, Log.LOG_DEBUG);
            random = new Random(seed);

        } else {
            random = new Random(parameters.size());
            if (log != null)
                log.log("Warning, I just created a non-deterministic Random object for you.", Log.LOG_ERR);
            else
                utilities.Log.getDefault().log("Warning, I just created a non-deterministic Random object for you.", Log.LOG_ERR);
        }
    }

    return random;
  }

  /**
   * Reponds to normal menu selections
   * @param e The event to look at
   */
  public void actionPerformed(java.awt.event.ActionEvent e) {
    Object command = e.getActionCommand();

    if (hasProperty(command)) {
        System.err.println(command.toString() + ": " + getProperty(command));
    } else {
        log.log("Whoops, can't find property " + command, Log.LOG_WARNING, FAC_STATE);
    }
  }

  /**
   * Returns true if the component has a popupmenu.
   */
  public boolean hasPopupMenu() {
      return true;
  }

  /**
   * Responds to mouse clicks
   */
  public void processMouseEvent(MouseEvent e) {
    if ((e.getID() == MouseEvent.MOUSE_PRESSED) || e.isPopupTrigger()) {
        MenuItem m;

        try {
            for (int i = 0; true; i++) {
                m = menu.getItem(i);
                //m = (MenuItem)menu.getComponent(i);
                m.removeActionListener(this);
            }
        } catch (ArrayIndexOutOfBoundsException ex) {
        }
        menu.removeAll();

        m = new MenuItem("Print all parameters");
        m.addActionListener(new ActionListener() {
                public void actionPerformed(java.awt.event.ActionEvent e) {
                    printParameterInfo();
                }
            });
        menu.add(m);
        menu.addSeparator();

        Enumeration en = new SortEnumeration(properties());
        for (int i = 0; (en.hasMoreElements() && (i < nextmenustart + menulen)); i++) {
            Object o = en.nextElement();
            if (i < nextmenustart) continue;
            m = new MenuItem(o.toString());
            m.addActionListener(this);
            menu.add(m);
        }
        if (en.hasMoreElements()) nextmenustart += menulen;
        else nextmenustart = 0;
        
        menu.show(this, e.getX(), e.getY());
    }
  }
  public void mouseClicked(MouseEvent e) { processMouseEvent(e); }
  public void mousePressed(MouseEvent e) { processMouseEvent(e); }
  public void mouseReleased(MouseEvent e) { processMouseEvent(e); }
  public void mouseDragged(MouseEvent e) { processMouseEvent(e); }
  public void mouseEntered(MouseEvent e) { }
  public void mouseExited(MouseEvent e) { }
  public void mouseMoved(MouseEvent e) { }
}
