/* Copyright (C) 2005, University of Massachusetts, Multi-Agent Systems Lab
 * See LICENSE for license information
 */

/***************************************************************************************
 * WindowManager.java
 *
 * manifest: Name: agent/simplest/WindowManager.class
 * manifest: Java-Bean: True
 ***************************************************************************************/

package agent.simplest;

import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;

import agent.base.*;

/**
 * The WindowManager component provides a consistent way to create and organize windows (JFrames)
 * in your agent.  Instead of creating a frame object directly, you can call <TT>createWindow</TT>
 * which creates the JFrame, registers it with the WindowManager, sets the visibility flag
 * to be the same as the WinVis parameter, and returns the reference.  At runtime,
 * the window will then be listed in the WindowManager's window menu, which appears when
 * you click on the WindowManager component box.
 * <P>
 * If you still need to create the window yourself, you may do so, and register it
 * with the <TT>addWindow</TT> method directly.
 * @see #createWindow
 * @see #addWindow
 */
public class WindowManager extends AgentComponent implements ActionListener, MouseListener, ItemListener, PropertyEventListener, WindowListener {
  static final long serialVersionUID = 1234567890L; // kludge static ser_id.  Bad bad Bryan!
  protected State state = null;
  protected static Hashtable wmapping = new Hashtable();
  protected static Hashtable mmapping = new Hashtable();
  private static boolean winvis;
  private static PopupMenu menu = new PopupMenu("Window List");
  private static int count = 0;

  public static final String WINVIS = "WinVis";

  /**
   * Default constructor
   */
  public WindowManager() {

    addDependency("State");
    State.addParameterInfo(WINVIS, "Boolean", "If true, registered windows will be visibile by default", Boolean.FALSE);

    // Listen for mouses
    addMouseListener(this);

    // Make default menu
    MenuItem m;

    m = new MenuItem("Show All");
    m.addActionListener(this);
    menu.add(m);

    m = new MenuItem("Hide All");
    m.addActionListener(this);
    menu.add(m);

    menu.addSeparator();

    add(menu);
  }

  /**
   * Init
   */
  public void init() {
    state = (State)State.findComponent("State");

    state.addPropertyEventListener(this);
  }

  /**
   * Default window visibility state accessors
   */
  public boolean getWinVis() { return winvis; }
  public void setWinVis(boolean w) { 
    winvis = w;
    if (state != null) state.setProperty(WINVIS, new Boolean(winvis));
  }

  /**
   * Creates a JFrame and registers it.
   * @param title The title of the frame
   * @return the new JFrame
   */
  public JFrame createWindow(String title) {
    JFrame f = new JFrame(title);
    addWindow(f);
    return f;
  }

  /**
   * Adds a window to the window list, setting its un/hidden
   * status.
   * @param w The window to add
   */
  public void addWindow(Window w) {

    // Add it
    String key = "Window__" + count++;
    CheckboxMenuItem m = new CheckboxMenuItem(getWindowTitle(w), winvis);
    w.addWindowListener(this);
    m.setActionCommand(key);
    m.addItemListener(this);
    menu.add(m);

    wmapping.put(key, w);
    mmapping.put(w, m);

    setWindowState(w, winvis);
  }

  /**
   * Removes a window from the window list
   * @param w The window to remove
   */
  public void removeWindow(Window w) {

    // Remove it
    CheckboxMenuItem m = (CheckboxMenuItem)mmapping.remove(w);
    w.removeWindowListener(this);
    m.removeItemListener(this);
    menu.remove(m);
    wmapping.remove(m.getActionCommand());
  }

    /**
     * Gets a window by its title
     * @param t The title to search for
     * @return The window, or null if not found
     */
    public Window getWindow(String t) {
        Enumeration ws = wmapping.elements();
        while (ws.hasMoreElements()) {
            Window w = (Window)ws.nextElement();
            if (t.equals(getWindowTitle(w)))
                return w;
        }
        return null;
    }

  /**
   * Returns the title of a window
   * @param w The the window
   * @return The window's title
   */
  public String getWindowTitle(Window w) {
    String title = null;

    if (w instanceof JFrame)
      title = ((JFrame)w).getTitle();
    else if (w instanceof Frame)
      title = ((Frame)w).getTitle();
    else if (w instanceof Dialog)
      title = ((Dialog)w).getTitle();

    if (title == null)
      title = new String(w.toString());

    return title;
  }

  /**
   * Sets the visibility of a window
   * @param w The window
   * @param visible The desired visibility status
   */
  public void setWindowState(Window w, boolean visible) {

    if (w != null) {
        w.setVisible(visible);
        if (visible) {
            w.setVisible(true);
        } else {
            w.setVisible(true);
        }
        CheckboxMenuItem m = (CheckboxMenuItem)mmapping.get(w);
        m.setState(visible);
    }
  }

  /**
   * Toggles the visibility of a window
   * @param w The window
   */
  public void toggleWindowState(Window w) {

    if (w != null) {
      setWindowState(w, !w.isShowing());
    }
  }

  /**
   * Shows all the known windows
   */
  public void showAllWindows() {
    Enumeration ws = wmapping.elements();
    while (ws.hasMoreElements()) {
      Window w = (Window)ws.nextElement();
      if (!w.isShowing())
	setWindowState(w, true);
    }
  }

  /**
   * Hides all the known windows
   */
  public void hideAllWindows() {
    Enumeration ws = wmapping.elements();
    while (ws.hasMoreElements()) {
      Window w = (Window)ws.nextElement();
      if (w.isShowing())
	setWindowState(w, false);
    }
  }

  /**
   * Returns true if the component has a popupmenu.
   */
  public boolean hasPopupMenu() {
      return true;
  }

  /**
   * Responds to mouse clicks
   */
  public void processMouseEvent(MouseEvent e) {
    if ((e.getID() == MouseEvent.MOUSE_PRESSED) || e.isPopupTrigger()) {
       menu.show(this, e.getX(), e.getY());
    }
  }
  public void mouseClicked(MouseEvent e) { processMouseEvent(e); }
  public void mousePressed(MouseEvent e) { processMouseEvent(e); }
  public void mouseReleased(MouseEvent e) { processMouseEvent(e); }
  public void mouseDragged(MouseEvent e) { processMouseEvent(e); }
  public void mouseEntered(MouseEvent e) { }
  public void mouseExited(MouseEvent e) { }
  public void mouseMoved(MouseEvent e) { }

  /**
   * Reponds to normal menu selections
   * @param e The event to look at
   */
  public void actionPerformed(java.awt.event.ActionEvent e) {
    String command = e.getActionCommand();

    if (command.equals("Show All")) {
      showAllWindows();
    } else if (command.equals("Hide All")) {
      hideAllWindows();
    }
  }

  /**
   * Reponds to checkbox menu selections
   * @param e The event to look at
   */
  public void itemStateChanged(ItemEvent e) {

    CheckboxMenuItem m = (CheckboxMenuItem)e.getItemSelectable();
    String command = m.getActionCommand();
    setWindowState((Window)wmapping.get(command), m.getState());
  }

    /**
     * Watch for window events.
     */
    public void windowActivated(WindowEvent e) { }
    public void windowClosed(WindowEvent e) { }
    public void windowClosing(WindowEvent e) { setWindowState(e.getWindow(), false); }
    public void windowDeactivated(WindowEvent e) { }
    public void windowDeiconified(WindowEvent e) { }
    public void windowIconified(WindowEvent e) { }
    public void windowOpened(WindowEvent e) { }

  /**
   * Watch for interesting property changes.  
   * @param e The event to look at
   */
  public void propertyChanged(PropertyEvent e) { }
  public void propertyRemoved(PropertyEvent e) { }
  public void propertyAdded(PropertyEvent e) {

    if (e.getKey().toString().equalsIgnoreCase(WINVIS)) {
      winvis = ((Boolean)e.getProperty()).booleanValue();
    }
  }
}
